#ifndef CLASS_H
#define CLASS_H

#include <cmm/vartype.h>
#include <cmm/context.h>
#include <cmm/value.h>

class Value ;
class Variable ;

class Class {
  public:
    struct ClassMemberTemplate {
	VarType iv_type ;
	int index ;
	Symbol *name_sym ;
    } ;
  public:
    Class    *superclass ;
    Variable *classnamep ;
    int offset ;
    int num_members ;
    int num_cvs_methods ;
    int num_instvars ;
    ClassMemberTemplate *class_template ;
    Value *class_vars_or_methods ;
    int have_disposer ;
  public:
    int FindMyEntry       (Symbol *, VarType *) ;
    Class *FindClass      (Symbol *, Value **) ;
    Value *FindMethodOrStaticOfClass (Symbol *, VarType *) ;
    void SetMethod        (int slot, Context& ctx) ;
    void SetCreator	  (int slot, Context& ctx) ;
    void SetAcMethod      (int slot, Arr& arr) ;
    int  SetBuiltinMethod (int slot, BuiltinMethod meth) ;
    int	 SetStaticMember  (int slot, Value& v) ;
    Value *FindMethod     (Symbol *, VarType *) ;
    void ShowSpec  () ;
    void DumpClass () ;
    Value *FindMethodOrStatic (Symbol *, VarType *) ;
} ;

class Instance {
  public:
    Class *classp ;
    int num_slots ;
    Value *instance_values ;
  private:
    void ShowValue (Class *classp) ;
  public:
    inline Value *FindMember (Symbol *name_sym) { VarType mtype ;
				return FindMember (classp, name_sym, &mtype) ; }
    Value *FindMember (Symbol *classname, Symbol *name) ;
    Value *FindMember (Class *, Symbol *, VarType *) ;
    Value *FindMember (Class *baseclass, Symbol *name_sym, Value **cp, VarType *mtype) {
	int dum ;
	return FindMember (baseclass, name_sym, cp, mtype, &dum) ;
    }
    Value *FindMember (Class *, Symbol *, Value **, VarType *, int *) ;
    inline Value *FindMemberByIndex (Value *classp, VarType vtype, int index) {
	if (vtype == VT_Private || vtype == VT_Public)
	    return &instance_values [index] ;
	else
	    return &classp->classp->class_vars_or_methods [index] ;
    }
    //----------------------------------------------------------------
    // member functions for assignment with class
    //----------------------------------------------------------------
    inline int SetValue (Symbol *member, int iv)
    	{ Value v ; v = iv ; return SetValue (member, v) ; }
    inline int SetValue (Symbol *member, double dv)
    	{ Value v ; v = dv ; return SetValue (member, v) ; }
    inline int SetValue (Symbol *member, char *sv)
    	{ Value v ; v = sv ; return SetValue (member, v) ; }
    inline int SetValue (Symbol *member, void *vv)
    	{ Value v ; v = (void *)vv; return SetValue (member, v);}
    inline int SetValue (Symbol *member, Value& v) 
        { Value *vp = FindMember (member) ;
	  if (!vp) return -1 ; *vp = v ; return 0 ; } ;
    inline int SetValue (Symbol *member)
	{ Value *vp = FindMember (member) ;
	  vp->Init () ; return 0 ; }
    //----------------------------------------------------------------
    // member functions for assignment with class
    //----------------------------------------------------------------
    inline int SetValue (Symbol *classname, Symbol *member, int iv)
    	{ Value v ; v = iv ; return SetValue (member, v) ; }
    inline int SetValue (Symbol *classname, Symbol *member, double dv)
    	{ Value v ; v = dv ; return SetValue (member, v) ; }
    inline int SetValue (Symbol *classname, Symbol *member, char *sv)
    	{ Value v ; v = sv ; return SetValue (member, v) ; }
    inline int SetValue (Symbol *classname, Symbol *member, void *vv)
    	{ Value v ; v = (Symbol *)vv; return SetValue (member, v); }
    inline int SetValue (Symbol *classname, Symbol *member, Value& v) 
    	{ Value *vp = FindMember (classname, member) ;
	  if (!vp) return -1 ; *vp = v ; return 0 ; } ;
    //----------------------------------------------------------------
    // class checking
    //----------------------------------------------------------------
    inline int FindClass (Symbol *cnamesym) {
	Value *cnamep ;
	return (int)FindClass (cnamesym, &cnamep) ;
    }
    Class *FindClass (Symbol *, Value **) ;
    Class *FindClass (Class *, Symbol *, Value **) ;
    Value *SelectClass (Class *) ;
    inline Class *MyClass (Class *cc) {
	Class *cp ;
	for (cp = classp ; cp ; cp = cp->superclass) {
	    if (cp == cc)
		return cp ;
	}
	return NULL ;
    }
    void DumpInstance () ;
    void ShowValue () ;
} ;

#endif
