//----------------------------------------------------------------
// cmm.h: Cmm Interpretter
//						(c) Ono Tetsuo
//----------------------------------------------------------------
#ifndef CMM_H
#define CMM_H

#include <stdio.h>
#include <cmm/config.h>
#include <cmm/context.h>
#include <cmm/value.h>
#include <cmm/stack.h>
#include <cmm/heap.h>
#include <cmm/vartype.h>
#include <cmm/class.h>
#include <cmm/symbol.h>

#define INSTPREFIX 0x01

#define ARG_PREFIX '$'
#define EXPR_DELIM ';'

#define D_TRUE  1.0
#define D_FALSE 0.0
#define I_TRUE  1
#define I_FALSE 0

//----------------------------------------------------------------
// Error codes
//----------------------------------------------------------------
#define NO_ERROR		0
#define LEX_ERROR		1
#define SYNTAX_ERROR		2
#define DIVISON_BY_0		3
#define INVALID_WORK_VAR	4
#define EXT_VAR_ERROR		5
#define NO_EXT_PARSER		6
#define ALREADY_DEFINED		7
#define TOO_MANY_FUNCS		8
#define UNDEFINED_IDENTIFIER	9
#define STACK_OVERFLOW		10
#define INVALID_STRING		11
#define NOT_ENOUGH_MEMORY	12
#define INVALID_EXPRESSION	13
#define INVALID_STRING_OPERATION 14
#define CANNOT_LOAD_FILE	15
#define UNDEFINED_CLASS		16
#define UNDEFINED_METHOD	17
#define UNDEFINED_MEMBER	18
#define INVALID_ARRAY_INDEX	19
#define INVALID_OBJECT		20
#define CANNOT_ACCESS_MEMBER	21
#define BAD_FUNCTION_OR_METHOD  22
#define INVALID_MEMBER		23
#define COMPILER_NOT_INSTALLED  24
#define INVALID_NEW		25
#define INVALID_SUPER_CLASS	26
#define FUNCALL_ERROR		27
#define METHODCALL_ERROR	28
#define TOO_MANY_CREATOR	29
#define CANNOT_LOAD_PACKAGE	30
#define BUILTIN_CALL_ERROR	31

#include <cmm/symbol.h>
#include <cmm/vartype.h>

#define ACODE_HEADER "#<!!-ACODE-!-ACODE-!-ACODE-!!>"

//----------------------------------------------------------------
// Cmm Interpretter
//----------------------------------------------------------------
class CmmInterp {
  public:
    // g[N
    enum Token {
	T_Number, T_IntNumber, T_String, T_Void, T_Symbol,
	T_WorkVar,
	T_ColonColon,  T_PointTo,
	T_Not,  T_PP,    T_MM,  T_BitRev,
	T_Mult, T_Div,   T_Mod, T_Band,
	T_Plus, T_Minus, T_Lsh, T_Rsh, T_Xor, T_Bor, 
	T_Eq,  T_Neq, T_Lt, T_Gt, T_Le, T_Ge, 
	T_And, T_Or,
	T_Assign,    T_RefAssign, T_AddAssign,  T_SubAssign,
	T_LshAssign, T_RshAssign, T_MultAssign,	T_DivAssign,
	T_ModAssign, T_OrAssign,  T_AndAssign,	T_XorAssign,
	T_Question,   T_Colon,
	T_If, T_Else, T_While, T_For, T_Return, T_Break, T_Continue,
	T_Func,	T_Vardef, T_Class, 
	T_Word,  
	T_ArgPrefix, T_ArgList, T_ArgCnt, T_BoundArg, T_Position,
	T_Variable,  T_Array,
	T_ArrElem,   T_Macro, T_Load, T_Package, T_New,
	T_Method,    T_Hiden, T_Private, T_Public, T_Static,
	T_Eval,      T_Err,
	T_Super,     T_Pause, T_Location, T_Exit, T_This, T_Self,
	T_Lbl, T_Rbl, T_Lpar, T_Rpar, T_Lbr, T_Rbr,  T_Comma,
	T_LLbr,      T_RRbr,  
	T_Try,	     T_Catch, T_Throw, T_Finally,
	T_Eos,       T_Eof,
	T_EndOfToken
    } ;
    enum AcState { Ac_InMain, Ac_InFunc, Ac_InMethod } ;
    enum ExecMode { ExInterp, ExAcInterp, ExCompiler } ;
    // ͂œulvĂ߂ union
    union LexInfo {
	Value  *valp ;
	double value ;
	int    ival ;
	char   *sval ;
	Symbol *symbol ;
    } ;
  private:
    // \e[ũf[^\
    struct ResWordTable {
	char *name ;
	Token token ;
    } ;
    // Location Information
    // ݎs̕ϐ́uʒuvǗ邽߂ class
    class LocInfo {
      public:
	VarType vtype ;
	Value *instancep ;
	Value *classp ;
	Value *varp ;
	inline LocInfo ()    { vtype = VT_Const ; instancep = NULL ;
			       classp = NULL ; varp = NULL ; }
	inline void Reset () { vtype = VT_Const ; instancep = NULL ;
			       classp = NULL ; varp = NULL ; }
    } ;
    void InitStorage (int, int) ;
    void InitStorage (Variable *, int, char *, int, StaticSymbolTable *symtbl = NULL) ;
  protected:
    ExecMode	exec_mode ;
    int 	has_compiler ;	// compiler Ă邩̃tO
    Context	*prog ;		// ͒̃\[XR[h̃obt@ƈʒu
    int		lastPC ;
    int		errorCode;	// ͒ɔG[̃R[h
    Context	errorCtx ;	// G[ʒu
    int 	argcnt ;
    int 	argindex ;
    Value	*thisptr ;
    Value	*classptr ;
    Value	*classRoot ;
    Value	*classClass ;
    Value	*intClass ;
    Value	*realClass ;
    Value	*arrayClass ;
    Value	*stringClass ;
    Value	*metaClass ; // to be defined
    Value	*exceptionClass ;
    Value	*voidClass ;
    Value	*contextClass ;
    Value	*symbolClass ;
  protected:
    // static tables
    static      char         *errorMes     [] ;
    static	ResWordTable reservedWords [] ;
    static	char	     builtInSources[] ;
    static	char	     creatorName[] ;
    static	char	     nullCreator[] ;
    static 	char	     disposerName[] ;
    static	char	     exceptionClassName[] ;
    Symbol	**resWordSymTable ;
    Symbol	*creatorSym ;
    Symbol	*disposerSym ;
    Symbol	*exceptionClassSym ;
  public:
    Value	*arrayOutOfBoundsException ;
    Value	*divisionByZeroException ;
    Value	*typeMismatchException ;
    Value	*argNumberException ;
    Value	*fileLoadException ;
    Value	*packageLoadException ;
    int		ThrowSystemException (Value *exception) ;
  protected:
    Value	*namep ;
    Symbol	*varsymbol ;
    char	*varname ;
    int		max_namelen ;
    Value	*stringp ;
    int		max_strlen ;
  public:
    // Symbol table
    SymbolTable *symbolTable ;
    // memory storage, Heap and stack
    VarStack	varstack  ;
    Variable    *globalLim ;
    Heap	heap_area ;
    //=== contructors
    CmmInterp (Variable *, int, char *, int, StaticSymbolTable *symtbl = NULL) ;
    CmmInterp (int, int) ;
    ~CmmInterp () {} ;
    SetStaticSymbolTable (StaticSymbolTable *symtbl) {
	symbolTable = symtbl ;
    }
  public:
    //================================================================
    // public methods
    //================================================================
    //--- initializer
    void Init () ;
    void InitEnv () ;
    void AddBuiltin  (char *, BuiltinFunc) ; // Builtin Function Handling
    int  AddBiMethod (char *, char *, BuiltinMethod) ;
    int  SetStaticMember (Symbol *, Symbol *, Value&) ;
    int  SetStaticMember (char *classname, char *member, Value& v)
		{   return SetStaticMember (symbolTable->GetSymbol (classname),
					    symbolTable->GetSymbol (member), v) ; }
    inline int SetStaticMember (char *classname, char *member, int iv)
	    { Value v ; v = iv ; return SetStaticMember (classname, member, v) ; }
    inline int SetStaticMember (char *classname, char *member, double dv)
	    { Value v ; v = dv ; return SetStaticMember (classname, member, v) ; }
    inline int SetStaticMember (char *classname, char *member, char *string)
	    { Value v ; v = string ; return SetStaticMember (classname, member, v) ; }
    
    Value *LookupVar (Symbol *sym, VarType *vtype)
			{ int is_member ; Value *cp ;
			  return LookupVar (sym, NULL, vtype, &is_member, &cp) ; }
    Value *LookupVar (Symbol *sym)
			{ VarType vtype ; int is_member ; Value *cp;
			  return LookupVar (sym, NULL, &vtype, &is_member, &cp) ; } ;
    Value *LookupVar (Symbol *sym, Class *classp, VarType *vtype, int *is_member, Value **cp) ;
    inline Value *FindMember (Value *ip, char *member)	{
	if (ip->IsInstance())
	    return ip->instance->FindMember (symbolTable->GetSymbol(member)) ;
	return NULL ;
    }
    inline Value *FindMember (Value *ip, char *classname, char *member) {
	if (!ip->IsInstance())
	    return NULL ;
	return ip->instance->FindMember (symbolTable->GetSymbol (classname),
					 symbolTable->GetSymbol (member)) ;
    }
    inline Class *FindClass (Value *ip, char *baseclass, char *classname, Value **valpp) {
	if (ip->IsInstance()) {
	    Symbol *class_sym = symbolTable->GetSymbol (classname) ;
	    Value *baseClass = (!baseclass) ? NULL : varstack.LookupVar (baseclass) ;
	    Class *baseclassp ;

	    if (baseClass && !baseClass->IsClass()) {
		return NULL ;
	    }
	    baseclassp = (baseClass) ? baseClass->classp : NULL ;
	    return ip->instance->FindClass (baseclassp, class_sym, valpp) ;
	}
	return NULL ;
    }
    //----------------------------------------------------------------
    // Assigment Functions for Instance First arg is Value *
    //----------------------------------------------------------------
    inline int SetValue (Value *ip, char *member, Value& v) {
	if (!ip->IsInstance())
	    return -1 ;
	ip->instance->SetValue (symbolTable->GetSymbol(member), v) ;
	return 0 ;
    }
    inline int SetValue (Value *ip, char *member, int iv)  {
	Value v ; v = iv ; return SetValue (ip, member, v) ;
    }
    inline int SetValue (Value *ip, char *member, double dv) {
	Value v ; v = dv ; return SetValue (ip, member, v) ;
    }
    inline int SetValue (Value *ip, char *member, char *sv) {
	Value v ; v = sv ; return SetValue (ip, member, v) ;
    }
    inline int SetValue (Value *ip, char *member, void *vv) {
	Value v ; v = (void *)vv ; return SetValue (ip, member, v) ;
    }
    //----------------------------------------------------------------
    // Assigment Functions for Instance -- First arg is Instance *
    //----------------------------------------------------------------
    inline int SetValue (Instance *ip, char *member, Value& v) {
	ip->SetValue (symbolTable->GetSymbol(member), v) ;
	return 0 ;
    }
    inline int SetValue (Instance *ip, char *member, int iv)  {
	Value v ; v = iv ; return SetValue (ip, member, v) ;
    }
    inline int SetValue (Instance *ip, char *member, double dv) {
	Value v ; v = dv ; return SetValue (ip, member, v) ;
    }
    inline int SetValue (Instance *ip, char *member, char *sv) {
	Value v ; v = sv ; return SetValue (ip, member, v) ;
    }
    inline int SetValue (Instance *ip, char *member, void *vv) {
	Value v ; v = (void *)vv ; return SetValue (ip, member, v) ;
    }
    //----------------------------------------------------------------
    inline Value *LookupVar (Symbol *sym, int hashv, Class *classp,
			     VarType *vtype, int *is_member,
			     Value **cpp, int *index)
   {
	Value *varp, *var2p ;
	*is_member = 0 ;
	*cpp = NULL ;
	varp = varstack.LookupVar (sym, hashv) ;
	if (!thisptr) {
	    *vtype = VT_NormalVar ;
	    return varp ;
	}
	// thisptr uāvꍇ: 
        if (!varp || varp <= thisptr) {
	    var2p = NULL ;
	    if (thisptr->IsInstance()) {
		Value *cp ;
		var2p = thisptr->instance->FindMember (classp, sym, &cp, vtype, index) ;
		if (var2p)
		    *cpp = cp ;
	    } else if (thisptr->IsClass()) {
		var2p = thisptr->classp->FindMethodOrStatic (sym, vtype) ;
		*cpp = NULL ;
		*index = -1 ;
	    }
	    if (!var2p)
		return varp ;
	    *is_member = 1 ;
	    return var2p ;
	}
	*vtype = VT_NormalVar ;
	return varp ;
    }
    Value *AddGlobalVar (char *) ;
    Value *AddGlobalVar (char *, double) ;
    Value *AddGlobalVar (char *, char *) ;
    Value *AddGlobalVar (char *, void *) ;
    Value *AddGlobalVar (char *, Value &) ;
    Value *AddGlobalVar (char *, int) ;
    Value AllocArray (int size) {
	return heap_area.AllocArray (size, size, &varstack) ;
    }
    char *AllocString (char *string) {
	return heap_area.AllocString (string, &varstack) ;
    }
    Symbol *GetSymbol (char *name) {
	return symbolTable->GetSymbol (name) ;
    }
    //--- executor    
    int         Execute (char *) ;
    int		ExecuteApp (char *buffer) { globalLim = varstack.Top() ;
					    return Execute (buffer) ; }
    int		SetSource (char *) ;
    //-- 
    int CommandArg (int, char **) ;
    // Inter class interface for HEAP
    int DisposeInstance (Instance *) ;	
    //-- error 
    inline int  errorno   () { return errorCode ; } ;
    inline char *errormes () { return errorMes[errorCode] ; }
    int EvalString      (char *) ;
    //-- lexical analyzer
    Token       Lex     (LexInfo *) ;
    inline void PutBack ()      { prog->Set(lastPC) ; }
    Value    *LexName   ()      { return namep ; }
    Instance *NewInstance (char *) ;
    Instance *NewInstance (char *, Value *args, int numargs) ; 
    inline Context *GetContext () { return prog ; }
    int Call (char *, char *, char *, Value *, int, Value *) ;
    int Call (Value *, Value *, Value *, int, Value *) ;
    int CallMethod (Value *, Value *, Value *, Value *, int, Value *) ;
    int SetErrorCode(int)  ;
    int SetErrorCode (int no, Context& ctx) ;
    int ExecInterp (char *) ;
    Value	*FixSpecialClass (Value *v) {
				return ((v->IsInteger  ()) ? intClass
					: (v->IsReal   ()) ? realClass
					: (v->IsString ()) ? stringClass
					: (v->IsArray  ()) ? arrayClass
					: (v->IsVoid   ()) ? voidClass  
					: (v->IsContext()) ? contextClass
					: (v->IsSymbol())  ? symbolClass  : NULL) ;
			    }
    int		IsSpecialClass (Value *classp) {
			return ((classp == intClass) || (classp == realClass)
				|| (classp == stringClass) || (classp == arrayClass)
				|| (classp == voidClass)) ;
		    }
    virtual int ExecJIT (char *) ;
    virtual int ExecCompiler (char *) ;
  private:
    //==================================================
    // Methods for Interpretter
    //==================================================
    BuiltinFunc FindBuiltin (char *, int *) ;
    inline Token CheckNext (char c, Token iftrue, Token iffalse)
    				{ if (prog->PeekC() != c) return iffalse ;
				  prog->Adv() ;
				  return iftrue  ; }
  private:
    int Term           (Token, LexInfo *, LocInfo *, int) ;
    int Factor         (Token, LexInfo *, LocInfo *, int) ;
    int Expression     (Token, LexInfo *, LocInfo *, int) ;
    int Condition2     (Token, LexInfo *, LocInfo *, int) ;
    int Condition1     (Token, LexInfo *, LocInfo *, int) ;
    int Condition0     (Token, LexInfo *, LocInfo *, int) ;
    int Conditional    (Token, LexInfo *, LocInfo *, int) ;
    int Assign	       (Token, LexInfo *, LocInfo *, int) ;
    int IfStatement    (Token, LexInfo *, LocInfo *, int) ;
    int WhileStatement (Token, LexInfo *, LocInfo *, int) ;
    int ForStatement   (Token, LexInfo *, LocInfo *, int) ;
    int TryCatchStatement (Token, LexInfo *, LocInfo *, int) ;
    int ClassDef       (Token, LexInfo *, LocInfo *, int) ;
    int MethoDef       (Token, LexInfo *, LocInfo *, int) ;
    int VarDef	       (Token, LexInfo *, LocInfo *, int) ;
    int FuncDef        (Token, LexInfo *, LocInfo *, int, int) ;
    int FunCall	       (Token, LexInfo *, LocInfo *, int) ;
    int FunCall	       (Value *entp, LocInfo *lip)
    				{ return FunCall (NULL, entp, lip) ; }
    int FunCall	       (char *strp, LocInfo *lip)
    				{ return FunCall (strp, NULL, lip) ; }
    int StartCreator   (Value *classp, Value *entry) ;
    int CallCreator    (Value *classp, Value *entry, int numargs) ;
    int FunCall	       (char *, Value *, LocInfo *) ;
    int Eval	       (Token, LexInfo *, LocInfo *, int) ;
    int Load	       (Token, LexInfo *, LocInfo *, int) ;
    int Package	       (Token, LexInfo *, LocInfo *, int) ;
    inline int Statement (Token t, LexInfo *lip, LocInfo *locp, int skip) 
			    { return Statement (t, lip, locp, skip, 1) ; } ;
    int Statement      (Token, LexInfo *, LocInfo *, int, int) ;
    int ExecMacro      (Value *, int) ;
    int ExpandMacro    (Value *, int) ;
    int ArrIndex       (Token, int, int) ;
    Value *TraceArray  (Token, LexInfo *, Value *, int) ;
    Value *Lvalue      (Value *, int, int) ;
    Value *PointArgument (int) ;
    Value *PointVar    (Token, char *, int, LocInfo *) ;
    int VarValue       (Token, LexInfo *, int, LocInfo *) ;
    int Grow           (Value *, int, int) ;
    int CheckAndGrow   (Value *, int) ;
    inline int AddChar (Value *vs, char c)  { return AddStr (vs, &c, 1) ; }
    int AddStr (Value *, char *) ;
    int AddStr (Value *, char *, int) ;
    int AddStr (Value *, Context *, int, int) ;
    int AddStr (Value *, Value *) ;
    int AddStrQuoted (Value *, Value *) ;
    int SkipBody (int) ;
    int SkipCreatorBody (int) ;
    int SkipArgList () ;
    int PushArgs (int, Value *, int *numargsp = NULL) ;
    int SetNames (int, int, int) ;
  protected:
    int Program	       (Value *valp) ;
    int Program	       (Value *valp, LocInfo *locp) ;
    inline Context *SetContext (char *src, int pc, int subtype = CTX_PROGRAM)
    				{  Context ctx ; Value *vp ;
				   ctx.Set (src, pc) ;
				   vp = varstack.TmPush (ctx) ;
				   vp->subtype = subtype ;
				   return &vp->context ; }
    int CallContext (Context *, int, Value *, Value *, LocInfo *) ;
    int CallCreator (Value *, Value *, Value *, int) ;
    Instance *CreateInstance (Class *c) ;
    int CallDisposer (Value *, Value *, Value *) ;
    int LoadText (char *, char **) ;
    int LoadPackage (char *) ;
    int Debug () ;
    virtual int ExecACode (char *) { fprintf (stderr, "No Compiler !\n") ;
				     return -1 ; }
    virtual void InitBuiltinSource () ;
    
    Value *ReadSource (char *) ;
    int  AllocMethodSlot (Value *, Symbol *, VarType) ;
    int  FixLocInfo (LocInfo *) ;
    void PrintException (Value *exception) ;
    char *FindCmmFile (char *path_varname, char *filename,
			char **suffixes, int sufcnt, int *filesize) ;
  public:
    void HandleSignal (Value *, int) ;
  public:
    // Compiler methods ... virtuals
    virtual int AcCompile    (char *, Value *) 				 {return 0;}
    virtual int AcCall       (Value *, int, Value *, Value *, LocInfo *) {return 0;}
    virtual int AcCallForm   (Value *, int, Value *, Value *, LocInfo *) {return 0;}
    virtual int AcFuncDef    (Token, LexInfo *, AcState, Value *, int)   {return 0;}
    virtual int AcMethoDef   (Token, LexInfo *, AcState, Value *)	 {return 0;}
    virtual int AcMethodBody  (Symbol *, Symbol *, Value *) {return 0;}
    virtual int AcCreatorBody (Symbol *, Symbol *, Value *) {return 0;}
    virtual int AcParseACode (char *, Value *)		 {return 0;}
    virtual int AcExecute    (Value *)			 {return 0;}
    virtual int AcExecForm   (Value *, LocInfo *) 	 {return 0;}
    virtual int AcCallCreator (Value *, Value *, Value *, int) { return 0; }
} ;

class CmmIStatic : public CmmInterp {
#define STACKSIZE 1024
#define HEAPSIZE  4096
#define SYMBOLSIZE 4096
    Variable vartable[STACKSIZE] ;
    char     heap [HEAPSIZE] ;
    char     symbolPool [SYMBOLSIZE] ;
    StaticSymbolTable symTable ;
  public:
    CmmIStatic () : CmmInterp (vartable, STACKSIZE, heap, HEAPSIZE,
			       (StaticSymbolTable *)1 /* dummy */) {
	symTable.Init (symbolPool, sizeof(symbolPool)) ;
	SetStaticSymbolTable (&symTable) ;
    }
} ;

#define NORMAL    0
#define BREAK     1
#define CONTINUE  2
#define RETURN    3
#define THROW	  4
#define EXIT	  -99

#ifdef CMM_INTERNAL
#define CHECKST(st) if (((st) < 0) || (st == EXIT)) return (st)
#define CHECKST0(st) if ((st) < 0) return NULL
#define CHECKST_EX(st) if ((st < 0) || st == EXIT || st == THROW) return st
#endif

typedef int (*Installer) (CmmInterp *) ;

#endif
