//
// cmmcomp.h: Cmm A-Cdoe Compiler/Executor
//
#ifndef CMMCOMP_H
#define CMMCOMP_H

#include <cmm/cmm.h>

class CmmComp : public CmmInterp {
    // Instructions
    enum AcInst {
	A_Nop, A_If, A_While, A_For, A_FuncDef, A_MethoDef, A_HidenDef,
	A_Funcall,
	A_Var, A_SelectMember, A_SelectClass, A_Const, A_Not,
	A_Negate, A_Block, A_SLBlock, A_Plus, A_Minus, A_Mult,
	A_Div, A_Mod, A_Rsh, A_Lsh, A_Eq, A_Neq, A_New, A_Lt, A_Gt,
	A_Le, A_Ge, A_Or, A_And, A_Bor, A_Band, A_Xor, A_BitRev,
	A_Assign, A_AddAssign, A_SubAssign, A_LshAssign, A_RshAssign,
	A_MultAssign, A_DivAssign, A_ModAssign, A_AndAssign,
	A_OrAssign, A_XorAssign, A_PPPre, A_MMPre, A_PPPost, A_MMPost,
	A_Return, A_Break, A_Exit, A_Continue, A_Eval, A_Arg,
	A_ArrIndex, A_ArgList, A_Load, A_Array, A_ArgCnt, A_Pause,
	A_MacroDef, A_ClassDef, A_VarDef, A_RefAssign, A_ThisPtr,
	A_GVar, A_LVar, A_ThisMember, A_NoMember,
	A_CreatorDef, A_Throw, A_TryCatch, A_Self, A_Position, A_BoundArg,
	A_Package, 
	A_EndInst,
    } ;
    static	char         *instNames    [] ;
  public:
    CmmComp (int stacksize, int heapsize) ;
    CmmComp (Variable *, int, char *, int, StaticSymbolTable *symtbl = NULL) ;
    void Init () ;
    int AcCompileAndGo (char * bufer, int show_inst = 0, int no_exec = 0) ;
    int AcCompileAndGoApp (char * buffer, int show_inst = 0, int no_exec = 0)
			    { globalLim = varstack.Top () ;
			      return AcCompileAndGo (buffer, show_inst, no_exec) ; }
    int AcCompile      (char *, Value *) ;
    int AcWriteACode   (char *, Value *, int int_inst = 0) ;
    int AcReadACode    (char *, Value *) ;
    int AcExecute      (Value *) ;
    int AcExecuteApp   (Value *form) { globalLim = varstack.Top () ;
				       return AcExecute (form) ; }
    int Execute (char *buffer)    { exec_mode = ExCompiler ;
				    return CmmInterp::Execute (buffer) ; } ;
    int ExecuteApp (char *buffer) { exec_mode = ExCompiler ;
				    return CmmInterp::ExecuteApp (buffer) ; } ;
    int AcDisCompile   (Value *) ;
    int ExecJIT (char *) ;
    int ExecCompiler (char *) ;
    int CompileFile (char *, Value *) ;
    int ExecACode (char *) ;
    void InitBuiltinSource () ;
    void HitInfo () ;
  private:
    Symbol	*instSyms[(int)A_EndInst] ;
    int 	var_hit  ;
    int		lvar_hit ;
    int		gvar_hit ;
    int		selmem_hit ;
    int		selmem_cache_hit ;
    int		this_mem_hit ;
  private:
    //===============================================
    // Methods for a-code compiler
    //===============================================
    Value *NewArray (int) ;
    Value *NewArray (int, int) ;
    Value *NewArray (int, AcInst) ;
    Value *NewArray (int, int, AcInst) ;
    // Byte code compiler
    int ByteCompileFunc  () ;
    int AcPointVar	 (Token, Value *, AcState, Value *) ;
    int AcVarAndValue    (Token, LexInfo *, Value *, AcState, Value *) ;
    int AcTerm           (Token, LexInfo *, Value *, AcState, Value *) ;
    int AcFactor         (Token, LexInfo *, Value *, AcState, Value *) ;
    int AcExpression     (Token, LexInfo *, Value *, AcState, Value *) ;
    int AcCondition2     (Token, LexInfo *, Value *, AcState, Value *) ;
    int AcCondition1     (Token, LexInfo *, Value *, AcState, Value *) ;
    int AcCondition0     (Token, LexInfo *, Value *, AcState, Value *) ;
    int AcConditional    (Token, LexInfo *, Value *, AcState, Value *) ;
    int AcAssign	 (Token, LexInfo *, Value *, AcState, Value *) ;
    int AcIfStatement    (Token, LexInfo *, AcState, Value *) ;
    int AcWhileStatement (Token, LexInfo *, AcState, Value *) ;
    int AcForStatement   (Token, LexInfo *, AcState, Value *) ;
    int AcClassDef       (Token, LexInfo *, AcState, Value *) ;
    int AcMethoDef       (Token, LexInfo *, AcState, Value *) ;
    int AcMethodBody     (Symbol *, Symbol *, Value *) ;
    int AcTryCatchStatement (Token, LexInfo *, AcState, Value *) ;
    int AcCreatorBody	 (Symbol *, Symbol *, Value *) ;
    int AcVarDef	 (Token, LexInfo *, AcState, Value *) ;
    int AcFuncDef        (Token, LexInfo *, AcState, Value *, int) ;
    int AcMacroDef	 (Token, LexInfo *, AcState, Value *) ;
    int AcFunCall	 (Token, LexInfo *, AcState, Value *) ;
    int AcEval		 (Token, LexInfo *, AcState, Value *) ;
    int AcLoad		 (Token, LexInfo *, AcState, Value *) ;
    int AcStatement	 (Token, LexInfo *, AcState, Value *) ;
    int AcProgram	 (Value *) ;
    int AcListInst	 (FILE *, Value *, int int_inst = 0) ;
    int AcShowInst 	 (FILE *, Value *, int, int) ;
    int AcFuncArg	 (Value *) ;
    int AcSuperArg	 (Value *) ;
    // a-code executer
    int AcExecForm	 (Value *, LocInfo *) ;
    int AcAdd		 (Value *, LocInfo *) ;
    int AcSub		 (Value *, LocInfo *) ;
    int AcMult		 (Value *, LocInfo *) ;
    int AcDiv		 (Value *, LocInfo *) ;
    int AcMod		 (Value *, LocInfo *) ;
    int AcRsh		 (Value *, LocInfo *) ;
    int AcLsh		 (Value *, LocInfo *) ;
    int AcEq		 (Value *, LocInfo *) ;
    int AcNeq		 (Value *, LocInfo *) ;
    int AcLt		 (Value *, LocInfo *) ;
    int AcGt		 (Value *, LocInfo *) ;
    int AcLe		 (Value *, LocInfo *) ;
    int AcGe		 (Value *, LocInfo *) ;
    int AcOr		 (Value *, LocInfo *) ;
    int AcAnd		 (Value *, LocInfo *) ;
    int AcAssign	 (Value *, LocInfo *) ;
    int AcRefAssign	 (Value *, LocInfo *) ;
    int AcAddAssign	 (Value *, LocInfo *) ;
    int AcSubAssign	 (Value *, LocInfo *) ;
    int AcLshAssign	 (Value *, LocInfo *) ;
    int AcRshAssign	 (Value *, LocInfo *) ;
    int AcMultAssign	 (Value *, LocInfo *) ;
    int AcDivAssign	 (Value *, LocInfo *) ;
    int AcModAssign	 (Value *, LocInfo *) ;
    int AcArgList	 (Value *, LocInfo *) ;
    int AcFuncall	 (Value *, LocInfo *) ;
    int AcMakeArray	 (Value *) ;
    int AcCallForm	 (Value *, int, Value *, Value *, LocInfo *) ;
    int AcPushArgs	 (Value *, Value *) ;
    int AcSetNames	 (Value *, int, int, int) ;
    int AcParseACode	 (char *, Value *) ;
    int AcParseForm	 (Value *) ;
    int AcLookup	 (char *) ;
    int AcLookup	 (Symbol *) ;
    int AcArgList	 (Value *) ;
    int AcSuperArgList   (Value *) ;
    int AcStartCreator	 (Value *, Value *, LocInfo *) ;
    int AcCallCreator	 (Value *, Value *, Value *, int numargs) ;
    int AcDisCompile	 (Value *, AcInst, int) ;
    int ThrowSystemException (Value *, Context&) ;
    Symbol *LexSymbolOrString () ;
} ;

#endif
