#ifndef HEAP_H
#define HEAP_H

#include <cmm/value.h>
#include <cmm/stack.h>
#include <cmm/class.h>

#define HP_PURE     0
#define HP_NONE	    1	
#define HP_STRING   2
#define HP_ARRAY    3
#define HP_INSTANCE 4
#define HP_CLASS    5

//
// class Heap: Cmm Interpretter ̃q[v̈񋟂
//
class Heap {
    class SegmentHeader {
      public:
	short mark ;
	short type ;
	int   size ;
    } ;
    class FreeSegmentHeader: public SegmentHeader {
      public:
	FreeSegmentHeader *next ;
    } ;
    int dynamic_allocate ;
    int max_size ;
    int used_size ;
    int rest_size ;
    char *heap_area ;
    char *areapt ;
    int alloc_count ;
    int alloc_free_count ;
    int gc_count ;
    int grow_count ;
    int sweep_count ;
    int gc1_count ;
    FreeSegmentHeader *freeList ;
    int  numFreeList ;
    int  disposing ;
    char *reservpt ;
    char *reserved_area ;
    int  reserv_size ;
    int  rest_of_reserved ;
    CmmInterp *cmmEnv ;
  public:
    Heap () ;
    ~Heap() ;
    void Init (char *, int) ;
    void Init (int) ;
    char *AllocString  (char *, VarStack *) ;
    char *AllocString  (int, VarStack *) ;
    char *AllocSegment (int, VarStack *) ;
    Instance *AllocInstance (int, VarStack *, Value **) ;
    Class    *AllocClass (VarStack *) ;
    int CheckSegment () ;
    Value NewArray (int, VarStack *) ;
    int  AllocArray (int, VarStack *, Value **) ;
    Value AllocArray (int, int, VarStack *) ;
    int  AddArray     (Value *, Value *, VarStack *, Value *) ;
    int  SegmentSize (Value *) ;
    int  CopyValue    (Value *, VarStack *, Value *) ;
    int  CopyArray    (Value *, VarStack *, Value *) ;
    int  ConsArray    (Value *, Value *, VarStack *, Value *) ;
    int  AppendArray  (Value *, Value *, VarStack *, Value *) ;
    int  PopArray     (Value *, int,     VarStack *, Value *) ;
    int  RemoveArray  (Value *, Value *, VarStack *, Value *) ;
    int  RemoveArray  (Value *, int,     VarStack *, Value *) ;
    void NotifyCmmEnv (CmmInterp *cenv) { cmmEnv = cenv ; }
    int  GarbageCollect (VarStack *) ;
    char *GarbageCollect (VarStack *, int) ;
    int  GCSweep    (VarStack *) ;
    int  Resize (int size, VarStack *) ;
    int  Expand (int size, VarStack *) ;
    void Dump () ;
    void BinDump () ;
    void Info () ;
    void DoMark (VarStack *, int) ;
  private:
    void SetupHeap (int) ;
    inline void ClearMark (char *heap_end) {
	char *markp ;
	for (markp = heap_area ;
	     markp < heap_end ;
	     markp += ((SegmentHeader *)markp)->size) 
	((SegmentHeader *)markp)->mark = 0 ;
    }
    inline char *GetFree (int) ;
    int  CopyArrayElements   (Value *, int, Value *, VarStack *) ;
    inline void MarkStack    (VarStack *, int, char *, char *) ;
    inline void MarkStackElement (Variable *, int, char *, char*) ;
    inline void MarkArray    (Value *, int, int, char *, char *) ;
    inline void MarkClass    (Class *, int, char *, char *) ;
    inline void MarkInstance (Instance *, int, char *, char *) ;
    inline void MovePointer  (VarStack *, char *, char *, int) ;
    inline void MoveValue    (Value *, char *, char *, int) ;
    inline void MoveArrPtr   (Value *, int, char *, char *, int) ;
    inline void MoveClass    (Class *, char *, char *, int) ;
    inline void MoveInstance (Instance *, char *, char *, int) ;
    inline void RemapPointer (char *, char *, char *, char *, VarStack *) ;
    inline void RemapArray   (Value *, int, char *, char *, char *) ;
    inline void RemapArrayElements (Value *, int, char *, char *, char *) ;
    inline void RemapClass   (Class *, char *, char *, char *) ;
    inline void RemapInstance (Instance *, char *, char *, char *) ;
    inline void MarkIt (char *ptr, int mark,
			       char *area_start, char *area_end)
    		{   SegmentHeader *hp ;
		    if (area_start <= ptr && ptr < area_end) {
			hp = (SegmentHeader *)(ptr - sizeof(SegmentHeader)) ;
			hp->mark = mark ;
		    }
		}
    inline int GetMark (char *ptr)
    		{ return ((SegmentHeader *)(ptr - sizeof(SegmentHeader)))->mark ; } ;
    inline void SetType (char *ptr, char type)
    		{ ((SegmentHeader *)(ptr - sizeof(SegmentHeader)))->type = type ; }
    inline char GetType (char *ptr)
		{ return ((SegmentHeader *)(ptr - sizeof(SegmentHeader)))->type ; }
    inline char *GetArea (int) ;
    inline int CallDisposer (char *) ;
    char *GetReserved (int) ;
} ;

#endif
