//
// stack.h: Cmm Interpretter
//
#ifndef STACK_H
#define STACK_H

#include <stdio.h>
#include <string.h>
#include <cmm/value.h>
#include <cmm/hash.h>

//
// class VarStack: Cmm Interpretter ̕ϐX^bN
//
class VarStack {
    int dynamic_allocate ;
    int stacksize ;
    int framep ;
    int stackp ;
    int argp ;
    int lsp ;
    Variable *stack ;
    Variable *hashTable [MAXHASH] ;
    SymbolTable *symbolTable ;
  private:
    void Fatal (char *) ;
  public:
    inline friend int Hash (char *s) ;
    VarStack ()  ;
    ~VarStack () ;
    void Init (Variable *stackarr, int size) ;
    void Init (int size) ;
    inline int Max()	       { return stacksize ; } 
    inline int OverFlow ()     { return (stackp >= lsp) ;   }
    inline int UnderFlow ()    { return (stackp <= 0) ; }
    inline Value *FrameRelative (int offset) { return &stack [framep + offset] ; }
    inline void LinkFp () { int newfp = stackp ; Push ("--link--", framep) ;  framep = newfp ; }
    inline void UnlinkFp () { stackp = framep ; Value x = stack[stackp] ; framep = x.intval ; Rehash() ;} ;
    inline void LinkAp ()  { int newargp = stackp ; Push ("--link--", argp) ;  argp = newargp ; }
    inline void UnlinkAp () { stackp = argp ; Value x = stack[stackp] ; argp = x.intval ; Rehash() ;} ;
    inline Value *LPush (Value* loc) {
				if (lsp - 1 <= stackp) {
				    Fatal ("Fatal: stack overflow!") ;
				    exit (30) ;
				}
				Value x ;
				x = loc ;
				stack [--lsp] = x ;
				return &stack [lsp] ;
			    }
    inline Value *LPush (Arr& ar) {
				if (lsp - 1 <= stackp) {
				    Fatal ("Fatal: stack overflow!") ;
				    exit (30) ;
				}
				Value x ;
				x =  ar ;
				stack [--lsp] = x ;
				return &stack [lsp] ;
			    }
    inline Value *LPush (Value& v) {
				if (lsp - 1 <= stackp) {
				    Fatal ("Fatal: stack overflow!") ;
				    exit (30) ;
				}
				stack [--lsp] = v ;
				return &stack [lsp] ;
			    }
    inline Value LPop  ()  { return stack [lsp++] ; }
    inline void  LPop1 ()     { lsp++ ; }
    inline void  LPop (int n) { lsp += n ; }
    inline int   InLStack (Value *vp) {
	return stack + lsp <= ((Variable *)vp) && ((Variable *)vp) < stack + stacksize ;
    }
    Value *LTop()          { return &stack [lsp] ; };

    //================================================================
    // String Interface
    //================================================================
    inline Variable *Push (char *name, Value& v) {
	int hashv ;
	Symbol *sym ;
	sym = symbolTable->GetSymbol (name) ;
	return Push (sym, v) ;
    }
    inline Variable *Push (char *name)
    				{ Value v ; return Push (name, v) ; }
    inline Variable *Push (char *name, char *str)
    				{ Value v ; v = str ; return Push (name, v) ; } 
    inline Variable *Push (char *name, double n)
    				{ Value v ; v = n   ; return Push (name, v) ; }
    inline Variable *Push (char *name, Context& ctx)
			    	{ Value v ; v = ctx ; return Push (name, v) ; }
    inline Variable *Push (char *name, BuiltinFunc ent)
    				{ Value v ; v = ent ; return Push (name, v) ; }
    inline Variable *Push (char *name, Variable **p)
    				{ Value v ; v = p   ; return Push (name, v) ; }
    inline Variable *Push (char *name, int i)
    				{ Value v ; v = i   ; return Push (name, v) ; }
    inline Variable *Push (char *name, Class *cp)
    				{ Value v ; v = cp  ; return Push (name, v) ; }
    inline Variable *Push (char *name, Instance *ip)
    				{ Value v ; v = ip  ; return Push (name, v) ; }
    inline Variable *Push (char *name, Value *vp)
    				{ Value v ; v = vp  ; return Push (name, v) ; }
    inline Variable *Push (char *name, Arr&  arr)
    				{ Value v ; v = arr ; return Push (name, v) ; }
    inline Variable *Push (char *name, Symbol *symbol)
				{ Value v ; v = symbol ; return Push (name, v); }
    //================================================================
    // Symbol Interface
    //================================================================
    inline Variable *Push (Symbol *sym, Value& v) {
	Variable *vp ;
	int hashv ;
	
	if (stackp >= lsp) {
	    Dump ("FATAL:: Dump Stack", 10) ;
	    Fatal ("stack overflow") ;
	}
	vp = stack + stackp ;
	vp->Set (sym, v) ;
	hashv = vp->sym->GetHashValue () ;
	vp->next = hashTable[hashv] ;
	hashTable[hashv] = vp ;
	stackp++ ;
	return vp ;
    }
    inline Variable *Push (Symbol *sym) { Value v ; return Push (sym, v) ; }
    inline Variable *Push (Symbol *sym, char *str)
    				{ Value v ; v = str ; return Push (sym, v) ; } 
    inline Variable *Push (Symbol *sym, double n)
    				{ Value v ; v = n   ; return Push (sym, v) ; }
    inline Variable *Push (Symbol *sym, Context& ctx)
			    	{ Value v ; v = ctx ; return Push (sym, v) ; }
    inline Variable *Push (Symbol *sym, BuiltinFunc ent)
    				{ Value v ; v = ent ; return Push (sym, v) ; }
    inline Variable *Push (Symbol *sym, Variable **p)
    				{ Value v ; v = p   ; return Push (sym, v) ; }
    inline Variable *Push (Symbol *sym, int i)
    				{ Value v ; v = i   ; return Push (sym, v) ; }
    inline Variable *Push (Symbol *sym, Class *cp)
    				{ Value v ; v = cp  ; return Push (sym, v) ; }
    inline Variable *Push (Symbol *sym, Instance *ip)
    				{ Value v ; v = ip  ; return Push (sym, v) ; }
    inline Variable *Push (Symbol *sym, Value *vp)
    				{ Value v ; v = vp  ; return Push (sym, v) ; }
    inline Variable *Push (Symbol *sym, Arr&  arr)
    				{ Value v ; v = arr ; return Push (sym, v) ; }
    inline Variable *Push (Symbol *sym, Symbol *symbol)
				{ Value v ; v = symbol ; return Push (sym, v); }
    //================================================================
    // Pop
    //================================================================
    inline void Pop (int n) {
	int newsp = stackp - n ;
	if (newsp < 0)
	    newsp = 0 ;
	Variable *sp ;
	Variable *spend ;
	for (sp = stack + stackp - 1, spend = stack + newsp ; sp >= spend ; sp--)
	{
	    if (sp->sym)
		hashTable[sp->sym->GetHashValue()] = sp->next ;
	}
	stackp = newsp ;
    }
    inline Value Pop () {
	Value x ;
	x = stack[stackp - 1] ;
	Pop (1) ;
	return x ;
    }
    inline void TmPop () { --stackp ; } ;
    inline void TmPop (int n) { stackp -= n ; }
    inline Value VPop () { return stack[--stackp] ; }
    void Dump  (const char *name, int bottom = 13) ;
    void Dump2 (const char *name, int lines) ;
    inline Variable *TmPush ()    { Value v ;
				    Variable *vp = &stack[stackp++] ;
				    *vp = v ;
				    vp->sym = NULL ;
				    return vp ;	 }
    inline Variable *TmPush (char *str)
				{ if (stackp >= lsp) return NULL ;
				  Variable *vp = &stack[stackp++] ;
				  vp->sym = NULL ;
				  vp->type = V_STRING ;
				  vp->sptr = str ;
				  return vp ; }	 
    inline Variable *TmPush (double n)
				{ if (stackp >= lsp) return NULL ;
				  Variable *vp = &stack[stackp++] ;
				  vp->sym = NULL ;
				  vp->type = V_REAL ;
				  vp->value = n ;
				  return vp ; }	 
    inline Variable *TmPush (Context& ctx)
				{ if (stackp >= lsp) return NULL ;
				  Variable *vp = &stack[stackp++] ;
				  vp->sym = NULL ;
				  vp->type = V_CONTEXT ;
				  vp->context = ctx ;
				  return vp ; }	 
    inline Variable *TmPush (BuiltinFunc ent)
				{ if (stackp >= lsp) return NULL ;
				  Variable *vp = &stack[stackp++] ;
				  vp->sym = NULL ;
				  vp->type = V_BUILTIN ;
				  vp->entry = ent ;
				  return vp ; }	 
    inline Variable *TmPush (int i)
				{ if (stackp >= lsp) return NULL ;
				  Variable *vp = &stack[stackp++] ;
				  vp->sym = NULL ;
				  vp->type = V_INTEGER ;
				  vp->intval = i ;
				  return vp ; }	 
    inline Variable *TmPush (Class *cp)
				{ if (stackp >= lsp) return NULL ;
				  Variable *vp = &stack[stackp++] ;
				  vp->sym = NULL ;
				  vp->type = V_CLASS ;
				  vp->classp = cp ;
				  return vp ; }	 
    inline Variable *TmPush (Instance *ip)
				{ if (stackp >= lsp) return NULL ;
				  Variable *vp = &stack[stackp++] ;
				  vp->sym = NULL ;
				  vp->type = V_INSTANCE ;
				  vp->instance = ip ;
				  return vp ; }	 
    inline Variable *TmPush (Value *valp) 
				{ if (stackp >= lsp) return NULL ;
				  Variable *vp = &stack[stackp++] ;
				  vp->sym = NULL ;
				  vp->type = V_LOCATION ;
				  vp->location = valp ;
				  return vp ; }	 
    inline Variable *TmPush (Arr&  arr)
				{ if (stackp >= lsp) return NULL ;
				  Variable *vp = &stack[stackp++] ;
				  vp->sym = NULL ;
				  vp->type = V_ARRAY ;
				  vp->arr  = arr ;
				  return vp ; }	 
    inline Variable *TmPush (Value& v) { if (stackp >= lsp) return NULL ;
						     Variable *vp = &stack[stackp++] ;
						     *vp = v ;
						     vp->sym = NULL;
						     return vp ; }
    inline Variable *TmPush (Symbol *symbol) {
				  if (stackp >= lsp) return NULL ;
				  Variable *vp = &stack[stackp++] ;
				  vp->sym = NULL ;
				  vp->type = V_SYMBOL ;
				  vp->symbol = symbol ;
				  return vp ; }	
    Value *SP () { return &stack [stackp] ; } ;
    inline Variable *Top     () { return &stack[stackp-1] ; }
    inline Value    *Top2    () { return &stack[stackp-2] ; }
    inline int	    VarIndex (Variable *v) { return (v - stack) ; }
    inline int      TopIndex () { return stackp - 1 ; } 
    inline Value    *Peek (int index) { return &stack[index] ; }
    inline Variable *PeekStack (int index) { return &stack[index] ; }
    inline Value    Get (int index)  { return stack[index] ; }
    inline int	    FramePtr () { return framep ; }
    inline int	    StackPtr () { return stackp ; }
    inline int	    ArgPtr   () { return argp ; }
    inline int	    LvalPtr  () { return lsp ; }
    inline void	    SetFramePtr (int fp) { framep = fp ; } ;
    inline void	    SetStackPtr (int sp) { Pop (stackp - sp) ; }
    inline int	    AvailableSize () { return stacksize - stackp ; } ;
    inline int	    FrameSize     () { return stackp - framep ; } ;
    inline int	    ArgSize       () { return stackp - argp - 1 ; } ;
    void Shrink (int from, int numshrink) ;
    Variable *GetFrame () { return &stack[framep] ; } ;
    Variable *GetArgBase () { return &stack[argp+1] ; } ;
    inline Value *LookupVar (Symbol *sym, int hashv) {
	Variable *ve ;
	for (ve = hashTable[hashv] ; ve ; ve = ve->next)
	    if (ve->sym == sym) 
		return ve ;
	return NULL ;
    }
    inline Value *LookupVar (char *name) {
	return LookupVar (symbolTable->GetSymbol(name)) ;
    }
    inline Value *LookupVar (Symbol *sym) {
	int hashv = sym->GetHashValue() ;
	Variable *ve = hashTable[hashv] ;
	for (; ve ; ve = ve->next) 
	    if (ve->sym == sym)
		return ve ;
	return NULL ;
    }
    Value *LookupVarInFrame (Symbol *) ;
    void SetName (int index, char *name) {
	SetSymbol (index, symbolTable->GetSymbol (name)) ;
    }
    void SetNameSimple (int index, char *name) {
	Symbol *sym = symbolTable->GetSymbol(name) ;
	SetSymSimple (index, sym) ;
    }
    void SetSymbol (int index, Symbol *symbol) ;
    inline void SetSymSimple (int index, Symbol *sym) {
	int hashv ;
	Variable *vp ;
	vp = &stack[index] ;
	vp->sym = sym ;
	hashv = sym->GetHashValue() ;
	vp->next = hashTable[hashv] ;
	hashTable[hashv] = vp ;
    }
    inline void SetSymSimple (int index, Symbol *sym, int hashv) {
	Variable *vp ;
	vp = &stack[index] ;
	vp->sym = sym ;
	vp->next = hashTable[hashv] ;
	hashTable[hashv] = vp ;
    }
    inline void SetNameSimple (int index, Symbol *sym, int hashv) {
	Variable *vp ;
	vp = &stack[index] ;
	vp->SetSymbol (sym) ;
	vp->next = hashTable[hashv] ;
	hashTable[hashv] = vp ;
    }
    inline int InStack (Value *vp) { return (stack <= (Variable *)vp
					     && (Variable *)vp < (stack + stackp)) ; }
    void Rehash () ;
    inline void NotifySymbolTable (SymbolTable *symt) { symbolTable = symt; }
} ;

#endif
