//----------------------------------------------------------------
// symbol.h: Cmm Interperter Symbol Manipulation
//				(c) Ono Tetsuo
//----------------------------------------------------------------
#ifndef SYMBOL_H
#define SYMBOL_H

#include <cmm/config.h>
#include <cmm/hash.h>

#define SYMTABLE_PAGE_SIZE 8192

#define SYM_RESERVED_WORD 0x01
#define SYM_ACODE	  0x02

//----------------------------------------------------------------
// class Symbol
// NOTE: An instance of class Symbol is specialized to be only
// mapped on the area of SymbolPage's pool member.
//----------------------------------------------------------------
class Symbol {
  public:
    // Methods
    inline void    Init(char *name) { length = strlen(name) ;
    				      strcpy (symbol, name) ;
				      symbol[length] = '\0' ;
				      next = NULL ;
				      flag = 0 ;
				      value = NULL ; }
    inline void    Init(char *name, int hv)
				    { length = strlen(name) ;
    				      strcpy (symbol, name) ;
				      symbol[length] = '\0' ;
				      next = NULL ;
				      hashv = hv ;
				      flag = 0 ;
				      value = NULL ; }
    inline void SetAttr (int f, void *v)  { flag = f ; value = v ; }
    inline void SetAttr (void *v)	  { value = v ; }
    inline int  GetFlag ()		  { return flag ; }
    inline void *GetValue ()		  { return value ; }
    inline void SetNext  (Symbol *p)	  { next = p ; }
    inline int  GetHashValue () 	  { return hashv ; }
    Symbol *GetNext () 		   	  { return next ; }
    //----------------------------------------------------------------
    // Data
    //----------------------------------------------------------------
  public:
    Symbol *next ;		// for Linear Link
    int    flag ;
    void   *value ;
    int    hashv ;		// for SymbolTable
    int    length ;
    char   symbol[1] ;
    // arbitrary size for symbol
} ;

class SymbolPage {
  public:
    SymbolPage () ;
    Symbol *AllocSymbol (char *, int hashv) ;
    inline void SetNext (SymbolPage *page) { next = page; }
    inline SymbolPage *GetNext () { return next ; }
    inline int PtrInPage (char *ptr)
	{ return (pool <= ptr) && (ptr < pool + SYMTABLE_PAGE_SIZE) ; }
  private:
    SymbolPage *next ;
    int  poolSize ;
    int  poolUsed ;
    int  poolRest ;
    char pool[SYMTABLE_PAGE_SIZE] ;
} ;

class SymbolTable {
  public:
    SymbolTable () ;
    ~SymbolTable () ;
    Symbol	*hashTable[MAXHASH] ;
    inline Symbol *FindSymbol  (char *name)
			{ return FindSymbol(Hash(name), name) ; }
    Symbol *FindSymbol (int, char *) ;
    Symbol *GetSymbol (char *name) ;
    Symbol *GetSymbol (char *name, int *) ;
  public:
    void  Dump () ;
    virtual Symbol *CreateSymbol (int, char *) 	   { return NULL ; } ;
    virtual int    PtrIsSymbol   (void *) 	   { return 0    ; } ;
    inline friend int Hash (char *) ;
} ;

class DynamicSymbolTable : public SymbolTable {
  public:
    DynamicSymbolTable () : SymbolTable () {
	poolHead = poolTail = new SymbolPage () ;
    }
    ~DynamicSymbolTable () ;
    SymbolPage *poolHead, *poolTail ;
    Symbol *CreateSymbol (int, char *) ;
    int PtrIsSymbol (void *) ;
} ;

class StaticSymbolTable : public SymbolTable {
  public:
    StaticSymbolTable () {
    }
    StaticSymbolTable (char *pool, int poolsize) : SymbolTable () {
	Init (pool, poolsize) ;
    }
    ~StaticSymbolTable () {}
    void Init (char *pool, int poolsize) {
	symbolPool = pool ;
	poolSize = poolsize ;
	used = 0 ;
	rest = poolsize ;
    }
    Symbol *CreateSymbol (int, char *) ;
    int PtrIsSymbol (void *) 	;
  private:
    char *symbolPool ;
    int  poolSize ;
    int  used ;
    int  rest ;
} ;

#endif
