//
// value.h
//
#ifndef VALUE_H
#define VALUE_H

#include <stdio.h>
#include <string.h>
#include <cmm/context.h>
#include <cmm/symbol.h>

class CmmInterp ;
class Value ;
class Variable ;
class Class ;
class Instance ;

typedef int (*BuiltinFunc) (CmmInterp *, int, Variable *, Value *) ;
typedef int (*BuiltinMethod) (CmmInterp *, Value *, int, Variable *, Value *) ;

enum ValType { V_INTEGER, V_VOID, V_REAL, V_STRING,  V_ARRAY , V_CONTEXT, V_BUILTIN, V_CUSTOM, V_CLASS, V_INSTANCE, V_LOCATION, V_BIMETHOD, V_ACODE, V_BSTREAM, V_SYMBOL } ;

// Arr zǗ邽߂̃f[^\
struct Arr {
    Value *arrp ;
    int numelems ;
} ;

struct BStream {
    int size ;
    char *stream ;
} ;

//
// Value lǗ邽߂̃NXB
//    gŒľ^Ă
class Value {
  public:
    ValType type:8 ;
    char    subtype:8 ;
    union {
	double value ;
	char   *sptr ;
	int    intval ;
	Arr    arr ;
	BStream  bstream ;
	Context   context ;
	BuiltinFunc   entry ;
	BuiltinMethod bimethod ;
	void     *voidp ;
	Class    *classp ;
	Instance *instance ;
	Value *location ;
	char   *pointer ;
	Symbol *symbol ;
    } ;
    inline Value ()	    { type = V_VOID ; subtype = 0 ; }
    inline void Init  ()    { type = V_VOID ; subtype = 0 ; }
    inline Value Void ()    { type = V_VOID   ; return *this ; }
    inline void InitArr (Value *v, int size)
			     { type = V_ARRAY ; arr.numelems = size ; arr.arrp = v ; } ;
    inline void InitACode (Value *v, int size)
			     { type = V_ACODE ; arr.numelems = size ; arr.arrp = v ; } ;
    inline char *GetString () {
	switch (type) {
	  case V_STRING: return sptr ;
	  case V_SYMBOL: return symbol->symbol ;
	  default:
	    return NULL ;
	}
    }
    inline int Vtype     ()  { return type ; }
    inline int IsVoid    ()  { return type == V_VOID   ; }
    inline int IsString  ()  { return type == V_STRING ; }
    inline int IsReal    ()  { return type == V_REAL   ; }
    inline int IsArray   ()  { return type == V_ARRAY  ; }
    inline int IsACode   ()  { return type == V_ACODE  ; }
    inline int IsARRAY   ()  { return type == V_ARRAY || type == V_ACODE ; }
    inline int IsContext ()  { return type == V_CONTEXT ; }
    inline int IsBuiltin ()  { return type == V_BUILTIN ; }
    inline int IsBiMethod()  { return type == V_BIMETHOD ; }
    inline int IsCustom  ()  { return type == V_CUSTOM ; } 
    inline int IsClass   ()  { return type == V_CLASS ; }
    inline int IsInteger ()  { return type == V_INTEGER ; }
    inline int IsInstance()  { return type == V_INSTANCE ; }
    inline int IsNumber  ()  { return type == V_INTEGER || type == V_REAL ; } ;
    inline int IsLocation () { return type == V_LOCATION ; }
    inline int IsBStream  () { return type == V_BSTREAM ; }
    inline int IsSymbol	  () { return type == V_SYMBOL ; }
    inline Value *ArrElem(int n) { if (n < 0 || n >= arr.numelems)
				       return NULL ;
				   return arr.arrp + n ; }
    inline int ArrSize  ()  { if (!IsArray()) return -1 ;
			      return arr.numelems ; }
    inline int True () {
	switch (type) {
	  case V_REAL:     return (value != 0.0)? 1 : 0 ;
	  case V_STRING:   return (int)(sptr) ;
	  case V_ARRAY:    return (arr.numelems > 0) ;
	  case V_BUILTIN:  return (int)entry ;
	  case V_BIMETHOD: return (int)bimethod ;
	  case V_INTEGER:  return intval ;
	  case V_CUSTOM:   return (int)voidp ;
	  case V_VOID:	   return 0 ;
	  case V_CLASS:	   return (int)classp ;
	  case V_INSTANCE: return (int)instance ;
	  case V_CONTEXT:  return (int)(context.text) ;
	  case V_SYMBOL:   return (int)symbol ; 
	  default:	   return 0 ;
	}
    }
    inline char   *sval() { return sptr  ; }
    inline int	  ival()  { return intval ; }
    inline double dval()  { return value ; }
    inline char     *StringVal () { return sptr   ; }
    inline int      IntVal      () { return intval ; }
    inline double   RealVal     () { return value ; }
    inline Instance *InstanceVal () { return instance ; }
    inline Class    *ClassVal   () { return classp ; }
    inline void     *CustomVal  () { return voidp ; }
    inline Symbol   *SymbolVal  () { return symbol ; }
    inline Value& operator[] (int n) { return arr.arrp[n] ; }
    inline void   operator=  (Arr&   a) { type = V_ARRAY  ; arr = a   ; }
    inline double operator=  (double d) { type = V_REAL   ; value = d ; return d ; }
    inline char   *operator= (char  *s) { type = V_STRING ; sptr  = s ; return sptr ; }
    inline Context& operator= (Context& ctx) { type = V_CONTEXT ; context = ctx ;
					       return context ; }
    inline void   operator= (void *p)     { type = V_CUSTOM ; voidp = p      ; }
    inline void   operator= (BuiltinFunc func)
					  { type = V_BUILTIN ; entry = func  ; }
    inline void   operator= (BuiltinMethod meth)
    					  { type = V_BIMETHOD ; bimethod = meth ; }
    inline void   operator= (Class *_c)   { type = V_CLASS ; classp = _c     ; }
    inline void   operator= (Instance *i) { type = V_INSTANCE ; instance = i ; }
    inline void   operator= (int i)       { type = V_INTEGER ; intval = i    ; }
    inline void   operator= (Value *loc)  { type = V_LOCATION ; location = loc ; }
    inline void   operator= (Symbol *sym) { type = V_SYMBOL ; symbol = sym ; }
    inline friend int operator== (Value&, Value&) ;
    inline friend int operator!= (Value&, Value&) ;
    inline friend int operator<  (Value&, Value&) ;
    inline friend int operator>  (Value&, Value&) ;
    inline friend int operator<= (Value&, Value&) ;
    inline friend int operator>= (Value&, Value&) ;
    inline int ForceInt () {   if (type == V_INTEGER) return intval ;
			       if (type == V_REAL)    return (int)value ;
			       if (type == V_STRING) {
				   int val ;
				   if (sscanf (sptr, "%d", &val) != 1)
				       return 0 ;
				   return val ;
			       }
			       return 0 ;
			   }
    inline double ForceReal () {   if (type == V_REAL)    return value ;
				   if (type == V_INTEGER) return (double)intval ;
				   if (type == V_STRING) {
				       double val ;
				       if (sscanf (sptr, "%lf", &val) != 1)
					   return 0.0 ;
				       return val ;
				   }
				   return 0.0 ;
			       }
    char   *ForceString () ;
    void   Print (char quote = '\0') ;
    void   Print (FILE *fp, char quote = '\0') ;
    void   Write (FILE *fp) ;
  private:
    friend int CompareArray (Value&, Value&) ;
} ;

inline int operator== (Value& v1, Value& v2) {
    if (v1.IsInteger() && v2.IsReal()) return v1.intval == v2.ForceInt() ;
    if (v1.IsReal() && v2.IsInteger()) return v1.value == v2.ForceReal() ;
    if (v1.Vtype() != v2.Vtype())      return 0 ;
    if (v1.IsString())                 return (strcmp (v1.sptr, v2.sptr) == 0) ;
    if (v1.IsArray() || v1.IsACode())  return CompareArray (v1, v2) ;
    if (v1.IsReal())		       return (v1.value == v2.value) ;
    if (v1.IsInteger()) 	       return (v1.intval == v2.intval) ;
    if (v1.IsBuiltin())                return v1.entry == v2.entry ;
    if (v1.IsVoid()) 		       return 1 ;
    if (v1.IsClass())		       return v1.classp == v2.classp ;
    if (v1.IsInstance()) 	       return v1.instance == v2.instance ;
    if (v1.IsSymbol())		       return v1.symbol  == v2.symbol ;
    return 0 ;
}

inline int operator!= (Value& v1, Value& v2) { return !(v1 == v2) ; }

inline int operator> (Value& v1, Value& v2)
{
    if (v1.IsInteger() && v2.IsReal()) {
	return v1.intval > v2.ForceInt() ;
    }
    if (v1.IsReal() && v2.IsInteger()) {
	return v1.value > v2.ForceReal() ;
    }
    if (v1.Vtype() != v2.Vtype())
	return 0 ;
    if (v1.IsString()) {
	return (strcmp (v1.sptr, v2.sptr) > 0) ;
    } else if (v1.IsReal()) {
	return v1.value > v2.value;
    } else if (v1.IsInteger()) {
	return v1.intval > v2.intval ;
    }
    return 0 ;
}

inline int operator< (Value& v1, Value& v2)
{
    if (v1.IsInteger() && v2.IsReal()) {
	return v1.intval < v2.ForceInt() ;
    }
    if (v1.IsReal() && v2.IsInteger()) {
	return v1.value < v2.ForceReal() ;
    }
    if (v1.Vtype() != v2.Vtype ())
	return 0 ;
    if (v1.IsString()) {
	return (strcmp (v1.sptr, v2.sptr) < 0) ;
    } else if (v1.IsReal()) {
	return v1.value < v2.value ;
    } else if (v1.IsInteger()) {
	return v1.intval < v2.intval ;
    }
    return 0 ;
}

inline int operator<= (Value& v1, Value& v2)
{
    int i ;

    if (v2.IsArray()) {
	for (i = 0 ; i < v2.arr.numelems ; i++) {
	    Value x = v2.arr.arrp[i] ;
	    if (v1 == x)
		return i + 1 ;
	}
	return 0 ;
    }
    if (v1.IsInteger() && v2.IsReal()) {
	return v1.intval <= v2.ForceInt() ;
    }
    if (v1.IsReal() && v2.IsInteger()) {
	return v1.value <= v2.ForceReal() ;
    }
    if (v1.Vtype() != v2.Vtype())
	return 0 ;
    if (v1.IsString()) {
	return (strcmp (v1.sptr, v2.sptr) <= 0) ;
    } else if (v1.IsReal()) {
	return v1.value <= v2.value ;
    } else if (v1.IsInteger()) {
	return v1.intval <= v2.intval ;
    }
    return 0 ;
}

inline int operator>= (Value& v1, Value& v2)
{
    if (v1.IsInteger() && v2.IsReal()) {
	return v1.intval >= v2.ForceInt() ;
    }
    if (v1.IsReal() && v2.IsInteger()) {
	return v1.value >= v2.ForceReal() ;
    }
    if (v1.Vtype() != v2.Vtype())
	return 0 ;
    if (v1.IsString()) {
	return (strcmp (v1.sptr, v2.sptr) >= 0) ;
    } else if (v1.IsReal()) {
	return v1.value >= v2.value ;
    } else if (v1.IsInteger()) {
	return v1.intval >= v2.intval ;
    }
    return 0 ;
}

// Variable NX
// ϐ̖OƒlǗBValue Class ̃TuNX    
class Variable : public Value {
  public:
    Symbol *sym ;
    Variable *next ;
  public:
    inline Variable () { next = 0 ; sym = NULL ; } ;
    inline void SetSymbol (Symbol *symbol) { sym  = symbol ; }
    inline int IsMySymbol (Symbol *symbol) { return (symbol == sym) ; }
    inline void operator= (Value& v)      { Value::operator= (v) ; } ;
    inline void operator= (Class *c)	   { Value::operator= (c) ; } ;
    inline void Set (Symbol *symbol, Value& v) { Value::operator= (v) ; sym = symbol ; }
} ;

#endif
