//
// cmmc.cc: Cmm A-Code compiler and executor
//					(c) Nj
//
#include <stdlib.h>
#include <cmm/cmmcomp.h>
#include <cmm/xpack.h>
#include <cmm/iopack.h>
#include <cmm/basicio.h>
#include <cmm/cmmsock.h>
#include <cmm/syscall.h>
#include <cmm/cmmmath.h>

#include "load_path.h"

char *myname ;

void Usage ()
{
    printf ("usage: %s [options] cmm_file [args...]\n\n", myname) ;
    printf ("    Cmm vOsBcmm_file ́A\n") ;
    printf ("	a). Cmm \\[X t@C\n") ;
    printf ("	b). Cmm A-Code t@C\n") ;
    printf ("    ̂ꂩBA-Code t@ĆAReadable FormatANumeric format\n") ;
    printf ("    ǂł悢B\n\n") ;
    printf ("    IvV:\n") ;
    printf ("	-interp        C^[v^[hł̎s\n") ;
    printf ("	-acode         sRpC[hł̎s\n") ;
    printf ("	-compiler      RpCs.\n") ;
    printf ("	-stack <size>  Cmm X^bN <size> Kbytes ɐݒ\n") ;
    printf ("	-heap  <size>  Cmm q[v <size> Kbytes ɐݒ\n") ;
    printf ("	-o <file>      RpC A-Code <file> Readable format ŏo\n") ;
    printf ("	-O <file>      RpC A-Code <file> Numeric format ŏo\n") ;
    printf ("	-discompile    A-Code tRpCĕWo͂ɏo\n") ;
    printf ("	-help	       ̃bZ[Wo\n") ;
    exit (1) ;
}

extern int __DUMP ;
extern int ___DUMP ;
extern int bcOn ;
extern int _GC1 ;
extern int _AR1 ;

void main (int ac, char **av)
{
    int i ;
    char *s ;
    enum ExMode { Interp, AcInterp, Compiler } ;
    ExMode exmode = Compiler ;
    CmmInterp *cmm ;
    char *writeTo ;
    int stack_size = 64 ;
    int heap_size  = 512 ;
    char *filename ;
    int int_inst = 0 ;
    int writeAcode = 0 ;
    int discompile = 0 ;

    myname = av[0] ;

    for (i = 1 ; i < ac ; i++) {
	s = av[i] ;
	if (*s != '-')
	    break ;

	if (!strncmp (s, "-acode", strlen(s)) || !strncmp (s, "-jit", strlen(s)))
	{
	    exmode = AcInterp ;
	} else if (!strncmp (s, "-interp", strlen(s))) {
	    exmode = Interp ;
	} else if (!strncmp (s, "-compiler", strlen(s))) {
	    exmode = Compiler ;
	} else if (!strncmp (s, "-gc0", strlen(s))) {
	    _GC1 = 0 ;
	} else if (!strncmp (s, "-ar0", strlen(s))) {
	    _AR1 = 0 ;
	} else if (!strncmp (s, "-help", strlen(s))) {
	    Usage () ;
	} else if (!strncmp (s, "-o", strlen(s))
		   || !strncmp (s, "-O", strlen(s)))
	{
	    i++ ;
	    if (i >= ac) {
		fprintf (stderr,
			 "-[oO] option requires a file name as a next argument\n") ;
		exit (1) ;
	    }
	    if (s[1] == 'O')
		int_inst = 1 ;
	    writeAcode++ ;
	    if (!strcmp(av[i], "-"))
		writeTo = NULL ;
	    else
		writeTo = av[i] ;
	} else if (!strncmp (s, "-dump", strlen(s))) {
	    __DUMP=1 ;
	} else if (!strncmp (s, "-discompile", strlen(s))) {
	    discompile++ ;
	} else if (!strncmp (s, "-heap", strlen(s))) {
	    i++ ;
	    if (i >= ac) {
		fprintf (stderr,
			 "-heap option requires a heap size(Kbytes) as a next argument\n") ;
		exit (1) ;
	    }
	    heap_size = atoi(av[i]) ;
	} else if (!strncmp (s, "-stack", strlen(s))) {
	    i++ ;
	    if (i >= ac) {
		fprintf (stderr,
			 "-stack option requires a stack size(Kbytes) as a next argument\n") ;
		exit (1) ;
	    }
	    stack_size = atoi(av[i]) ;
	}
    }
    if (i >= ac) {
	Usage () ;
    }
    filename = av[i] ;
    if (exmode == Interp) {
	CmmInterp *cmmi = new CmmInterp (stack_size * 1024, heap_size * 1024) ;

	InstallIOPack  (cmmi) ;
	InstallX       (cmmi) ;
	InstallBasicIO (cmmi) ;
	InstallCmmSock (cmmi) ;
	InstallSysCall (cmmi) ;
	InstallMath    (cmmi) ;

	cmmi->Execute (PATHCONF) ;
	cmmi->CommandArg (ac - i, av + i) ;
	cmmi->ExecInterp (filename) ;
    } else {
	CmmComp *cmmc = new CmmComp (stack_size * 1024, heap_size * 1024) ;
	Value acode ;

	cmmc->Execute (PATHCONF) ;
	InstallIOPack  (cmmc) ;
	InstallX       (cmmc) ;
	InstallBasicIO (cmmc) ;
	InstallCmmSock (cmmc) ;
	InstallSysCall (cmmc) ;
	InstallMath    (cmmc) ;
	if (writeAcode || discompile) {
	    cmmc->CompileFile (filename, &acode) ;
	}
	cmmc->CommandArg (ac - i, av + i) ;
	if (writeAcode) {
	    cmmc->AcWriteACode (writeTo, &acode, int_inst) ;
	} else if (discompile) {
	    cmmc->AcDisCompile (&acode) ;
	} else if (exmode == Interp) {
	    cmmc->ExecInterp (filename) ;
	} else if (exmode == AcInterp) {
	    cmmc->ExecJIT (filename) ;
	} else if (exmode == Compiler) {
	    cmmc->ExecCompiler (filename) ;
	}
    }
}
