#include <stdio.h>
#include <ctype.h>	
#include "y.tab.h"

static FILE *inFp ;
static char fileName[300] ;
static int lineNum ;
static char nameBuffer[512] ;
static char strBuffer[1024] ;
extern int yydebug ;

int second_op = 0 ;

struct {
	char *name ;
	int val ;
} reservedWords [] = {
	{ "break",    T_Break    },
        { "catch",    T_Catch	 },
	{ "class",    T_Class    },
	{ "continue", T_Continue },
	{ "else",     T_Else     },
	{ "eval",     T_Eval     },
	{ "exit",     T_Exit	 },
	{ "finally",  T_Finally  },
	{ "for",      T_For      },
	{ "func",     T_Func     },
	{ "hidden",   T_Hiden	 },
	{ "hiden",    T_Hiden	 },
	{ "if",       T_If       },
	{ "load",     T_Load     },
	{ "macro",    T_Macro    },
	{ "method",   T_Method   },
	{ "new",      T_New      },
	{ "pause",    T_Pause	 },
	{ "private",  T_Private  },
	{ "public",   T_Public   },
	{ "return",   T_Return   },
	{ "static",   T_Static   },
    	{ "throw",    T_Throw	 },
	{ "try",      T_Try	 },
	{ "var",      T_Vardef   },
	{ "void",     T_Void     },
	{ "while",    T_While    },
} ;

/*
   In linux standard ctype macros do not work correctly.
   if a value of char is lager than or equal to 0x80.
 */
#undef isspace
#undef isdigit

#define isspace(c) (((c) == ' ')||((c) == '\t')||((c) == '\n')||((c)=='\r'))
#define isdigit(c) (((c)>='0') && ((c)<='9'))

#ifndef EUC
#define iskanji(c)         ((unsigned char)(c) >= 0x80 && (unsigned char)(c) <= 0x9f ||\
			    (unsigned char)(c) >= 0xe0 && (unsigned char)(c) <= 0xef)
#else
#define iskanji(c)	((unsigned char)c & 0x80)
#endif

#define isoctdigit(c) (((c)>='0') && ((c)<='7'))

#define ishexdigit(c) ((((c)>='0') && ((c)<='9'))\
		       ||((c)>='a' && ((c)<='f'))\
		       ||((c)>='A' && ((c)<='F')))

#define HEXVAL(c) ((('0'<=(c))&&((c)<='9')) ? ((c)-'0')\
		   :(('a'<=(c))&&((c)<='f')) ? ((c)-'a'+10)\
		   : ((c)-'A'+10))

#define ARRSIZE(arr) ((sizeof (arr))/(sizeof (arr[0])))

StrCmpNoCase (char *str1, char *str2)
{
    char c1, c2 ;

    for ( ;; ) {
	c1 = *str1++ ;
	c2 = *str2++ ;
	if (isupper(c1))
	    c1 = tolower(c1) ;
	if (isupper(c2))
	    c2 = tolower(c2) ;
	if (c1 != c2)
	    return c1 - c2 ;
	if (c1 == '\0')
	    return 0 ;
    }
}

int FindKey (char *str)
{
    int i ;
    
    for (i = 0 ; reservedWords[i].name ; i++)
    {
	if (!strcmp (str, reservedWords[i].name))
	    return reservedWords[i].val ;
    }
    return -1 ;
}

char *make_string (char *str)
{
    extern char *malloc () ;
    char *p ;
    p = malloc (strlen(str) + 1) ;
    if (!p) {
	yyerror ("܂") ;
	exit (1) ;
    }
    strcpy (p, str) ;
    return p ;
}

/*  -- */
yylex ()
{
    int c ;
    int c2 ;

  again:
    while ((c = getc (inFp)) && isspace(c)) {
	if (c == '\n')
	    lineNum++ ;
    }
    if (c == EOF) {
	return 0 ;
    }
    c2 = getc (inFp) ;
    if (c == '#' || c == '/' && c2 == '/') {
	/*
	   Rg shell  '#' ܂́Ac++  '//' 
	   # ́An߂͂Ȃ̂ǁAUNIX ŁA#! g悤ɂ邽߂ɒǉ
	 */   
	while ((c = getc(inFp)) != EOF && c != '\n')
	    ;
	if (c == '\n') {
	    lineNum++ ;
	    goto again ;
	}
    }

#define CHKADV(c,c2,iftrue,iffalse) ((c != c2) ? (ungetc (c, inFp), iffalse) : iftrue)
    switch (c) {
      case '+':
	if (c2 == '+') { return T_PP ; }
	if (c2 == '=') { return T_AddAssign ; }
	ungetc (c2, inFp) ;
	return '+' ;
      case '-':
	if (c2 == '-') { return T_MM ; }
	if (c2 == '>') { return T_PointTo ; }
	if (c2 == '=') { return T_SubAssign ; }
	ungetc (c2, inFp) ;
	return '-' ;
      case '*':
	return CHKADV (c2, '=', T_MultAssign, '*') ;
      case '/':
	return CHKADV (c2, '=', T_DivAssign,  '/') ;
      case '=':
	return CHKADV (c2, '=', T_Eq, '=') ;
      case '!':
	return CHKADV (c2, '=', T_Neq, '!') ;
      case '<':
	if (c2 == '=') { return (!second_op) ? T_Le : T_2nd_Le ; }
	if (c2 == '<') { c2 = getc (inFp) ;
			 if (c2 == '=') { return T_LshAssign ; }
			 ungetc (c2, inFp) ;
			 return T_Lsh ; }
	ungetc (c2, inFp) ;
	return (!second_op)? '<' : T_2nd_Lt ;
      case '>':
	if (c2 == '=') { return (!second_op) ? T_Ge : T_2nd_Ge ; }
	if (c2 == '>') { c2 = getc (inFp) ;
			 if (c2 == '=') { return T_RshAssign ; }
			 ungetc (c2, inFp) ;
			 return T_Rsh ; }
	ungetc (c2, inFp) ;
	return (!second_op) ? '>' : T_2nd_Gt ;
      case '^':
	if (c2 == '=') { return T_XorAssign ; }
	ungetc (c2, inFp) ;
	return '^' ;
      case '|':
	if (c2 == '|') { return T_Or ; }
	if (c2 == '=') { return T_OrAssign ; }
	ungetc (c2, inFp) ;
	return '|' ;
      case '&':
	if (c2 == '&') { return T_And ; }
	if (c2 == '=') { return T_AndAssign ; }
	ungetc (c2, inFp) ;
	return '&' ;
      case '$':
	if      (c2 == '#') { return T_ArgCnt  ; }
	else if (c2 == '*') { return T_ArgList ; }
	ungetc (c2, inFp) ;
	return  '$' ; 
      case '%':
	return CHKADV (c2, '=', T_ModAssign, '%') ;
      case ':':
	if (c2 == ':') { return T_ColonColon ; }
	if (c2 == '=') { return T_RefAssign ; }
	ungetc (c2, inFp) ;
	return ':' ;
      case '{':
	if (c2 == '{') { return T_LLbr ; }
	ungetc (c2, inFp) ;
	return '{' ;
      case '}':
	if (c2 == '}') { return T_RRbr ; }
	ungetc (c2, inFp) ;
	return '}' ;
      case '(':
      case ')':
      case '?':
      case '[':
      case ']':
      case ',':
      case '@':
      case '`':
      case '~':
	ungetc (c2, inFp) ;
	return c ;
      case '"':
      case '\'':
	{
	    /*  */
	    int delim = c ;
	    char *heapp ;
	    char *s0 ;
	    int  len ;
	    int i = 0 ;
	    
	    c = c2 ;
	    for (len = 0 ; c != delim ; c = getc(inFp)) {
		if (!c || c == '\n') {
		    if (c == '\n')
			lineNum++ ;
		}
		if (iskanji (c)) {
		    strBuffer[i++] = c ;
		    strBuffer[i++] = getc (inFp) ;
		} else if (c == '\\') {
		    c = getc(inFp) ;
		    switch (c) {
		      case 'n':
			strBuffer[i++] = '\n' ; break ;
		      case 'r':
			strBuffer[i++] = '\r' ; break ;
		      case 'b':
			strBuffer[i++] = '\b' ; break ;
		      case 'f':
			strBuffer[i++] = '\f' ; break ;
		      case 't':
			strBuffer[i++] = '\t' ; break ;
		      default:
			strBuffer[i++] = c    ; break ;
		    }
		} else {
		    strBuffer[i++] = c ;
		}
	    }
	    strBuffer[i] = '\0' ;
	    if (yydebug)
		printf ("STRING %s\n", strBuffer) ;
	    yylval.sval = strBuffer ;
#if 0
	    if (!(heapp = heap_area.AllocSegment (len+1, &varstack))) {
		SetErrorCode (NOT_ENOUGH_MEMORY) ;
		return T_Err ;
	    }
	    for (s0 = heapp, prog->Set(strstart) ;  (c = prog->GetC()) != delim ; ) {
		if (iskanji (c)) {
		    *s0++ = c ;
		    *s0++ = prog->GetC() ;
		} else if (c == '\\') {
		    c = prog->GetC() ;
		    switch (c) {
		      case 'n':
			*s0++ = '\n' ; break ;
		      case 'r':
			*s0++ = '\r' ; break ;
		      case 'b':
			*s0++ = '\b' ; break ;
		      case 'f':
			*s0++ = '\f' ; break ;
		      case 't':
			*s0++ = '\t' ; break ;
		      default:
			*s0++ = c    ; break ;
		    }
		} else {
		    *s0++ = c ;
		}
	    }
	    *s0 =  '\0' ;
	    lp->sval = heapp ;
#endif
	    return T_String ;
	}
      case '0': case '1':
      case '2': case '3':
      case '4': case '5':
      case '6': case '7':
      case '8': case '9':
    	{
	    /* 萔 */
	    int ival = 0 ;
	    double dval ;
	    double base = 0.1 ;
	    
	    if (c == '0' && (c2 == 'x' || c2 == 'X')) {
		c = getc (inFp) ;
		for (; c && ishexdigit(c) ; c = getc(inFp)) {
		    ival = 16 * ival + HEXVAL(c) ;
		}
		ungetc (c, inFp) ;
		yylval.ival = ival ;	     
		return T_IntNumber ;
	    } else if (c == '0' && isoctdigit(c2)) {
		c = c2 ;
		for (; c && ishexdigit(c) ; c = getc(inFp)) {
		    ival = 8 * ival + c - '0' ;
		}
		ungetc (c, inFp) ;
		yylval.ival = ival ;	     
		return T_IntNumber ;
	    } else {
		ungetc (c2, inFp) ;
		for ( ; c && isdigit (c) ; c = getc(inFp)) {
		    ival = 10 * ival + c - '0' ;
		}
		if (c != '.') {
		    yylval.ival = ival ;
		    ungetc (c, inFp) ;
		    return T_IntNumber ;
		}
		dval = (double)ival ;
		for (c = getc (inFp) ;
		     c && isdigit(c) ;
		     c = getc (inFp), base *= 0.1)
		{
		    dval += base * (c - '0') ;
		}
		ungetc (c, inFp) ;
		yylval.value = dval ;
		return T_Number ;
	    }
	}
      case '.':
	if (!isdigit(c2)) {
	    ungetc (c2, inFp) ;
	    return T_PointTo ;
	} else {
	    double dval = 0.0 ;
	    double base = 0.1 ;
	    for (c = c2 ; c && isdigit(c) ; c = getc(inFp), base *= 0.1) {
		dval += base * (c - '0') ;
	    }
	    ungetc (c, inFp) ;
	    yylval.value = dval ;
	    return T_Number ;
	}
      case ';':
	ungetc (c2, inFp) ;
	return c ;
      default:
	if (isalpha(c) || c == '_' || iskanji(c)) {
	    /* O */
	    int i ;
	    char *s, *s0 ;
	    int l, r, m ;
	    int cmp ;
	    char *name ;

	    ungetc (c2, inFp) ;
	    for (i = 0 ; isalnum(c) || c == '_' || iskanji(c) ; c = getc(inFp)) {
		if (iskanji (c)) {
		    nameBuffer[i++] = c ;
		    nameBuffer[i++] = getc (inFp) ;
		} else {
		    nameBuffer[i++] = c ;
		}
	    }
	    nameBuffer[i] = '\0' ;
	    if (yydebug)
		fprintf (stderr, "### nameBuffer ## %s\n", nameBuffer) ;

	    ungetc (c, inFp) ;
	    name = nameBuffer ;
	    
	    l = 0 ;
	    r = ARRSIZE(reservedWords) - 1 ;
	    
	    while (l <= r) {
		m = (l + r)/2 ;
		for (s = name, s0 = reservedWords[m].name ; ; s++, s0++) {
		    if ((cmp = (*s - *s0)) != 0)
			break ;
		    if (*s == '\0')
			return reservedWords[m].val ;
		}
		if (cmp < 0) {
		    r = m - 1 ;
		} else {
		    l = m + 1 ;
		}
	    }
	    yylval.sval = nameBuffer ;
	    return T_Word ;
	}
	return T_Err ;
    }
    /*NOT REACHED*/
}

#if 0
yyerror (mes)
	char *mes ;
{
	fprintf (stderr, "%s: %s (%dsڕt)\n", fileName, mes, lineNum) ;
	return 0 ;
}
#endif

int LexOpen (char *filename)
{
	if (!filename) {
		inFp = stdin ;
		strcpy (fileName, "<stdin>") ;
		return 1 ;
	}
	if (!(inFp = fopen (filename, "r"))) {
		return 0 ;
	}
	if (strlen (filename) < 300)
		strcpy (fileName, filename) ;
	else
		strcpy (fileName, "<DEF_FILE>") ;
	lineNum = 1 ;
	return 1 ;
}

int LexClose ()
{
	if (inFp != stdin)
		fclose (inFp) ;
}
	
yyerror (mes)
	char *mes ;
{
	fprintf (stderr, "%s: %s (%dsڕt)\n", fileName, mes, lineNum) ;
	return 0 ;
}


