/*
 * cmmp: Cmm parser
 */
%{
#include <stdio.h>
%}

%union {
    int ival ;
    double value ;
    char *sval ;
} ;

%token T_Number
%token T_IntNumber
%token T_While
%token T_Func
%token T_Return
%token T_Break
%token T_Continue
%token T_Word
%token T_For
%token T_ArgCnt
%token T_ArgList
%token T_ArgPrefix
%token T_Vardef
%token T_Load
%token T_Void
%token T_Class
%token T_New
%token T_Method
%token T_Hiden
%token T_Private
%token T_Public
%token T_Static
%token T_String
%token T_Eval
%token T_Pause
%token T_Exit
%token T_This
%token T_Macro
%token T_Err
%token T_LLbr
%token T_RRbr
%token T_Try
%token T_Catch
%token T_Finally
%token T_Throw

%token T_If
%token T_Else

%left   T_ColonColon T_PointTo
%right  '!'    T_PP  T_MM   '~'
%left   T_Mult T_Div T_Mod  T_Band
%left   T_Lsh  T_Rsh T_Plus T_Minus  T_Bor T_Xor
%left   T_Eq   T_Neq T_Lt   T_Gt     T_Le  T_Ge
%left   T_And
%left   T_Or
%right  '?' ':'

%right T_Assign    T_AddAssign  T_SubAssign T_LshAssign
%right T_RshAssign T_MultAssign T_DivAssign T_ModAssign
%right T_RefAssign T_XorAssign  T_AndAssign T_OrAssign

%start cmm_program

%%

cmm_program:
	  cmm_statement
	| cmm_program cmm_statement
	;

cmm_statement:
	  assignment_expression
	| assignment_expression ';'
	| ';'
	;

decl_expression:
	  vardef
	| funcdef
	| classdef
	| methodef
	;

/* VARDEF */
vardef:
	  T_Vardef var_decl_list
	;

var_decl_list:
	  var_decl
	| var_decl_list ',' var_decl
	;

var_decl:
	  T_Word
	| T_Word '='         assignment_expression
	| T_Word T_RefAssign assignment_expression
	;

/* CLASSDEF */
classdef:
	  T_Class T_Word superclass_specifier '{' member_decl_list '}'
	;

superclass_specifier:

	| ':' T_Word
	;

member_decl_list:
	  member_decl
	| member_decl_list member_decl 
	;

member_decl:
	  T_Public  member_name_list_optional_delim
	| T_Private member_name_list_optional_delim
	| T_Static  member_name_list_optional_delim
	| T_Method  method_definition_list_optional_delim
	| T_Hiden   method_definition_list_optional_delim
	;

member_name_list_optional_delim:
	  member_name_list
	| member_name_list ';'
	;

member_name_list:
	  T_Word
	| member_name_list ',' T_Word
	;

method_definition_list_optional_delim:
	  method_definition_list
	| method_definition_list ';'
	;

method_definition_list:
	  method_name_or_method_name_and_body
	| method_definition_list ',' method_name_or_method_name_and_body
	;

method_name_or_method_name_and_body:
	  T_Word 
	| T_Word '(' arg_name_list_or_null ')' assignment_expression
	;

/* FUNCDEF */
funcdef:
	  T_Func T_Word '(' arg_name_list_or_null ')' cmm_statement
	;

/* METHODEF */
methodef:
	  T_Method T_Word T_ColonColon T_Word '(' arg_name_list_or_null ')'
		assignment_expression
	;

arg_name_list_or_null:

	| arg_name_list
	;

arg_name_list:
	  T_Word
	| arg_name_list ',' T_Word
	;

control_expression:
	  if_expression
	| while_expression
	| for_expression
	| block_expression
	| no_scope_block_expression
	| T_Break ';'
	| T_Break assignment_expression
	| T_Continue
	| T_Return ';'
	| T_Return assignment_expression
	;

block_expression:
	'{' expression_list_or_null '}'
	;

no_scope_block_expression:
	T_LLbr expression_list_or_null T_RRbr
	;

expression_list_or_null:

	| expression_list
	;

expression_list :
	  cmm_statement
	| expression_list cmm_statement
	;

if_expression:
	  T_If '(' assignment_expression ')' cmm_statement
	| T_If '(' assignment_expression ')' cmm_statement T_Else cmm_statement
    	;

while_expression:
	  T_While '(' assignment_expression ')' cmm_statement
	;

for_expression:
	  T_For '(' cmm_expression_list_or_null ';'
		    cmm_expression_or_null ';'
		    cmm_expression_list_or_null ')'
		cmm_statement
	;

cmm_expression_list_or_null:

	| cmm_expression_list
	;

cmm_expression_list:
	  assignment_expression
	| cmm_expression_list ',' assignment_expression
	;

cmm_expression_or_null:

	| assignment_expression
	;

assignment_expression:
	  conditional_expression
	| locator assignment_operator assignment_expression
	;

assignment_operator:
	  '='
	| T_AddAssign
	| T_SubAssign
	| T_MultAssign
	| T_DivAssign
	| T_ModAssign
	| T_LshAssign
	| T_RshAssign
	| T_RefAssign
	| T_AndAssign
	| T_OrAssign	
	| T_XorAssign
	;

conditional_expression:
	  or_expression
	| or_expression '?' or_expression ':' conditional_expression
	;

or_expression:
	  and_expression
	| and_expression  T_Or or_expression 
	;

and_expression:
	  compare_expression
	| compare_expression T_And and_expression 
	;

compare_expression:
	  add_sub_expression
	| add_sub_expression compare_operator compare_expression 
	;

compare_operator:
	  T_Eq
	| T_Neq
	| '<'
	| '>'
	| T_Le
	| T_Ge
	;

add_sub_expression:
	  mult_div_expression
	| mult_div_expression '+'   add_sub_expression
	| mult_div_expression '-'   add_sub_expression
	| mult_div_expression T_Lsh add_sub_expression
	| mult_div_expression T_Rsh add_sub_expression
	| mult_div_expression '|'   add_sub_expression
	| mult_div_expression '^'   add_sub_expression
	;

mult_div_expression:
	  rvalue
	| mult_div_expression '*' rvalue 
	| mult_div_expression '/' rvalue 
	| mult_div_expression '%' rvalue 
	| mult_div_expression '&' rvalue
	;

rvalue:
	  cmm_special_expression
	| unary_expression
	;

cmm_special_expression:
	  T_Load '(' assignment_expression ')'
	| T_Eval '(' assignment_expression ')'
	| T_New	   assignment_expression
	| T_Pause
	| T_Exit 
	;

unary_expression:
	  const_var_locator
	| pre_inc_dec_unary
	| prefix_unary
	| decl_expression
 	| control_expression
	;

pre_inc_dec_unary:
	  T_PP locator
	| T_MM locator
	;

prefix_unary:
	  '!' const_var_locator
	| '-' const_var_locator
	| '+' const_var_locator
	| '~' const_var_locator
	;

const_var_locator:
	  locator
	| constant_locator
	;

locator:
	  var_locator
	| var_locator selector_list
	;

var_locator:
	  '(' assignment_expression ')'
	| T_Word
	| T_ArgCnt
	| T_ArgList
	| '$' '[' assignment_expression ']'
	| '$' T_Word
	| member_selector
	| class_selector
	| T_This
	;

selector_list:
	  selector
	| selector_list selector
	| T_PP
	| T_MM
	;

selector:
	  '[' assignment_expression ']'
	| member_selector
	| class_selector
	| '(' arg_list_or_null ')'
	;

member_selector:
	  T_PointTo T_Word
	| T_PointTo '(' assignment_expression ')'
	;

class_selector:
	  T_ColonColon T_Word
	;

constant_locator:
	  constant
	| constant constant_selector_list
	;

constant:
	  T_IntNumber
	| T_Number
	| T_Void
	| '[' element_list ']'
	| T_String
	| '`' T_Word
	| '@' T_Word
	;

constant_selector_list:
	  constant_selector
	| constant_selector_list constant_selector
	;

constant_selector:
	  T_PointTo T_Word
	| T_ColonColon T_Word
	| '(' arg_list_or_null ')' 
	;

element_list:
	  cmm_expression_or_null
	| element_list ',' cmm_expression_or_null
	;

arg_list_or_null:

	| arg_list
	;

arg_list:
	  assignment_expression
	| assignment_expression ',' arg_list
	;

%%	
main (int argc, char **argv)
{
    char *filename = NULL ;
    int tok ;
    int i ;
    

    for (i = 1 ; i < argc ; i++) {
	if (argv[i][0] != '-') {
	    break ;
	}
	if (!strncmp (argv[i], "-debug", strlen(argv[i]))) {
	    yydebug++ ;
	}
    }
    if (i >= argc) {
	fprintf (stderr, "Usage : %s -debug cmm_source_file\n",
		 argv[0]) ;
	exit (1) ;
    }
    filename = argv[i] ;
    if (!LexOpen (filename)) {
	fprintf (stderr, "cannot open file %s\n", filename) ;
	exit (1) ;
    }
    yyparse () ;
    LexClose () ;
}
