//----------------------------------------------------------------//
// basicio.cc: Cmm Builtin Function Library			  //
//	UNIX system call interface.				  //
//		open, creat, close				  //
//		read, write,					  //
//----------------------------------------------------------------//
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <cmm/cmm.h>

//----------------------------------------------------------------
// read (filedesc, maxlength)
//	description:
//		Read string from the specified file descriptor up
//		to maxlength. Currently it can read only string.
//	parameters:
//		filedesc ... file descriptor for read (integer).
//		maxlength .. max length of read
//	returns:
//		string ..... when read string.
//		void  ...... read nothging.
//----------------------------------------------------------------
static int __read (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 2)
	return 0 ;
    if (!arg[0].IsInteger() || !arg[1].IsInteger())
	return 0 ;
    char *buffer ;
    int nbytes ;
    buffer = cenv->heap_area.AllocSegment (arg[1].intval + 1, &cenv->varstack) ;
    if (!buffer) {
	fprintf (stderr, "read: cannot allocate memory\n") ;
	return -1 ;
    }
    nbytes = read (arg[0].intval, buffer, arg[1].intval) ;
    if (nbytes <= 0) {
	return 0 ;
    }
    *retval = buffer ;
    return 0 ;
}

//----------------------------------------------------------------
// write (filedesc, string, strlength)
//	description:
//		write string to the file specified file
//		descriptor
//	parameters:
//		filedesc ... file descriptor (integer)
//		string  .... the string to write (string)
//		strlength .. the length of the string (integer)
//	returns:
//		[success] number of bytes written
//		[failure] -1
//----------------------------------------------------------------
static int __write (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 3)
	return 0 ;
    if (!arg[0].IsInteger () || !arg[1].IsString() || !arg[2].IsInteger())
	return 0 ;
    int nbytes ;
    nbytes = write (arg[0].intval, arg[1].sptr, arg[2].intval) ;
    *retval = nbytes ;
    return 0 ;
}

//----------------------------------------------------------------
// open (pathname, flags)
// open (pathname, flags, mode)
//	description:
//		attempts to open a file and return a file descriptor
//	parameters:
//		pathname .... the pathname of the filename to open
//		flags ....... flags to determine open mode
//		mode ........ permissions
//----------------------------------------------------------------
static int __open (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 2 && args != 3)
	return 0 ;
    if (!arg[0].IsString() || !arg[1].IsInteger())
	return 0 ;
    if (args == 3) {
	if (!arg[2].IsInteger())
	    return 0 ;
    }
    int fd ;
    if (args == 2)
	fd = open (arg[0].sptr, arg[1].intval) ;
    else
	fd = open (arg[0].sptr, arg[1].intval, arg[2].intval) ;
    *retval = fd ;
    return 0 ;
}

static int __creat (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 2)
	return 0 ;
    if (!arg[0].IsString() || !arg[1].IsInteger())
	return 0 ;
    int fd ;
    fd = creat (arg[0].sptr, arg[1].intval) ;
    *retval = fd ;
    return 0 ;
}

static int __close (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 1)
	return 0 ;
    if (!arg[0].IsInteger())
	return 0 ;
    close (arg[0].intval) ;
    return 0 ;
}

static int __unlink (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 1)
	return 0 ;
    if (!arg[0].IsString())
	return 0 ;
    unlink (arg[0].sptr) ;
    *retval = 0 ;
    return 0 ;
}

static int __pipe (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    int fdesc[2] ;
    Value r ;
    int st ;

    retval->Init() ;
    st = pipe (fdesc) ;
    if (st < 0)
	return 0 ;
    r = cenv->AllocArray(2) ;
    r.arr.arrp[0].intval = fdesc[0] ;
    r.arr.arrp[1].intval = fdesc[1] ;
    *retval = r ;
    return 0 ;
}

void InstallBasicIO (CmmInterp *cmm)
{
    cmm->AddGlobalVar ("O_CREAT",	O_CREAT) ;
    cmm->AddGlobalVar ("O_RDONLY",	O_RDONLY) ;
    cmm->AddGlobalVar ("O_WRONLY",	O_WRONLY) ;
    cmm->AddGlobalVar ("O_RDWR",	O_RDWR) ;
    cmm->AddGlobalVar ("O_EXCL",	O_EXCL) ;
    cmm->AddGlobalVar ("O_NOCTTY",	O_NOCTTY) ;
    cmm->AddGlobalVar ("O_TRUNC",	O_TRUNC) ;
    cmm->AddGlobalVar ("O_APPEND",	O_APPEND) ;
    cmm->AddGlobalVar ("O_NDELAY",	O_NDELAY) ;
#ifdef O_SYNC
    cmm->AddGlobalVar ("O_SYNC",	O_SYNC) ;
#endif
    cmm->AddBuiltin   ("creat",		__creat) ;
    cmm->AddBuiltin   ("open",		__open) ;
    cmm->AddBuiltin   ("read",		__read) ;
    cmm->AddBuiltin   ("write",		__write) ;
    cmm->AddBuiltin   ("close",		__close) ;
    cmm->AddBuiltin   ("unlink",	__unlink) ;
}
