//----------------------------------------------------------------
// class.cc:: Cmm class support 
//----------------------------------------------------------------
#include <cmm/class.h>

int Class::FindMyEntry (Symbol *name_sym, VarType *vtype)
{
    int i ;

    for (i = 0 ; i < num_members ; i++) {
	if (class_template[i].name_sym == name_sym) {
	    *vtype = class_template[i].iv_type ;
	    return i ;
	}
    }
    return -1 ;
}

Class *Class::FindClass (Symbol *name_sym, Value **classvp)
{
    Class *cp ;

    for (cp = this ; cp ; cp = cp->superclass) {
	if (cp->classnamep->sym == name_sym) {
	    *classvp = cp->classnamep ;
	    return cp ;
	}
    }
    return NULL ;
}

Class *Instance::FindClass (Symbol *classname_sym, Value **classvp)
{
    Class *cp ;

    for (cp = classp ; cp ; cp = cp->superclass) {
	if (cp->classnamep->sym == classname_sym) {
	    *classvp = cp->classnamep ;
	    return cp ;
	}
    }
    return NULL ;
}

Value *Class::FindMethodOrStaticOfClass (Symbol *name_sym, VarType *vtype)
{
    int index ;

    index = FindMyEntry (name_sym, vtype) ;
    if (index >= 0) {
	if (*vtype != VT_ClassVar
	    && *vtype != VT_Method
	    && *vtype != VT_BiMethod
	    && *vtype != VT_Hiden
	    && *vtype != VT_BiHiden)
	{
	    return NULL ;
	}
	return &class_vars_or_methods [class_template[index].index] ;
    }
    return NULL ;
}

Value *Class::FindMethodOrStatic (Symbol *name_sym, VarType *vtype)
{
    int index ;
    Class *cp ;

    for (cp = this ; cp ; cp = cp->superclass) {
	index = cp->FindMyEntry (name_sym, vtype) ;
	if (index >= 0) {
	    if (*vtype != VT_ClassVar
		&& *vtype != VT_Method
		&& *vtype != VT_BiMethod
		&& *vtype != VT_Hiden
		&& *vtype != VT_BiHiden)
	    {
		return NULL ;
	    }
	    return &cp->class_vars_or_methods [cp->class_template[index].index] ;
	}
    }
    return NULL ;
}

Value *Class::FindMethod (Symbol *name_sym, VarType *vtype)
{
    Class *cp ;
    Value *vp ;
    
    for (cp = this ; cp ; cp = cp->superclass) {
	if ((vp = cp->FindMethodOrStatic (name_sym, vtype))
	    && (*vtype == VT_Method || *vtype == VT_Hiden))
	{
	    return vp ;
	}
    }
    return NULL ;
}

Class *Instance::FindClass (Class *baseClass, Symbol *name_sym, Value **classvp)
{
    Class *cp ;
    cp = classp ;
    if (baseClass) {
	for (; cp && cp != baseClass ; cp = cp->superclass)
	    ;
	if (!cp)
	    return NULL ;
    }
    for ( ; cp ; cp = cp->superclass) {
	if (cp->classnamep->sym == name_sym) {
	    *classvp = cp->classnamep ;
	    return cp ;
	}
    }
    return NULL ;
}

Value *Instance::SelectClass (Class *thisclass)
{
    Class *cp ;

    for (cp = classp ; cp && cp != thisclass ; cp = cp->superclass) 
	;
    if (!cp)
	return classp->classnamep ;
    return cp->classnamep ;
}

void Class::SetMethod (int slot, Context& ctx)
{
    if (slot < 0 || slot >= num_members)
	return ;
    int idx = class_template[slot].index ;
    class_vars_or_methods[idx] = ctx ;
    class_vars_or_methods[idx].subtype = CTX_METHOD ;
}

void Class::SetCreator (int slot, Context& ctx)
{
    if (slot < 0 || slot >= num_members)
	return ;
    int idx = class_template[slot].index ;
    class_vars_or_methods[idx] = ctx ;
    class_vars_or_methods[idx].subtype = CTX_CREATOR ;
}

void Class::SetAcMethod (int slot, Arr& arr)
{
    if (slot < 0 || slot >= num_members)
	return ;
    int idx = class_template[slot].index ;
    class_vars_or_methods[idx] = arr ;
}

int Class::SetBuiltinMethod (int slot, BuiltinMethod meth)
{
    if (slot < 0 || slot >= num_members)
	return -1 ;
    int idx = class_template[slot].index ;
    class_vars_or_methods[idx] = meth ;
    return 0 ;
}

int Class::SetStaticMember (int slot, Value& v)
{
    if (slot < 0 || slot > num_members)
	return -1 ;
    int idx = class_template[slot].index ;
    class_vars_or_methods[idx] = v ;
    return 0 ;
}

void Class::DumpClass ()
{
    printf ("<%lx super> member::", superclass) ;
    int i ;
    
    for (i = 0 ; i < num_members ; i++) {
	printf ("<%lx>%s ", class_template[i].name_sym, class_template[i].name_sym->symbol) ;
    }
    printf ("// INSTVAR (offset = %d num_instvras = %d)",
	    offset, num_instvars) ;
    printf ("// CV MTH (%d) ", num_cvs_methods) ;
    for (i = 0 ; i < num_cvs_methods ; i++) {
	class_vars_or_methods[i].Print() ;
    }
    printf ("\n") ;
}

void Class::ShowSpec ()
{
    printf ("Class : %s\n", classnamep->sym->symbol) ;
    printf ("    superclass = %s\n", superclass ? superclass->classnamep->sym->symbol : "None") ;
    printf ("    defined members = %d\n", num_members) ;
    printf ("	 class offset = %d,  number of private/public vars = %d\n",
	    offset, num_instvars) ;
    int i ;
    ClassMemberTemplate *cp ;
    
    for (i = 0 ; i < num_members ; i++) {
	cp = &class_template[i] ;
	printf ("	%s:  type (%s)  index (%d)\n",
		cp->name_sym->symbol, 
		((cp->iv_type == VT_Method) ? "method"
		 : (cp->iv_type == VT_Hiden) ? "hiden"
		 : (cp->iv_type == VT_Public) ? "public"
		 : (cp->iv_type == VT_Private) ? "private"
		 : (cp->iv_type == VT_ClassVar) ? "static"
		 : (cp->iv_type == VT_BiMethod) ? "BImethod"
		 : "???"),
		cp->index) ;
    }
}

void Instance::DumpInstance ()
{
    int i ;
    printf ("classp <%lx> ", classp) ;
    for (i = 0 ; i < num_slots ; i++)
	instance_values[i].Print () ;
}

Value *Instance::FindMember (Class *baseclass, Symbol *name_sym, VarType *mtype)
{
    int num_members ; 
    Class::ClassMemberTemplate *tpend ;
    Class::ClassMemberTemplate *tpp ;
    Class *cp ;
    char *np0, *np1 ;

    if (!baseclass)
	baseclass = classp ;

    for (cp = baseclass ; cp ; cp = cp->superclass) {
	int l = 0  ;
	int r = cp->num_members - 1 ;
	int m ;
	Symbol *sym ;

	tpp = cp->class_template ;
	while (r >= l) {
	    m = (l + r) >> 1 ;
	    sym = tpp[m].name_sym ;
	    if (name_sym < sym) {
		r = m - 1 ;
	    } else if (name_sym > sym) {
		l = m + 1 ;
	    } else {
		tpp += m ;
		*mtype = tpp->iv_type ;
		if (tpp->iv_type == VT_ClassVar
		    || tpp->iv_type == VT_Method
		    || tpp->iv_type == VT_Hiden)
		{
		    return &cp->class_vars_or_methods [tpp->index] ;
		} else {
		    return &instance_values [cp->offset + tpp->index] ;
		}
	    }
	}
#if 0
	tpp = cp->class_template ;
	tpend = tpp + cp->num_members ;
	for ( ; tpp < tpend ; tpp++) {
	    if (tpp->name_sym == name_sym) {
		*mtype = tpp->iv_type ;
		if (tpp->iv_type == VT_ClassVar
		    || tpp->iv_type == VT_Method
		    || tpp->iv_type == VT_Hiden)
		{
		    return &cp->class_vars_or_methods[tpp->index] ;
		} else {
		    return &instance_values[cp->offset + tpp->index] ;
		}
	    }
	}
#endif
    }
    return NULL ;
}

Value *Instance::FindMember (Symbol *classname_sym, Symbol *name_sym)
{
    Class *cp ;
    Value *classvp ;
    VarType vtype ;

    cp = FindClass (classname_sym, &classvp) ;
    if (!cp)
	return NULL ;
    return FindMember (cp, name_sym, &vtype) ;
}

Value *Instance::FindMember (Class *baseclass, Symbol *name_sym,
			     Value **cpp, VarType *mtype, int *index)
{
    int num_members ;
    Class::ClassMemberTemplate *tpend ;
    Class::ClassMemberTemplate *tpp ;
    Class *cp ;
    char *np0, *np1 ;
    Symbol *sym ;

    if (!baseclass)
	baseclass = classp ;
    for (cp = baseclass ; cp ; cp = cp->superclass) {
	int l = 0  ;
	int r = cp->num_members - 1 ;
	int m ;

	tpp = cp->class_template ;
	while (r >= l) {
	    m = (l + r) >> 1 ;
	    sym = tpp[m].name_sym ;
	    if (name_sym < sym) {
		r = m - 1 ;
	    } else if (name_sym > sym) {
		l = m + 1 ;
	    } else {
		tpp += m ;
		*mtype = tpp->iv_type ;
		*cpp = cp->classnamep ;
		if (tpp->iv_type == VT_ClassVar
		    || tpp->iv_type == VT_Method
		    || tpp->iv_type == VT_Hiden)
		{
		    *index = tpp->index ;
		    return &cp->class_vars_or_methods [tpp->index] ;
		} else {
		    *index = cp->offset + tpp->index ;
		    return &instance_values [cp->offset + tpp->index] ;
		}
	    }
	}
#if 0
	tpp = cp->class_template ;
	tpend = tpp + cp->num_members ;
	for ( ; tpp < tpend ; tpp++) {
	    if (tpp->name_sym == name_sym) {
		*mtype = tpp->iv_type ;
		*cpp = cp->classnamep ;
		if (tpp->iv_type == VT_ClassVar
		    || tpp->iv_type == VT_Method
		    || tpp->iv_type == VT_Hiden)
		{
		    *index = tpp->index ;
		    return &cp->class_vars_or_methods [tpp->index] ;
		} else {
		    *index = cp->offset + tpp->index ;
		    return &instance_values [cp->offset + tpp->index] ;
		}
	    }
	}
#endif	
    }
    return NULL ;
}

#if 0
Value *Instance::FindMemberByIndex (Value *classp, VarType vtype, int index)
{
    Class *cp ;

    for (cp = this->classp ; cp ; cp = cp->superclass) {
	if (classp->classp == cp) {
	    break ;
	}
    }
    if (!cp) {
	printf ("INVALID class ineritance !!\n") ; 
    }
    if (vtype == VT_Private || vtype == VT_Public)
	return &instance_values [index] ;
    else
	return &classp->classp->class_vars_or_methods [index] ;
}
#endif

void Instance::ShowValue ()
{
    ShowValue (classp) ;
}

void Instance::ShowValue (Class *classp)
{
    int i ;

    if (classp->superclass) {
	ShowValue (classp->superclass) ;
    }
    printf ("from class %s\n", classp->classnamep->sym->symbol) ;
    printf ("    method/static\n") ;
    for (i = 0 ; i < classp->num_members ; i++) {
	Class::ClassMemberTemplate *cp = &classp->class_template[i] ;
	if (cp->iv_type == VT_Method) {
	    printf ("	%s: (method)\n", cp->name_sym->symbol) ;
	} else if (cp->iv_type == VT_Hiden) {
	    printf ("	%s: (hiden)\n", cp->name_sym->symbol) ;
	} else if (cp->iv_type == VT_BiMethod) {
	    printf ("   %s: (Builtin Method)\n", cp->name_sym->symbol) ;
	} else if (cp->iv_type == VT_ClassVar) {
	    printf ("	%s: (static) ", cp->name_sym->symbol) ;
	    classp->class_vars_or_methods[cp->index].Print() ;
	    printf ("\n") ;
	}
    }
    printf ("    private/public\n") ;
    for (i = 0 ; i < classp->num_members ; i++) {
	Class::ClassMemberTemplate *cp = &classp->class_template[i] ;
	if (cp->iv_type == VT_Private || cp->iv_type == VT_Public) {
	    printf ("	%s: ", cp->name_sym->symbol) ;
	    printf ("%s ", (cp->iv_type == VT_Private) ? "private" : "public") ;
	    instance_values[classp->offset + cp->index].Print() ;
	    printf ("\n") ;
	}
    }
}

/*
 * Local variables:
 * c-indent-level: 4
 * c-continued-statement-offset: 4
 * c-brace-offset: -4
 * c-argdecl-indent: 2
 * c-label-offset:  -2
 * c-auto-newline:   nil
 * end:
 */
