//----------------------------------------------------------------
// cmmcomp.cc: Cmm A-Code Compiler and Executor
//					(c) Nj
//----------------------------------------------------------------
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef linux
#include <sys/file.h>
#endif
#include <stdlib.h>

#ifdef MSVC_2
#include <fcntl.h>
#include <io.h>
#endif

#define CMM_INTERNAL
#include <cmm/class.h>
#include <cmm/cmmcomp.h>

#define ECTX(form) ((form)->arr.arrp[1].context)

#ifndef MSVC_2
extern "C" {
    int read (int, char *, int) ;
    int close (int) ;
} ;
#endif

//---BEGIN NAMES---
char *CmmComp::instNames[] = {
	"Nop",
	"If",
	"While",
	"For",
	"FuncDef",
	"MethoDef",
	"HidenDef",
	"Funcall",
	"Var",
	"SelectMember",
	"SelectClass",
	"Const",
	"Not",
	"Negate",
	"Block",
	"SLBlock",
	"Plus",
	"Minus",
	"Mult",
	"Div",
	"Mod",
	"Rsh",
	"Lsh",
	"Eq",
	"Neq",
	"New",
	"Lt",
	"Gt",
	"Le",
	"Ge",
	"Or",
	"And",
	"Bor",
	"Band",
	"Xor",
	"BitRev",
	"Assign",
	"AddAssign",
	"SubAssign",
	"LshAssign",
	"RshAssign",
	"MultAssign",
	"DivAssign",
	"ModAssign",
	"AndAssign",
	"OrAssign",
	"XorAssign",
	"PPPre",
	"MMPre",
	"PPPost",
	"MMPost",
	"Return",
	"Break",
	"Exit",
	"Continue",
	"Eval",
	"Arg",
	"ArrIndex",
	"ArgList",
	"Load",
	"Array",
	"ArgCnt",
	"Pause",
	"MacroDef",
	"ClassDef",
	"VarDef",
	"RefAssign",
	"ThisPtr",
	"GVar",
	"LVar",
	"ThisMember",
	"NoMember",
	"CreatorDef",
	"Throw",
	"TryCatch",
	"Self",
	"Position",
	"BoundArg",
	"Package",
} ;


//---END NAMES---

// Compiler builtins
static int _HitInfo (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ((CmmComp *)cenv)->HitInfo () ;
    *retval = 0 ;
    return 0 ;
}

static int _Compile (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    if (!argp[0].IsString()) {
	return -1 ;
    }
    int st ;
    Value acode ;
    if (!cenv->AcCompile) {
	cenv->SetErrorCode (COMPILER_NOT_INSTALLED) ;
	*retval = NULL ;
	return -1 ;
    } else {
	st = cenv->AcCompile (argp[0].sptr, &acode) ;
    }
    if (st < 0)
	retval->Init () ;
    *retval = acode ;
    return 0 ;
}

static int _Exec (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    int st ;

    if (!argp[0].IsArray ()) {
	return -1 ;
    }
    if (!cenv->AcExecute) {
	cenv->SetErrorCode (COMPILER_NOT_INSTALLED) ;
	*retval = NULL ;
	return -1 ;
    } else {
	st = cenv->AcExecute (&argp[0]) ;
    }
    if (st < 0)
	return -1 ;
    *retval = cenv->varstack.Pop  () ;
    return 0 ;
}

CmmComp::CmmComp (int stacksize, int heapsize)
	: CmmInterp (stacksize, heapsize)
{
    Init () ;
}

CmmComp::CmmComp (Variable *stack, int stacksize,
		  char *heap, int heapsize, StaticSymbolTable *symtbl = NULL)
	: CmmInterp (stack, stacksize, heap, heapsize, symtbl)
{
    Init () ;
}

void CmmComp::InitBuiltinSource ()
{
    Value acform ;
    int st ;

    st = AcCompile (builtInSources, &acform) ;
    st = AcExecute (&acform) ;
}

void CmmComp::Init ()
{
    int i ;

    var_hit  = 0 ;
    lvar_hit = 0 ;
    gvar_hit = 0 ;
    selmem_hit = 0 ;
    selmem_cache_hit = 0 ;
    this_mem_hit = 0 ;
    has_compiler = 1 ;
    exec_mode = ExCompiler ;
    varstack.Push ("__compiler__", 1) ;
    varstack.Push ("hit_info", _HitInfo) ;
    varstack.Push ("compile", _Compile) ;
    varstack.Push ("fexec", _Exec) ;

    for (i = 0 ; i < A_EndInst ; i++)
	instSyms[i] = symbolTable->GetSymbol (instNames[i]) ;
    
    globalLim = varstack.Top () ;
}

//=== index definition for A_ClassDef
#define AC_CLASSNAME  2
#define AC_SUPERCLASS 3
#define AC_INSTVARS   4
#define AC_CLASSVARS  5
#define AC_MEMBERS    6

#define ARRSIZE(arr) (sizeof(arr)/sizeof(arr[0]))
int CmmComp::AcLookup (char *mnemonic)
{
    int i ;
    for (i = 0 ; i < ARRSIZE(instNames) ; i++) {
	if (!strcmp(mnemonic, instNames[i])) {
	    return i ;
	}
    }
    return -1 ;
}

int CmmComp::AcLookup (Symbol *instsymbol)
{
    int i ;
    for (i = 0 ; i < ARRSIZE(instNames) ; i++) {
	if (instsymbol == instSyms[i]) {
	    return i ;
	}
    }
    return -1 ;
}

//================================================================
// ACode compiler
//================================================================

//----------------------------------------------------------------
// CmmComp::NewArray -- A-Code i[z𐶐邽߂̃\bhQ
//----------------------------------------------------------------
Value *CmmComp::NewArray (int n)
{
    Value *arr ;
    Value x ;
    if (heap_area.AllocArray (n, &varstack, &arr) < 0)
	return NULL ;
    x.InitArr (arr, n) ;
    return varstack.TmPush (x) ;
}

Value *CmmComp::NewArray (int n, AcInst inst)
{
    Value *vp ;
    if (n < 2)
	return NULL ;
    vp = NewArray (n) ;
    if (!vp)
	return NULL ;
    vp->arr.arrp[0] = (int)inst ;
    vp->arr.arrp[1] = *prog ;
    return vp ;
}

Value *CmmComp::NewArray (int n, int reserved)
{
    Value x ;
    x = heap_area.AllocArray (n, reserved, &varstack) ;
    if (x.IsVoid())
	return NULL ;
    return varstack.TmPush (x) ;
}

Value *CmmComp::NewArray (int n, int reserved, AcInst inst)
{
    Value *vp ;
    vp = NewArray (n, reserved) ;
    if (!vp)
	return NULL ;
    vp->arr.arrp[0] = (int)inst ;
    vp->arr.arrp[1] = *prog ;
    return vp ;
}

//----------------------------------------------------------------
// CmmComp::AcStatement -- ̃RpC
//----------------------------------------------------------------
int CmmComp::AcStatement (Token t, LexInfo *lip, AcState state, Value *form)
{
    int st = 0 ;
    Value expr ;
    Value *subform ;

    switch (t) {
      case T_If:
	st = AcIfStatement    (t, lip, state, form) ;
	break ;
      case T_While:
	st = AcWhileStatement (t, lip, state, form) ;
	break ;
      case T_For:
	st = AcForStatement   (t, lip, state, form) ;
	break ;
      case T_Try:
	st = AcTryCatchStatement (t, lip, state, form) ;
	break ;
      case T_Vardef:
	st = AcVarDef   (t, lip, state, form) ;
	break ;
      case T_Class:
	st = AcClassDef (t, lip, state, form) ;
	break ;
      case T_Func:
	st = AcFuncDef  (t, lip, state, form, 1) ;
	break ;
      case T_Macro:
	st = AcFuncDef  (t, lip, state, form, 0) ;
	break ;
      case T_Method:
      case T_Hiden:
	st = AcMethoDef (t, lip, state, form) ;
	break ;
      case T_Load:
      case T_Package:
	subform = NewArray (3, (t == T_Load) ? A_Load : A_Package) ;
	t = Lex (lip) ;
	st = AcAssign (t, lip, NULL, state, &expr) ;
	CHECKST (st) ;
	subform->arr.arrp[2] = expr ;
	*form = varstack.Pop() ;
	break ;
      case T_Pause:
	subform = NewArray (2, A_Pause) ;
	*form = varstack.Pop () ;
	break ;
      case T_Exit:
      case T_Break:
      case T_Return:
	subform = NewArray (2, 3, ((t == T_Return) ? A_Return
				   : (t == T_Break) ?  A_Break
				   : A_Exit)) ;
	t = Lex(lip) ;
	if (t != T_Eos) {
	    st = AcAssign (t, lip, NULL, state, &expr) ;
	    CHECKST(st) ;
	    heap_area.AddArray (subform, &expr, &varstack, subform) ;
	} else {
	    PutBack () ;
	}
	*form = varstack.VPop() ;
	break ;
      case T_Throw:
	subform = NewArray (2, A_Throw) ;
	t = Lex(lip) ;
	st = AcAssign (t, lip, NULL, state, &expr) ;
	CHECKST(st) ;
	heap_area.AddArray (subform, &expr, &varstack, subform) ;
	*form = varstack.VPop() ;
	break ;
      case T_Continue:
	subform = NewArray (2, A_Continue) ;
	*form = varstack.VPop () ;
	break ;
      case T_Lbr:
      case T_LLbr:
        {
	    int statmentcnt = 0 ;
	    Token startT = t ;

	    subform = NewArray (2, 24, ((startT == T_Lbr) ? A_Block : A_SLBlock)) ;
	    while (1) {
		t = Lex (lip) ;
		if (t == T_Eof)
		    return SetErrorCode (SYNTAX_ERROR) ;
		if (startT == T_Lbr && t == T_Rbr)
		    break ;
		else if (startT == T_LLbr && t == T_RRbr)
		    break ;
		st = AcStatement (t, lip, state, &expr) ;
		if (st == BREAK || st == RETURN) {
		    if (st == BREAK)
			st = 0 ;
		    break ;
		}
		CHECKST(st) ;
		if (expr.arr.arrp[0].intval != (int)A_Nop) {
		    heap_area.AddArray (subform, &expr, &varstack, subform) ;
		    statmentcnt++ ;
		}
	    }
	    if (statmentcnt == 0) {
		subform->arr.arrp[0].intval = (int)A_Nop ;
		*form = varstack.VPop () ;
	    } else if (statmentcnt == 1) {
		*form = subform->arr.arrp[2] ;
		varstack.VPop () ;
	    } else {
		*form = varstack.VPop() ;
	    }
	}
	break ;
      case T_Eos:
	st = 0 ;
	subform = NewArray (2, A_Nop) ;
	*form = varstack.Pop () ;
	break ;
      case T_Eof:
	st = 0 ;
	break ;
      case T_Word:   case T_WorkVar:
      case T_Plus:   case T_Minus:     case T_Not:    case T_Eval:
      case T_Number: case T_IntNumber: case T_String: case T_Symbol:
      case T_Lpar:   case T_ArgPrefix: case T_Void:
      case T_ColonColon: case T_PointTo: case T_Lbl:
      case T_ArgCnt: case T_ArgList: case T_PP: case T_MM:
      case T_BitRev: case T_This: case T_Self: case T_BoundArg:
	st = AcAssign (t, lip, NULL, state, form) ;
	break ;
      case T_Err:
      default:
	return SetErrorCode (SYNTAX_ERROR) ;
    }
    t = Lex (lip) ;
    if (t != T_Eos) {
	PutBack () ;
    }
    return st ;
}

// [A_If [cond form] [true form] ([false form])] 
int CmmComp::AcIfStatement (Token t, LexInfo *lip, AcState state, Value *form)
{
    Value *iform ;
    Value expr ;
    int st ;

    iform = NewArray(5, A_If) ;
    if (Lex (lip) != T_Lpar)
	return SetErrorCode (SYNTAX_ERROR) ;
    t = Lex (lip) ;
    st = AcAssign (t, lip, NULL, state, &expr) ;
    if (Lex (lip) != T_Rpar)
	return SetErrorCode (SYNTAX_ERROR) ;
    iform->arr.arrp[2] = expr ;
    CHECKST(st) ;
    t = Lex (lip) ;
    st = AcStatement (t, lip, state, &expr) ;
    iform->arr.arrp[3] = expr ;
    CHECKST (st) ;
    t = Lex (lip) ;
    if (t == T_Else) {
	Value elseform ;
	t = Lex (lip) ;
	st = AcStatement (t, lip, state, &elseform) ;
	CHECKST (st) ;
	iform->arr.arrp[4] = elseform ;
    } else {
	PutBack () ;
    }
    *form = varstack.Pop() ;
    return st ;
}

// [A_While [condform] [execform]]
int CmmComp::AcWhileStatement (Token t, LexInfo *lip, AcState state, Value *form)
{
    int st ;
    Value *wform ;
    Value expr ;

    wform = NewArray (4, A_While) ;
    if (Lex (lip) != T_Lpar)
	return SetErrorCode (SYNTAX_ERROR) ;
    t = Lex (lip) ;
    st = AcAssign (t, lip, NULL, state, &expr) ;
    CHECKST(st) ;
    wform->arr.arrp[2] = expr ;
    if (Lex (lip) != T_Rpar)
	return SetErrorCode (SYNTAX_ERROR) ;

    t = Lex (lip) ;
    expr.Init () ;
    st = AcStatement (t, lip, state, &expr) ;
    CHECKST(st) ;

    wform->arr.arrp[3] = expr ;
    *form = varstack.Pop() ;
    return st ;
}

// [A_For [init] [cond] [reset] [body]]
int CmmComp::AcForStatement (Token t, LexInfo *lip, AcState state, Value *form)
{
    int st ;
    Value *fform ;
    Value *subform ;
    Value expr ;
    Value x ;

    fform = NewArray (6, A_For) ;

    subform = NewArray (0, 3) ;
    if (Lex (lip) != T_Lpar) {
	return SetErrorCode (SYNTAX_ERROR) ;
    }
    // ݒ 
    int exprcnt ;
    for (exprcnt = 0 ; ; exprcnt++) {
	t = Lex (lip) ;
	if (t == T_Eos) {
	    if (exprcnt != 0)
		return SetErrorCode (SYNTAX_ERROR) ;
	    break ;
	}
	st = AcAssign (t, lip, NULL, state, &expr) ;
	CHECKST(st) ;
	if (heap_area.AddArray (subform, &expr, &varstack, subform) < 0)
	    return -1 ;
	t = Lex (lip) ;
	if (t == T_Comma)
	    continue ;
	if (t == T_Eos)
	    break ;
	return SetErrorCode (SYNTAX_ERROR) ;
    }
    x = varstack.Pop() ;
    fform->arr.arrp[2] = x ;
    // [v 
    t = Lex (lip) ;
    if (t != T_Eos) {
	st = AcAssign (t, lip, NULL, state, &expr) ;
	CHECKST(st) ;
	fform->arr.arrp[3] = expr ;
	t = Lex (lip) ;
	if (t != T_Eos)
	    return SetErrorCode (SYNTAX_ERROR) ;
    } else {
	expr.Init ();
	fform->arr.arrp[3] = expr ;
    }
    // Đݒ 
    subform = NewArray (0, 3) ;
    for (exprcnt = 0 ; ; exprcnt++) {
	t = Lex (lip) ;
	if (t == T_Rpar) {
	    if (exprcnt != 0)
		return SetErrorCode (SYNTAX_ERROR) ;
	    break ;
	}
	st = AcAssign (t, lip, NULL, state, &expr) ;
	CHECKST(st) ;
	if (heap_area.AddArray (subform, &expr, &varstack, subform) < 0)
	    return -1 ;
	t = Lex (lip) ;
	if (t == T_Comma)
	    continue ;
	if (t == T_Rpar)
	    break ;
	return SetErrorCode (SYNTAX_ERROR) ;
    }
    x = varstack.Pop() ;
    fform->arr.arrp[4] = x ;
    
    t = Lex (lip) ;
    st = AcAssign (t, lip, NULL, state, &expr) ;
    CHECKST(st) ;
    fform->arr.arrp[5] = expr ;
    *form = varstack.Pop() ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcClassDef -- NX`̉́B
//
//    [Classdef <LOC> classname superclass
//    	num_of_private/public num_of_method/static
//	[[`private member], [`method' name], [`static, name]]]
//----------------------------------------------------------------
int CmmComp::AcClassDef (Token t, LexInfo *lip, AcState state, Value *form)
{
    Value *cdefp = NewArray (7, A_ClassDef) ;
    char *s ;
    Symbol *classname_sym ;
    Symbol *member_sym ;

    // NX̎擾
    t = Lex (lip) ;
    if (t != T_Word)
	return SetErrorCode (SYNTAX_ERROR) ;
    /* s = heap_area.AllocString (varname , &varstack) ; */
    classname_sym = lip->symbol ;
    cdefp->arr.arrp[2] = classname_sym ;
    // X[p[NX̎擾
    t = Lex (lip) ;
    if (t == T_Colon) {
	t = Lex (lip) ;
	if (t != T_Word)
	    return SetErrorCode (SYNTAX_ERROR) ;
	/* s = heap_area.AllocString (varname , &varstack) ; */
	cdefp->arr.arrp[3] = lip->symbol ;
	t = Lex (lip) ;
    }
    if (t != T_Lbr)
	return SetErrorCode (SYNTAX_ERROR) ;
    cdefp->arr.arrp[4] = 0 ;
    cdefp->arr.arrp[5] = 0 ;

    int num_instvars = 0 ;
    int num_cvs_methods = 0 ;
    Value *members = NewArray (0, 30) ;
    Value *vp ;
    Token mtype ;
    int st ;
    int has_creator = 0 ;
    int is_creator = 0 ;
    int creator_defined = 0 ;

    for (;;) {
	mtype = t = Lex (lip) ;
	if (t == T_Rbr)
	    break ;
	t = Lex (lip) ;
	for (;;) {
	    if (t != T_Word)
		return SetErrorCode (SYNTAX_ERROR) ;
	    member_sym = lip->symbol; 
	    if (mtype == T_Private || mtype == T_Public) {
		// CX^Xo -- private, public
		num_instvars++ ;
		vp = NewArray (2) ;
		vp->arr.arrp[0] = (int)((mtype == T_Private) ? VT_Private : VT_Public) ;
		vp->arr.arrp[1] = member_sym ;
		heap_area.AddArray (members, vp, &varstack, members) ;
		varstack.Pop () ;
	    } else if (mtype == T_Method || mtype == T_Hiden || mtype == T_Static) {
		// NXo -- \bhA`AX^eBbN
		Value methform ;
		num_cvs_methods++ ;
		vp = NewArray (2, 3) ;
		vp->arr.arrp[0] = (int)((mtype == T_Method) ? VT_Method
					: (mtype == T_Hiden)  ? VT_Hiden
					: VT_ClassVar) ;
		vp->arr.arrp[1] = member_sym ;
		is_creator = 0 ;
		// NG[^̃`FbN
		if (lip->symbol == creatorSym) {
		    if (mtype != T_Method)
			return SetErrorCode (SYNTAX_ERROR) ;
		    is_creator = 1 ;
		    has_creator++ ;
		    if (has_creator > 1)
			return SetErrorCode (TOO_MANY_CREATOR) ;
		}
		if (mtype == T_Method || mtype == T_Hiden) {
		    if (Lex (lip) == T_Lpar) {
			// \bh̖{̂`Ă
			PutBack () ;

			if (is_creator) {
			    creator_defined = 1 ;
			    st = AcCreatorBody (classname_sym, member_sym, &methform) ;
			} else {
			    st = AcMethodBody (classname_sym, member_sym, &methform) ;
			}
			if (mtype == T_Hiden)
			    methform.arr.arrp[0] = (int)A_HidenDef ;
			CHECKST(st) ;
			heap_area.AddArray (vp, &methform, &varstack, vp) ;
		    } else if (is_creator) {
			PutBack () ;
			// NG[^{̂`ĂȂBftHgNG[^`
			Context *possave = prog ;

			// ftHgNG[^RpC
			prog = SetContext (nullCreator, 0) ;  // stackp => +2
			st = AcCreatorBody (classname_sym, member_sym, &methform) ;
			CHECKST(st) ;
			heap_area.AddArray (vp, &methform, &varstack, vp) ;
			// ǂݎʒuƂɖ߂
			prog = possave ;
			varstack.TmPop (1) ;
		    } else {
			PutBack () ;
		    }
		} else if (mtype == T_Static) {
		    if (Lex (lip) == T_Assign) {
			// X^eBbNȍ
			Value vform ;
			t = Lex (lip) ;
			st = AcAssign (t, lip, NULL, state, &vform) ;
			CHECKST(st) ;
			heap_area.AddArray (vp, &vform, &varstack, vp) ;
		    } else {
			PutBack () ;
		    }
		}
		heap_area.AddArray (members, vp, &varstack, members) ;
		varstack.Pop () ;
	    }
	    t = Lex (lip) ;
	    if (t == T_Comma) {
		t = Lex (lip) ;
	    } else {
		if (t == T_Static || t == T_Private
		    || t == T_Public || t == T_Method
		    || t == T_Rbr || t == T_Hiden)
		{
		    PutBack () ;
		}
		break ;
	    }
	}
    }
    if (has_creator == 0 || !creator_defined) {
	// NG[^{̂`ĂȂBftHgNG[^`
	Context *possave = prog ;
	Value name ;
	Value creatorForm ;
	vp = NewArray (3) ;		  // stackp => +1
	vp->arr.arrp[0] = (int)VT_Method ;
	vp->arr.arrp[1] = creatorSym ;

	// ftHgNG[^RpC
	prog = SetContext (nullCreator, 0) ;  // stackp => +2
	st = AcCreatorBody (classname_sym, creatorSym, &creatorForm) ;
	CHECKST(st) ;
	vp->arr.arrp[2] = creatorForm ;
	heap_area.AddArray (members, vp, &varstack, members) ;

	// ǂݎʒuƂɖ߂
	prog = possave ;
	varstack.Pop (2) ;

	num_cvs_methods++ ;
    }
    cdefp->arr.arrp[4] = num_instvars ;
    cdefp->arr.arrp[5] = num_cvs_methods ;
    cdefp->arr.arrp[6] = *members ;
    varstack.Pop (2) ;
    *form = *cdefp ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcTryCatchStatement -- try - catch \
//----------------------------------------------------------------
int CmmComp::AcTryCatchStatement (Token t, LexInfo *lip, AcState state, Value *form)
{
    Value *tcform ;
    Value expr ;
    Value *exceptionform ;
    int found_finally = 0 ;
    int st ;

    // [A_TryCatch <SRC> body [[[ExceptionClass obj] exception-body]
    //	0	    1     2    [[ExceptionClass obj] exception-body] 
    //			       [[ExceptionClass obj] exception-body] // 3
    //				:				]
    //		        Finally-body  4]
    tcform = NewArray (5, A_TryCatch) ;
    t = Lex (lip) ;
    st = AcStatement (t, lip, state, &expr) ;
    tcform->arr.arrp[2] = expr ;

    exceptionform = NewArray (0, 1) ;

    for (;;) {
	t = Lex (lip) ;
	if (t != T_Catch && t != T_Finally) {
	    PutBack () ;
	    break ;
	}
	if (t == T_Catch) {
	    Value *subform = NewArray (2) ;
	    Value *class_instance_name = NewArray (2) ;
	    char *name  ;

	    if (found_finally)
		return SetErrorCode (SYNTAX_ERROR) ;
	    t = Lex (lip) ;
	    if (t != T_Lpar)
		return SetErrorCode (SYNTAX_ERROR) ;
	    t = Lex (lip) ;
	    if (t != T_Word)
		return SetErrorCode (SYNTAX_ERROR) ;
#if 0
	    name = heap_area.AllocString (varname , &varstack) ;
	    if (!name)
		return SetErrorCode (NOT_ENOUGH_MEMORY) ;
	    class_instance_name->arr.arrp[0] = name ;
#endif
	    class_instance_name->arr.arrp[0] = lip->symbol ;

	    t = Lex (lip) ;
	    if (t != T_Word)
		return SetErrorCode (SYNTAX_ERROR) ;
	    t = Lex (lip) ;
	    if (t != T_Rpar)
		return SetErrorCode (SYNTAX_ERROR) ;
#if 0
	    name = heap_area.AllocString (varname , &varstack) ;
	    class_instance_name->arr.arrp[1] = name ;
#endif
	    class_instance_name->arr.arrp[1] = lip->symbol ;
	    subform->arr.arrp[0] = class_instance_name->arr ;
	    varstack.TmPop () ;

	    t = Lex(lip) ;
	    st = AcAssign (t, lip, NULL, state, &expr) ;
	    CHECKST(st) ;
	    subform->arr.arrp[1] = expr ;
	    heap_area.AddArray (exceptionform, subform, &varstack, exceptionform) ;
	    varstack.TmPop () ;
	} else if (t == T_Finally) {
	    found_finally = 1 ;
	    t = Lex (lip) ;
	    st = AcAssign (t, lip, NULL, state, &expr) ;
	    tcform->arr.arrp[4] = expr ;
	}
    }
    tcform->arr.arrp[3] = exceptionform->arr ;

    *form = *tcform ;
    varstack.TmPop (2) ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcVarDef -- ϐ`̉́B
//----------------------------------------------------------------
int CmmComp::AcVarDef (Token t, LexInfo *lip, AcState state, Value *form)
{
    //
    // [A_VarDef [varname initexpr] [varname initexpr]]
    //
    Value *formp ;
    LexInfo li ;
    Value *subarr ;
    int st ;

    formp = NewArray (2, 8, A_VarDef) ;
    for (;;) {
	subarr = NewArray (2) ;
	t = Lex (&li) ;
	if (t != T_Word)
	    return SetErrorCode (SYNTAX_ERROR) ;
	subarr->arr.arrp[0] = li.symbol ;
	t = Lex (&li) ;
	if (t == T_Assign || t == T_RefAssign) {
	    Value expr ;
	    t = Lex (&li) ;
	    st = AcAssign (t, &li, NULL, state, &expr) ;
	    subarr->arr.arrp[1] = expr ;
	    t = Lex (&li) ;
	}
	heap_area.AddArray (formp, subarr, &varstack, formp) ;
	varstack.Pop () ;
	if (t != T_Comma) {
	    PutBack () ;
	    break ;
	}
    }
    *form = varstack.Pop () ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcFuncDef -- ֐`̃RpC
//----------------------------------------------------------------
int CmmComp::AcFuncDef (Token t, LexInfo *lip, AcState state, Value *form, int is_func)
{
    Value *fformp ;
    Value expr ;
    int st ;
    char *s ;

    fformp = NewArray (5, A_FuncDef) ;
    if (Lex (lip) != T_Word)
	return SetErrorCode (SYNTAX_ERROR) ;
    fformp->arr.arrp[2] = lip->symbol ;

    st = AcFuncArg (&expr) ; CHECKST(st) ;
    fformp->arr.arrp[3] = expr ;
    t = Lex (lip) ;
    st = AcStatement (t, lip, Ac_InFunc, &expr) ;
    CHECKST (st) ;
    fformp->arr.arrp[4] = expr ;

    *form = varstack.Pop () ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcMethodBody -- \bh`̃RpC
//----------------------------------------------------------------
int CmmComp::AcMethodBody (Symbol *classname_sym, Symbol *methodname_sym, Value *form)
{
    Value *fformp ;
    Value expr ;
    int st ;
    Token t ;
    LexInfo li ;

    fformp = NewArray (6, A_MethoDef) ;
    fformp->arr.arrp[2] = classname_sym ;
    fformp->arr.arrp[3] = methodname_sym ;

    st = AcFuncArg (&expr) ; CHECKST (st) ;
    fformp->arr.arrp[4] = expr ;
    t = Lex (&li) ;
    st = AcStatement (t, &li, Ac_InMethod, &expr) ;
    CHECKST (st) ;
    fformp->arr.arrp[5] = expr ;
    *form = varstack.Pop () ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcCreatorBody -- NG[^̃RpC
//----------------------------------------------------------------
int CmmComp::AcCreatorBody (Symbol *classname_sym, Symbol *methodname_sym, Value *form)
{
    Value *fformp ;
    Value expr ;
    int st ;
    Token t ;
    LexInfo li ;

    fformp = NewArray (7, A_CreatorDef) ;
    fformp->arr.arrp[2] = classname_sym ;
    fformp->arr.arrp[3] = methodname_sym ;

    st = AcFuncArg (&expr) ; CHECKST (st) ;
    fformp->arr.arrp[4] = expr ;
    t = Lex (&li) ;
    if (t == T_Colon) {
	Value superarg ;

	AcSuperArg (&superarg) ; CHECKST(st) ;
	fformp->arr.arrp[5] = superarg ;
	t = Lex (&li) ;
    } else {
	fformp->arr.arrp[5].Init() ;
    }
    st = AcStatement (t, &li, Ac_InMethod, &expr) ;
    CHECKST (st) ;
    fformp->arr.arrp[6] = expr ;
    *form = varstack.Pop () ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcMethoDef -- O`̃\bh/NG[^̃RpC
//----------------------------------------------------------------
int CmmComp::AcMethoDef (Token t, LexInfo *lip, AcState state, Value *form)
{
    Value *fformp ;
    Value expr ;
    int idx ;
    int st ;
    int is_creator = 0 ;

    fformp = NewArray (6, ((t == T_Method) ? A_MethoDef : A_HidenDef)) ;
    if (Lex (lip) != T_Word)
	return SetErrorCode (SYNTAX_ERROR) ;
    fformp->arr.arrp[2] = lip->symbol ;

    t = Lex (lip) ;
    if (t != T_ColonColon)
	return SetErrorCode (SYNTAX_ERROR) ;
    if (Lex (lip) != T_Word)
	return SetErrorCode (SYNTAX_ERROR) ;

    if (lip->symbol == creatorSym) {
	// NG[^̏ꍇ
	Value v ;

	is_creator = 1 ;
	heap_area.AddArray (fformp, &v, &varstack, fformp) ;
	fformp->arr.arrp[0] = A_CreatorDef ;
	fformp->arr.arrp[3] = lip->symbol ;
    } else {
	fformp->arr.arrp[3] = lip->symbol ;
    }
    idx = 4 ;
    st = AcFuncArg (&expr) ; CHECKST (st) ;
    fformp->arr.arrp[idx++] = expr ;

    t = Lex (lip) ;
    if (is_creator) {
	if (t == T_Colon) {
	    // X[p[NXNG[^ւ̈̒`
	    st = AcSuperArg (&expr) ; CHECKST (st) ;
	    fformp->arr.arrp[idx] = expr ;
	    t = Lex (lip) ;
	} else {
	    fformp->arr.arrp[idx].Init () ;
	}
	++idx ;
    }
    st = AcStatement (t, lip, Ac_InMethod, &expr) ;
    CHECKST (st) ;
    fformp->arr.arrp[idx] = expr ;
    *form = varstack.Pop () ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcFuncArg -- ֐, \bḧ`̃RpC
//----------------------------------------------------------------
int CmmComp::AcFuncArg (Value *formp)
{
    Token t ;
    LexInfo li ;
    Value *argp = NewArray (0, 7) ;
    Value argname ;
    int argcnt ;
    Value hashv ;

    t = Lex (&li) ;
    if (t != T_Lpar)
	return SetErrorCode (SYNTAX_ERROR) ;
    for (argcnt = 0 ; ; argcnt++) {
	t = Lex (&li) ;
	if (t == T_Rpar && argcnt == 0)
	    break ;
	if (t != T_Word)
	    return SetErrorCode (SYNTAX_ERROR) ;
	hashv = li.symbol->GetHashValue() ;
	argname = li.symbol ;
	heap_area.AddArray (argp, &argname, &varstack, argp) ;
	heap_area.AddArray (argp, &hashv, &varstack, argp) ;
	t = Lex (&li) ;
	if (t == T_Rpar)
	    break ;
	if (t != T_Comma)
	    return SetErrorCode (SYNTAX_ERROR) ;
    }
    *formp = varstack.Pop() ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcSuperArg -- X[p[NXNG[^ւ̈̃RpC
//----------------------------------------------------------------
int CmmComp::AcSuperArg (Value *form)
{
    Value *argp = NewArray (0, 4) ;
    Value expr ;
    Token t ;
    LexInfo li ;
    int st ;

    t = Lex (&li) ;
    if (t != T_Lpar)
	return SetErrorCode (SYNTAX_ERROR) ;
    t = Lex (&li) ;
    if (t != T_Rpar) {
	for (;;) {
	    st = AcAssign (t, &li, NULL, Ac_InMethod, &expr) ; CHECKST(st) ;
	    heap_area.AddArray (argp, &expr, &varstack, argp) ;
	    t = Lex (&li) ;
	    if (t == T_Rpar)
		break ;
	    if (t != T_Comma)
		return SetErrorCode (SYNTAX_ERROR) ;
	    t = Lex (&li) ;
	}
    }
    *form = varstack.Pop () ;
    return 0 ;
}

//================================================================
// ̃RpĈ߂̃\bhQ
//================================================================

//----------------------------------------------------------------
// CmmComp::AcAssign -- 
//----------------------------------------------------------------
int CmmComp::AcAssign (Token t, LexInfo *lip, Value *vbase, AcState state, Value *form)
{
    Value *aform ;
    Value lform ;
    Value rform ;
    Value *lformp ;
    int st ;
    Token opr ;
    AcInst binst ;

    st = AcConditional (t, lip, vbase, state, &lform) ;
    CHECKST(st) ;
    lformp = varstack.TmPush (lform) ;
    opr = Lex (lip) ;
    switch (opr) {
      case T_Assign:	 binst = A_Assign     ; break ;
      case T_RefAssign:  binst = A_RefAssign  ; break ;
      case T_AddAssign:	 binst = A_AddAssign  ; break ;
      case T_SubAssign:	 binst = A_SubAssign  ; break ;
      case T_MultAssign: binst = A_MultAssign ; break ;
      case T_DivAssign:	 binst = A_DivAssign  ; break ;
      case T_ModAssign:	 binst = A_ModAssign  ; break ;
      case T_LshAssign:	 binst = A_LshAssign  ; break ;
      case T_RshAssign:	 binst = A_RshAssign  ; break ;
      case T_OrAssign:   binst = A_OrAssign   ; break ;
      case T_AndAssign:  binst = A_AndAssign  ; break ;
      case T_XorAssign:  binst = A_XorAssign  ; break ;
      default:
	PutBack () ;
	*form = varstack.Pop () ;
	return 0 ;
    }
    aform = NewArray (4, binst) ;
    aform->arr.arrp[2] = *lformp ;
    t = Lex (lip) ;
    st = AcAssign (t, lip, NULL, state, &rform) ;
    CHECKST(st) ;
    aform->arr.arrp[3] = rform ;
    *form = varstack.Pop() ;
    varstack.Pop() ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcConditional -- ? : R
//----------------------------------------------------------------
int CmmComp::AcConditional (Token t, LexInfo *lip, Value *vbase, AcState state, Value *form)
{
    Value *formp ;
    Value cdform ;
    Value *cdformp ;
    Value tform ;
    Value fform ;
    int st ;

    st = AcCondition0 (t, lip, vbase, state, &cdform) ;
    cdformp = varstack.TmPush (cdform) ;
    CHECKST(st) ;
    t = Lex (lip) ;
    if (t != T_Question) {
	PutBack () ;
	*form = varstack.Pop () ;
	return 0 ;
    }
    formp = NewArray (5, A_If) ;
    formp->arr.arrp[2] = *cdformp ;

    t = Lex (lip) ;
    st = AcCondition0 (t, lip, NULL, state, &tform) ;
    CHECKST (st) ;
    formp->arr.arrp[3] = tform ;

    if (Lex (lip) != T_Colon)
	return SetErrorCode (SYNTAX_ERROR) ;
    t = Lex (lip) ;
    st = AcConditional (t, lip, NULL, state, &fform) ;
    CHECKST (st) ;
    formp->arr.arrp[4] = fform ;
    *form = varstack.Pop () ;
    varstack.Pop() ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcCondition0 -- And 
//----------------------------------------------------------------
int CmmComp::AcCondition0 (Token t, LexInfo *lip, Value *vbase, AcState state, Value *form)
{
    int st ;
    Value rform ;
    Value *rformp ;
    Value lform ;
    Value *cform ;

    st = AcCondition1 (t, lip, vbase, state, &rform) ;
    CHECKST(st) ;
    while (1) {
	rformp = varstack.TmPush (rform) ; // avoid GC
	t = Lex (lip) ;
	if (t != T_And) {
	    PutBack () ;
	    *form = varstack.Pop () ;
	    return 0 ;
	}
	cform  = NewArray (4, A_And) ;
	t = Lex (lip) ;
	st = AcCondition1 (t, lip, NULL, state, &lform) ;
	cform->arr.arrp[2] = *rformp ;
	cform->arr.arrp[3] = lform ;
	rform = varstack.Pop() ;
	varstack.Pop() ;
    }
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcCondition1 -- Or 
//----------------------------------------------------------------
int CmmComp::AcCondition1 (Token t, LexInfo *lip, Value *vbase, AcState state, Value *form)
{
    int st ;
    Value rform ;
    Value *rformp ;
    Value lform ;
    Value *cform ;

    st = AcCondition2 (t, lip, vbase, state, &rform) ;
    CHECKST(st) ;
    while (1) {
	rformp = varstack.TmPush (rform) ; // avoid GC
	t = Lex (lip) ;
	if (t != T_Or) {
	    PutBack () ;
	    *form = varstack.Pop () ;
	    return 0 ;
	}
	cform = NewArray (4, A_Or) ;
	t = Lex (lip) ;
	st = AcCondition2 (t, lip, NULL, state, &lform) ;
	cform->arr.arrp[2] = *rformp ;
	cform->arr.arrp[3] = lform ;
	rform = varstack.Pop() ;
	varstack.Pop() ;
    }
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcCondition0 -- r
//----------------------------------------------------------------
int CmmComp::AcCondition2 (Token t, LexInfo *lip, Value *vbase, AcState state, Value *form)
{
    int st ;
    Value lform, rform,   rrform ;
    Value *lformp, *rformp, *rrformp ;
    Value *cform ;
    AcInst binst ;
    int tri_term ;
    Token opr2 ;

    st = AcExpression (t, lip, vbase, state, &lform) ;
    CHECKST(st) ;

    while (1) {
	lformp = varstack.TmPush (lform) ; // PUSH +1
	t = Lex (lip) ;
	switch (t) {
	  case T_Eq:    binst = A_Eq  ; break ;
	  case T_Neq:   binst = A_Neq ; break ;
	  case T_Le:    binst = A_Le  ; break ;
	  case T_Lt:    binst = A_Lt  ; break ;
	  case T_Ge:    binst = A_Ge  ; break ;
	  case T_Gt:    binst = A_Gt  ; break ;
	  default:
	    PutBack () ;
	    *form = varstack.Pop () ;
	    return 0 ;
	}
	tri_term = 0 ;
	t = Lex (lip) ;
	st = AcExpression (t, lip, NULL, state, &rform) ;
	rformp = varstack.TmPush (rform) ; // PUSH +2

	if (binst == A_Gt || binst == A_Ge || binst == A_Lt || binst == A_Le) {
	    opr2 = Lex (lip) ;
	    if ((binst == A_Lt || binst == A_Le) && (opr2 == T_Lt || opr2 == T_Le))
		tri_term = 1 ;
	    else if ((binst == A_Gt || binst == A_Ge) && (opr2 == T_Gt || opr2 == T_Ge))
		tri_term = 1 ;
	    else {
		PutBack () ;
		tri_term = 0 ;
	    }
	}
	if (!tri_term) {
	    cform = NewArray (4, binst) ;
	    cform->arr.arrp[2] = *lformp ;
	    cform->arr.arrp[3] = *rformp ;
	    lform = varstack.Pop () ;
	    varstack.Pop (2) ;
	} else {
	    AcInst binst2  = (opr2 == T_Lt) ? A_Lt
				: (opr2 == T_Le) ? A_Le
				    : (opr2 == T_Gt) ? A_Gt
					: A_Ge ;
	    Value *cform1, *cform2, *andform ;
	    t = Lex (lip) ;
	    st = AcExpression (t, lip, NULL, state, &rrform) ;
	    CHECKST(st) ;
	    rrformp = varstack.TmPush (rrform) ; // PUSH +3
	    cform1 = NewArray (4, binst) ;		 // PUSH +4
	    cform1->arr.arrp[2] = *lformp ;
	    cform1->arr.arrp[3] = *rformp ;
	    cform2 = NewArray (4, binst2) ;		 // PUSH +5
	    cform2->arr.arrp[2] = *rformp ;
	    cform2->arr.arrp[3] = *rrformp ;
	    andform = NewArray (4, A_And) ;		 // PUSH +6
	    andform->arr.arrp[2] = *cform1 ;
	    andform->arr.arrp[3] = *cform2 ;
	    lform = varstack.Pop () ;
	    varstack.Pop (5) ;
	}

    }
}

//----------------------------------------------------------------
// CmmComp::AcExpression -- AVtgAbit-or, bit-xor 
//----------------------------------------------------------------
int CmmComp::AcExpression (Token t, LexInfo *lip, Value *vbase, AcState state, Value *form)
{
    int st ;
    Value rform, lform ;
    Value *rformp ;
    Value *cform ;
    AcInst binst ;
    LexInfo li ;

    st = AcFactor (t, lip, vbase, state, &rform) ;
    CHECKST(st) ;
    while (1) {
	rformp = varstack.TmPush (rform) ;
	t = Lex (lip) ;
	if      (t == T_Plus)  binst = A_Plus ;
	else if (t == T_Minus) binst = A_Minus ;
	else if (t == T_Rsh)   binst = A_Rsh ;
	else if (t == T_Lsh)   binst = A_Lsh ;
	else if (t == T_Bor)   binst = A_Bor ;
        else if (t == T_Xor)   binst = A_Xor ;
	else {
	    PutBack () ;
	    *form = varstack.Pop () ;
	    return 0 ;
	}
	cform = NewArray (4, binst) ;
	t = Lex (&li) ;
	st = AcFactor (t, &li, NULL, state, &lform) ;
	cform->arr.arrp[2] = *rformp ;
	cform->arr.arrp[3] = lform ;
	rform = varstack.Pop () ;
	varstack.Pop () ;
    }
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcFactor -- 揜A]Abit-and 
//----------------------------------------------------------------
int CmmComp::AcFactor (Token t, LexInfo *lip, Value *vbase, AcState state, Value *form)
{
    int st ;
    Value rform, lform ;
    Value *rformp ;
    Value *cform ;
    AcInst binst ;
    LexInfo li ;

    st = AcTerm (t, lip, vbase, state, &rform) ;
    CHECKST(st) ;
    while (1) {
	rformp = varstack.TmPush (rform) ;
	t = Lex (lip) ;
	if      (t == T_Mult) binst = A_Mult ;
	else if (t == T_Div)  binst = A_Div ;
	else if (t == T_Mod)  binst = A_Mod ;
	else if (t == T_Band) binst = A_Band ;
	else {
	    PutBack () ;
	    *form = varstack.Pop () ;
	    return 0 ;
	}
	cform = NewArray (4, binst) ;
	t = Lex (&li) ;
	st = AcTerm (t, &li, NULL, state, &lform) ;
	cform->arr.arrp[2] = *rformp ;
	cform->arr.arrp[3] = lform ;
	rform = varstack.Pop () ; // rform = *cform ;
	varstack.Pop () ;
    }
}

//----------------------------------------------------------------
// CmmComp::AcTerm -- P
//----------------------------------------------------------------
int CmmComp::AcTerm (Token t, LexInfo *lip, Value *vbase, AcState state, Value *form)
{
    int st ;
    int unaryop = 0 ;
    Value *basep = NULL ;
    Value *uarrp ;
    Value uarr ;
    AcInst uinst ;
    Value xform ;
#define PRE_INC  0x01
#define PRE_DEC  0x02
#define UMINUS   0x04
#define UNOT     0x08
#define UREV	 0x10
#define PREOPR   (PRE_INC|PRE_DEC)

    // check unary operator
    if (t == T_PP) {
	unaryop |= PRE_INC ;
    } else if (t == T_MM) {
	unaryop |= PRE_DEC ;
    } else if (t == T_Minus) {
	unaryop |= UMINUS ;
    } else if (t == T_Not) {
	unaryop |= UNOT ;
    } else if (t == T_BitRev) {
	unaryop |= UREV ;
    } else if (t == T_Plus) {
	// ignore unary plus
	t = Lex (lip) ;
    }
    if (unaryop) // PZqꍇ́Aǂ 
	t = Lex (lip) ;
    st = AcVarAndValue (t, lip, vbase, state, &xform) ;  CHECKST(st) ;
    basep = varstack.TmPush (xform) ;
    t = Lex (lip) ;
    if (unaryop) {
	if      (unaryop & PRE_INC) uinst = A_PPPre ;
	else if (unaryop & PRE_DEC) uinst = A_MMPost ;
	else if (unaryop & UNOT   ) uinst = A_Not ;
	else if (unaryop & UMINUS ) uinst = A_Negate ;
	else if (unaryop & UREV)    uinst = A_BitRev ;

	uarrp = NewArray (3, uinst) ;
	uarrp->arr.arrp[2] = *basep ;
	uarr = varstack.Pop() ;
	varstack.Pop () ;
	basep = varstack.TmPush (uarr) ;
    }
    if (t == T_PP || t == T_MM) {
	if (unaryop & PREOPR)
	    return SetErrorCode (SYNTAX_ERROR) ;
	if (t == T_PP)      uinst = A_PPPost ;
	else if (t == T_MM) uinst = A_MMPost ;
	uarrp = NewArray (3, uinst) ;
	uarrp->arr.arrp[2] = *basep ;
	uarr = varstack.Pop() ;
	varstack.Pop () ;
	basep = varstack.TmPush (uarr) ;
    } else {
	PutBack () ;
    }
    *form = varstack.Pop () ;
    return 0 ;
}

//----------------------------------------------------------------
// CmmComp::AcVarAndValue -- ()A萔AoA
//----------------------------------------------------------------
int CmmComp::AcVarAndValue (Token t, LexInfo *lip, Value *basep, AcState state, Value *form)
{
    LexInfo li ;
    Value expr ;
    int st ;
    int readnext = 1 ;
    char *vname ;

    if (basep && (t != T_Word && t != T_Lpar))
	return SetErrorCode (SYNTAX_ERROR) ;
    switch (t) {
      case T_Lpar:
	// ( JbRBċAI AcAssign Ă
	t = Lex (&li) ;
	st = AcAssign (t, &li, basep, state, &expr) ;  CHECKST(st) ;
	basep = varstack.TmPush (expr) ;
	t = Lex (&li) ;
	if (t != T_Rpar)
	    return SetErrorCode (SYNTAX_ERROR) ;
	break ;
      case T_Word:
	// ϐ
	if (!basep) {
	    basep = NewArray (5, A_Var) ;
	    basep->arr.arrp[2] = lip->symbol ;
	    basep->arr.arrp[3] = lip->symbol->GetHashValue() ;
	} else {
	    Value *newbasep ;

	    newbasep = NewArray (7, A_SelectMember) ;
	    newbasep->arr.arrp[2] = *basep ;
	    newbasep->arr.arrp[3] = lip->symbol ;
	    basep = newbasep ;
	}
	break ;
      case T_IntNumber:
	basep = NewArray (3, A_Const) ;
	basep->arr.arrp[2] = lip->ival ;
	break ;
      case T_Number:
	basep = NewArray (3, A_Const) ;
	basep->arr.arrp[2] = lip->value  ;
	break ;
      case T_Void:
	basep = NewArray (3, A_Const) ;
	basep->arr.arrp[2].Init () 	;
	break ;
      case T_Symbol:
	basep = NewArray (3, A_Const) ;
	basep->arr.arrp[2] = lip->symbol ;
	break ;
      case T_String:
	{
	    Value *strp ;

	    strp = varstack.TmPush (lip->sval) ;
	    basep = NewArray (3, A_Const) ;
	    basep->arr.arrp[2] = *strp ;

	    Value v ;
	    v = varstack.VPop () ;
	    varstack.TmPop () ;
	    basep = varstack.TmPush (v) ;
	}
	break ;
      case T_This:
	if (state != Ac_InMethod)
	    return SetErrorCode (INVALID_EXPRESSION) ;
	basep = NewArray (2, A_ThisPtr) ;
	break ;
      case T_Self:
	if (state != Ac_InMethod)
	    return SetErrorCode (INVALID_EXPRESSION) ;
	basep = NewArray (2, A_Self) ;
	break ;
      case T_ColonColon:
      case T_PointTo:
	if (state != Ac_InMethod)
	    return SetErrorCode (INVALID_EXPRESSION) ;
	basep = NewArray (2, A_ThisPtr) ;
	readnext = 0 ;
	break ;
      case T_New:
	basep = NewArray (3, A_New) ;
	t = Lex (&li) ;
	st = AcAssign (t, &li, NULL, state, &expr) ;
	basep->arr.arrp[2] = expr ;
	*form = varstack.VPop () ;
	return st ;
      case T_Eval:
	basep = NewArray (3, A_Eval) ;
	if (Lex (&li) != T_Lpar)
	    return SetErrorCode (SYNTAX_ERROR) ;
	t = Lex (&li) ;
	st = AcAssign (t, &li, NULL, state, &expr) ;
	if (Lex (&li) != T_Rpar)
	    return SetErrorCode (SYNTAX_ERROR) ;
	basep->arr.arrp[2] = expr ;
	break ;
      case T_ArgList:
	basep = NewArray (2, A_ArgList) ;
	break ;
      case T_ArgCnt:
	basep = NewArray (2, A_ArgCnt) ;
	break ;
      case T_BoundArg:
	basep = NewArray (3, A_BoundArg) ;
	t = Lex (&li) ;
	if (t != T_Word) {
	    return SetErrorCode (SYNTAX_ERROR) ;
	}
	basep->arr.arrp[2] = li.symbol ;
	break ;
      case T_Position:
	basep = NewArray (2, A_Position) ;
	break ;
      case T_ArgPrefix:
	t = Lex (&li) ;
	if (t == T_Lbl) {
	    basep = NewArray (3, A_Arg) ;
	    t = Lex (&li) ;
	    st = AcAssign (t, &li, NULL, state, &expr) ;
	    CHECKST(st) ;
	    if (Lex (&li) != T_Rbl)
		return SetErrorCode (SYNTAX_ERROR) ;
	    basep->arr.arrp[2] = expr ;
	} else if (t == T_Word) {
	    basep = NewArray (3, A_NoMember) ;
	    basep->arr.arrp[2] = li.symbol ;
	} else {
	    return SetErrorCode (SYNTAX_ERROR) ;
	}
	break ;
      case T_Lbl:
	{
	    // [ ... ] z萔
	    basep = NewArray (3, A_Array) ;
	    Value *subarr = NewArray (0, 10) ;
	    t = Lex (&li) ;
	    if (t != T_Rbl) {
		for ( ; ; t = Lex (&li)) {
		    if (t == T_Rbl || t == T_Comma) {
			Value *nullp = NewArray (3, A_Const) ;

			nullp->arr.arrp[2].Init() ;
			heap_area.AddArray (subarr, nullp, &varstack, subarr) ;
			varstack.Pop () ;
			if (t == T_Rbl)
			    break ;
		    } else {
			st = AcAssign (t, &li, NULL, state, &expr) ;
			CHECKST(st) ;
			heap_area.AddArray (subarr, &expr, &varstack, subarr) ;
			t = Lex (&li) ;
			if (t == T_Rbl)
			    break ;
			if (t != T_Comma)
			    return SetErrorCode (SYNTAX_ERROR) ;
		    }
		}
	    }
	    basep->arr.arrp[2] = *subarr ;
	    varstack.VPop () ;
	}
	break ;
      default:
	st = AcStatement (t, lip, state, &expr) ; CHECKST(st) ;
	*form = expr ;
	return st ;
    }
    if (readnext)
	t = Lex (&li) ;

    Value *formp ;
    Value *subp ;
    Value x ;

    for ( ; ; t = Lex (&li)) {
	switch (t) {
	  case T_Lbl:
	    t = Lex (&li) ;
	    st = AcAssign (t, &li, NULL, state, &expr) ; CHECKST(st) ;
	    t = Lex (&li) ;
	    if (t != T_Rbl)
		return SetErrorCode (SYNTAX_ERROR) ;
	    subp = varstack.TmPush (expr) ;
	    formp = NewArray (4, A_ArrIndex) ;
	    formp->arr.arrp[2] = *basep ;
	    formp->arr.arrp[3] = *subp ;
	    x = varstack.VPop () ;
	    varstack.TmPop (2) ;
	    basep = varstack.TmPush (x) ;
	    break ;
	  case T_PointTo:
	    // oI
	    t = Lex (&li) ;
	    if (t == T_Word) {
		formp = NewArray (7, A_SelectMember) ;
		formp->arr.arrp[2] = *basep ;
		formp->arr.arrp[3] = li.symbol ;
		x = varstack.VPop () ;
		varstack.TmPop () ;
		basep = varstack.TmPush (x) ;
	    } else if (t == T_Lpar) {
		t = Lex (&li) ;
		st = AcAssign (t, &li, basep, state, &expr) ;
		t = Lex (&li) ;
		if (t != T_Rpar)
		    return SetErrorCode (SYNTAX_ERROR) ;
		varstack.TmPop () ;
		basep = varstack.TmPush (expr) ;
	    }
	    break ;
	  case T_ColonColon:
	    // NXXR[vI
	    t = Lex (&li) ;
	    if (t != T_Word)
		return SetErrorCode (SYNTAX_ERROR) ;
	    formp = NewArray (5, A_SelectClass) ;
	    formp->arr.arrp[2] = *basep ;
	    formp->arr.arrp[3] = li.symbol ;
	    formp->arr.arrp[4] = li.symbol->GetHashValue() ;
	    x = varstack.VPop ();
	    varstack.TmPop () ;
	    basep = varstack.TmPush (x) ;
	    break ;
	  case T_Lpar:
	    // ֐Ăт
	    subp  = NewArray (0, 5) ;
	    formp = NewArray (4, A_Funcall) ;
	    t = Lex (&li) ;
	    if (t != T_Rpar) {
		for (;; t = Lex (&li)) {
		    st = AcAssign (t, &li, NULL, state, &expr) ; CHECKST(st) ;
		    heap_area.AddArray (subp, &expr, &varstack, subp) ;
		    t = Lex (&li) ;
		    if (t == T_Rpar)
			break ;
		    if (t != T_Comma)
			return SetErrorCode (SYNTAX_ERROR) ;
		}
	    }
	    formp->arr.arrp[2] = *basep ;
	    formp->arr.arrp[3] = *subp ;
	    x = varstack.VPop () ;
	    varstack.TmPop (2) ;
	    basep = varstack.TmPush (x) ;
	    break ;
	  default:
	    PutBack () ;
	    *form = varstack.VPop () ;
	    return 0 ;
	}
    }
}

//----------------------------------------------------------------
// CmmComp::AcShowInst -- A-code t@C`o
//----------------------------------------------------------------
int CmmComp::AcShowInst (FILE *fp, Value *x, int level, int int_inst)
{
#define AC_INDENT 2
    if (!fp)
	fp = stdout ;
    if (!x->IsArray()) {
	printf ("NOT ARRAY !!") ;
	x->Print () ;
	// printf ("\n") ;
	return  0 ;
    }
    Value head ;
    Value e ;
    int nlevel = -1 ;
    int i = 2 ;
    int donl = 0 ;
    int elemend ;

    elemend = x->arr.numelems ;
    if (x->arr.numelems == 0) {
	fprintf (fp, "[]") ;
	return 0 ;
    }
    head = x->arr.arrp[0] ;
    if (head.IsArray()) {
	AcShowInst (fp, &head, level+1, int_inst) ;
	i = 1 ;
    } else {
	int j ;
	for (j = 0 ; j < level * AC_INDENT ; j++)
	    putc (' ', fp) ;
	if (!int_inst)
	    fprintf (fp, "[%s ", instNames[head.intval]) ;
	else
	    fprintf (fp, "[%d ", head.intval) ;
#if 0
	if (!x->arr.arrp[1].IsContext()) {
	    printf ("NO CONTEXT !!\n") ;
	} else {
	    x->arr.arrp[1].Print () ;
	}
#endif
	AcInst inst = (AcInst)head.intval ;

	switch (inst) {
	  case A_While:
	    {
		AcShowInst (fp, &x->arr.arrp[2], -1, int_inst) ;
		fprintf (fp, "\n") ;
		nlevel = level + 1 ;
		i = 3 ;
		donl++ ;
	    }
	    break ;
	  case A_For:
	    {
		Value *xp = &x->arr.arrp[2] ;
		int j ;

		fprintf (fp, "[") ;
		for (j = 0 ; j < xp->arr.numelems ; j++)
		    AcShowInst (fp, &xp->arr.arrp[j], -1, int_inst) ;
		fprintf (fp, "]") ;
		if (x->arr.arrp[3].IsArray())
		    AcShowInst (fp, &x->arr.arrp[3], -1, int_inst) ;
		else
		    x->arr.arrp[3].Write(fp) ;
		xp = &x->arr.arrp[4] ;
		fprintf (fp, "[") ;
		for (j = 0 ; j < xp->arr.numelems ; j++)
		    AcShowInst (fp, &xp->arr.arrp[j], -1, int_inst) ;
		fprintf (fp, "]") ;
		fprintf (fp, "\n") ;
		nlevel = level+1 ;
		i = 5 ;
		donl++ ;
	    }
	    break ;
	  case A_If:
	    {
		AcShowInst (fp, &x->arr.arrp[2], -1, int_inst) ;
		fprintf (fp, "\n") ;
		nlevel = level+1 ;
		i = 3 ;
		donl++ ;
		AcShowInst (fp, &x->arr.arrp[3], nlevel, int_inst) ;
		if (x->arr.arrp[4].IsArray()) {
		    AcShowInst (fp, &x->arr.arrp[4], nlevel, int_inst) ;
		}
		i = 5 ;
	    }
	    break ;
	  case A_Block:
	  case A_SLBlock:
	    {
		fprintf (fp, "\n") ;
		nlevel = level+1 ;
		donl++ ;
	    }
	    break ;
	  case A_FuncDef:
	    {
		nlevel = level+1 ;
		x->arr.arrp[2].Write (fp) ;
		x->arr.arrp[3].Write (fp) ;
		fprintf (fp, "\n") ;
		donl++ ;
		i = 4 ;
	    }
	    break ;
	  case A_MethoDef:
	  case A_HidenDef:
	    {
		x->arr.arrp[2].Write (fp) ;
		x->arr.arrp[3].Write (fp) ;
		x->arr.arrp[4].Write (fp) ;
		fprintf (fp, "\n") ;
		i = 5 ;
		nlevel = level+1 ;
		donl++ ;
	    }
	    break ;
	  case A_CreatorDef:
	    {
		x->arr.arrp[2].Write (fp) ;
		x->arr.arrp[3].Write (fp) ;
		x->arr.arrp[4].Write (fp) ;
		if (x->arr.arrp[5].IsVoid()) {
		    x->arr.arrp[5].Write (fp) ;
		} else if (x->arr.arrp[5].IsArray()) {
		    fprintf (fp, "[") ;
		    for (j = 0 ; j < x->arr.arrp[5].arr.numelems ; j++) {
			AcShowInst (fp, &x->arr.arrp[5].arr.arrp[j], -i, int_inst) ;
		    }
		    fprintf (fp, "]") ;
		}
		fprintf (fp, "\n") ;
		i = 6 ;
		nlevel = level + 1 ;
		donl++ ;
	    }
	    break ;
	  case A_Funcall:
	    {
		AcShowInst (fp, &x->arr.arrp[2], -1, int_inst) ;
		Value xarg = x->arr.arrp[3] ;
		int j ;
		fprintf (fp, "[") ;
		for (j = 0 ; j < xarg.arr.numelems ; j++) {
		    AcShowInst (fp, &xarg.arr.arrp[j], -1, int_inst) ;
		}
		fprintf (fp, "]") ;
		nlevel = -1 ;
		i = 4 ;
	    }
	    break ;
	  case A_Array:
	    {
		Value xarg = x->arr.arrp[2] ;
		fprintf (fp, "[") ;
		for (j = 0 ; j < xarg.arr.numelems ; j++) {
		    AcShowInst (fp, &xarg.arr.arrp[j], -1, int_inst) ;
		}
		fprintf (fp, "]") ;
		i = 4 ;
	    }
	    break ;
	  case A_ClassDef:
	    {
		x->arr.arrp[2].Write (fp) ; putc (' ', fp) ;
		x->arr.arrp[3].Write (fp) ; putc (' ', fp) ;
		x->arr.arrp[4].Write (fp) ; putc (' ', fp) ;
		x->arr.arrp[5].Write (fp) ; putc (' ', fp) ;
		int num_members = x->arr.arrp[4].intval + x->arr.arrp[5].intval ;
		Value member ;

		member = x->arr.arrp[6] ;
		fprintf (fp, "[") ;
		for (i = 0 ; i < num_members ; i++) {
		    Value a_member ;
		    a_member = member[i] ;
		    putc ('[', fp) ;
		    if (int_inst) {
			fprintf (fp, "%d ", a_member[0].intval) ;
			a_member[1].Write(fp) ;
			if (a_member.arr.numelems == 3) {
			    AcShowInst (fp, &a_member[2], -1, int_inst) ;
			}
		    } else if ((VarType)a_member[0].intval == VT_Private) {
			fprintf (fp, "private ") ;
			a_member[1].Write(fp) ;
		    } else if ((VarType)a_member[0].intval == VT_Public) {
			fprintf (fp, "public ") ;
			a_member[1].Write(fp) ;
		    } else if ((VarType)a_member[0].intval == VT_ClassVar) {
			fprintf (fp, "static ") ;
			a_member[1].Write(fp) ;
			if (a_member[2].IsArray())
			    AcShowInst (fp, &a_member[2], level, int_inst) ;
		    } else {
			if ((VarType)a_member[0].intval == VT_Method)
			    fprintf (fp, "method ") ;
			else
			    fprintf (fp, "hiden ") ;
			a_member[1].Write(fp) ;
			if (a_member.arr.numelems == 3) {
			    AcShowInst (fp, &a_member[2], -1, int_inst) ;
			}
		    }
		    putc (']', fp) ;
		}
		fprintf (fp, "]\n") ;
		i = 7 ;
		break ;
	    }
	  case A_VarDef:
	    for (i = 2 ; i < x->arr.numelems ; i++) {
		Value *ap = &x->arr.arrp[i] ;
		putc ('[', fp) ;
		ap->arr.arrp[0].Write (fp) ;
		if (ap->arr.arrp[1].IsArray()) {
		    AcShowInst (fp, &ap->arr.arrp[1], level+1, int_inst) ;
		}
		putc (']', fp) ;
	    }
	    break ;
	  case A_TryCatch:
	    {
		AcShowInst (fp, &x->arr.arrp[2], level+1, int_inst) ;
		Value *catch_blocks = &x->arr.arrp[3] ;

		putc ('[', fp) ;
		for (j = 0 ; j < catch_blocks->arr.numelems ; j++) {
		    Value *a_catch_block = &catch_blocks->arr.arrp[j] ;
		    putc ('[', fp) ;
		    a_catch_block->arr.arrp[0].Write (fp) ;
		    AcShowInst (fp, &a_catch_block->arr.arrp[1], level+1, int_inst) ;
		    putc (']', fp) ;
		}
		putc (']', fp) ;
		if (x->arr.arrp[4].IsArray()) {
		    AcShowInst (fp, &x->arr.arrp[4], level+1, int_inst) ;
		} else {
		    x->arr.arrp[4].Write(fp) ;
		}
		i = x->arr.numelems ;
	    }
	    break ;
	  case A_LVar:
	    fprintf (fp, "%d", x->arr.arrp[2].intval) ;
	    i = x->arr.numelems ;
	    break ;
	  case A_GVar:
	    fprintf (fp, "%d", (int)((Variable *)x->arr.arrp[2].location -
				     varstack.PeekStack(0))) ;
	    i = x->arr.numelems ;
	    break ;
	  case A_SelectMember:
	    i = elemend = 4 ;
	    AcShowInst (fp, &x->arr.arrp[2], nlevel, int_inst) ;
	    x->arr.arrp[3].Write (fp) ;
	    if (x->arr.arrp[4].IsLocation ()) {
		fprintf (fp, "0x%x %d %d",
			 x->arr.arrp[4].location,
			 x->arr.arrp[5].intval,
			 x->arr.arrp[6].intval) ;
	    }
	    break ;
	  case A_Var:
	    elemend = 4 ;
	    break ;
	  case A_BoundArg:
	    elemend = 3 ;
	    break ;
	  case A_ThisMember:
	    i = elemend = 0 ;
	    fprintf (fp, "0x%x %d %d",
		     (int)x->arr.arrp[2].location,
		     x->arr.arrp[3].intval,
		     x->arr.arrp[4].intval) ;
	    break ;
	  default:
	    ;
	}
    }
    for ( ; i < elemend ; i++) {
	e = x->arr.arrp[i] ;
	if (e.IsArray()) {
	    AcShowInst (fp, &e, nlevel, int_inst) ;
	} else if (e.IsLocation()) {
	    fprintf (fp, "%d", varstack.VarIndex ((Variable *)e.location)) ;
	} else {
	    if (i > 0)
		putc (' ', fp);
	    e.Write (fp) ;
	}
    }
    if (donl) {
	int j ;
	for (j = 0 ; j < level * AC_INDENT ; j++)
	    putc (' ', fp) ;
    }
    fprintf (fp, "]") ;
    if (level >= 0)
	fprintf (fp, "\n") ;
    return 0 ;
}

int CmmComp::AcProgram (Value *form)
{
    int    st = 0 ;
    Token  t ;
    LexInfo li ;
    Value expr ;

    Value *progp = NewArray (0, 20) ;

    while (1) {
	t = Lex (&li) ;
	if (t == T_Eof)
	    break ;
	if (t == T_Eos)
	    continue ;
	expr.Init () ;
	st = AcStatement (t, &li, Ac_InMain, &expr) ;
	CHECKST(st) ;
	heap_area.AddArray (progp, &expr, &varstack, progp) ;
    }
    *form = varstack.Pop () ; // MODIFIED
    return st ;
}

inline int CmmComp::AcMakeArray (Value *form)
{
    Value *formp ;
    Value *arrp ;
    Value retv ;
    int st ;
    int i ;
    int nelems = form->arr.numelems ;
    LocInfo loc ;

    formp = varstack.TmPush (*form) ;
    arrp = NewArray (nelems) ;
    for (i = 0 ; i < nelems ; i++) {
	st = AcExecForm (&formp->arr.arrp[i], &loc) ;
	CHECKST_EX(st) ;
	arrp->arr.arrp[i] = varstack.VPop () ;
    }
    retv = varstack.VPop () ;
    varstack.TmPop () ;
    varstack.TmPush (retv) ;
    return 0 ;
}

int __debugOn = 0 ;
//----------------------------------------------------------------
// CmmComp::AcExecForm -- Execute one A-code
//	ǵAǂł switch B
//----------------------------------------------------------------
extern int CompareClassTemplate (const void *, const void *) ;

int CmmComp::AcExecForm (Value *form, LocInfo *locp)
{
    Value *formp ;
    AcInst binst ;
    LocInfo loc ;
    Value tf ;
    int i ;
    Value retv ;
    Value *classp ;
    int idx ;
    int shrink ;
    Value *vp ;
    int is_member ;
    Value *cp ;
    VarType vtype ;

    int st = 0 ;

#if 0
    if (!form->IsArray()) {
	printf ("Ececutor Fatal :: Invalid Form\n") ;
	varstack.Dump ("INVALID Form !!") ;
	return -1 ;
	varstack.Push ("--const--", *form) ;
	return 0 ;
    }
#endif
    binst = (AcInst)form->arr.arrp[0].intval ;
#if 0
    if (__debugOn)
	AcShowInst (stdout, form, 0, 0) ;
#endif
    switch (binst) {
      case A_Nop:
	varstack.TmPush () ;
	return 0 ;
      case A_If:
	{
	    formp = varstack.LPush (form->arr) ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST(st) ;
	    tf = varstack.VPop () ;
	    if (tf.True()) {
		locp->Reset () ;
		st = AcExecForm (&formp->arr.arrp[3], locp) ; CHECKST(st) ;
	    } else if (formp->arr.arrp[4].IsArray()) {
		locp->Reset () ;
		st = AcExecForm (&formp->arr.arrp[4], locp) ; CHECKST(st) ;
	    } else {
		locp->Reset () ;
		retv.Init () ;
		varstack.TmPush (retv) ;
	    }
	    varstack.LPop1 () ;
	    return st ;
	}
      case A_While:
	{
	    Value *condform ;
	    Value *bodyform ;

	    condform = varstack.LPush (form->arr.arrp[2]) ;
	    bodyform = varstack.LPush (form->arr.arrp[3]) ;

	    varstack.TmPush () ;
	    for (;;) {
		loc.Reset () ;
		st = AcExecForm (condform, &loc) ;
		CHECKST_EX(st) ;
		tf = varstack.VPop () ;
		if (!tf.True())
		    break ;
		varstack.TmPop () ;
		locp->Reset () ;
		st = AcExecForm (bodyform, locp) ;  CHECKST_EX(st) ;
		if (st == BREAK || st == RETURN || st == THROW) {
		    if (st == BREAK)
			st = 0 ;
		    break ;
		}
	    }
	    varstack.LPop (2) ;
	    return st ;
	}
      case A_For:
	{
	    Value *initform ;
	    Value *condform ;
	    Value *bodyform ;
	    Value *resetform ;

	    initform  = varstack.LPush (form->arr.arrp[2]) ;
	    condform  = varstack.LPush (form->arr.arrp[3]) ;
	    resetform = varstack.LPush (form->arr.arrp[4]) ;
	    bodyform  = varstack.LPush (form->arr.arrp[5]) ;

	    for (i = 0 ; i < initform->arr.numelems ; i++) {
		loc.Reset () ;
		st = AcExecForm (&initform->arr.arrp[i], &loc) ; CHECKST_EX(st) ;
	    }

	    varstack.TmPop (i) ;
	    varstack.TmPush () ;
	    loc.Reset () ;

	    if (!condform->IsArray()) {
		//
		// [vȂꍇ́A[v̎sȗ
		//
		for (;;) {
		    varstack.TmPop () ;
		    locp->Reset () ;
		    st = AcExecForm (bodyform, locp) ;  CHECKST_EX(st) ;
		    if (st == BREAK || st == RETURN || st == THROW) {
			if (st == BREAK)
			    st = 0 ;
			break ;
		    }
		    for (i = 0 ; i < resetform->arr.numelems ; i++) {
			loc.Reset () ;
			st = AcExecForm (&resetform->arr.arrp[i], &loc) ; CHECKST_EX(st) ;
		    }
		    varstack.TmPop (i) ;
		}
	    } else {
		for (;;) {
		    loc.Reset () ;
		    st = AcExecForm (condform, &loc) ; CHECKST_EX(st) ;
		    tf = varstack.VPop () ;
		    if (!tf.True())
			break ;

		    varstack.TmPop () ;
		    locp->Reset () ;
		    st = AcExecForm (bodyform, locp) ; CHECKST_EX(st) ;
		    if (st == BREAK || st == RETURN || st == THROW) {
			if (st == BREAK)
			    st = 0 ;
			break ;
		    }
		    for (i = 0 ; i < resetform->arr.numelems ; i++) {
			loc.Reset () ;
			st = AcExecForm (&resetform->arr.arrp[i], &loc) ; CHECKST_EX(st) ;
		    }
		    varstack.TmPop (i) ;
		}
	    }
	    varstack.LPop (4) ;
	    return st ;
	}
      case A_FuncDef:
	{
	    formp = varstack.TmPush (form->arr) ;

	    Symbol *s = formp->arr.arrp[2].symbol ;
	    varstack.SetSymSimple (varstack.TopIndex(), s) ;
	    varstack.TmPush (*formp) ;
	    return 0 ;
	}
      case A_VarDef:
	formp = varstack.LPush (form->arr) ;
	{
	    for (i = 2 ; i < formp->arr.numelems ; i++) {
		Value *avarp ;
		avarp = &formp->arr.arrp[i] ;
		if (avarp->arr.arrp[1].IsArray()) {
		    loc.Reset () ;
		    st = AcExecForm (&avarp->arr.arrp[1], &loc) ; CHECKST_EX(st) ;
		    varstack.SetSymSimple (varstack.TopIndex(),
					   formp->arr.arrp[i].arr.arrp[0].symbol) ;
		} else {
		    varstack.Push (avarp->arr.arrp[0].symbol) ;
		}
	    }
	    varstack.LPop1 () ;
	    Value *vp = varstack.Top () ;
	    varstack.TmPush (*vp) ;
	    return 0 ;
	}
      case A_ClassDef:
	formp = varstack.LPush (form->arr) ;
	{
	    Class    *classp ;
	    Variable *varp ;
	    int      instvar_idx = 0 ;
	    int      classvar_idx = 0 ;
	    int      i ;
	    int num_members ;
	    int num_instvars = formp->arr.arrp[4].intval ;
	    int num_cvs_methods = formp->arr.arrp[5].intval ;
#define MEMBERS 6

	    num_members = num_instvars + num_cvs_methods ;
	    Symbol *namesym = formp->arr.arrp[2].symbol ;

	    varp   = varstack.Push (namesym) ;
   	    classp = heap_area.AllocClass (&varstack) ;
	    if (!classp)
		return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(formp)) ;
	    classp->superclass 	    = NULL ;
	    classp->offset = 0 ;
	    classp->num_members     = 0 ;
	    classp->num_instvars    = 0 ;
	    classp->num_cvs_methods = 0 ;
	    classp->class_template  = NULL ;
	    classp->class_vars_or_methods = NULL ;
	    classp->classnamep      = varp ;

	    *varp = classp ;

	    if (formp->arr.arrp[3].IsSymbol()) {
		Value *superclassp ;
		if (!(superclassp = varstack.LookupVar (formp->arr.arrp[3].symbol))
		    || !superclassp->IsClass())
		    return SetErrorCode (UNDEFINED_CLASS, ECTX(formp)) ;
		if (IsSpecialClass (superclassp))
		    return SetErrorCode (INVALID_SUPER_CLASS, ECTX(formp)) ;
		classp->superclass = superclassp->classp ;
		classp->offset = superclassp->classp->offset
				    + superclassp->classp->num_instvars ;
	    } else {
		classp->superclass = classClass->classp ;
		classp->offset = 0 ;
	    }

	    Class::ClassMemberTemplate *class_member_p ;
	    class_member_p
		= (Class::ClassMemberTemplate *)heap_area.AllocSegment
		    (sizeof(Class::ClassMemberTemplate) * num_members, &varstack) ;
	    if (!class_member_p)
		return SetErrorCode (NOT_ENOUGH_MEMORY) ;
	    varp->classp->class_template = class_member_p ;

	    if (num_cvs_methods > 0) {
		Value *cvs_method_p ;

		cvs_method_p = (Value *)heap_area.AllocSegment
				    (sizeof(Value) * num_cvs_methods, &varstack) ;
		if (!cvs_method_p) {
		    return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(formp)) ;
		}
		for (i = 0 ; i < num_cvs_methods ; i++) {
		    cvs_method_p[i].Init() ;
		}
		varp->classp->class_vars_or_methods = cvs_method_p ;
	    }
	    int num_initializers = 0 ;

	    for (i = 0 ; i < num_members ; i++) {
		Value member  ;
		VarType vtype ;

		member = formp->arr.arrp[MEMBERS].arr.arrp[i] ;
		vtype = (VarType)member[0].intval ;

		varp->classp->class_template[i].iv_type = vtype ;
		varp->classp->class_template[i].name_sym = member[1].symbol ;

		if (vtype == VT_Private || vtype == VT_Public)
		    varp->classp->class_template[i].index = instvar_idx++ ;
		else {
		    if (member[1].symbol == disposerSym) {
			varp->classp->have_disposer = 1 ;
		    }
		    if ((vtype == VT_Method || vtype == VT_Hiden)
			&& member.arr.numelems == 3)
		    {
			varp->classp->class_vars_or_methods[classvar_idx] = member[2].arr ;
		    }

		    else if (vtype == VT_ClassVar
			       && member.arr.numelems == 3)
		    {
			varp->classp->class_vars_or_methods[classvar_idx] = member.arr.arrp[2] ;
			varp->classp->class_vars_or_methods[classvar_idx].subtype
			    = CTX_STATIC_INITIALIZER ;
			num_initializers++ ;  
		    }
		    varp->classp->class_template[i].index = classvar_idx++ ;
		}
	    }
	    varp->classp->num_members = num_members ;
	    varp->classp->num_cvs_methods = num_cvs_methods ;
	    varp->classp->num_instvars = num_instvars ;
	    //
	    // ooCiT[`邽߂Ƀ\[gĂ
	    //
	    qsort (varp->classp->class_template,
		   (size_t)varp->classp->num_members,
		   (size_t)sizeof (Class::ClassMemberTemplate),
		   CompareClassTemplate) ;

	    if (num_initializers > 0) {
		// Static R[h̎s
		Value *classptrsave = classptr ;
		Value *thisptrsave = thisptr ;
		int i ;
		int st ;
		LocInfo loc ;

		classptr = varp ;
		thisptr = varp ;
		for (i = 0  ; i < classvar_idx ; i++) {
		    if (varp->classp->class_vars_or_methods[i].subtype != CTX_STATIC_INITIALIZER)
			continue ;
		    loc.Reset () ;
		    st = AcExecForm (&varp->classp->class_vars_or_methods[i], &loc) ;
		    CHECKST_EX(st) ;
		    varp->classp->class_vars_or_methods[i] = varstack.VPop () ;
		    varp->classp->class_vars_or_methods[i].subtype = CTX_NONE ;
		}
		classptr = classptrsave ;
		thisptr  = thisptrsave ;
	    }
	    locp->Reset () ;
	    varstack.LPop1 () ;
	    varstack.TmPush (*varp) ;
	}
	return 0 ;
      case A_MethoDef:
      case A_HidenDef:
	classp = varstack.LookupVar (form->arr.arrp[2].symbol) ;
	if (!classp->IsClass())
	    return SetErrorCode (UNDEFINED_CLASS, ECTX(form)) ;
	idx = classp->classp->FindMyEntry (form->arr.arrp[3].symbol,
					   &locp->vtype) ;
	if (idx < 0) {
	    locp->vtype = (binst == A_MethoDef) ? VT_Method : VT_Hiden ;
	    idx = AllocMethodSlot (classp, form->arr.arrp[3].symbol,
				   locp->vtype) ;
	    if (idx < 0)
		return -1 ;
	}
	if (form->arr.arrp[3].symbol == disposerSym) {
	    classp->classp->have_disposer = 1 ;
	}
	if (locp->vtype != VT_Method && locp->vtype != VT_Hiden)
	    return SetErrorCode (INVALID_MEMBER, ECTX(form)) ;
	if (idx < 0)
	    return SetErrorCode (UNDEFINED_MEMBER, ECTX(form)) ;
	classp->classp->SetAcMethod (idx, form->arr) ;
	locp->Reset () ;
	varstack.TmPush(*form) ;
	return 0 ;
      case A_Funcall:
	shrink = varstack.StackPtr () ;
	formp = varstack.TmPush (form->arr) ;
	{
	    int framepsave = varstack.FramePtr () ;
	    LocInfo loc ;
	    LocInfo retloc ;
	    Value 	  *entp ;
	    Value	  *acfuncp ;
	    int numargs = 0 ;
	    int newfp ;
	    Context *calladr ;
	    Value	  x ;
	    int constThis = 0 ;

	    st = AcExecForm (&(formp->arr.arrp[2]), &loc) ; CHECKST_EX(st) ;
	    if (loc.instancep && varstack.InLStack (loc.instancep)) {
		constThis++ ;
	    }
	    entp = varstack.Top () ;
	    switch (entp->type) {
	      case V_CONTEXT:
		if (entp->subtype == CTX_MACRO) {
		    ;
		} else if (entp->subtype == CTX_METHOD) {
		    if (!loc.instancep)
			return SetErrorCode (SYNTAX_ERROR, ECTX(form)) ;
		    varstack.TmPush (entp->context) ;
		    calladr = &(varstack.Top()->context) ;
		    newfp = varstack.StackPtr () ;
		    varstack.TmPush (*loc.instancep) ;
		    numargs++ ;
		    for (i = 0 ; i < formp->arr.arrp[3].arr.numelems ; i++) {
			LocInfo xloc ;
			st = AcExecForm (&(formp->arr.arrp[3].arr.arrp[i]), &xloc) ;
			CHECKST_EX(st) ;
			numargs++ ;
		    }
		    varstack.SetFramePtr (newfp) ;
		    st = CallContext (calladr, numargs, loc.instancep, loc.classp,
				      &retloc) ;
		} else if (entp->subtype == CTX_CREATOR) {
		    Instance *ip ;
		    Value newinstance ;
		    Value *newip ;
		    Value *entry ;

		    ip = CreateInstance (loc.classp->classp) ;
		    if (!ip)
			return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(form)) ;
		    newinstance = ip ;

		    varstack.TmPush (entp->context) ;
		    // calladr = &(varstack.Top()->context) ;
		    entry = varstack.Top () ;
		    newfp = varstack.StackPtr () ;
		    varstack.TmPush (newinstance) ;
		    newip = varstack.Top () ;
		    numargs++ ;
		    for (i = 0 ; i < formp->arr.arrp[3].arr.numelems ; i++) {
			LocInfo xloc ;
			st = AcExecForm (&(formp->arr.arrp[3].arr.arrp[i]), &xloc) ;
			CHECKST_EX(st) ;
			numargs++ ;
		    }
		    varstack.SetFramePtr (newfp) ;
		    st = CallCreator (loc.classp, entry, newip, numargs) ;
		    CHECKST_EX(st) ;
		    varstack.TmPush (*newip) ;
		} else {
		    varstack.TmPush (entp->context) ;
		    calladr = &(varstack.Top()->context) ;
		    newfp = varstack.StackPtr () ;
		    for (i = 0 ; i < formp->arr.arrp[3].arr.numelems ; i++) {
			LocInfo xloc ;
			st = AcExecForm (&(formp->arr.arrp[3].arr.arrp[i]), &xloc) ;
			CHECKST_EX(st) ;
			numargs++ ;
		    }
		    varstack.SetFramePtr (newfp) ;
		    st = CallContext (calladr, numargs, NULL, NULL, &retloc) ;
		}
		CHECKST_EX(st) ;
		x = varstack.VPop () ;
		break ;
	      case V_STRING:
		calladr = SetContext (entp->sptr, 0) ;
		newfp = varstack.StackPtr () ;
		if (loc.vtype == VT_Method || loc.vtype == VT_Hiden) {
		    varstack.TmPush (*loc.instancep) ;
		    numargs++ ;
		    for (i = 0 ; i < formp->arr.arrp[3].arr.numelems ; i++) {
			LocInfo xloc ;
			st = AcExecForm (&(formp->arr.arrp[3].arr.arrp[i]), &xloc) ;
			CHECKST_EX(st) ;
			numargs++ ;
		    }
		    varstack.SetFramePtr (newfp) ;
		    st = CallContext (calladr, numargs, loc.instancep, loc.classp, &retloc) ;
		} else {
		    for (i = 0 ; i < formp->arr.arrp[3].arr.numelems ; i++) {
			LocInfo xloc ;
			st = AcExecForm (&(formp->arr.arrp[3].arr.arrp[i]), &xloc) ;
			CHECKST_EX(st) ;
			numargs++ ;
		    }
		    varstack.SetFramePtr (newfp) ;
		    st = CallContext (calladr, numargs, NULL, NULL, &retloc) ;
		}
		CHECKST_EX(st) ;
		x = varstack.VPop () ;
		break ;
	      case V_BIMETHOD:
		newfp = varstack.StackPtr () ;
		varstack.TmPush (*loc.instancep) ;
		numargs++ ;
		for (i = 0 ; i < formp->arr.arrp[3].arr.numelems ; i++) {
		    LocInfo xloc ;
		    st = AcExecForm (&(formp->arr.arrp[3].arr.arrp[i]), &xloc) ;
		    CHECKST_EX(st) ;
		    numargs++ ;
		}
		varstack.SetFramePtr (newfp) ;
		st = (*entp->bimethod) (this, loc.instancep,
					varstack.FrameSize(), varstack.GetFrame(), &x) ;
		if (st < 0)
		    return SetErrorCode (METHODCALL_ERROR, ECTX(formp)) ;
		break ;
	      case V_BUILTIN:
		newfp = varstack.StackPtr () ;
		for (i = 0 ; i < formp->arr.arrp[3].arr.numelems ; i++) {
		    LocInfo xloc ;
		    st = AcExecForm (&(formp->arr.arrp[3].arr.arrp[i]), &xloc) ;
		    CHECKST_EX(st) ;
		    numargs++ ;
		}
		varstack.SetFramePtr (newfp) ;
		st = (*entp->entry) (this, varstack.FrameSize(), varstack.GetFrame(), &x) ;
		if (st < 0)
		    return SetErrorCode (FUNCALL_ERROR, ECTX(formp)) ;
		break ;
	      case V_ARRAY:
	      case V_ACODE:
		if ((AcInst)entp->arr.arrp[0].intval == A_CreatorDef) {
		    if (!loc.classp || !loc.classp->IsClass) {
			fprintf (stderr, "Invalid consturctor call ??????\n") ;
			return -1 ;
		    } else{
			st = AcStartCreator (formp, entp, &loc) ; CHECKST_EX(st) ;
		    }
		} else {
		    acfuncp = varstack.TmPush (entp->arr) ;
		    newfp = varstack.StackPtr () ;
		    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden) {
			varstack.TmPush (*loc.instancep) ;
			numargs++ ;
			for (i = 0 ; i < formp->arr.arrp[3].arr.numelems ; i++) {
			    LocInfo xloc ;
			    st = AcExecForm (&(formp->arr.arrp[3].arr.arrp[i]), &xloc) ;
			    CHECKST_EX(st) ;
			    numargs++ ;
			}
			varstack.SetFramePtr (newfp) ;
			st = AcCallForm (acfuncp, numargs, loc.instancep, loc.classp, &retloc) ;
		    } else {
			for (i = 0 ; i < formp->arr.arrp[3].arr.numelems ; i++) {
			    LocInfo xloc ;
			    st = AcExecForm (&(formp->arr.arrp[3].arr.arrp[i]), &xloc) ;
			    CHECKST_EX(st) ;
			    numargs++ ;
			}
			varstack.SetFramePtr (newfp) ;
			st = AcCallForm (acfuncp, numargs, NULL, NULL, &retloc) ;
		    }
		}
		CHECKST_EX(st) ;
		x = varstack.VPop () ;
		break ;
	      default:
		return SetErrorCode (INVALID_EXPRESSION, ECTX(formp)) ;
	    }
	    if (constThis) {
		varstack.LPop () ;
		constThis = 0 ;
	    }
	    // t[|C^AX^bN|C^̕A
	    if (st == THROW) {
		Value v ;
		v = ECTX(formp) ;
		varstack.Top()->instance->SetValue (symbolTable->GetSymbol("pos"), v) ;
	    } else if (st != EXIT) {
		varstack.SetFramePtr (framepsave) ;
		varstack.SetStackPtr (shrink) ;
		// ^[lZbg 
		varstack.TmPush (x) ;
		*locp = retloc ;
		return NORMAL ;
	    }
	    return st ;
	}
      case A_TryCatch:
	{
	    int spsave = varstack.StackPtr () ;
	    Value *retp ;
	    Value *einstance ;
	    LocInfo loc ;
	    int do_finally = 0 ;

	    // try p[g̎s
	    formp = varstack.LPush (form->arr) ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ;
	    retp = varstack.Top () ;
	    if (st == 0)
		do_finally = 1 ;
	    if (st < 0 || st == EXIT) {
		varstack.LPop () ;
		return st ;
	    }
	    varstack.SetStackPtr (spsave) ;
	    varstack.TmPush (*retp) ;
	    if (st == THROW) {
		// OꂽBŃLb`邩`FbNB 
		// ܂AO Exception NX̔hNX`FbN
		einstance = varstack.Top () ;
		if (!einstance->IsInstance()
		    || !einstance->instance->classp->FindClass (exceptionClassSym, &vp))
		{
		    return SetErrorCode (INVALID_OBJECT, ECTX(formp)) ;
		}
		// Ŏ󂯎O`FbN
		Value *catchlist = &formp->arr.arrp[3] ;
		for (i = 0 ; i < catchlist->arr.numelems ; i++) {
		    Value *ex_class_and_var = &catchlist->arr.arrp[i].arr.arrp[0] ;
		    Value *class_name = &ex_class_and_var->arr.arrp[0] ;
		    Value *var_name    = &ex_class_and_var->arr.arrp[1] ;
		    Value *exception_class ;
		    Value *body = &catchlist->arr.arrp[i].arr.arrp[1] ;
		    Value *real_classp ;

		    // ONX̃`FbN
		    exception_class = varstack.LookupVar (class_name->symbol) ;
		    if (!exception_class || !exception_class->IsClass())
			return SetErrorCode (UNDEFINED_CLASS, ECTX(formp)) ;
		    // OCX^XAŎw肳ĂONXA 
		    // 邢́A̔hNX`FbN 
		    if (!einstance->instance->FindClass
				(NULL,
				 exception_class->classp->classnamep->sym,
				 &real_classp))
		    {
			// Ŏ󂯎Oł͂ȂB𒲂ׂ
			continue ;
		    }
		    // 󂯎OCX^Xɕϐ{B
		    varstack.SetSymSimple (spsave, var_name->symbol) ;
		    loc.Reset () ;
		    st = AcExecForm (body, &loc) ;
		    CHECKST_EX(st) ;
		    do_finally = 1 ;
		    retp = varstack.Top () ;
		    varstack.SetStackPtr (spsave) ;
		    varstack.TmPush (*retp) ;
		    break ;
		}
	    }
	    if (formp->arr.arrp[4].IsArray() && do_finally) {
		// Execute 'finally' block
		varstack.Pop () ;
		loc.Reset () ;
		st = AcExecForm (&formp->arr.arrp[4], &loc) ;
	    }
	    varstack.LPop () ;
	    return st ;
	}
	break ;
      case A_NoMember:
	vp = varstack.LookupVar (form->arr.arrp[2].symbol) ;
	locp->vtype = VT_NormalVar ;
	if (!vp) {
	    SetErrorCode (UNDEFINED_IDENTIFIER, ECTX(form)) ;
	    return -1 ;
	}
	goto fix_loc_info ;
      case A_Var:
	var_hit++ ;
	vp = LookupVar (form->arr.arrp[2].symbol,
			form->arr.arrp[3].intval,
			(classptr) ? classptr->classp : NULL,
			&locp->vtype, &is_member, &cp, &idx) ;
	if (!vp)
	    return SetErrorCode (UNDEFINED_IDENTIFIER, ECTX(form)) ;
	if (!is_member) {
	    if (vp < globalLim) {
		form->arr.arrp[0] = (int)A_GVar ;
		form->arr.arrp[2] = (Value *)vp ;
	    }
	}
	locp->varp = vp ;
	if (vp->IsClass()) {
	    locp->instancep = vp ;
	    locp->classp = vp ;
	} else if (vp->IsInstance()) {
	    locp->instancep = vp ;
	    locp->classp = vp->instance->classp->classnamep ;
	} else if (locp->vtype == VT_Method || locp->vtype == VT_BiMethod
		   || locp->vtype == VT_Hiden || locp->vtype == VT_BiHiden)
	{
	    if (!locp->instancep) {
		if (!thisptr)
		    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
		locp->instancep = thisptr ;
		locp->classp    = cp ;
	    } else {
		locp->classp = locp->instancep->classp->classnamep ;
	    }
	} else {
	    locp->instancep = NULL ;
	    locp->classp = NULL ;
	}
	varstack.TmPush (*vp) ;
	return 0 ;
      case A_GVar:
	gvar_hit++ ;
	vp = form->arr.arrp[2].location ;
	locp->vtype = VT_NormalVar ;
	goto fix_loc_info ;
      case A_LVar:
	lvar_hit++ ;
	vp = varstack.FrameRelative (form->arr.arrp[2].intval + argcnt) ;
	locp->vtype = VT_NormalVar ;
      fix_loc_info:
	locp->varp = vp ;
	if (vp->IsClass()) {
	    locp->instancep = vp ;
	    locp->classp = vp ;
	} else if (vp->IsInstance()) {
	    locp->instancep = vp ;
	    if (thisptr && (thisptr->instance == vp->instance))
		locp->classp = classptr ;
	    else
		locp->classp = vp->instance->classp->classnamep ;
	} else if (locp->vtype == VT_Method || locp->vtype == VT_BiMethod
		   || locp->vtype == VT_Hiden || locp->vtype == VT_BiHiden)
	{
	    if (!locp->instancep) {
		if (!thisptr)
		    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
		locp->instancep = thisptr ;
		locp->classp    = classptr ;
	    } else {
		locp->classp = locp->instancep->classp->classnamep ;
	    }
	} else {
	    locp->instancep = NULL ;
	    locp->classp = NULL ;
	}
	varstack.TmPush (*vp) ;
	return 0 ;
      case A_ThisMember:
	this_mem_hit++ ;
	locp->instancep = thisptr ;
	locp->classp = classptr ;

	cp = form->arr.arrp[2].location ;
	locp->vtype = (VarType)form->arr.arrp[3].intval ;
	vp = thisptr->instance->FindMemberByIndex
				    (cp, locp->vtype, form->arr.arrp[4].intval) ;
	if (!vp)
	    return SetErrorCode (UNDEFINED_MEMBER, ECTX(form)) ;
	locp->varp = vp ;
	if (vp->IsClass()) {
	    locp->classp = vp ;
	} else if (vp->IsInstance()) {
	    locp->classp = vp->instance->classp->classnamep ;
	} else if (locp->vtype != VT_Method && locp->vtype != VT_BiMethod
		   && locp->vtype != VT_Hiden && locp->vtype != VT_BiHiden)
	{
	    locp->classp = NULL ;
	    locp->instancep = NULL ;
	}
	varstack.TmPush (*vp) ;
	return 0 ;
      case A_SelectMember:
	formp = varstack.TmPush (form->arr) ;
	st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	if (!loc.varp) {
	    loc.varp = varstack.LPush (*varstack.Top ()) ;
	}
	varstack.TmPop () ;
	vp = loc.varp ;
	if (vp->IsInstance()) {
	    //    0          1        2      3        
	    // [SelectMember <CONTXT> [Expr] "member"]
	    //
	    loc.instancep = vp ;
	    selmem_hit++ ;
	    loc.varp = vp->instance->FindMember
		(loc.classp ? (loc.classp->classp): vp->instance->classp,
		 formp->arr.arrp[3].symbol,
		 &cp,
		 &vtype,
		 &idx) ;
	    if (!loc.varp)
		return SetErrorCode (UNDEFINED_MEMBER, ECTX(form)) ;
	    formp->arr.arrp[4] = cp ;
	    formp->arr.arrp[5] = (int)vtype ;
	    formp->arr.arrp[6] = (int)idx ;
	    loc.vtype = vtype ;

	    if ((vtype == VT_Private || vtype == VT_Hiden || vtype == VT_BiHiden)
		&& (!thisptr || vp->instance != thisptr->instance))
	    {
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(formp)) ;
	    }
	    if (loc.varp->IsInstance()) {
		loc.instancep = loc.varp ;
		// loc.classp = loc.varp->instance->classp->classnamep ;
		loc.classp = cp ;
	    } else if (loc.varp->IsClass()) {
		loc.instancep = loc.varp ;
		loc.classp = loc.varp ;
	    } else if (vtype != VT_Method && vtype != VT_BiMethod
		       && vtype != VT_Hiden && vtype != VT_BiHiden)
	    {
		loc.classp = NULL ;
		loc.instancep = NULL ;
	    } else {
		loc.classp = cp ;
	    }
	} else if (vp->IsClass()) {
	    loc.classp = vp->classp->classnamep ;
	    loc.varp   = vp->classp->FindMethodOrStatic
				(formp->arr.arrp[3].symbol, &vtype) ;
	    if (!loc.varp)
		return SetErrorCode (UNDEFINED_MEMBER, ECTX(form)) ;
	    loc.vtype = vtype ;
	    if (loc.varp->IsInstance()) {
		loc.classp = loc.varp->instance->classp->classnamep ;
	    } else if (loc.varp->IsClass()) {
		loc.instancep = loc.varp ;
		loc.classp = loc.varp ;
	    } else if (vtype != VT_Method && vtype != VT_BiMethod
		       && vtype != VT_Hiden && vtype != VT_BiHiden)
	    {
		loc.classp = NULL ;
		loc.instancep = NULL ;
	    }
	} else {
	    if (!loc.classp)
		loc.classp = FixSpecialClass (loc.varp) ;
	    if (!loc.classp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    loc.instancep = loc.varp ;
	    loc.varp = loc.classp->classp->FindMethod
		(formp->arr.arrp[3].symbol,
		 &loc.vtype) ;
	}
	if (!loc.varp)
	    return SetErrorCode (UNDEFINED_MEMBER, ECTX(form)) ;
	if (loc.varp->IsInstance()) {
	    loc.instancep = loc.varp ;
	    loc.classp = loc.varp->instance->classp->classnamep ;
	}
	*locp = loc ;
	*((Value *)varstack.Top()) = *loc.varp ;
	return 0 ;
      case A_SelectClass:
	{
	    formp = varstack.TmPush (form->arr) ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp) {
		loc.varp = varstack.LPush (*varstack.Top()) ;
	    }
	    if (!loc.classp) {
		loc.classp = FixSpecialClass (loc.varp) ;
	    }
	    varstack.TmPop () ;

	    Value *v ;
	    v = varstack.LookupVar (formp->arr.arrp[3].symbol,
				    formp->arr.arrp[4].intval) ;
	    if (!v || !v->IsClass())
		return SetErrorCode (UNDEFINED_CLASS, ECTX(form)) ;

	    if (!loc.varp->IsInstance()) {
		Class *cp ;
		cp = loc.classp->classp->FindClass (formp->arr.arrp[3].symbol,
						    &loc.classp) ;
		if (!cp)
		    return SetErrorCode (UNDEFINED_CLASS, ECTX(form)) ;
		loc.classp = cp->classnamep ;
	    } else {
		loc.classp = loc.varp->instance->SelectClass (v->classp) ;
	    }
	    if (!loc.classp)
		return SetErrorCode (UNDEFINED_CLASS, ECTX(form)) ;
	    *locp = loc ;
	    *((Value *)varstack.Top()) = *loc.classp ;
	}
	return 0 ;
      case A_Const:
	locp->Reset () ;
	vp = varstack.TmPush (form->arr.arrp[2]) ;
	return 0 ;
      case A_Array:
	formp = varstack.TmPush (form->arr) ;
	locp->Reset () ;
	st = AcMakeArray (&formp->arr.arrp[2]) ;
	CHECKST_EX (st) ;
	retv = varstack.VPop () ;
	varstack.TmPop () ;
	varstack.TmPush (retv) ;
	locp->classp = arrayClass ;
	return st ;
      case A_ArrIndex:
	formp = varstack.TmPush (form->arr) ;  // +1 
	{
	    Value *basep ;
	    
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ; // +2
	    basep = varstack.Top() ;

	    LocInfo iloc ;
	    if (!basep->IsArray()) {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    st = AcExecForm (&formp->arr.arrp[3], &iloc) ;		// +3
	    idx = (varstack.VPop()).ForceInt () ;			// +2
	    if (idx < 0 || idx >= basep->arr.numelems) {
		return ThrowSystemException (arrayOutOfBoundsException, ECTX(form)) ;
	    }
	    locp->Reset () ;
	    locp->varp = &basep->arr.arrp[idx] ;
	    if (locp->varp->IsInstance()) {
		locp->instancep = locp->varp ;
		locp->classp = locp->varp->instance->classp->classnamep ;
	    } else if (locp->varp->IsClass()) {
		locp->classp = locp->varp->classp->classnamep ;
	    }
	    varstack.TmPop (2) ;
	    varstack.TmPush (*locp->varp) ;
	}
	return 0 ;
      case A_Not:
	st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	vp = varstack.Top () ;
	*(Value *)vp = (int)((vp->True()) ? I_FALSE : I_TRUE) ;
	return st ;
      case A_Negate:
	st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	vp = varstack.Top () ;
	if (vp->IsInteger())    vp->intval = -1 * vp->intval ;
	else if (vp->IsReal())  vp->value = -1.0 * vp->value ;
	else
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	return st ;
      case A_BitRev:
	st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	vp = varstack.Top () ;
	if (vp->IsInteger())    vp->intval = ~vp->intval ;
	else
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	return st ;
      case A_Block:
	formp = varstack.LPush (form->arr) ;
        {
	    int spsave = varstack.StackPtr () ;
	    varstack.TmPush () ;
	    for (i = 2 ; i < formp->arr.numelems ; i++) {
		loc.Reset () ;
		varstack.TmPop () ;
		st = AcExecForm (&formp->arr.arrp[i], &loc) ;
		CHECKST_EX(st) ;
		if (st != NORMAL) {
		    break ;
		}
	    }
	    *locp = loc ;
	    retv = varstack.VPop () ;
	    varstack.LPop1 () ;
	    varstack.SetStackPtr (spsave) ;
	    varstack.TmPush (retv) ;
	    return st ;
	}
      case A_SLBlock:
	formp = varstack.LPush (form->arr) ;
        {
	    varstack.TmPush () ;
	    for (i = 2 ; i < formp->arr.numelems ; i++) {
		loc.Reset () ;
		varstack.TmPop () ;
		st = AcExecForm (&formp->arr.arrp[i], &loc) ;
		CHECKST_EX(st) ;
		if (st != NORMAL) {
		    break ;
		}
	    }
	    *locp = loc ;
	    varstack.LPop1 () ;
	    // retv = varstack.VPop () ;
	    // varstack.SetStackPtr (spsave) ;
	    // *((Value *)varstack.Top()) = retv ;
	    return st ;
	}
      case A_Plus:
	formp = varstack.TmPush (form->arr) ;
        {
	    Value *x1, *x2, xx ;
	    int st ;
	    LocInfo loc ;
	    ValType vt1 ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    x1 = varstack.Top () ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x2 = varstack.Top () ;
	    if ((vt1 = x1->type) == V_INTEGER) {
		xx = x1->intval + x2->ForceInt() ;
	    } else if (vt1 == V_REAL) {
		xx = x1->value + x2->ForceReal() ;
	    } else if (vt1 == V_STRING) {
		int len1 = strlen(x1->sval()) ;
		int len2 = strlen(x2->ForceString()) ;
		char *snew ;

		snew = heap_area.AllocSegment (len1 + len2 + 1, &varstack) ;
		if (!snew)
		    return SetErrorCode (NOT_ENOUGH_MEMORY) ;
		strcpy (snew, x1->sptr) ;
		strcat (snew, x2->ForceString()) ;
		xx = snew ;
	    } else if (vt1 == V_ARRAY && x2->type == V_ARRAY) {
		if (heap_area.ConsArray (x1, x2, &varstack, &xx) < 0)
		    return SetErrorCode (NOT_ENOUGH_MEMORY) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    varstack.TmPop (3) ;
	    locp->Reset () ;
	    varstack.TmPush (xx) ;
	    return st ;
	}
      case A_Minus:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value *x1, *x2, xx ;
	    int st ;
	    LocInfo loc ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    x1 = varstack.Top () ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x2 = varstack.Top () ;
	    if (x1->IsInteger()) {
		xx = x1->intval - x2->ForceInt() ;
	    } else if (x1->IsReal()) {
		xx = x1->value - x2->ForceReal() ;
	    } else if (x1->IsArray ()) {
		if (heap_area.RemoveArray (x1, x2, &varstack, &xx) < 0) {
		    return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(form)) ;
		}
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    varstack.TmPop (3) ;
	    locp->Reset () ;
	    varstack.TmPush (xx) ;
	    return st ;
	}
      case A_Mult:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value *x1, *x2, xx ;
	    int st ;
	    LocInfo loc ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    x1 = varstack.Top () ;
	    *locp = loc ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x2 = varstack.Top () ;
	    if (x1->IsInteger()) {
		xx = x1->intval * x2->ForceInt() ;
	    } else if (x1->IsReal()) {
		xx = x1->value * x2->ForceReal() ;
	    } else if (x1->IsString ()) {
		char *stmp ;
		int slen = strlen (x1->sptr) ;
		int nums = x2->ForceInt () ;
		stmp = heap_area.AllocSegment (slen * nums + 1, &varstack) ;
		if (!stmp)
		    return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(form)) ;
		stmp[0] = '\0' ;
		int ii, jj ;
		for (ii = jj = 0 ; ii < nums ; ii++, jj += slen)
		    strcpy (stmp + jj, x1->sptr) ;
		xx = stmp ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    varstack.TmPop (3) ;
	    locp->Reset () ;
	    varstack.TmPush (xx) ;
	    return st ;
	}
      case A_Div:
	formp = varstack.TmPush (form->arr) ;
        {
	    Value *x1, *x2, xx ;
	    int st ;
	    LocInfo loc ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    x1 = varstack.Top () ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x2 = varstack.Top () ;
	    if (x1->IsInteger()) {
		int ix2 ;
		if ((ix2 = x2->ForceInt()) == 0)
		    return ThrowSystemException (divisionByZeroException, ECTX(form)) ;
		xx = x1->intval / ix2 ;
	    } else if (x1->IsReal()) {
		double rx2 ;
		if ((rx2 = x2->ForceReal()) == 0.0)
		    return ThrowSystemException (divisionByZeroException, ECTX(form)) ;
		xx = x1->value / rx2 ;
	    } else if (x1->IsArray()) {
		int idx = x2->ForceInt() ;

		if (idx < 0 || idx >= x1->arr.numelems) {
		    return ThrowSystemException (arrayOutOfBoundsException, ECTX(form)) ;
		}
		if (heap_area.RemoveArray (x1, x2->ForceInt(), &varstack, &xx) < 0)
		    return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(form)) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    varstack.TmPop (3) ;
	    varstack.TmPush (xx) ;
	    locp->Reset () ;
	    return 0 ;
	}
      case A_Mod:
	formp = varstack.TmPush (form->arr) ;
        {
	    Value *x1, *x2, xx ;
	    int st ;
	    LocInfo li ;

	    st = AcExecForm (&formp->arr.arrp[2], locp) ; CHECKST_EX(st) ;
	    x1 = varstack.Top () ;
	    st = AcExecForm (&formp->arr.arrp[3], &li ) ; CHECKST_EX(st) ;
	    x2 = varstack.Top () ;
	    if (x1->IsInteger()) {
		int ix2 ;
		if ((ix2 = x2->ForceInt()) == 0)
		    return ThrowSystemException (divisionByZeroException, ECTX(form)) ;
		xx = x1->intval % ix2 ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    varstack.TmPop (3) ;
	    locp->Reset () ;
	    varstack.TmPush (xx) ;
	    return 0 ;
	}
      case A_Rsh:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value *x1, *x2, xx ;
	    int st ;
	    LocInfo loc ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    x1 = varstack.Top () ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x2 = varstack.Top () ;

	    if (x1->IsInteger()) {
		xx = x1->intval >> x2->ForceInt () ;
	    } else if (x1->IsArray()) {
		Value *ap ;
		int nelems = x1->arr.numelems - x2->ForceInt () ;
		if (nelems > x1->arr.numelems) {
		    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
		}
		if (nelems < 0)
		    nelems = 0 ;
		if (heap_area.AllocArray (nelems, &varstack, &ap) < 0)
		    return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(form)) ;
		for (i = 0 ; i < nelems ; i++) {
		    heap_area.CopyValue (&x1->arr.arrp[i], &varstack, &ap[i]) ;
		}
		xx.InitArr (ap, nelems) ;
	    }
	    varstack.TmPop (3) ;
	    locp->Reset () ;
	    varstack.TmPush (xx) ;
	    return st ;
	}
      case A_Lsh:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value *x1, *x2, xx ;
	    int st ;
	    LocInfo loc ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    x1 = varstack.Top () ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x2 = varstack.Top () ;

	    if (x1->IsInteger() && (x2->IsNumber() || x2->IsString())) {
		xx = x1->intval << x2->ForceInt () ;
	    } else if (x1->IsArray()) {
		if (heap_area.AppendArray (x1, x2, &varstack, &xx) < 0)
		    return SetErrorCode (NOT_ENOUGH_MEMORY) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    varstack.TmPop (3) ;
	    locp->Reset () ;
	    varstack.TmPush (xx) ;
	    return st ;
	}
      case A_Band:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value *x1, *x2, xx ;
	    int st ;
	    LocInfo loc ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    x1 = varstack.Top () ;
	    *locp = loc ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x2 = varstack.Top () ;
	    if (x1->IsInteger() && x2->IsInteger ()) {
		xx = (x1->intval & x2->intval) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    varstack.TmPop (3) ;
	    locp->Reset () ;
	    varstack.TmPush (xx) ;
	    return st ;
	}
      case A_Bor:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value *x1, *x2, xx ;
	    int st ;
	    LocInfo loc ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    x1 = varstack.Top () ;
	    *locp = loc ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x2 = varstack.Top () ;
	    if (x1->IsInteger() && x2->IsInteger ()) {
		xx = (x1->intval | x2->intval) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    varstack.TmPop (3) ;
	    locp->Reset () ;
	    varstack.TmPush (xx) ;
	    return st ;
	}
      case A_Xor:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value *x1, *x2, xx ;
	    int st ;
	    LocInfo loc ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    x1 = varstack.Top () ;
	    *locp = loc ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x2 = varstack.Top () ;
	    if (x1->IsInteger() && x2->IsInteger ()) {
		xx = (x1->intval ^ x2->intval) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    varstack.TmPop (3) ;
	    locp->Reset () ;
	    varstack.TmPush (xx) ;
	    return st ;
	}
      case A_Eq:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value x1, x2, xx ;
	    int st ;
	    LocInfo li ;

	    st = AcExecForm (&formp->arr.arrp[2], &li) ; CHECKST_EX(st) ;
	    st = AcExecForm (&formp->arr.arrp[3], &li) ; CHECKST_EX(st) ;
	    x2 = varstack.VPop () ;
	    x1 = varstack.VPop () ;
	    xx = (int)(x1 == x2) ;
	    locp->Reset () ;
	    *((Value *)varstack.Top()) = xx ;
	    return st ;
	}
      case A_Neq:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value x1, x2, xx ;
	    int st ;
	    LocInfo li ;

	    st = AcExecForm (&formp->arr.arrp[2], &li) ; CHECKST_EX(st) ;
	    st = AcExecForm (&formp->arr.arrp[3], &li) ; CHECKST_EX(st) ;
	    x2 = varstack.VPop () ;
	    x1 = varstack.VPop () ;
	    xx = (int)(x1 != x2) ;
	    locp->Reset () ;
	    *((Value *)varstack.Top()) = xx ;
	    return st ;
	}
      case A_Lt:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value x1, x2, xx ;
	    int st ;
	    LocInfo li ;

	    st = AcExecForm (&formp->arr.arrp[2], locp) ; CHECKST_EX(st) ;
	    st = AcExecForm (&formp->arr.arrp[3], &li ) ; CHECKST_EX(st) ;
	    x2 = varstack.VPop () ;
	    x1 = varstack.VPop () ;
	    xx = (int)(x1 < x2) ;
	    locp->Reset () ;
	    *((Value *)varstack.Top()) = xx ;
	    return st ;
	}
      case A_Gt:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value x1, x2, xx ;
	    int st ;
	    LocInfo li ;

	    st = AcExecForm (&formp->arr.arrp[2], locp) ; CHECKST_EX(st) ;
	    st = AcExecForm (&formp->arr.arrp[3], &li ) ; CHECKST_EX(st) ;
	    x2 = varstack.VPop () ;
	    x1 = varstack.VPop () ;
	    xx = (int)(x1 > x2) ;
	    locp->Reset () ;
	    *((Value *)varstack.Top()) = xx ;
	    return st ;
	}
      case A_Le:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value x1, x2, xx ;
	    int st ;
	    LocInfo li ;

	    st = AcExecForm (&formp->arr.arrp[2], locp) ; CHECKST_EX(st) ;
	    st = AcExecForm (&formp->arr.arrp[3], &li ) ; CHECKST_EX(st) ;
	    x2 = varstack.VPop () ;
	    x1 = varstack.VPop () ;
	    xx = (int)(x1 <= x2) ;
	    locp->Reset () ;
	    *((Value *)varstack.Top()) = xx ;
	    return st ;
	}
      case A_Ge:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value x1, x2, xx ;
	    int st ;
	    LocInfo li ;

	    st = AcExecForm (&formp->arr.arrp[2], &li) ; CHECKST_EX(st) ;
	    st = AcExecForm (&formp->arr.arrp[3], &li) ; CHECKST_EX(st) ;
	    x2 = varstack.VPop () ;
	    x1 = varstack.VPop () ;
	    xx = (int)(x1 >= x2) ;
	    locp->Reset () ;
	    *((Value *)varstack.Top()) = xx ;
	    return st ;
	}
      case A_Or:
	formp = varstack.TmPush (form->arr) ;
	{
	    int st ;
	    LocInfo li ;
	    Value   x ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], locp) ; CHECKST_EX(st) ;
	    x = varstack.VPop () ;
	    if (x.True()) {
		varstack.TmPop () ;
		varstack.TmPush(1) ;
		return st ;
	    }
	    st = AcExecForm (&formp->arr.arrp[3], &li ) ; CHECKST_EX(st) ;
	    x = varstack.VPop () ;
	    *((Value *)varstack.Top()) = x ;
	    locp->Reset () ;
	    return st ;
	}
      case A_And:
	formp = varstack.TmPush (form->arr) ;
	{
	    int st ;
	    LocInfo li ;
	    Value   x ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], locp) ; CHECKST_EX(st) ;
	    x = varstack.VPop () ;
	    if (!x.True()) {
		*((Value *)varstack.Top()) = 0 ;
		return st ;
	    }
	    st = AcExecForm (&formp->arr.arrp[3], &li ) ; CHECKST_EX(st) ;
	    x = varstack.VPop () ;
	    *((Value *)varstack.Top()) = (x.True()) ? 1 : 0 ;
	    locp->Reset () ;
	    return st ;
	}
      case A_New:
	{
	    Value xclass ;
	    st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    xclass = varstack.VPop () ;
	    if (!xclass.IsClass ())
		return SetErrorCode (UNDEFINED_CLASS, ECTX(form)) ;
	    if (IsSpecialClass (xclass.classp->classnamep))
		return SetErrorCode (INVALID_NEW, ECTX(form)) ;
	    Instance *ip = CreateInstance (xclass.classp) ;
	    if (!ip)
		return -1 ;
	    varstack.TmPush (ip) ;
	}
	return st ;
      case A_Assign:
	formp = varstack.LPush (form->arr) ;
        {
	    Value *xp ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    varstack.TmPop () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    xp = varstack.Top () ;
	    if (xp->IsString () || xp->IsARRAY()) {
		Value rvalue ;
		heap_area.CopyValue (xp, &varstack, &rvalue) ;
		*lvalp->location = rvalue ;
	    } else {
		*lvalp->location = *xp ;
	    }
	    locp->Reset () ;
	    locp->varp = lvalp->location ;
	    varstack.LPop (2) ;
	}
	return st ;
      case A_RefAssign:
	formp = varstack.LPush (form->arr) ;
        {
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    varstack.TmPop () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    *lvalp->location = *(Value *)varstack.Top() ;
	    locp->Reset () ;
	    *locp = loc ;
	    varstack.LPop (2) ;
	}
	return st ;
      case A_AddAssign:
	formp = varstack.LPush (form->arr) ;
        {
	    Value *x0, *x ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;
	    Value rval ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden)
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(form)) ;
	    x0 = varstack.Top () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x = varstack.Top () ;

	    if (x0->IsReal()) {
		rval = x0->value + x->ForceReal () ;
	    } else if (x0->IsInteger()) {
		rval = x0->intval + x->ForceInt () ;
	    } else if (x0->IsString()) {
		char *s, *stmp ;

		s = x->ForceString () ;
		stmp = heap_area.AllocSegment (strlen (x0->sval()) + strlen(s) +1,
					       &varstack) ;
		if (!stmp)
		    return SetErrorCode (NOT_ENOUGH_MEMORY) ;
		strcpy (stmp, lvalp->location->sptr) ;
		strcat (stmp, x->sptr) ;
		rval = stmp ;
	    } else if (x0->IsArray () && x->IsArray()) {
		if (heap_area.ConsArray (x0, x, &varstack, &rval) < 0)
		    return SetErrorCode (NOT_ENOUGH_MEMORY) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    *(lvalp->location) = rval ;
	    locp->varp = lvalp->location ;

	    varstack.LPop (2) ;
	    varstack.TmPop (2) ;
	    varstack.TmPush (rval) ;
	}
	return st ;
      case A_SubAssign:
	formp = varstack.LPush (form->arr) ;
	{
	    Value *x0, *x ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;
	    Value rval ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden)
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(form)) ;
	    x0 = varstack.Top () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x = varstack.Top () ;

	    if (x0->IsReal()) {
		rval = x0->value - x->ForceReal () ;
	    } else if (x0->IsInteger()) {
		rval = x0->intval - x->ForceInt () ;
	    } else if (x0->IsArray()) {
		if (heap_area.RemoveArray (x0, x, &varstack, &rval) < 0) {
		    return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(form)) ;
		}
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    *(lvalp->location) = rval ;
	    locp->varp = lvalp->location ;

	    varstack.TmPop (2) ;
	    varstack.TmPush (rval) ;
	    varstack.LPop (2) ;
	}
	return 0 ;
      case A_LshAssign:
	formp = varstack.LPush (form->arr) ;
	{
	    Value *x0, *x ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;
	    Value rval ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden)
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(form)) ;
	    x0 = varstack.Top () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x = varstack.Top () ;

	    if (x0->IsInteger()) {
		rval = x0->intval << x->ForceInt () ;
	    } else if (x0->IsArray()) {
		if (heap_area.AppendArray (x0, x, &varstack, &rval) < 0)
		    return SetErrorCode (NOT_ENOUGH_MEMORY) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    *(lvalp->location) = rval ;
	    locp->varp = lvalp->location ;

	    varstack.TmPop (2) ;
	    varstack.TmPush (rval) ;
	    varstack.LPop (2) ;
	}
	return st ;
      case A_RshAssign:
	formp = varstack.LPush (form->arr) ;
	{
	    Value *x0, *x ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;
	    Value rval ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden)
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(form)) ;
	    x0 = varstack.Top () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x = varstack.Top () ;

	    if (x0->IsInteger()) {
		rval = x0->intval >> x->ForceInt () ;
	    } else if (x0->IsArray()) {
		Value *ap ;
		int nelems = x0->arr.numelems - x->ForceInt ();
		if (nelems < 0)
		    nelems = 0 ;
		if (heap_area.AllocArray (nelems, &varstack, &ap) < 0)
		    return SetErrorCode (NOT_ENOUGH_MEMORY) ;
		for (i = 0 ; i < nelems ; i++)
		    heap_area.CopyValue (&x0->arr.arrp[i], &varstack, &ap[i]) ;
		rval.InitArr (ap, nelems) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    *(lvalp->location) = rval ;
	    locp->varp = lvalp->location ;

	    varstack.TmPop (2) ;
	    varstack.TmPush (rval) ;
	    varstack.LPop (2) ;
	}
	return st ;
      case A_MultAssign:
	formp = varstack.LPush (form->arr) ;
	{
	    Value *x0, *x ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;
	    Value rval ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden)
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(form)) ;
	    x0 = varstack.Top () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x = varstack.Top () ;

	    if (x0->IsReal()) {
		rval = x0->value * x->ForceReal () ;
	    } else if (x0->IsInteger()) {
		rval = x0->intval * x->ForceInt () ;
	    } else if (x0->IsArray()) {
		if (heap_area.AddArray (x0, x, &varstack, &rval) < 0)
		    return SetErrorCode (NOT_ENOUGH_MEMORY) ;
	    } else if (x0->IsString()) {
		char *stmp ;
		int slen = strlen (x0->sptr) ;
		int nums = x->ForceInt () ;
		stmp = heap_area.AllocSegment (slen * nums + 1, &varstack) ;
		if (!stmp)
		    SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(form)) ;
		int ii, jj ;
		for (ii = jj = 0 ; ii < nums ; ii++, jj += slen)
		    strcpy (stmp + jj, x0->sptr) ;
		rval = stmp ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    *(lvalp->location) = rval ;
	    locp->varp = lvalp->location ;

	    varstack.TmPop (2) ;
	    varstack.TmPush (rval) ;
	    varstack.LPop (2) ;
	}
	return st ;
      case A_DivAssign:
	formp = varstack.LPush (form->arr) ;
	{
	    Value *x0, *x ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;
	    Value rval ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden)
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(form)) ;
	    x0 = varstack.Top () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x = varstack.Top () ;

	    if (x0->IsReal()) {
		if (x->ForceReal() == 0.0)
		    return ThrowSystemException (divisionByZeroException, ECTX(form)) ;
		rval = x0->value / x->ForceReal () ;
	    } else if (x0->IsInteger()) {
		if (x->ForceInt() == 0)
		    return ThrowSystemException (divisionByZeroException, ECTX(form)) ;
		rval = x0->intval / x->ForceInt () ;
	    } else if (x0->IsArray ()) {
		if (heap_area.RemoveArray (x0, x->ForceInt(), &varstack, &rval) < 0)
		    return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(form)) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    *(lvalp->location) = rval ;
	    locp->varp = lvalp->location ;

	    varstack.TmPop (2) ;
	    varstack.TmPush (rval) ;
	    varstack.LPop (2) ;
	}
	return st ;
      case A_ModAssign:
	formp = varstack.LPush (form->arr) ;
	{
	    Value *x0, *x ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;
	    Value rval ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden)
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(form)) ;
	    x0 = varstack.Top () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x = varstack.Top () ;

	    if (x0->IsInteger()) {
		if (x->ForceInt() == 0)
		    return ThrowSystemException (divisionByZeroException, ECTX(form)) ;
		rval = x0->intval % x->ForceInt () ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    *(lvalp->location) = rval ;
	    locp->varp = lvalp->location ;

	    varstack.TmPop (2) ;
	    varstack.TmPush (rval) ;
	    varstack.LPop (2) ;
	}
	return 0 ;
      case A_OrAssign:
	formp = varstack.LPush (form->arr) ;
	{
	    Value *x0, *x ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;
	    Value rval ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden)
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(form)) ;
	    x0 = varstack.Top () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x = varstack.Top () ;

	    if (x0->IsInteger() && x->IsInteger()) {
		rval = (x0->intval | x->intval ) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    *(lvalp->location) = rval ;
	    locp->varp = lvalp->location ;

	    varstack.TmPop (2) ;
	    varstack.TmPush (rval) ;
	    varstack.LPop (2) ;
	}
	return 0 ;
      case A_AndAssign:
	formp = varstack.LPush (form->arr) ;
	{
	    Value *x0, *x ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;
	    Value rval ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden)
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(form)) ;
	    x0 = varstack.Top () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x = varstack.Top () ;

	    if (x0->IsInteger() && x->IsInteger()) {
		rval = (x0->intval & x->intval ) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    *(lvalp->location) = rval ;
	    locp->varp = lvalp->location ;
	    
	    varstack.TmPop (2) ;
	    varstack.TmPush (rval) ;
	    varstack.LPop (2) ;
	}
	return 0 ;
      case A_XorAssign:
	formp = varstack.LPush (form->arr) ;
	{
	    Value *x0, *x ;
	    int st ;
	    LocInfo loc ;
	    Value *lvalp ;
	    Value rval ;

	    locp->Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    if (!loc.varp)
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    if (loc.vtype == VT_Method || loc.vtype == VT_Hiden)
		return SetErrorCode (CANNOT_ACCESS_MEMBER, ECTX(form)) ;
	    x0 = varstack.Top () ;
	    lvalp = varstack.LPush (loc.varp) ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[3], &loc) ; CHECKST_EX(st) ;
	    x = varstack.Top () ;

	    if (x0->IsInteger() && x->IsInteger()) {
		rval = (x0->intval ^ x->intval ) ;
	    } else {
		return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	    }
	    *(lvalp->location) = rval ;
	    locp->varp = lvalp->location ;

	    varstack.TmPop (2) ;
	    varstack.TmPush (rval) ;
	    varstack.LPop (2) ;
	}
	return 0 ;
      case A_PPPre:
	st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	varstack.TmPop () ;
	if (!loc.varp)
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	if (loc.varp->IsInteger()) {
	    ++(loc.varp->intval) ;
	} else if (loc.varp->IsReal()) {
	    ++(loc.varp->value) ;
	} else {
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	}
	varstack.TmPush (*loc.varp) ;
	return st ;
      case A_MMPre:
	st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	varstack.TmPop () ;
	if (!loc.varp)
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	if (loc.varp->IsInteger()) {
	    --(loc.varp->intval) ;
	} else if (loc.varp->IsReal()) {
	    --(loc.varp->value) ;
	} else {
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	}
	varstack.TmPush (*loc.varp) ;
	return st ;
      case A_PPPost:
	st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	if (!loc.varp)
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	if (loc.varp->IsInteger()) {
	    (loc.varp->intval)++ ;
	} else if (loc.varp->IsReal()) {
	    (loc.varp->value)++ ;
	} else {
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	}
	return st ;
      case A_MMPost:
	st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	if (!loc.varp)
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	if (loc.varp->IsInteger()) {
	    (loc.varp->intval)-- ;
	} else if (loc.varp->IsReal()) {
	    (loc.varp->value)-- ;
	} else {
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	}
	return st ;
      case A_Return:
	if (form->arr.numelems == 3) {
	    st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    *locp = loc ;
	} else {
	    locp->Reset () ;
	    varstack.TmPush () ;
	}
	return RETURN ;
      case A_Throw:
	st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	return THROW ;
      case A_Break:
	if (form->arr.numelems == 3) {
	    st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    *locp = loc ;
	} else {
	    locp->Reset () ;
	    varstack.TmPush () ;
	}
	return BREAK ;
      case A_Exit:
	if (form->arr.numelems == 3) {
	    st = AcExecForm (&form->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    *locp = loc ;
	} else {
	    varstack.TmPush () ;
	}
	return EXIT ;
      case A_Continue:
	varstack.TmPush () ;
	return CONTINUE ;
      case A_Eval:
	formp = varstack.LPush (form->arr) ;
	{
	    Value evs ;
	    int shrink0 ;
	    int shrinkcnt ;
	    char *evalStr ;

	    shrink0 = varstack.StackPtr() ;
	    if (formp->arr.arrp[2].IsArray()) {
		st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
		evs = varstack.VPop () ;
	    } else if (formp->arr.arrp[2].IsString()) {
		evs = formp->arr.arrp[2] ;
	    } else if (formp->arr.arrp[2].IsSymbol()) {
		evs = formp->arr.arrp[2].symbol->symbol ;
	    } else {
		return SetErrorCode (SYNTAX_ERROR, ECTX(formp)) ;
	    }
	    if (evs.IsString ())
		evalStr = evs.sptr ;
	    else if (evs.IsSymbol())
		evalStr = evs.symbol->symbol ;
	    else
		return SetErrorCode (SYNTAX_ERROR, ECTX(formp)) ;

	    ExecMode exec_mode_save = exec_mode ;
	    exec_mode = ExAcInterp ;
	    Context *lastProgContext = prog ;
	    int lastLastPC = lastPC ;
	    prog = SetContext (evalStr, 0) ;
	    lastPC = 0 ;
	    shrinkcnt = varstack.StackPtr() - shrink0 ;
	    loc.Reset () ;
	    st = Program (&retv, &loc) ;
	    if (st < 0) {
		errorCtx = *prog ;
		return -1 ;
	    }
	    varstack.Shrink (shrink0, shrinkcnt) ;
/*
	    retv = *(Value *)varstack.Top () ;
*/
	    varstack.TmPush (retv) ;
	    varstack.LPop1 () ;
	    prog = lastProgContext ;
	    *locp = loc ;
	    lastPC = lastLastPC ;
	    exec_mode = exec_mode_save ;
	}
	return 0 ;
      case A_Load:
	formp = varstack.LPush (form->arr) ;
	{
	    Value vname ;
	    Value acform ;
	    char *buffer ;
	    int loaded ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    vname = varstack.VPop () ;

	    if (!vname.IsString())
		return ThrowSystemException (typeMismatchException, ECTX(formp)) ;
	    if ((loaded = LoadText (vname.sptr, &buffer)) < 0)  {
		fprintf (stderr, "%s: ", vname.sptr) ;
		return ThrowSystemException (fileLoadException, ECTX(formp)) ;
	    }
	    if (loaded) {
		varstack.TmPush (1) ;
	    } else {
		Context *lastProgContext = prog ;
		int     lastLastPC = lastPC ;

		if (!strncmp (buffer, ACODE_HEADER, strlen(ACODE_HEADER))) {
		    st = AcParseACode (buffer, &acform) ; CHECKST_EX(st) ;
		} else {
		    st = AcCompile (buffer, &acform) ; CHECKST_EX(st) ;
		}
		st = AcExecute (&acform) ;
		prog   = lastProgContext ;
		lastPC = lastLastPC ;
	    }
	    varstack.LPop1 () ;
	    return st ;
	}
      case A_Package:
	formp = varstack.LPush (form->arr) ;
	{
	    Value vname ;
	    Value acform ;
	    int loaded ;
	    char *pkgname ;

	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    vname = varstack.VPop () ;
	    if (!vname.IsString()) {
		return ThrowSystemException (typeMismatchException, ECTX(formp)) ;
	    }
	    if ((loaded = LoadPackage (vname.sptr)) < 0)  {
		return ThrowSystemException (packageLoadException, ECTX(formp)) ;
	    }
	    varstack.TmPush (1) ;
	    varstack.LPop1 () ;
	    return st ;
	}
      case A_Pause:
	Debug () ;
	return 0 ;
      case A_ArgCnt:
	varstack.TmPush (argcnt) ;
	return 0 ;
      case A_BoundArg: 
        {
	    Variable *vp ;
	    Value v ;
	    Variable *argstart = varstack.PeekStack (argindex) ;

	    loc.Reset () ;
	    vp = (Variable *)varstack.LookupVar (form->arr.arrp[2].symbol) ;
	    if (!vp) {
		v.Init() ; 
	    } else if (vp < argstart || vp >= argstart + argcnt) {
		v = 0 ;
	    } else {
		v = 1 ;
	    }
	    varstack.TmPush (v) ;
	}
	return 0 ;
      case A_Position:
	varstack.TmPush (ECTX(form)) ;
	return 0 ;
      case A_Arg:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value x ;
	    loc.Reset () ;
	    st = AcExecForm (&formp->arr.arrp[2], &loc) ; CHECKST_EX(st) ;
	    x = varstack.VPop () ;
	    if (!x.IsNumber())
		return SetErrorCode (INVALID_ARRAY_INDEX, ECTX(form)) ;
	    int index = x.ForceInt() ;
	    if (argindex < 0)
		return SetErrorCode (SYNTAX_ERROR, ECTX(form)) ;
	    if (index < 0 || index >= argcnt)
		return SetErrorCode (INVALID_ARRAY_INDEX, ECTX(form)) ;
	    *((Value *)varstack.Top()) = *varstack.Peek (argindex + index) ;
	}
	return 0 ;
      case A_ArgList:
	formp = varstack.TmPush (form->arr) ;
	{
	    Value *arr ;
	    Value x ;
	    int i, j ;

	    if (argindex < 0)
		return SetErrorCode (SYNTAX_ERROR, ECTX(form)) ;
	    if (heap_area.AllocArray (argcnt, &varstack, &arr) < 0)
		return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(form)) ;
	    x.InitArr (arr, argcnt) ;
	    for (i = 0, j = argindex ; i < argcnt ; i++, j++) {
		arr[i] = *varstack.Peek (j) ;
	    }
	    *((Value *)varstack.Top()) = x ;
	    return 0 ;
	}
      case A_ThisPtr:
	if (!thisptr)
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	varstack.TmPush (*thisptr) ;
	locp->varp = thisptr ;
	locp->instancep = thisptr ;
	locp->classp    = classptr ;
	locp->vtype = VT_NormalVar ;
	return 0 ;
      case A_Self:
	if (!thisptr) 
	    return SetErrorCode (INVALID_EXPRESSION, ECTX(form)) ;
	varstack.TmPush (*thisptr) ;
	locp->varp = thisptr ;
	locp->instancep = thisptr ;
	if (thisptr->IsInstance()) {
	    locp->classp = thisptr->instance->classp->classnamep ;
	} else if (thisptr->IsClass()) {
	    locp->classp = thisptr ;
	} else {
	    locp->classp = FixSpecialClass (thisptr) ;
	}
	locp->vtype = VT_NormalVar ;
	return 0 ;
      default:
	return SetErrorCode (SYNTAX_ERROR) ;
    }
}

int CmmComp::AcStartCreator (Value *formp, Value *entp, LocInfo *locp)
{
    int i ;
    int st ;
    int framepsave ;
    int stackpsave ;
    int numargs ;
    int newfp ;
    Value *creatorp ;
    Instance *ip ;

    ip = CreateInstance (locp->classp->classp) ;
    if (!ip)
	return SetErrorCode (NOT_ENOUGH_MEMORY, ECTX(formp)) ;
    varstack.TmPush (ip) ;

    Value *instance = varstack.Top () ;

    locp->instancep = instance ;
    framepsave = varstack.FramePtr () ;
    stackpsave = varstack.StackPtr () ;

    creatorp = varstack.TmPush (entp->arr) ;
    newfp = varstack.StackPtr () ;

    varstack.TmPush (ip) ;
    numargs = formp->arr.arrp[3].arr.numelems ;
    for (i = 0 ; i < numargs ; i++) {
	LocInfo xloc ;
	st = AcExecForm (&(formp->arr.arrp[3].arr.arrp[i]), &xloc) ;
	CHECKST_EX(st) ;
    }
    numargs++ ; // increment for this pointer
    varstack.SetFramePtr (newfp) ;

    st = AcCallCreator (locp->classp, entp, locp->instancep, numargs) ;
    CHECKST_EX(st) ;

    varstack.SetFramePtr (framepsave) ;
    varstack.SetStackPtr (stackpsave) ;
    return st ;
}

int CmmComp::AcCallCreator (Value *classp, Value *entry, Value *instancep, int numargs)
{
    int st ;
    int   argindexsave ;
    int   argcntsave   ;
    Value *thisptrsave  ;
    Value *classptrsave ;
    LocInfo loc ;

    int startsp = varstack.FramePtr () ;
    int endpp = startsp + numargs ;
    int pp = startsp ;
    int i = 0 ;
    Arr *argp ;

    if (entry) {
	argindexsave = argindex ;
	argcntsave   = argcnt ;
	thisptrsave  = thisptr ;
	classptrsave = classptr ;

	argindex     = startsp ;
	argcnt       = numargs ;
	// ֐̎Ɖ̃oCh
	varstack.SetSymSimple (pp, symbolTable->GetSymbol("this")) ;
	argp = &entry->arr.arrp[4].arr ;
	pp++ ;
	for (i = 0 ; pp < endpp && i < argp->numelems ; pp++, i += 2)
	    varstack.SetSymSimple (pp, argp->arrp[i].symbol, argp->arrp[i+1].intval) ;
	thisptr  = varstack.Peek (startsp) ;
	classptr = classp ;
    }
    if (classp->classp->superclass) {
	int stackpsave = varstack.StackPtr () ;
	int framepsave = varstack.FramePtr () ;
	int superargs = 0 ;
	Value *superCreator ;
	VarType vtype ;

	varstack.SetFramePtr (varstack.StackPtr()) ;
	superargs = 1 ;
	varstack.TmPush (*instancep) ;

	if (entry->arr.arrp[5].IsArray()) {
	    for (i = 0 ; i < entry->arr.arrp[5].arr.numelems ; i++) {
		LocInfo xloc ;
		st = AcExecForm (&(entry->arr.arrp[5].arr.arrp[i]), &xloc) ;
		CHECKST_EX(st) ;
		superargs++ ;
	    }
	}
	superCreator = classp->classp->superclass->FindMethod (creatorSym, &vtype) ;
	if (superCreator) {
	    if (superCreator->IsContext()) {
		// X[p[NXNG[^A\[XR[hő
		// B
		st = CallCreator (classp->classp->superclass->classnamep,
				  superCreator, instancep, superargs) ;
	    } else {
		st = AcCallCreator (classp->classp->superclass->classnamep,
				    superCreator, instancep, superargs) ;
	    }
	} else {
	    st = AcCallCreator (classp->classp->superclass->classnamep,
				superCreator, instancep, superargs) ;
	}
	CHECKST_EX(st) ;
	// t[|C^AX^bN|C^̕A 
	varstack.SetFramePtr (framepsave) ;
	varstack.SetStackPtr (stackpsave) ;
    }
    if (entry) {
	loc.vtype     = VT_Method ;
	loc.classp    = classp ;
	loc.instancep = instancep ;
	loc.varp      = entry ;
	st = AcExecForm (&entry->arr.arrp[6], &loc) ;
	CHECKST_EX(st) ;

	argindex = argindexsave ;
	argcnt   = argcntsave ;
	thisptr  = thisptrsave ;
	classptr = classptrsave ;
    }
    return NORMAL  ;
}

int CmmComp::AcCallForm (Value *acfuncp, int numargs, Value *thisp, Value *classp, LocInfo *locp)
{
    int st ;
    int   argindexsave ;
    int   argcntsave   ;
    Value *thisptrsave  ;
    Value *classptrsave ;
    int startsp = varstack.FramePtr () ;
    int endpp = startsp + numargs ;
    int pp = startsp ;
    int i = 0 ;
    Arr *argp ;
    //
    //	varstack.TmPush (argindex) ;
    //  varstack.TmPush (argindex) ;
    //  varstack.TmPush (thisptr) ;
    //  varstack.TmPush (classptr) ;
    //
    argindexsave = argindex ;
    argcntsave   = argcnt ;
    thisptrsave  = thisptr ;
    classptrsave = classptr ;

    argindex     = startsp ;
    argcnt       = numargs ;
    // ֐̎Ɖ̃oCh
    if (thisp) {
	if ((AcInst)acfuncp->arr.arrp[0].intval == A_CreatorDef) {
	    // NG[^
	    thisptr = varstack.Peek (startsp) ;
	    if (thisptr->IsClass()) {
		// NG[^̓NX\bhȂ̂ŁAthis ́ANX
		// `wĂB̈ʒuɁACX^X
		// ZbgĂANG[^\bhĂяoB
		Instance *ip = CreateInstance (classp->classp) ;
		*thisptr = ip ;
		st = AcCallCreator (classp, acfuncp, thisptr, numargs) ;
		varstack.TmPush (*thisptr) ;
	    }
	} else {
	    locp->Reset () ;
	    varstack.SetSymSimple (pp, symbolTable->GetSymbol("this")) ;
	    argp = &acfuncp->arr.arrp[4].arr ;
	    pp++ ;
	    for (i = 0 ; pp < endpp && i < argp->numelems ; pp++, i += 2)
		varstack.SetSymSimple (pp, argp->arrp[i].symbol,
				       argp->arrp[i+1].intval) ;
	    thisptr  = varstack.Peek (startsp) ;
	    classptr = classp ;
	    st = AcExecForm (&acfuncp->arr.arrp[5], locp) ;
	}
    } else {
	argp = &acfuncp->arr.arrp[3].arr ;
	for (i = 0 ; pp < endpp && i < argp->numelems ; pp++, i += 2)
	    varstack.SetSymSimple (pp, argp->arrp[i].symbol, argp->arrp[i+1].intval) ;
	thisptr = NULL ;
	st = AcExecForm (&acfuncp->arr.arrp[4], locp) ;
    }
    CHECKST_EX(st) ;
    argindex = argindexsave ;
    argcnt   = argcntsave ;
    thisptr  = thisptrsave ;
    classptr = classptrsave ;
    //	
    //  classptr = (varstack.VPop()).location ;
    //  thisptr  = (varstack.VPop()).location ;
    //  argcnt   = (varstack.VPop()).intval ;
    //	argindex = (varstack.VPop()).intval ;
    //
    return st ;
}

//================================================================
// A-code executor
//================================================================
int CmmComp::AcExecute (Value *form)
{
    Value *baseform ;
    LocInfo loc ;
    int i ;
    int st = 0 ;
    int shrink ;
    Value retv ;
    int nelems ;

    baseform = varstack.TmPush (form->arr) ;
    shrink = varstack.TopIndex () ;
    varstack.TmPush () ;
    nelems = baseform->arr.numelems ;
    for (i = 0 ; i < nelems ; i++) {
	loc.Reset () ;
	varstack.TmPop () ;

	st = AcExecForm (&baseform->arr.arrp[i], &loc) ;
	if (st < 0 || st == RETURN || st == THROW)
	    break ;
    }
    if (st < 0 && st != EXIT) {
	if (errorCtx.text)
	    errorCtx.PrintError (errorMes[errorCode]) ;
	else
	    printf ("%s\n", errorMes[errorCode]) ;
	return -1 ;
    } else if (st == THROW) {
	PrintException (varstack.Top()) ;
    }
    retv = varstack.VPop () ;
    varstack.Shrink (shrink, 1) ;
    varstack.TmPush (retv) ;
    return 0 ;
}

int CmmComp::AcListInst (FILE *fp, Value *form, int int_inst)
{
    int i ;

    if (!form->IsArray())
	return -1 ;
    fprintf (fp, "%s\n", ACODE_HEADER) ;
    for (i = 0 ; i < form->arr.numelems ; i++)
	AcShowInst (fp, &form->arr.arrp[i], 0, int_inst) ;
    return 0 ;
}

int CmmComp::AcCompile (char *buf, Value *form)
{
    int st ;

    prog = SetContext (buf, 0) ;
    lastPC = 0 ;
    st = AcProgram (form) ;
    varstack.TmPop () ;
    if ((st < 0 && st != EXIT) || errorCode ) {
	prog->PrintError (errorMes[errorCode]) ;
    }
    return st ;
}

int CmmComp::AcCompileAndGo (char *buf, int show_inst, int no_exec)
{
    int st ;
    Value form ;

    st = AcCompile (buf, &form) ;

    Value *formp = varstack.TmPush (form) ;

    if (st < 0)
	return st ;
    if (show_inst > 0) {
	printf ("========= INSTRUCTIONS ==================\n") ;
	AcListInst (stdout, formp) ;
	printf ("========= END OF INSTRUCTIONS ==================\n") ;
    }
    if (!no_exec) {
	st = AcExecute (formp) ;
    }
    if (show_inst < 0) {
	printf ("========= INSTRUCTIONS ==================\n") ;
	AcListInst (stdout, formp) ;
	printf ("========= END OF INSTRUCTIONS ==================\n") ;
    }
    varstack.TmPop () ;
    return st ;
}

//
// ACode Translator
//
int CmmComp::AcParseACode (char *buffer, Value *formp)
{
    prog = SetContext (buffer, 0) ;

    Value *acode = NewArray (0, 50) ;
    Value form ;
    int st ;

    while (1) {
	st = AcParseForm (&form) ;
	if (st < 0)
	    break ;
	if (st > 0)
	    break ;
	heap_area.AddArray (acode, &form, &varstack, acode) ;
    }
    if (st < 0) {
	prog->PrintError (errorMes [errorCode]) ;
    }
    *formp = varstack.Pop () ;
    varstack.Pop () ;
    return 0 ;
}

//================================================================
// A-code Loader
// Translate A-code file form to A-code internal Form
//================================================================
int CmmComp::AcParseForm (Value *formp)
{
    Token t ;
    LexInfo li ;
    Value *basep ;
    Value *slist ;
    Value *arglist ;
    Value subform ;
    Symbol *sym ;
    int st = 0 ;
    int mnemonic ;
    AcInst inst ;

    t = Lex (&li) ;
    if (t == T_Eof)
	return 1 ;
    if (t != T_Lbl)
	return SetErrorCode (SYNTAX_ERROR) ;

    t = Lex(&li) ;
    if (t == T_IntNumber) {
	inst = (AcInst)li.ival ;
    } else if (t == T_Word || t == T_Symbol) {
	inst = (AcInst)AcLookup (li.symbol) ;
    } else {
	return SetErrorCode (SYNTAX_ERROR) ;
    }
    switch (inst) {
      case A_Nop:
	basep = NewArray (2, inst) ;
	break ;
      case A_If:
	basep = NewArray (4, 5, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[3] = subform.arr ;
	t = Lex (&li) ;
	PutBack () ;
	if (t == T_Lbl) {
	    st = AcParseForm (&subform) ; CHECKST(st) ;
	    heap_area.AddArray (basep, &subform, &varstack, basep) ;
	}
	break ;
      case A_While:
	basep = NewArray (4, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[3] = subform ;
	break ;
      case A_For:
	basep = NewArray (6, inst) ;
	slist = NewArray (0, 4) ;
	if (Lex (&li) != T_Lbl)
	    return SetErrorCode (SYNTAX_ERROR) ;
	// ݒ
	for (;;) {
	    t = Lex (&li) ;
	    if (t == T_Rbl)
		break ;
	    if (t == T_Lbl) {
		PutBack () ;
		st = AcParseForm (&subform) ; CHECKST(st) ;
		heap_area.AddArray (slist, &subform, &varstack, slist) ;
	    }
	}
	basep->arr.arrp[2] = slist->arr ;
	varstack.Pop () ;

	//--------
	if (Lex (&li) == T_Void)
	    basep->arr.arrp[3].Init() ;
	else {
	    PutBack() ;
	    st = AcParseForm (&subform) ; CHECKST(st) ;
	    basep->arr.arrp[3] = subform ;
	}
	slist = NewArray (0, 10) ;
	if (Lex (&li) != T_Lbl)
	    return -1 ;
	while (1) {
	    t = Lex (&li) ;
	    if (t == T_Rbl)
		break ;
	    if (t == T_Lbl) {
		PutBack () ;
		st = AcParseForm (&subform) ; CHECKST(st) ;
		heap_area.AddArray (slist, &subform, &varstack, slist) ;
	    }
	}
	basep->arr.arrp[4] = slist->arr ;
	varstack.Pop() ;

	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[5] = subform ;
	break ;
      case A_TryCatch:
	{
	    Value expr ;

	    basep = NewArray (5, A_TryCatch) ;
	    st = AcParseForm (&subform) ; CHECKST(st) ;
	    basep->arr.arrp[2] = subform ;

	    if (Lex (&li) != T_Lbl)
		return -1 ;
	    Value *catchlist = NewArray (0, 2) ;
	    while (1) {
		t = Lex (&li) ;
		if (t == T_Rbl)
		    break ;
		if (t != T_Lbl)	return SetErrorCode (SYNTAX_ERROR) ;
		// OubN̉
		Value *exception_block = NewArray (2) ;
		// ONXƂ̃CX^X󂯂ϐ𓾂
		Value *class_and_instance = NewArray (2) ;

		if (Lex (&li) != T_Lbl) {
		    return SetErrorCode (SYNTAX_ERROR) ;
		}
		if (!(sym = LexSymbolOrString ()))
		    return SetErrorCode (SYNTAX_ERROR) ;
		class_and_instance->arr.arrp[0] = sym ;
		if (Lex (&li) != T_Comma)
		    return SetErrorCode (SYNTAX_ERROR) ;
		if (!(sym = LexSymbolOrString ()))
		    return SetErrorCode (SYNTAX_ERROR) ;
		class_and_instance->arr.arrp[1] = sym ;
		exception_block->arr.arrp[0] = *class_and_instance ;
		if (Lex (&li) != T_Rbl)
		    return SetErrorCode (SYNTAX_ERROR) ;
		t = Lex (&li) ;
		if (t != T_Lbl) 
		    return SetErrorCode (SYNTAX_ERROR) ;
		// O󂯎ƂɎs{
		PutBack () ;
		st = AcParseForm (&expr) ; CHECKST(st) ;
		exception_block->arr.arrp[1] = expr ;
		heap_area.AddArray (catchlist, exception_block, &varstack, catchlist) ;
		varstack.TmPop(2) ;
		// ̗OubNI
		t = Lex (&li) ;
		if (t != T_Rbl)
		    return SetErrorCode (SYNTAX_ERROR) ;
	    }
	    basep->arr.arrp[3] = catchlist->arr ;
	    t = Lex (&li) ;
	    if (t == T_Lbl) {
		// finally ubN
		PutBack () ;
		st = AcParseForm (&subform) ;
		basep->arr.arrp[4] = subform ;
	    } else if (t == T_Void) {
		// finally ubNȂ
		basep->arr.arrp[4].Init() ;
	    } else {
		return SetErrorCode (SYNTAX_ERROR) ;
	    }
	    varstack.TmPop () ;
	}
	break ;
      case A_FuncDef:
	basep = NewArray (5, inst) ;
	if (!(sym = LexSymbolOrString()))
	    return SetErrorCode (SYNTAX_ERROR) ;
	basep->arr.arrp[2] = sym ;
	st = AcArgList (&subform) ; CHECKST(st) ;
	basep->arr.arrp[3] = subform ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[4] = subform ;
	break ;
      case A_MethoDef:
      case A_HidenDef:
	basep = NewArray (6, inst) ;
	if (!(sym = LexSymbolOrString()))
	    return SetErrorCode (SYNTAX_ERROR) ;
	basep->arr.arrp[2] = sym ;
	if (!(sym = LexSymbolOrString()))
	    return SetErrorCode (SYNTAX_ERROR) ;
	basep->arr.arrp[3] = sym ;
	st = AcArgList (&subform) ; CHECKST(st) ;
	basep->arr.arrp[4] = subform ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[5] = subform ;
	break ;
      case A_CreatorDef:
	basep = NewArray (7, inst) ;
	if (!(sym = LexSymbolOrString()))
	    return SetErrorCode (SYNTAX_ERROR) ;
	basep->arr.arrp[2] = sym ;
	if (!(sym = LexSymbolOrString()))
	    return SetErrorCode (SYNTAX_ERROR) ;
	basep->arr.arrp[3] = sym ;
	st = AcArgList (&subform) ; CHECKST(st) ;
	basep->arr.arrp[4] = subform ;
	st = AcSuperArgList (&subform) ;
	basep->arr.arrp[5] = subform ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[6] = subform ;
	break ;
      case A_VarDef:
	{
	    basep = NewArray (2, 6, A_VarDef) ;
	    for (;;) {
		t = Lex (&li) ;
		if (t == T_Rbl)
		    break ;
		if (t != T_Lbl) 
		    return -1 ;
		Value *subarr = NewArray (2) ;

		if (!(sym = LexSymbolOrString ()))
		    return SetErrorCode(SYNTAX_ERROR) ;
		subarr->arr.arrp[0] = sym ;
		t = Lex (&li) ;
		if (t != T_Rbl) {
		    PutBack () ;
		    Value subform ;
		    AcParseForm (&subform) ; CHECKST(st) ;
		    subarr->arr.arrp[1] = subform ;
		    t = Lex (&li) ; 
		}
		heap_area.AddArray (basep, subarr, &varstack, basep) ;
		varstack.Pop () ;
	    }
	    PutBack () ;
	}
	break ;
      case A_ClassDef:
	{
	    basep = NewArray (7, inst) ;
	    if (!(sym = LexSymbolOrString()))
		return SetErrorCode (SYNTAX_ERROR) ;
	    basep->arr.arrp[AC_CLASSNAME] = sym ;
	    t = Lex (&li) ;
	    if (t == T_Void) {
		basep->arr.arrp[AC_SUPERCLASS].Init() ;
	    } else if (t == T_Symbol) {
		basep->arr.arrp[AC_SUPERCLASS] = li.symbol ;
	    } else if (t == T_String) {
		basep->arr.arrp[AC_SUPERCLASS] = symbolTable->GetSymbol(li.sval) ;
	    } else {
		return SetErrorCode (SYNTAX_ERROR) ;
	    }
	    if (Lex (&li) != T_IntNumber)
		return -1 ;
	    basep->arr.arrp[AC_INSTVARS] = li.ival ;
	    if (Lex (&li) != T_IntNumber)
		return -1 ;
	    basep->arr.arrp[AC_CLASSVARS] = li.ival ;

	    Value *members = NewArray (0, 10) ;
	    Token vtname ;
	    VarType vtype ;

	    if (Lex (&li) != T_Lbl)
		return -1 ;
	    for (;;) {
		t = Lex (&li) ;
		if (t == T_Rbl)
		    break ;
		if (t != T_Lbl)
		    return -1 ;
		vtname = Lex (&li) ;
		if (vtname == T_IntNumber)    vtype = (VarType)li.ival ;
		else if (vtname == T_Private) vtype = VT_Private ;
		else if (vtname == T_Public)  vtype = VT_Public ;
		else if (vtname == T_Static)  vtype = VT_ClassVar ;
		else if (vtname == T_Method)  vtype = VT_Method ;
		else if (vtname == T_Hiden)   vtype = VT_Hiden ;
		else
		    return -1 ;
		Value *aMember ;
		if (vtype != VT_Method && vtype != VT_Hiden) 
		    aMember = NewArray (2) ;
		else
		    aMember = NewArray (2, 3) ;
		aMember->arr.arrp[0] = (int)vtype ;
		if (!(sym = LexSymbolOrString()))
		    return SetErrorCode (SYNTAX_ERROR) ;
		aMember->arr.arrp[1] = sym ;
		if (vtype == VT_Method || vtype == VT_Hiden) {
		    t = Lex (&li) ;
		    PutBack () ;
		    if (t == T_Lbl) {
			Value body ;
			st = AcParseForm (&body) ; CHECKST(st) ;
			heap_area.AddArray (aMember, &body, &varstack, aMember) ;
		    }
		} else if (vtype == VT_ClassVar) {
		    t = Lex (&li) ;
		    PutBack () ;
		    if (t == T_Lbl) {
			Value initval ;
			st = AcParseForm (&initval) ; CHECKST(st) ;
			heap_area.AddArray (aMember, &initval, &varstack, aMember) ;
		    }
		}

		heap_area.AddArray (members, aMember, &varstack, members) ;
		varstack.Pop () ;
		t = Lex (&li) ;
		if (t != T_Rbl)
		    return -1 ;
	    }
	    basep->arr.arrp[MEMBERS] = members->arr ;
	    varstack.Pop () ;
	}
	break ;
      case A_Funcall:
	basep = NewArray (4, inst) ;
	st = AcParseForm (&subform) ;
	basep->arr.arrp[2] = subform ;

	if (Lex (&li) != T_Lbl)
	    return SetErrorCode (SYNTAX_ERROR) ;
	arglist = NewArray (0, 10) ;
	for (;;) {
	    t = Lex (&li) ;
	    if (t == T_Lbl) {
		PutBack () ;
		st = AcParseForm (&subform) ; CHECKST(st) ;
		heap_area.AddArray (arglist, &subform, &varstack, arglist) ;
	    } else if (t == T_Rbl) {
		break ;
	    }
	}
	basep->arr.arrp[3] = arglist->arr ;
	varstack.Pop () ;
	break ;
      case A_NoMember:
	basep = NewArray (3, inst) ;
	if (!(sym = LexSymbolOrString()))
	    return SetErrorCode (SYNTAX_ERROR) ;
	basep->arr.arrp[2] = sym ;
	break ;
      case A_Var:
	basep = NewArray (5, inst) ;
	if (!(sym = LexSymbolOrString()))
	    return SetErrorCode (SYNTAX_ERROR) ;
	basep->arr.arrp[2] = sym ;
	t = Lex (&li) ;
	if (t == T_IntNumber) {
	    basep->arr.arrp[3] = li.ival ;
	} else if (t == T_Rbl) {
	    PutBack () ;
	    basep->arr.arrp[3] = basep->arr.arrp[2].symbol->GetHashValue() ;
	} else {
	    return SetErrorCode (SYNTAX_ERROR) ;
	}
	break ;
      case A_SelectMember:
	basep = NewArray (7, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform ;
	if (!(sym = LexSymbolOrString()))
	    return SetErrorCode(SYNTAX_ERROR) ;
	basep->arr.arrp[3] = sym ;
	break ;
      case A_SelectClass:
	basep = NewArray (5, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform ;
	if (!(sym = LexSymbolOrString ()))
	    return SetErrorCode(SYNTAX_ERROR) ;
	basep->arr.arrp[3] = sym ;
	t = Lex (&li) ;
	if (t == T_IntNumber) {
	    basep->arr.arrp[4] = li.ival ;
	} else if (t == T_Rbl) {
	    PutBack () ;
	    basep->arr.arrp[4] = basep->arr.arrp[3].symbol->GetHashValue() ;
	} else {
	    return SetErrorCode (SYNTAX_ERROR) ;
	}
	break ;
      case A_Const:
	basep = NewArray (3, inst) ;
	t = Lex (&li) ;
	if (t == T_IntNumber) {
	    basep->arr.arrp[2] = li.ival ;
	} else if (t == T_Number) {
	    basep->arr.arrp[2] = li.value ;
	} else if (t == T_String) {
	    basep->arr.arrp[2] = li.sval ;
	} else if (t == T_Void) {
	    basep->arr.arrp[2].Init () ;
	} else if (t == T_Symbol) {
	    basep->arr.arrp[2] = li.symbol ;
	} else {
	    return SetErrorCode (SYNTAX_ERROR) ;
	}
	break ;
      case A_Not:
      case A_Negate:
      case A_PPPre:
      case A_MMPre:
      case A_PPPost:
      case A_MMPost:
	basep = NewArray (3, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
	break ;
      case A_AddAssign:
      case A_SubAssign:
      case A_LshAssign:
      case A_RshAssign:
      case A_MultAssign:
      case A_DivAssign:
      case A_ModAssign:
      case A_XorAssign:
      case A_AndAssign:
      case A_OrAssign:
	basep = NewArray (4, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[3] = subform.arr ;
	break ;
      case A_Block:
	basep = NewArray (2, 20, inst) ;
	while (1) {
	    t = Lex (&li) ;
	    PutBack () ;
	    if (t == T_Rbl) {
		break ;
	    } else {
		st = AcParseForm (&subform) ; CHECKST(st) ;
		heap_area.AddArray (basep, &subform, &varstack, basep) ;
	    }
	}
	break ;
      case A_Plus:
      case A_Minus:
      case A_Mult:
      case A_Div:
      case A_Mod:
      case A_Rsh:
      case A_Lsh:
      case A_Eq:
      case A_Neq:
      case A_Lt:
      case A_Gt:
      case A_Le:
      case A_Ge:
      case A_Or:
      case A_And:
      case A_Band:
      case A_Bor:
      case A_Xor:
      case A_Assign:
      case A_RefAssign:
	basep = NewArray (4, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[3] = subform.arr ;
	break ;
      case A_New:
	basep = NewArray (3, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
	break ;
      case A_Return:
      case A_Break:
      case A_Exit:
	t = Lex (&li) ;
	if (t == T_Rbl) {
	    PutBack () ;
	    basep = NewArray (2, inst) ;
	} else {
	    PutBack () ;
	    basep = NewArray (3, inst) ;
	    st = AcParseForm (&subform) ; CHECKST(st) ;
	    basep->arr.arrp[2] = subform.arr ;
	}
	break ;
      case A_Throw:
	basep = NewArray (3, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
	break ;
      case A_Continue:
	basep = NewArray (2, inst) ;
	break ;
      case A_Eval:
	basep = NewArray (3, inst) ;
	t = Lex (&li) ;
	if (t == T_String) {
	    basep->arr.arrp[2] = li.sval ;
	} else {
	    PutBack () ;
	    AcParseForm (&subform) ;
	    basep->arr.arrp[2] = subform.arr ;
	}
	break ;
      case A_Arg:
	basep = NewArray (3, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
	break ;
      case A_ArrIndex:
	basep = NewArray (4, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[3] = subform.arr ;
	break ;
      case A_ArgList:
	basep = NewArray (2, inst) ;
	break ;
      case A_Load:
	basep = NewArray (3, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
      case A_Package:
	basep = NewArray (3, inst) ;
	st = AcParseForm (&subform) ; CHECKST(st) ;
	basep->arr.arrp[2] = subform.arr ;
	break ;
      case A_Array:
      {
	basep = NewArray (3, inst) ;
	Value *arrp = NewArray (0, 20) ;
	if (Lex (&li) != T_Lbl)
	    return SetErrorCode (SYNTAX_ERROR) ;
	while (1) {
	    t = Lex (&li) ;
	    if (t == T_Rbl)
		break ;
	    PutBack () ;
	    st = AcParseForm (&subform) ;
	    heap_area.AddArray (arrp, &subform, &varstack, arrp) ;
	}
	basep->arr.arrp[2] = arrp->arr ;
	varstack.Pop () ;
	break ;
      }
      case A_ArgCnt:
	basep = NewArray (2, inst) ;
	break ;
      case A_Pause:
	basep = NewArray (2, inst) ;
	break ;
      case A_BoundArg:
	basep = NewArray (3, inst) ;
	t = Lex (&li) ;
	if (t == T_String) {
	    basep->arr.arrp[2] = li.sval ;
	} else {
	    return SetErrorCode (SYNTAX_ERROR) ;
	}
	break ;
      case A_MacroDef:
	break ;
      case A_ThisPtr:
	basep = NewArray (2, inst) ;
	break ;
      case A_Self:
	basep = NewArray (2, inst) ;
	break ;
      case A_Position:
	basep = NewArray (2, inst) ;
	break ;
      default:
	break ;
    }
    if (Lex (&li) != T_Rbl)
	return SetErrorCode (SYNTAX_ERROR) ;
    *formp = varstack.Pop () ;
    return 0 ;
}

// CmmComp::AcArgList -- ֐/\bḧ̃RpC
int CmmComp::AcArgList (Value *arglistp)
{
    Token t ;
    LexInfo li ;
    Value *arglist ;

    if (Lex (&li) != T_Lbl)
	return -1 ;
    arglist = NewArray (0, 10) ;
    t = Lex (&li) ;
    if (t == T_Rbl) {
	;
    } else {
	while (1) {
	    Value x ;
	    if (t == T_Symbol) {
		x = li.symbol ;
	    } else if (t == T_String) {
		x = symbolTable->GetSymbol (li.sval) ;
	    } else {
		return SetErrorCode (SYNTAX_ERROR) ;
	    }
	    heap_area.AddArray (arglist, &x, &varstack, arglist) ;
	    t = Lex (&li) ;
	    if (t != T_Comma)
		return -1 ;
	    t = Lex (&li) ;
	    if (t != T_IntNumber)
		return -1 ;
	    x = li.ival ;
	    heap_area.AddArray (arglist, &x, &varstack, arglist) ;
	    t = Lex (&li) ;
	    if (t == T_Rbl) {
		break ;
	    } else if (t == T_Comma) {
		t = Lex (&li) ;
	    } else {
		return SetErrorCode (SYNTAX_ERROR) ;
	    }
	}
    }
    *arglistp = varstack.Pop () ;
    return 0 ;
}

// NG[^̌p
int CmmComp::AcSuperArgList (Value *arglistp)
{
    Token t ;
    LexInfo li ;
    int st ;

    if ((t = Lex (&li)) == T_Void) {
	arglistp->Init() ;
	return 0 ;
    }
    if (t != T_Lbl)
	return -1 ;
    t = Lex (&li) ;
    if (t == T_Rbl) {
	;
    } else {
	Value *arglist ;
	arglist = NewArray (0, 10) ;
	
	PutBack () ;
	for (;;) {
	    Value x ;
	    st = AcParseForm (&x) ;
	    CHECKST(st) ;
	    heap_area.AddArray (arglist, &x, &varstack, arglist) ;
	    t = Lex (&li) ;
	    if (t == T_Rbl)
		break ;
	    PutBack () ;
	}
	*arglistp = *arglist ;
	varstack.TmPop() ;
    }
    return 0 ;
}

void CmmComp::HitInfo ()
{
    printf (">> Var = %d, LVar = %d, GVar = %d\n",
	    var_hit, lvar_hit, gvar_hit) ;
    printf (">> SelectMember = %d, SelMemCache = %d thisMember = %d\n",
	    selmem_hit, selmem_cache_hit,this_mem_hit) ;
}

int CmmComp::ExecACode (char *buffer)
{
    int st ;
    Value form ;

    st = AcParseACode (buffer, &form) ;
    if (st < 0)
	return -1 ;
    globalLim = varstack.Top () ;
    return AcExecute (&form) ;
}

int CmmComp::ExecJIT (char *filename)
{
    Value *vp ;

    vp = ReadSource (filename) ;
    if (!vp)
	return -1 ;
    if (!strncmp (vp->sptr, ACODE_HEADER, strlen(ACODE_HEADER))) {
	return ExecACode (vp->sptr) ;
    }
    exec_mode = ExAcInterp ;
    return ExecuteApp (vp->sptr) ;
}

int CmmComp::ExecCompiler (char *filename)
{
    Value *vp ;

    vp = ReadSource (filename) ;
    if (!vp)
	return -1 ;
    if (!strncmp (vp->sptr, ACODE_HEADER, strlen(ACODE_HEADER))) {
	return ExecACode (vp->sptr) ;
    }
    return AcCompileAndGoApp (vp->sptr) ;
}

int CmmComp::CompileFile (char *filename, Value *form)
{
    Value *vp ;
    int st ;

    vp = ReadSource (filename) ;
    if (!vp)
	return -1 ;
    if (!strncmp (vp->sptr, ACODE_HEADER, strlen (ACODE_HEADER))) {
	st = AcParseACode (vp->sptr, form) ;
	return st ;
    }
    st = AcCompile (vp->sptr, form) ;
    return st ;
}

int CmmComp::ThrowSystemException (Value *exception, Context& pos)
{
    Value v ;
    v = pos ;
    exception->instance->SetValue (symbolTable->GetSymbol("pos"), v) ;
    varstack.TmPush (*exception) ;
    return THROW ;
}

Symbol *CmmComp::LexSymbolOrString ()
{
    Token t ;
    LexInfo li ;
    
    t = Lex (&li) ;
    if (t == T_Symbol)
	return li.symbol ;
    else if (t == T_String)
	return symbolTable->GetSymbol (li.sval) ;
    else
	return NULL ;
}

