//----------------------------------------------------------------
// cmmmath.cc: Cmm Mathematical functions
//----------------------------------------------------------------
#include <math.h>
#include <cmm/cmm.h>
#include <cmm/cmmmath.h>

static int __sqrt (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = sqrt (arg[0].ForceReal()) ;
    return 0 ;
}

static int __sin (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = sin(arg[0].ForceReal()) ;
    return 0 ;
}

static int __cos (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = cos(arg[0].ForceReal()) ;
    return 0 ;
}

static int __tan (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = tan(arg[0].ForceReal()) ;
    return 0 ;
}

static int __asin (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = asin(arg[0].ForceReal()) ;
    return 0 ;
}

static int __acos (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = acos(arg[0].ForceReal()) ;
    return 0 ;
}

static int __atan (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = atan(arg[0].ForceReal()) ;
    return 0 ;
}

static int __atan2 (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = atan2(arg[0].ForceReal(), arg[1].ForceReal()) ;
    return 0 ;
}

static int __floor (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = floor (arg[0].ForceReal()) ;
    return 0 ;
}

static int __ceil (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = ceil (arg[0].ForceReal()) ;
    return 0 ;
}

static int __abs (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    int ival = arg[0].ForceInt() ;
    *retval = (ival < 0) ? -ival : ival ;
    return 0 ;
}

static int __fabs (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = fabs (arg[0].ForceReal()) ;
    return 0 ;
}

static int __exp (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = exp (arg[0].ForceReal()) ;
    return 0 ;
}

static int __log (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = log (arg[0].ForceReal()) ;
    return 0 ;
}

static int __log10 (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = log10 (arg[0].ForceReal()) ;
    return 0 ;
}

static int __pow (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = pow (arg[0].ForceReal(), arg[1].ForceReal()) ;
    return 0 ;
}

// Builtin Methods for Real Class
static int _Real_Sqrt (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = sqrt (thisp->ForceReal()) ;
    return 0 ;
}

static int _Real_Sin (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = sin(thisp->ForceReal()) ;
    return 0 ;
}

static int _Real_Cos (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = cos(thisp->ForceReal()) ;
    return 0 ;
}

static int _Real_Tan (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = tan(thisp->ForceReal()) ;
    return 0 ;
}

static int _Real_Asin (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = asin(thisp->ForceReal()) ;
    return 0 ;
}

static int _Real_Acos (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = acos(thisp->ForceReal()) ;
    return 0 ;
}

static int _Real_Atan (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = atan(thisp->ForceReal()) ;
    return 0 ;
}

static int _Real_Floor (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = floor (thisp->ForceReal()) ;
    return 0 ;
}

static int _Real_Ceil (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = ceil (thisp->ForceReal()) ;
    return 0 ;
}

static int _Real_Fabs (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = fabs (arg[0].ForceReal()) ;
    return 0 ;
}

static int _Real_Exp (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = exp (arg[0].ForceReal()) ;
    return 0 ;
}

static int _Real_Log (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = log (arg[0].ForceReal()) ;
    return 0 ;
}

static int _Real_Log10 (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = log10 (arg[0].ForceReal()) ;
    return 0 ;
}

static int _Real_Pow (CmmInterp *cenv, Value *thisp, int args, Variable *arg, Value *retval)
{
    *retval = pow (arg[0].ForceReal(), arg[1].ForceReal()) ;
    return 0 ;
}

void InstallMath (CmmInterp *cmm)
{
    cmm->AddGlobalVar ("PI",	  (double)3.14159265358979) ;
    cmm->AddGlobalVar ("PI_D2",   (double)1.57079632679490) ;
    cmm->AddGlobalVar ("PI_M3D2", (double)4.71238898038469) ;
    cmm->AddGlobalVar ("PI_M2",	  (double)6.28318530717958 ) ;
    cmm->AddBuiltin ("sqrt",	__sqrt) ;
    cmm->AddBuiltin ("sin",	__sin) ;
    cmm->AddBuiltin ("cos",	__cos) ;
    cmm->AddBuiltin ("tan",	__tan) ;
    cmm->AddBuiltin ("asin",	__asin) ;
    cmm->AddBuiltin ("acos",	__acos) ;
    cmm->AddBuiltin ("atan",	__atan) ;
    cmm->AddBuiltin ("atan2",	__atan2) ;
    cmm->AddBuiltin ("floor",	__floor) ;
    cmm->AddBuiltin ("ceil",	__ceil) ;
    cmm->AddBuiltin ("abs",	__abs) ;
    cmm->AddBuiltin ("fabs",	__fabs) ;
    cmm->AddBuiltin ("exp",	__exp) ;
    cmm->AddBuiltin ("log",	__log) ;
    cmm->AddBuiltin ("log10",	__log10) ;
    cmm->AddBuiltin ("pow",	__pow) ;
    cmm->AddBiMethod ("Real", "Sqrt",  _Real_Sqrt) ;
    cmm->AddBiMethod ("Real", "Cos",   _Real_Cos) ;
    cmm->AddBiMethod ("Real", "Sin",   _Real_Sin) ;
    cmm->AddBiMethod ("Real", "Tan",   _Real_Tan) ;
    cmm->AddBiMethod ("Real", "Acos",  _Real_Acos) ;
    cmm->AddBiMethod ("Real", "Asin",  _Real_Asin) ;
    cmm->AddBiMethod ("Real", "Atan",  _Real_Atan) ;
    cmm->AddBiMethod ("Real", "Floor", _Real_Floor) ;
    cmm->AddBiMethod ("Real", "Ceil",  _Real_Ceil) ;
    cmm->AddBiMethod ("Real", "Log",   _Real_Log) ;
    cmm->AddBiMethod ("Real", "Log10", _Real_Log10) ;
    cmm->AddBiMethod ("Real", "Pow",   _Real_Pow) ;
}

