//----------------------------------------------------------------
// cmmmisc.cc: A-Code environment
//		A-Code writer/reader-translator/discompiler
//					(c) 쏤X
//----------------------------------------------------------------
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef MSVC_2
#include <sys/file.h>
#endif
#ifdef MSVC_2
#include <fcntl.h>
#endif

#include <stdlib.h>

#define CMM_INTERNAL
#include <cmm/class.h>
#include <cmm/cmmcomp.h>

extern "C" {
    int read (int, char *, int) ;
    int close (int) ;
} ;

// Write A-code to a file
int CmmComp::AcWriteACode (char *fileName, Value *form, int int_inst)
{
    FILE *fp ;
    if (!fileName) {
	fp = stdout ;
    } else {
	fp = fopen (fileName, "w") ;
	if (!fp)
	    return -1 ;
    }
    AcListInst (fp, form, int_inst) ;
    if (fileName)
	fclose (fp) ;
    return 0 ;
}

// Read A-code from a file
int CmmComp::AcReadACode (char *fileName, Value *form)
{
    int fd ;
    int st ;
    struct stat fst ;
    char *buffer ;

    fd = open (fileName, O_RDONLY) ;
    if (fd < 0) {
	fprintf (stderr, "Cannot found file %s\n", fileName) ;
	return -1 ;
    }
    fstat (fd, &fst) ;
    buffer = heap_area.AllocSegment (fst.st_size + 1, &varstack) ;
    read (fd, buffer, fst.st_size) ;
    buffer [fst.st_size] = '\0' ;
    
    st = AcParseACode (buffer, form) ;

    close (fd) ;
    return st ;
}

int CmmComp::AcDisCompile(Value *form)
{
    int i ;
    int st ;
    AcInst lastinst = A_Nop ;

    for (i = 0 ; i < form->arr.numelems ; i++) {
	Value *a_code ;
	a_code = &form->arr.arrp[i] ;
	AcInst inst = (AcInst)a_code->arr.arrp[0].intval ;
	if (lastinst == A_Nop) {
	    lastinst = inst ;
	} else {
	    if ((lastinst == A_VarDef && inst != A_VarDef) 
		|| (lastinst != A_VarDef && inst == A_VarDef))
		printf ("\n") ;
	    lastinst = inst ;
	}
	st = AcDisCompile (a_code, A_Block, 0) ;
    }
    return st ;
}

//----------------------------------------------------------------
// CmmComp::AcDisCompile -- Convert a-code to Cmm source
//----------------------------------------------------------------
int CmmComp::AcDisCompile (Value *formp, AcInst upperinst, int level)
{
    AcInst binst ;
    int dopar = 0 ;
    char *opr ;
    int st = 0 ;
    int i ;
    Value *head ;
#define DC_INDENT 4
    if (!formp->IsArray()) {
	printf ("Ececutor Fatal :: Invalid Form\n") ;
	return -1 ;
    }
    head = &formp->arr.arrp[0] ;
    if (head->IsArray ()) {
	AcDisCompile (head, A_Nop, 0) ;
    }
    binst = (AcInst)head->intval ;

    if (binst != A_Block && binst != A_SLBlock && binst != A_If) {
	for (i = 0 ; i < level * DC_INDENT ; i++)
	    putchar (' ') ;
    }
    switch (binst) {
      case A_Nop:
	printf ("{}") ;
	break ;
      case A_VarDef:
	{
	    printf ("var ") ;
	    for (i = 2 ; i < formp->arr.numelems ; i++) {
		if (i > 2)
		    printf (", ") ;
		Value *avarp = &formp->arr.arrp[i] ;
		printf ("%s", avarp->arr.arrp[0].sptr) ;
		if (avarp->arr.numelems == 2 &&
		    avarp->arr.arrp[1].IsArray())
		{
		    printf (" = ") ;
		    st = AcDisCompile (&avarp->arr.arrp[1], binst, -1) ; CHECKST (st) ;
		}
	    }
	    printf (" ;\n") ;
	}
	return 0 ;
      case A_ClassDef:
	{
	    for (i = 0 ; i < level * DC_INDENT ; i++)
		putchar (' ') ;
	    printf ("class ") ;
	    printf ("%s", formp->arr.arrp[2].sptr) ;
	    if (formp->arr.arrp[3].IsString()) {
		printf (": %s", formp->arr.arrp[3].sptr) ;
	    }
	    printf ("{\n") ;
	    Value *members = &formp->arr.arrp[6] ;
	    for (i = 0 ; i < members->arr.numelems ; i++) {
		Value *amember = &members->arr.arrp[i] ;
		Value *bodyp = NULL ;
		int j ;

		for (j = 0 ; j < (level + 1) * DC_INDENT ; j++)
		    putchar (' ') ;
		if (amember->arr.arrp[0].intval == (int)VT_Private) {
		    printf ("private %s", amember->arr.arrp[1].sptr) ;
		} else if (amember->arr.arrp[0].intval == (int)VT_Public) {
		    printf ("public %s", amember->arr.arrp[1].sptr) ;
		} else if (amember->arr.arrp[0].intval == (int)VT_ClassVar) {
		    printf ("static %s", amember->arr.arrp[1].sptr) ;
		} else if (amember->arr.arrp[0].intval == (int)VT_Method) {
		    printf ("method %s", amember->arr.arrp[1].sptr) ;
		    if (amember->arr.arrp[2].IsArray()) {
			int is_creator ;
			    
			bodyp = &amember->arr.arrp[2] ;

			is_creator = (bodyp->arr.arrp[0].intval == (int)A_CreatorDef) ; 
			
			printf ("(") ;
			for (j = 0 ; j < bodyp->arr.arrp[4].arr.numelems ; j += 2) {
			    if (j > 0)
				printf (", ") ;
			    printf ("%s", bodyp->arr.arrp[4].arr.arrp[j].sptr) ;
			}
			printf (")") ;
			if (is_creator) {
			    if (bodyp->arr.arrp[5].IsArray()) {
				printf (" : (") ;
				for (j = 0 ; j < bodyp->arr.arrp[5].arr.numelems ; j++) {
				    if (j > 0)
					printf (", ") ;
				    st = AcDisCompile (&bodyp->arr.arrp[5].arr.arrp[j],
						       A_CreatorDef, -1) ;
				}
				printf (")\n") ;
			    } else {
				printf ("\n") ;
			    }
			    st = AcDisCompile (&bodyp->arr.arrp[6], A_MethoDef, level+2) ;
			} else {
			    printf ("\n") ; 
			    st = AcDisCompile (&bodyp->arr.arrp[5], A_MethoDef, level+2) ;
			}
		    }
		}
		if (!bodyp) {
		    printf (" ;\n") ;
		}
	    }
	    for (i = 0 ; i < level * DC_INDENT ; i++)
		putchar (' ') ;
	    printf ("}\n") ;
	}
	break ;
      case A_If:
      {
	if (upperinst != A_Nop) {
	    for (i = 0 ; i < level * DC_INDENT ; i++)
		putchar (' ') ;
	}
	printf ("if (") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	printf (") ") ;
	
	Value *thenp = &formp->arr.arrp[3] ;
	Value *elsep = &formp->arr.arrp[4] ;
	AcInst thenhead = (AcInst)thenp->arr.arrp[0].intval ;
	
	if (thenhead != A_Block && thenhead != A_SLBlock && level >= 0) {
	    if (thenhead == A_If) {
		printf ("{") ;
	    }
	    printf ("\n") ;
	}
	st = AcDisCompile (&formp->arr.arrp[3], binst, (level < 0) ? level: level+1) ; CHECKST(st) ;
	if (thenhead == A_If) {
	    for (i = 0 ; i < level * DC_INDENT ; i++)
		putchar (' ') ;
	    printf ("}") ;
	}
	if (elsep->IsArray ()) {
	    if ((AcInst)thenp->arr.arrp[0].intval != (int)A_Block
		&& (AcInst)thenp->arr.arrp[0].intval != (int)A_Block)
	    {
		if (level < 0 || thenhead == A_If)
		    putchar (' ') ;
		else 
		    for (i = 0 ; i < level * DC_INDENT ; i++)
			putchar (' ') ;
		printf ("else ") ;
	    } else {
		printf (" else ") ;
	    }
	    Value *elsep = &formp->arr.arrp[4] ;
	    AcInst elsehead = (AcInst)elsep->arr.arrp[0].intval ;
	    if (elsehead == A_If) {
		st = AcDisCompile (elsep, A_Nop, level) ;
	    } else {
		if (elsehead != A_Block && elsehead != A_SLBlock && level >= 0)
		    printf ("\n") ;
		st = AcDisCompile (&formp->arr.arrp[4], A_Nop, (level < 0) ? level : level+1) ;
	    }
	    CHECKST(st) ;
	} else {
	    if (thenhead == A_Block || thenhead == A_SLBlock || thenhead == A_If)
		printf ("\n") ;
	}
	return 0 ;
      }
      case A_While:
	printf ("while (") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	printf (") ") ;
	st = AcDisCompile (&formp->arr.arrp[3], binst, level+1) ; CHECKST(st) ;
	return 0 ;
	break ;
      case A_For:
	printf ("for (") ;
	for (i = 0 ; i < formp->arr.arrp[2].arr.numelems ; i++) {
	    if (i > 0)
		printf (", ") ;
	    st = AcDisCompile (&formp->arr.arrp[2].arr.arrp[i], binst, -1) ; CHECKST (st) ;
	}
	printf ("; ") ;
	if (formp->arr.arrp[3].IsArray()) {
	    st = AcDisCompile (&formp->arr.arrp[3], binst, -1) ;
	    CHECKST(st) ;
	}
	printf ("; ") ;
	for (i = 0 ; i < formp->arr.arrp[4].arr.numelems ; i++) {
	    if (i > 0)
		printf (", ") ;
	    st = AcDisCompile (&formp->arr.arrp[4].arr.arrp[i], binst, -1) ; CHECKST (st) ;
	}
	printf (") ") ;
	if (formp->arr.arrp[5].arr.arrp[0].intval != (int)A_Block
	    && formp->arr.arrp[5].arr.arrp[0].intval != (int)A_SLBlock)
	{
	    printf ("\n") ;
	}
	st = AcDisCompile (&formp->arr.arrp[5], binst, level+1) ;
	CHECKST(st) ;
	return 0 ;
//	if (formp->arr.arrp[5].arr.arrp[0].intval == (int)A_Block
//		&& formp->arr.arrp[5].arr.arrp[0].intval == (int)A_Block)
//	    return 0 ;
	break ;
      case A_TryCatch:
	{
	    int j ;
	    printf ("try ") ;
	    if (formp->arr.arrp[2].arr.arrp[0].intval != (int)A_Block
		&& formp->arr.arrp[2].arr.arrp[0].intval != (int)A_SLBlock)
	    {
		printf ("\n") ;
	    }
	    st = AcDisCompile (&formp->arr.arrp[2], binst, level+1) ;
	    Value *catchlist = &formp->arr.arrp[3] ;
	    for (i = 0 ; i < catchlist->arr.numelems ; i++) {
		Value *catch_condition = &catchlist->arr.arrp[i].arr.arrp[0] ;
		Value *catch_body      = &catchlist->arr.arrp[i].arr.arrp[1] ;
		for (j = 0 ; j < level * DC_INDENT ; j++)
		    putchar (' ') ;
		printf ("catch (%s %s) ",
			catch_condition->arr.arrp[0].sptr,
			catch_condition->arr.arrp[1].sptr) ;
		if (catch_body->arr.arrp[0].intval != (int)A_Block
		    && catch_body->arr.arrp[0].intval != (int)A_SLBlock)
		{
		    printf ("\n") ;
		}
		st = AcDisCompile (catch_body, binst, level+1) ;
	    }
	    if (formp->arr.arrp[4].IsArray()) {
		for (j = 0 ; j < level * DC_INDENT ; j++)
		    putchar (' ') ;
		printf ("finally") ;
		if (formp->arr.arrp[4].arr.arrp[0].intval != (int)A_Block
		    && formp->arr.arrp[4].arr.arrp[0].intval != (int)A_SLBlock)
		{
		    printf ("\n") ;
		}
		st = AcDisCompile (&formp->arr.arrp[4], binst, level+1) ;
	    }
	} 
	break ;
      case A_FuncDef:
	printf ("func %s (", formp->arr.arrp[2].sptr) ;
	for (i = 0 ; i < formp->arr.arrp[3].arr.numelems ; i += 2) {
	    if (i > 0)
		printf (", ") ;
	    printf ("%s", formp->arr.arrp[3].arr.arrp[i].sptr) ;
	}
	printf (")\n") ;
	st = AcDisCompile (&formp->arr.arrp[4], binst, level+1) ;
	printf ("\n") ;
	return 0 ;
      case A_MethoDef:
	printf ("method %s::%s (",
		formp->arr.arrp[2].sptr, formp->arr.arrp[3].sptr) ;
	for (i = 0 ; i < formp->arr.arrp[4].arr.numelems ; i += 2) {
	    if (i > 0)
		printf (", ") ;
	    printf ("%s", formp->arr.arrp[4].arr.arrp[i].sptr) ;
	}
	printf (")\n") ;
	st = AcDisCompile (&formp->arr.arrp[5], binst, level+1) ;
	break ;
      case A_CreatorDef:
	printf ("method %s::%s (",
		formp->arr.arrp[2].sptr, formp->arr.arrp[3].sptr) ;
	for (i = 0 ; i < formp->arr.arrp[4].arr.numelems ; i += 2) {
	    if (i > 0)
		printf (", ") ;
	    printf ("%s", formp->arr.arrp[4].arr.arrp[i].sptr) ;
	}
	printf (")\n") ;
	if (formp->arr.arrp[5].IsArray()) {
	    printf (": (") ;
	    for (i = 0 ; i < formp->arr.arrp[5].arr.numelems ; i++) {
		if (i > 0)
		    printf (", ") ;
		AcDisCompile (&formp->arr.arrp[5].arr.arrp[i], binst, level+1) ;
	    }
	    printf (")\n") ;
	}
	st = AcDisCompile (&formp->arr.arrp[5], binst, level+1) ;
	break ;
      case A_Funcall:
      {
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ;
	printf ("(") ;
	Value *argp = &formp->arr.arrp[3] ;
	for (i = 0 ; i < argp->arr.numelems ; i++) {
	    if (i > 0)
		printf (", ") ;
	    st = AcDisCompile (&argp->arr.arrp[i], binst, -1) ;
	}
	printf (")") ;
	break ;
      }
      case A_Var:
	if (formp->arr.arrp[2].IsString()) {
	    printf ("%s", formp->arr.arrp[2].sptr) ;
	} else if (formp->arr.arrp[2].IsLocation()) {
	    printf ("@%d",
		    varstack.VarIndex ((Variable *)formp->arr.arrp[2].location)) ;
	}
	break ;
      case A_SelectMember:
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	printf ("->") ;
	printf ("%s", formp->arr.arrp[3].sptr) ;
	break ;
      case A_ThisPtr:
	for (i = 0 ; i < level * DC_INDENT ; i++)
	    putchar (' ') ;
	// printf ("this") ;
	return 0 ;
	break ;
      case A_Self:
	for (i = 0 ; i < level * DC_INDENT ; i++)
	    putchar (' ') ;
	printf ("self") ;
	return 0 ;
      case A_SelectClass:
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	printf ("::") ;
	printf ("%s", formp->arr.arrp[3].sptr) ;
	break ;
      case A_Const:
	formp->arr.arrp[2].Print ('"') ;
	break ;
      case A_Array:
	{
	    Value *arp ;

	    arp = &formp->arr.arrp[2] ;
	    printf ("[") ;
	    for (i = 0 ; i < arp->arr.numelems ; i++) {
		if (i > 0)
		    printf (", ") ;
		st = AcDisCompile (&arp->arr.arrp[i], binst, -1) ; CHECKST(st) ;
	    }
	    printf ("]") ;
	}
	break ;
      case A_ArrIndex:
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	printf ("[") ;
	st = AcDisCompile (&formp->arr.arrp[3], binst, -1) ; CHECKST(st) ;
	printf ("]") ;
	break ;
      case A_Not:
	printf ("!") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	break ;
      case A_Negate:
	printf ("-") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	break ;
      case A_BitRev:
	printf ("~") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	break ;
      case A_Block:
      case A_SLBlock:
	{
	    AcInst theinst = binst ;
	    if (upperinst != A_For && upperinst != A_While && upperinst != A_If && upperinst != A_Nop) {
		for (i = 0 ; i < (level - 1)* DC_INDENT ; i++)
		    putchar (' ') ;
	    }
	    if (theinst == A_Block)
		printf ("{\n") ;
	    else
		printf ("{{\n") ;
	    AcInst lastinst = A_Nop ;
	    
	    for (i = 2 ; i < formp->arr.numelems ; i++) {
		Value *aCode = &formp->arr.arrp[i] ;
		AcInst inst = (AcInst)aCode->arr.arrp[0].intval ;
		if (lastinst == A_Nop) {
		    lastinst = inst ;
		} else {
		    if ((lastinst == A_VarDef && inst != A_VarDef) 
			|| (lastinst != A_VarDef && inst == A_VarDef))
			printf ("\n") ;
		    lastinst = inst ;
		}
		st = AcDisCompile (aCode, binst, level) ;
	    }
	    for (i = 0 ; i < (level-1) * DC_INDENT ; i++)
		putchar (' ') ;
	    if (theinst == A_Block)
		printf ("}") ;
	    else
		printf ("}}") ;
	    if (upperinst == A_If)
		return 0 ;
	}
	break ;
      case A_Mult:
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ;
	printf (" * ") ;
	st = AcDisCompile (&formp->arr.arrp[3], binst, -1) ;
	break ;
      case A_Div:
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ;
	printf (" / ") ;
	st = AcDisCompile (&formp->arr.arrp[3], binst, -1) ;
	break ;
      case A_Mod:
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ;
	printf (" %% ") ;
	st = AcDisCompile (&formp->arr.arrp[3], binst, -1) ;
	break ;
      case A_Plus:
      case A_Minus:
      case A_Rsh:
      case A_Lsh:
	dopar = 0 ;
	if (upperinst == A_Mult || upperinst == A_Div || upperinst == A_Mod) {
	    dopar = 1 ;
	    printf ("(") ;
	}
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ;
	opr = (binst == A_Plus) ? "+"
	    : (binst == A_Minus)  ? "-"
		: (binst == A_Rsh) ? "<<"
		    :  ">>" ;
	printf (" %s ", opr) ;
	st = AcDisCompile (&formp->arr.arrp[3], binst, -1) ;
	if (dopar)
	    printf (")") ;
	break ;
      case A_Eq:
      case A_Neq:
      case A_Lt:
      case A_Gt:
      case A_Le:
      case A_Ge:
	dopar = 0 ;
	if (upperinst == A_Mult || upperinst == A_Div || upperinst == A_Mod
	    || upperinst == A_Plus || upperinst == A_Minus
	    || upperinst == A_Lsh  || upperinst == A_Rsh)
	{
	    dopar = 1 ;
	    printf ("(") ;
	}
	opr = (binst == A_Eq)  ? "=="
	    : (binst == A_Neq) ? "!="
		: (binst == A_Lt) ? "<"
		    : (binst == A_Gt) ? ">"
			: (binst == A_Le) ? "<="
			    : ">=" ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ;
	printf (" %s ", opr) ;
	st = AcDisCompile (&formp->arr.arrp[3], binst, -1) ;
	if (dopar)
	    printf (")") ;
	break ;
      case A_And:
      case A_Band:
	{
	    dopar = 0 ;
	    if (upperinst == A_Mult || upperinst == A_Div || upperinst == A_Mod
		|| upperinst == A_Plus || upperinst == A_Minus
		|| upperinst == A_Lsh  || upperinst == A_Rsh
		|| upperinst == A_Eq   || upperinst == A_Neq
		|| upperinst == A_Lt   || upperinst == A_Gt
		|| upperinst == A_Le   || upperinst == A_Ge)
	    {
		dopar = 1 ;
		printf ("(") ;
	    }
	    st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ;
	    if (binst == A_And)
		printf (" && ") ;
	    else
		printf (" & ") ;
	    st = AcDisCompile (&formp->arr.arrp[3], binst, -1) ;
	    if (dopar)
		printf (")") ;
	}
	break ;
      case A_Or:
      case A_Bor:
      case A_Xor:
	dopar = 0 ;
	if (upperinst == A_Mult || upperinst == A_Div || upperinst == A_Mod
	    || upperinst == A_Plus || upperinst == A_Minus
	    || upperinst == A_Lsh  || upperinst == A_Rsh
	    || upperinst == A_Eq   || upperinst == A_Neq
	    || upperinst == A_Lt   || upperinst == A_Gt
	    || upperinst == A_Le   || upperinst == A_Ge
	    || upperinst == A_And)
	{
	    dopar = 1 ;
	    printf ("(") ;
	}
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ;
	if (binst == A_Or)
	    printf (" || ") ;
	else if (binst == A_Bor)
	    printf (" | ") ;
	else if (binst == A_Xor)
	    printf (" ^ ") ;
	st = AcDisCompile (&formp->arr.arrp[3], binst, -1) ;
	if (dopar)
	    printf (")") ;
	break ;
      case A_Assign:
      case A_AddAssign:
      case A_SubAssign:
      case A_LshAssign:
      case A_RshAssign:
      case A_MultAssign:
      case A_DivAssign:
      case A_ModAssign:
      case A_RefAssign:
      case A_OrAssign:
      case A_AndAssign:
      case A_XorAssign:
	dopar = 0 ;
	if (upperinst == A_Mult || upperinst == A_Div || upperinst == A_Mod
	    || upperinst == A_Plus || upperinst == A_Minus
	    || upperinst == A_Lsh  || upperinst == A_Rsh
	    || upperinst == A_Eq   || upperinst == A_Neq
	    || upperinst == A_Lt   || upperinst == A_Gt
	    || upperinst == A_Le   || upperinst == A_Ge
	    || upperinst == A_And  || upperinst == A_Or)
	{
	    dopar = 1 ;
	    printf ("(") ;
	}
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ;
	printf (" %s ",
		(binst == A_Assign) ? "="
		: (binst == A_RefAssign)  ? ":="
		: (binst == A_AddAssign)  ? "+="
		: (binst == A_SubAssign)  ? "-="
		: (binst == A_MultAssign) ? "*="
		: (binst == A_DivAssign)  ? "/="
		: (binst == A_ModAssign)  ? "%="
		: (binst == A_LshAssign)  ? "<<="
		: (binst == A_RshAssign)  ? ">>="
		: (binst == T_AndAssign)  ? "&="
		: (binst == T_OrAssign)   ? "|="
		: (binst == T_XorAssign)  ? "^="
		: "??") ;
	st = AcDisCompile (&formp->arr.arrp[3], binst, -1) ;
	if (dopar)
	    printf (")") ;
	break ;
      case A_New:
	printf ("new ") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	break ;
      case A_PPPre:
	printf ("++") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	break ;
      case A_MMPre:
	printf ("--") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	break ;
      case A_PPPost:
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	printf ("++") ;
	break ;
      case A_MMPost:
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	printf ("--") ;
	break ;
      case A_Return:
      case A_Break:
      case A_Exit:
	if (binst == A_Return)
	    printf ("return") ;
	else if (binst == A_Break)
	    printf ("break") ;
	else
	    printf ("exit") ;
	if (formp->arr.numelems == 3) {
	    printf (" ") ;
	    st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	}
	printf (" ;\n") ;
	return 0 ;
      case A_Continue:
	printf ("continue") ;
	break ;
      case A_Eval:
	printf ("eval (") ;
	if (formp->arr.arrp[2].IsString()) {
	    putchar ('"') ;
	    formp->arr.arrp[2].Print() ;
	    putchar ('"') ;
	} else {
	    st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	}
	printf (")") ;
	break ;
      case A_Load:
	printf ("load (") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	printf (")") ;
	break ;
      case A_Package:
	printf ("package (") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	printf (")") ;
	break ;
      case A_Pause:
	printf ("pause\n") ;
	break ;
      case A_ArgCnt:
	printf ("$#") ;
	break ;
      case A_BoundArg:
	printf ("$?%s", formp->arr.arrp[2].sptr) ;
	break ;
      case A_Arg:
	printf ("$[") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ;
	printf ("]") ;
	break ;
      case A_Position:
	printf ("$$") ;
	break ;
      case A_ArgList:
	printf ("$*") ;
	break ;
      case A_Throw:
	printf ("throw ") ;
	st = AcDisCompile (&formp->arr.arrp[2], binst, -1) ; CHECKST(st) ; 
	break ;
      default:
	return SetErrorCode (SYNTAX_ERROR) ;
    }
    if (upperinst == A_Block || upperinst == A_SLBlock
	|| ((upperinst == A_FuncDef || upperinst == A_MethoDef) && binst != A_Block)) {
	printf (" ;") ;
    }
    if (level >= 0)
	printf ("\n") ;
    return 0 ;
}

