//----------------------------------------------------------------
// heap.cpp : Cmm C^[v^: q[vGAǗ
//----------------------------------------------------------------
#include <stdio.h>
#include <ctype.h>
#include <cmm/class.h>
#include <cmm/value.h>
#include <cmm/stack.h>
#include <cmm/heap.h>
#include <cmm/cmm.h>

#ifndef NULL
#define NULL 0
#endif

#define DEBUG_DUMP

int __DUMP = 0 ;

int _GC1 = 1 ;
int _AR1 = 1 ;

#define IN_AREA(p,start,end) (((start)<=(char *)(p)) && ((char *)(p)<(end)))
#define MARKIT(ptr, m, area_start, area_end)\
	{((SegmentHeader *)(((char *)ptr)-sizeof(SegmentHeader)))->mark = m ;}

#define RESERVED_FOR_DISPOSE (1024*5)

//----------------------------------------------------------------
// q[vGÅ{\
//    Cmm ̃q[vGÅ{Iȍ\͈ȉ̂悤Ȃ
//
//      heap_area
//              |
//              v
// [U[   +---------------------------+ -------------------
// n|C^ | SegmentHeader (mark, size)|  ^  ^           ^
//   |          +...........................+ size|           |
//   +--------->| ^[U[̗vTCY |  |  |           |
//              | v                         |  v  |           |
// [U[   +---------------------------+ --- | used_size |
// n|C^ | SegmentHeader	(mark, size)|  ^  |           |
//   |          +...........................+  |  |           |
//   +--------->|                           |  |  |           |
//              |                           |  v  |           | max_size
//          --- +---------------------------+ --- |           |
//         /    | SegmentHeader (mark, size)|  ^  |           |
//  q[v|     +...........................+  |  |           |
//  ZOg  |                           |  |  |           |
//        |     |                           |  |  |           |
//         \    |                           |  v  v           |
//          --- +---------------------------+ ------          |
//              |                           |  ^              |
//	        |                           |  |              |
//	        |                           |  | rest_size    |
//reserved_area |                           |  |              |
//   |          |                           |  v              v
//   |          +---------------------------+ -------------------
//   +--------->|SegmentHeader (mark, size) |                 ^
//	        |...........................|                 |
// 		|                           |                 |reserv_size
//	        |                           |		      |
//	        +---------------------------+ ---	      |
// reservpt---->|                           |  ^              |
//	        |                           |rest_of_reserved |
//	        |                           |  v              v
//	        +---------------------------+ -------------------
// 
// {Iɂ́Aheap_area + used_size ̈ʒuɁA
// 
//	<[U[̗vTCY> + sizeof(SegmentHeader)
// 
// ̃q[vZOgmۂÃZOg̍ŏ̕ɁAZO
// gwb_uAheap_area + used_size + sizeof(SegmentHeader) ̃|
// C^[U[ɓnB
//
// reserved_area ́AfBX|[UĂяo̎Ɏg\̈̐擪
// ւ̃|C^ŁAfBX|[UĂяoɂ́AZOg
// BĂяoIAAďB
//----------------------------------------------------------------
Heap::Heap()
{
    dynamic_allocate = 0 ;
    max_size  = 0 ;
    used_size = 0 ;
    rest_size = 0 ;
    alloc_count = gc_count = grow_count = gc1_count = sweep_count = alloc_free_count = 0 ;
    heap_area = NULL ;
}

Heap::~Heap()
{
    if (dynamic_allocate)
	delete heap_area ;
}

void Heap::Init (char *area, int size)
{
    dynamic_allocate = 0 ;
    heap_area = area ;
    if (size < RESERVED_FOR_DISPOSE * 4) {
	fprintf (stderr, "Heap area is too Small\n") ;
	exit (1) ;
    }
    SetupHeap (size) ;
}

void Heap::Init (int size)
{
    if (dynamic_allocate && heap_area) {
	delete heap_area ;
    }
    dynamic_allocate = 1 ;
    heap_area = new char [size + RESERVED_FOR_DISPOSE] ;
    SetupHeap (size + RESERVED_FOR_DISPOSE) ;
}

void Heap::SetupHeap (int size)
{
    int i ;

    max_size = size - RESERVED_FOR_DISPOSE ;
    used_size = 0 ;
    rest_size = size - RESERVED_FOR_DISPOSE ;
    for (i = 0 ; i < size ; i++)
	heap_area[i] = 0 ;
    freeList  = NULL ;
    disposing = 0 ;

    reserved_area = heap_area + max_size ;
    reservpt = reserved_area ;
    reserv_size = RESERVED_FOR_DISPOSE ;
    rest_of_reserved = reserv_size ;
}

int Heap::CopyValue (Value *v, VarStack *stack, Value *retv)
{
    Value xx ;
    char *s ;
    Value *vp ;
    Value *destp ;

    vp = stack->TmPush (*v) ;
    destp = stack->TmPush (xx) ;

    if (v->IsString()) {
	s = AllocString (vp->sptr, stack) ;
	if (!s)
	    return -1 ;
	*destp = s ;
    } else if (v->IsARRAY ()) {
	if (CopyArray (vp, stack, destp) < 0)
	    return -1 ;
    } else {
	*destp = *vp ;
    }
    *retv = *destp ;
    stack->TmPop (2) ;
    return 0 ;
}

char *Heap::AllocString (char *string, VarStack *stack)
{
    char *strptr ;
    Value *xp ;

    //
    // string q[vGÃ|C^̏ꍇAȉ AllocSegment 
    // Ŕ Grabage Collection ɂĂ̒l\
    // ̂ŁA܂AX^bNɂ̃|C^mۂAGarbage
    // Collection ŖɂȂȂ悤ɂB
    //
    xp = stack->TmPush (string) ;
    strptr = AllocSegment (strlen(string) + 1, stack) ;
    if (!strptr)
	return NULL ;
    strcpy (strptr, xp->sptr) ;
    SetType (strptr, HP_STRING) ;
    stack->TmPop (1) ;
    return strptr ;
}

char *Heap::AllocString (int len, VarStack *stack)
{
    char *pt ;

    pt = AllocSegment (len, stack) ;
    if (!pt)
	return NULL ;
    SetType (pt, HP_STRING) ;
    return pt ;
}

Class *Heap::AllocClass (VarStack *stack)
{
    Class *cp ;

    cp = (Class *)AllocSegment (sizeof(Class), stack) ;
    SetType ((char *)cp, HP_CLASS) ;
    return cp ;
}

Instance *Heap::AllocInstance (int num_members, VarStack *stack, Value **memberslot)
{
    Instance *ip ;
    int i ;

    ip = (Instance *)AllocSegment (sizeof(Instance) + sizeof(Value) * num_members, stack) ;
    SetType ((char *)ip, HP_INSTANCE) ;

    Value *vp = (Value *)((char *)ip + sizeof(Instance)) ;

    for (i = 0 ; i < num_members ; i++)
	vp[i].Init() ;
    *memberslot = vp ;
    return ip ;
}

Value Heap::NewArray (int size, VarStack *stack)
{
    Value *arr ;
    Value ret ;

    if (AllocArray (size, stack, &arr) < 0) {
	ret.Init () ;
    } else {
	ret.InitArr (arr, size) ;
    }
    return ret ;
}

int Heap::AllocArray (int size, VarStack *stack, Value **arrp)
{
    Value *retv ;
    int i ;

    if (size == 0) {
	*arrp = NULL ;
	return 0 ;
    }
    retv = (Value *)AllocSegment (sizeof(Value) * size, stack) ;
    if (!retv)
	return -1 ;
    SetType ((char *)retv, HP_ARRAY) ;
    for (i = 0 ; i < size ; i++)
	retv[i].Init() ;
    *arrp = retv ;
    return 0 ;
}

//----------------------------------------------------------------
//  AllocArray ́A̔z̃TCY傫Ȃł낤Ƃ
// \肵āȂȕ̃GAmۂȂz𐶐邽
// ̃\bhłB̗\zւ̗vf̒ǉ́AHeap::AddArray ōs
// B
// ́AA-Code RpC AppendArray gďƂɁA]
// ɂAAllocSegment pAɏd߂ Heap::AddArray 
// Ƃɗpӂꂽ\bhłBgƂɂAq[vGA
// ɑ΂ARpC̃AP[V 1/50 xɂ܂ŉ
// 邱ƂłI
//----------------------------------------------------------------
Value Heap::AllocArray (int size, int reserve, VarStack *stack)
{
    if (_AR1) {
	Value ret ;
	Value *arrp ;
	if (reserve < size)
	    reserve = size ;
	if (reserve == 0) {
	    ret.arr.numelems = 0 ;
	    ret.arr.arrp = NULL ;
	    return ret ;
	}
	arrp = (Value *)AllocSegment (sizeof(Value) * reserve, stack) ;
	if (!arrp)
	    ret.Init () ;
	SetType ((char *)arrp, HP_ARRAY) ;
	int i ;
	for (i = 0 ; i < reserve ; i++)
	    arrp[i].Init ();
	ret.InitArr (arrp, size) ;
	return ret ;
    } else {
	Value ret ;
	Value *arrp ;
	if (size > 0) {
	    arrp = (Value *)AllocSegment (sizeof(Value) * size, stack) ;
	    if (!arrp)
		ret.Init () ;
	} else {
	    arrp = NULL ;
	}
	SetType ((char *)arrp, HP_ARRAY) ;
	int i ;
	for (i = 0 ; i < size ; i++)
	    arrp[i].Init ();
	ret.InitArr (arrp, size) ;
	return ret ;
    }
}

int Heap::AddArray (Value *a, Value *b, VarStack *stack, Value *retv)
{
    if (!_AR1) {
	return AppendArray (a, b, stack, retv) ;
    }
    int nelems = a->arr.numelems ;
    int reselems ;
    Value *ap, *bp ;
    Value newv ;
    Value *retp ;

    ap = stack->TmPush (*a) ;
    bp = stack->TmPush (*b) ;

    if (!ap->arr.arrp) {
	newv = AllocArray (0, 10, stack) ;
	retp = stack->TmPush (newv.arr) ;
	// CopyValue (bp, stack, &bcopied) ;
	retp->arr.arrp[0] = *bp ; // bcopied ;
	retp->arr.numelems = 1 ;
	*retv = stack->VPop () ;
	stack->TmPop (2) ;
	return 0 ;
    }
    // q[vGAɎۂɊmۂĂ郁TCYA̔z
    // ő\vf𓾂
    reselems = SegmentSize (ap) / sizeof(Value) ;
    if (nelems < reselems) {
	//
	// ܂Aq[vɔzu΂v̗]T
	//
	// CopyValue (bp, stack, &bcopied) ;
	ap->arr.arrp[nelems] = *bp ; // bcopied ;
	ap->arr.numelems = nelems+1 ;
	*retv = *ap ;
	stack->TmPop (2) ;
	return 0  ;
    } else {
	int i ;
	//
	// q[vɗ]TȂBVɁu傫ȁvz̈mۂA
	// Ɍ݂̔zlRs[B
	//
	newv = AllocArray (ap->arr.numelems, ap->arr.numelems + 24, stack) ;
	if (newv.IsVoid()) {
	    return -1 ;
	}
	for (i = 0 ; i < ap->arr.numelems ; i++) {
	    newv.arr.arrp[i] = ap->arr.arrp[i] ;
	}
	retp = stack->TmPush (newv.arr) ;
	// CopyValue (bp, stack , &bcopied) ;
	retp->arr.arrp[i] = *bp ; // bcopied ;
	retp->arr.numelems = ap->arr.numelems + 1;
	*retv = stack->VPop () ;
	stack->TmPop (2) ;
	return 0 ;
    }
}

int Heap::CopyArray (Value *v, VarStack *stack, Value *retv)
{
    Value *ap ;
    Value *destp ;
    Value *srcp ;
    Value xx ;
    int st ;

    srcp = stack->TmPush (*v) ;

    if (AllocArray (v->arr.numelems, stack, &ap) < 0)
	return -1 ;
    xx.InitArr (ap, v->arr.numelems) ;
    destp = stack->TmPush (xx) ; // to avoid garbage collection
    st = CopyArrayElements (destp, 0, srcp, stack) ;
    *retv = *destp ;
    stack->TmPop (2) ;
    return st ;
}

int Heap::CopyArrayElements (Value *xp, int offset, Value *v, VarStack *stack)
{
    int i, j ;
    char *s ;
    Value *ap ;

    for (i = offset, j = 0 ; j < v->arr.numelems ; i++, j++) {
	if (v->arr.arrp[j].IsString()) {
	    if (!(s = AllocString (strlen(v->arr.arrp[j].sptr) + 1, stack)))
		return -1 ;
	    xp->arr.arrp[i] = s ;
	    strcpy (s, v->arr.arrp[j].sptr) ;
	} else if (v->arr.arrp[j].IsARRAY()) {
	    int nelems = v->arr.arrp[j].arr.numelems ;
	    Value *subsrc, src ;
	    Value *subdest, dest ;
#ifdef DEBUG_DUMP
	    if (__DUMP) {
		printf ("## Heap area = %x ##", (int)this->heap_area) ;
		stack->Dump ("============ copy subarry 1===============") ;
	    }
#endif
	    AllocArray (nelems, stack, &ap) ;
	    dest.InitArr (ap, nelems) ;
	    subdest = stack->TmPush (dest) ;

	    src = v->arr.arrp[j] ;
	    subsrc = stack->TmPush (src) ;

	    CopyArrayElements (subdest, 0, subsrc, stack) ;

	    xp->arr.arrp[i] = *subdest ;
#ifdef DEBUG_DUMP
	    if (__DUMP) {
		printf ("## Heap area = %x ##", (int)this->heap_area) ;
		stack->Dump ("=== copy subarry 2===============") ;
	    }
#endif
	    stack->Pop(2) ;
	} else {
	    xp->arr.arrp[i] = v->arr.arrp[j] ;
	}
    }
    return 0 ;
}

int Heap::ConsArray (Value *a, Value *b, VarStack *stack, Value *retv)
{
    Value xx ;
    Value *ap ;
    Value *destp ;
    Value *srcp1 ;
    Value *srcp2 ;
    int nums = a->arr.numelems + b->arr.numelems ;
    int st ;

    srcp1 = stack->TmPush (*a) ;
    srcp2 = stack->TmPush (*b) ;

    if (AllocArray (nums, stack, &ap) < 0)
	return -1 ;
    xx.InitArr (ap, nums) ;
    destp = stack->TmPush (xx) ;

    st = CopyArrayElements (destp, 0, srcp1, stack) ;
    if (st < 0)
	return -1 ;
    st = CopyArrayElements (destp, srcp1->arr.numelems, srcp2, stack) ;
    *retv =  *destp ;
    stack->TmPop(3) ;
    return st ;
}

int Heap::AppendArray (Value *a, Value *b, VarStack *stack, Value *retv)
{
    int st ;
    Value *srcp1, *srcp2, *destp ;
    Value xx  ;
    Value *ap ;
    Value rv ;
    int nums = a->arr.numelems+1 ;

    srcp1 = stack->TmPush (*a) ;
    srcp2 = stack->TmPush (*b) ;
    if (AllocArray (nums, stack, &ap) < 0)
	return -1 ;
    xx.InitArr (ap, nums) ;
    destp = stack->TmPush (xx) ;
    st = CopyArrayElements (destp, 0, srcp1, stack) ;
    if (st < 0)
	return -1 ;
    CopyValue (srcp2, stack, &rv) ;
    destp->arr.arrp[nums-1] = rv ;
    stack->TmPop (3) ;
    *retv = *destp ;
    return st ;
}

int Heap::PopArray (Value *a, int n, VarStack *stack, Value *retv)
{
    Value *arrp ;
    int nelems ;
    Value *ap ;
    Value *destp ;
    int i ;

    arrp = stack->TmPush (*a) ;
    if ((nelems = arrp->arr.numelems - n) < 0)
	nelems = 0 ;
    if (AllocArray (nelems, stack, &ap) < 0)
	return -1 ;
    destp = stack->TmPush () ;
    destp->InitArr (ap, nelems) ;
    for (i = 0 ; i < nelems ; i++) {
	Value v ;
	CopyValue (&arrp->arr.arrp[i], stack, &v) ;
	destp->arr.arrp[i] = v ;
    }
    *retv = stack->VPop () ;
    stack->TmPop () ;
    return 0 ;
}

int Heap::RemoveArray (Value *a, Value *v, VarStack *stack, Value *retv)
{
    int i, j, iindex = -1 ;
    Value *arrp, *vp ;

    arrp = stack->TmPush (*a) ;
    vp   = stack->TmPush (*v) ;

    for (iindex = -1, i = 0 ; i < arrp->arr.numelems ; i++) {
	if (arrp->arr.arrp[i] == *vp) {
	    iindex = i ;
	    break ;
	}
    }
    if (iindex < 0) {
	Value ret ;
	CopyValue (a, stack, &ret) ;
	stack->TmPop (2) ;
	*retv = ret ;
    } else {
	int nelems = arrp->arr.numelems - 1;
	Value *destp, *ap ;

	if (AllocArray (nelems, stack, &ap) < 0)
	    return -1 ;
	destp = stack->TmPush () ;
	destp->InitArr (ap, nelems) ;

	for (i = j = 0 ; i < arrp->arr.numelems ; i++) {
	    if (i == iindex)
		continue ;
	    Value elem ;
	    CopyValue (&a->arr.arrp[i], stack, &elem) ;
	    destp->arr.arrp[j++] = elem ;
	}
	*retv = *destp ;
	stack->TmPop (3) ;
    }
    return 0 ;
}

int Heap::RemoveArray (Value *a, int idx, VarStack *stack, Value *retv)
{
    int i, j ;
    Value *arrp ;

    if (idx < 0 || idx >= a->arr.numelems) {
	Value ret ;
	CopyValue (a, stack, &ret) ;
	*retv = ret ;
    } else {
	int nelems = a->arr.numelems - 1;
	Value *destp, *ap ;

	arrp = stack->TmPush (*a) ;
	if (AllocArray (nelems, stack, &ap) < 0)
	    return -1 ;
	destp = stack->TmPush () ;
	destp->InitArr (ap, nelems) ;

	for (i = j = 0 ; i < arrp->arr.numelems ; i++) {
	    if (i == idx)
		continue ;
	    Value elem ;
	    CopyValue (&a->arr.arrp[i], stack, &elem) ;
	    destp->arr.arrp[j++] = elem ;
	}
	*retv = *destp ;
	stack->TmPop (2) ;
    }
    return 0 ;
}

#define SEG_MIN_SIZE 128

inline char *Heap::GetFree (int size)
{
    int segsize = size + sizeof(SegmentHeader) ;
    FreeSegmentHeader *freep ;
    FreeSegmentHeader *prevp ;

    for (prevp = NULL, freep = freeList ;
	 freep ;
	 prevp = freep, freep = freep->next)
    {
	if (freep->size >= segsize) {
	    if (freep->size - segsize > sizeof(SegmentHeader) + SEG_MIN_SIZE) {
		// split free segment
		FreeSegmentHeader *hp2 = (FreeSegmentHeader *)((char *)freep + segsize) ;
		hp2->size = freep->size - segsize ;
		hp2->mark = 0 ;
		hp2->type = HP_PURE ;
		hp2->next = freep->next ;
		freep->size = segsize ;
		if (!prevp)
		    freeList = hp2 ;
		else
		    prevp->next = hp2 ;
	    } else {
		numFreeList-- ;
		if (!prevp) {
		    freeList = freep->next ;
		} else {
		    prevp->next = freep->next ;
		}
	    }
	    alloc_free_count++ ;
	    freep->mark = 0 ;
	    freep->type = HP_NONE ;
	    return (char *)freep + sizeof(SegmentHeader) ;
	}
    }
    return NULL ;
}

inline char *Heap::GetArea (int size)
{
    int area_size = sizeof (Heap::SegmentHeader) + size ;
    char *area_ptr ;

    if (rest_size < area_size)
	return NULL ;
    alloc_count++ ;
    area_ptr = heap_area + used_size ;
    ((Heap::SegmentHeader *)area_ptr)->mark = 0 ;
    ((Heap::SegmentHeader *)area_ptr)->type = HP_NONE ;
    ((Heap::SegmentHeader *)area_ptr)->size = area_size  ;
    used_size += area_size ;
    rest_size -= area_size ;
    return area_ptr + sizeof(Heap::SegmentHeader) ;
}

inline void Heap::MarkClass (Class *cl, int mark, char *area_start, char *area_end)
{
    int i ;

    if (!cl)
	return ;
    if (cl->superclass) {
	if (!GetMark ((char *)cl->superclass)) {
	    MarkIt ((char *)cl->superclass, mark, area_start, area_end) ;
	    MarkClass (cl->superclass, mark, area_start, area_end) ;
	}
    }
    if (cl->class_template) {
	if (!GetMark ((char *)cl->class_template)) {
	    MarkIt ((char *)cl->class_template, mark, area_start, area_end) ;
	}
    }
    if (cl->class_vars_or_methods) {
	if (!GetMark ((char *)cl->class_vars_or_methods)) {
	    MarkIt ((char *)cl->class_vars_or_methods, mark, area_start, area_end) ;
	    MarkArray (cl->class_vars_or_methods,
		       cl->num_cvs_methods,
		       mark, area_start, area_end) ;
	}
    }
}

inline void Heap::MarkInstance (Instance *ip, int mark, char *area_start, char *area_end)
{
    if (!ip)
	return ;
    if (ip->classp) {
	if (!GetMark ((char *)ip->classp)) {
	    MarkIt ((char *)ip->classp, mark, area_start, area_end) ;
	    MarkClass (ip->classp, mark, area_start, area_end) ;
	}
    }
    if (ip->instance_values) {
	MarkArray (ip->instance_values, ip->num_slots, mark, area_start, area_end) ;
    }
}

inline void Heap::MarkStackElement (Variable *vp, int mark, char *area_start, char *area_end)
{
    char *areap ;
    Heap::SegmentHeader *hp ;

    if (vp->IsString()) {
	areap = vp->sval() ;
    } else if (vp->IsARRAY()) {
	areap = (char *)vp->arr.arrp ;
    } else if (vp->IsContext()) {
	areap = vp->context.text ;
    } else if (vp->IsClass()) {
	areap = (char *)vp->classp ;
    } else if (vp->IsInstance ()) {
	areap = (char *)vp->instance ;
    } else {
	return ;
    }
    if (area_start <= areap && areap < area_end) {
	// q[v̈ȊOwĂ|C^͖B
	hp = (Heap::SegmentHeader *)(areap - sizeof(Heap::SegmentHeader)) ;
	hp->mark = mark ;
    }
    if (vp->IsARRAY()) {
	MarkArray (vp->arr.arrp, vp->arr.numelems, mark, area_start, area_end) ;
    } else if (vp->IsClass()) {
	MarkClass (vp->classp, mark, area_start, area_end) ;
    } else if (vp->IsInstance()) {
	MarkInstance (vp->instance, mark, area_start, area_end) ;
    }
}

void Heap::DoMark (VarStack *stack, int mark)
{
    char *heap_end = heap_area + used_size ;

    MarkStack (stack, mark, heap_area, heap_end) ;
}

//----------------------------------------------------------------
// MarkStack :: X^bNQƂĂq[vZOg̃GAwb
// _Ƀ}[Nݒ肷B
//----------------------------------------------------------------
inline void Heap::MarkStack (VarStack *stack, int mark, char *area_start, char *area_end)
{
    int i ;
    Variable *vp ;

    for (i = 0 ; i < stack->StackPtr() ; i++) {
	vp = stack->PeekStack(i) ;
	MarkStackElement (vp, mark, area_start, area_end) ;
    }
    for (i = stack->LvalPtr() ; i < stack->Max() ; i++) {
	vp = stack->PeekStack(i) ;
	MarkStackElement (vp, mark, area_start, area_end) ;
    }
}

//----------------------------------------------------------------
// MarkArray:: z^f[^̊evfŁAq[vZOgQƂĂ
// ̂ɂāÃq[vZOg̃GAwb_Ƀ}[N
//----------------------------------------------------------------
inline void Heap::MarkArray (Value *v, int numelems, int mark, char *area_start, char *area_end)
{
    Value *ap ;
    Value *endap ;

    if (!v)
	return ;

    ap = v ;
    endap = v + numelems ;
    for ( ; ap < endap ; ap++) {
	if (ap->IsString() && IN_AREA(ap->sptr, area_start, area_end)) {
	    if (!GetMark (ap->sptr)) {
		MarkIt (ap->sptr, mark, area_start, area_end) ;
	    }
	} else if (ap->IsARRAY() /* && IN_AREA(ap->arr.arrp, area_start, area_end) */) {
	    // z͕Kq[v̒ɂ
	    if (ap->arr.arrp) {
		// z̗vfA܂zȂ炻ׂă}[N
		if (!GetMark ((char *)ap->arr.arrp)) {
		    MarkIt ((char *)ap->arr.arrp, mark, area_start, area_end) ;
		    MarkArray (ap->arr.arrp, ap->arr.numelems, mark, area_start, area_end) ;
		}
	    }
	} else if (ap->IsContext() && IN_AREA(ap->context.text, area_start, area_end)) {
	    if (!GetMark (ap->context.text)) {
		MarkIt ((char *)ap->context.text, mark, area_start, area_end) ;
	    }
	} else if (ap->IsClass() /* && IN_AREA(ap->classp, area_start, area_end) */) {
	    if (!GetMark ((char *)ap->classp)) {
		MarkIt ((char *)ap->classp, mark, area_start, area_end) ;
		MarkClass (ap->classp, mark, area_start, area_end) ;
	    }
	} else if (ap->IsInstance() /* && IN_AREA(ap->instance, area_start, area_end) */) {
	    if (!GetMark ((char *)ap->instance)) {
		MarkIt ((char *)ap->instance, mark, area_start, area_end) ;
		MarkInstance (ap->instance, mark, area_start, area_end) ;
	    }
	}
    }
}

//----------------------------------------------------------------
// MovePointer :: X^bÑf[^ŁAړΏۂƂȂq[vZOg
// QƂĂ̂Aړ̃q[vZOgQƂ悤ɁA
// X^bÑ|C^̒lύXB
//----------------------------------------------------------------
inline void Heap::MovePointer (VarStack *stack, char *ptr, char *limit, int movesize)
{
    int i ;
    Variable *vp ;

    for (i = 0 ; i < stack->StackPtr() ; i++) {
	vp = stack->PeekStack (i) ;
	MoveValue (vp, ptr, limit, movesize) ;
    }
    for (i = stack->LvalPtr() ; i < stack->Max() ; i++) {
	vp = stack->PeekStack (i) ;
	if (vp->IsLocation()) {
	    if (ptr <= (char *)vp->location && (char *)vp->location < limit) {
		vp->location = (Value *)((char *)vp->location - movesize) ;
	    }
	} else {
	    MoveValue (vp, ptr, limit, movesize) ;
	}
    }
}

inline void Heap::MoveArrPtr (Value *v, int numelems, char *ptr, char *limit, int movesize)
{
    int i ;

    for (i = 0 ; i < numelems ; i++) {
	MoveValue (&v[i], ptr, limit, movesize) ;
    }
}

//----------------------------------------------------------------
// MoveValue:: |C^̕tς
// ړΏۂƂȂq[vZOgQƂĂ̂ɂāA̎Q
// ړ̃q[vZOgɂȂ悤Ƀ|C^̒lύXB
// QƐ悪z/Class/Instancȅꍇ́Aɂ̐ɁAړΏۃq[v
// ZOg\̂ŁAċAIɈړΏۂTB
//
// ܂Az/Class/Instance ́ÁAX^bN邢́A̔zv
// fQƂĂꍇB̏ꍇAړ̈ʒuړΏ۔
// ͓ɂꍇ
//
//    		(ptr < heap_segment_size - movesize < limit)
//
// QƈʒuύXs|C^AɂxʒuύXsĂ
// B邽߂ɁAʒuύXsq[vZOgɑ΂ẮA
// GAwb_̃}[N 2ɂāA}[N 2 ̔zɂẮAړ
// sȂ悤ɂĂB
//----------------------------------------------------------------
inline void Heap::MoveValue (Value *vp, char *ptr, char *limit, int movesize)
{
    Heap::SegmentHeader *hp ;

    if (vp->IsString()) {
	if (ptr <= vp->sptr &&  vp->sptr < limit) {
	    vp->sptr -= movesize ;
	}
    } else if (vp->IsARRAY()) {
#if 0
	if (vp->arr.numelems == 0)
	    return ;
#endif
	if (!vp->arr.arrp)
	    return ;
	hp = (Heap::SegmentHeader *)((char *)vp->arr.arrp - sizeof(Heap::SegmentHeader)) ;
	if (hp->mark == 1) {
	    hp->mark++ ;
	    MoveArrPtr (vp->arr.arrp, vp->arr.numelems, ptr, limit, movesize) ;
	}
	if (ptr <= (char *)vp->arr.arrp && (char *)vp->arr.arrp < limit) {
	    vp->arr.arrp = (Value *)((char *)vp->arr.arrp - movesize) ;
	}
    } else if (vp->IsClass()) {
	if (vp->classp) {
	    hp = (Heap::SegmentHeader *)((char *)vp->classp - sizeof(Heap::SegmentHeader)) ;
	    if (hp->mark == 1) {
		hp->mark++ ;
		MoveClass (vp->classp, ptr, limit, movesize) ;
	    }
	    if (ptr <= (char *)vp->classp && (char *)vp->classp < limit) {
		vp->classp = (Class *)((char *)vp->classp - movesize) ;
	    }
	}
    } else if (vp->IsInstance ()) {
	if (vp->instance) {
	    hp = (Heap::SegmentHeader *)((char *)vp->instance - sizeof(Heap::SegmentHeader)) ;
	    if (hp->mark == 1) {
		hp->mark++ ;
		MoveInstance (vp->instance, ptr, limit, movesize) ;
	    }
	    if (ptr <= (char *)vp->instance && (char *)vp->instance < limit) {
		vp->instance = (Instance *)((char *)vp->instance - movesize) ;
	    }
	}
    } else if (vp->IsContext()) {
	if (ptr <= vp->context.text && vp->context.text < limit) {
	    vp->context.text -= movesize ;
	}
    } else if (vp->IsLocation()) {
	if (ptr <= (char *)vp->location &&  (char *)vp->location < limit) {
	    vp->location = (Value *)((char *)vp->location - movesize) ;
	}
    }
}

inline void Heap::MoveClass (Class *cl, char *ptr, char *limit, int movesize)
{
    Heap::SegmentHeader *hp ;
    int i ;

    if (cl->superclass) {
	hp = (Heap::SegmentHeader *)((char *)cl->superclass - sizeof(Heap::SegmentHeader)) ;
	if (hp->mark == 1) {
	    hp->mark++ ;
	    MoveClass (cl->superclass, ptr, limit, movesize) ;
	}
	if (ptr <= (char *)cl->superclass && (char *)cl->superclass < limit)
	{
	    cl->superclass = (Class *)((char *)cl->superclass - movesize) ;
	}
    }
    if (cl->class_template) {
	hp = (Heap::SegmentHeader *)((char *)cl->class_template - sizeof(Heap::SegmentHeader)) ;
	if (hp->mark == 1) {
	    hp->mark++ ;
#if 0
	    for (i = 0 ; i < cl->num_members ; i++) {
		Class::ClassMemberTemplate *tp = &cl->class_template[i] ;
		if (ptr <= tp->name && tp->name < limit) {
		    tp->name -= movesize ;
		}
	    }
#endif
	}
	if (ptr <= (char *)cl->class_template
	    && (char *)cl->class_template < limit)
	{
	    cl->class_template =
		(Class::ClassMemberTemplate *)((char *)cl->class_template
					       - movesize) ;
	}
    }
    if (cl->class_vars_or_methods) {
	hp = (Heap::SegmentHeader *)((char *)cl->class_vars_or_methods
				     - sizeof (Heap::SegmentHeader)) ;
	if (hp->mark == 1) {
	    hp->mark++ ;
	    MoveArrPtr (cl->class_vars_or_methods,
			cl->num_cvs_methods, ptr, limit, movesize) ;
	}
	if (ptr <= (char *)cl->class_vars_or_methods
	    && (char *)cl->class_vars_or_methods < limit)
	{
	    cl->class_vars_or_methods
		= (Value *)((char *)cl->class_vars_or_methods - movesize) ;
	}
    }
}

inline void Heap::MoveInstance (Instance *ip, char *ptr, char *limit, int movesize)
{
    Heap::SegmentHeader *hp ;

    if (ip->classp) {
	hp = (Heap::SegmentHeader *)((char *)ip->classp - sizeof (Heap::SegmentHeader)) ;
	if (hp->mark == 1) {
	    hp->mark++ ;
	    MoveClass (ip->classp, ptr, limit, movesize) ;
	}
	if (ptr <= (char *)ip->classp && (char *)ip->classp < limit) {
	    ip->classp = (Class *)((char *)ip->classp - movesize) ;
	}
    }
    if (ip->instance_values) {
	MoveArrPtr (ip->instance_values, ip->num_slots, ptr, limit, movesize) ;
	if (ptr <= (char *)ip->instance_values
	    && (char *)ip->instance_values < limit)
	{
	    ip->instance_values = (Value *)((char *)ip->instance_values - movesize) ;
	}
    }
}

//----------------------------------------------------------------
// RemapPointer: X^bÑf[^ŁAÂq[v̈QƂĂ
// ̂AVq[v̈QƂ悤ɁA|C^̒lύXB
//----------------------------------------------------------------
inline void Heap::RemapPointer (char *ptr, char *limit, char *newheap, char *markend, VarStack *stack)
{
    Variable *vp ;
    int i ;

    for (i = 0 ; i < stack->StackPtr () ; i++) {
	vp = stack->PeekStack (i) ;
	if (vp->IsString ()) {
	    if (vp->sptr < ptr || vp->sptr >= limit)
		continue ;
	    vp->sptr = newheap + (vp->sptr - ptr) ;
	} else if (vp->IsARRAY()) {
	    if ((char *)vp->arr.arrp < ptr || (char *)vp->arr.arrp >= limit)
		continue ;
	    Value *newarr = (Value *)(newheap + ((char *)vp->arr.arrp - ptr)) ;
	    RemapArray (newarr, vp->arr.numelems, ptr, limit, newheap) ;
	    vp->arr.arrp = newarr ;
	} else if (vp->IsContext()) {
	    if (vp->context.text < ptr || vp->context.text >= limit)
		continue ;
	    vp->context.text = newheap + (vp->context.text - ptr) ;
	} else if (vp->IsClass()) {
	    if ((char *)vp->classp < ptr || (char *)vp->classp >= limit)
		continue ;
	    Class *newcl = (Class *)(newheap + ((char *)vp->classp - ptr)) ;
	    RemapClass (newcl, ptr, limit, newheap) ;
	    vp->classp = newcl ;
	} else if (vp->IsInstance()) {
	    if ((char *)vp->instance < ptr || (char *)vp->instance >= limit)
		continue ;
	    Instance *newip = (Instance *)((char *)vp->instance - ptr + newheap) ;
	    RemapInstance (newip, ptr, limit, newheap) ;
	    vp->instance = newip ;
	}
    }
    for (i = stack->LvalPtr() ; i < stack->Max () ; i++) {
	vp = stack->PeekStack (i) ;
	if (vp->IsLocation ()) {
	    Value *newvp = (Value *)((char *)vp->location - ptr + newheap) ;
	    if ((char *)vp->location < ptr || (char *)vp->location >= limit)
		continue ;
	    vp->location = newvp ;
	} else if (vp->IsARRAY ()) {
	    if ((char *)vp->arr.arrp < ptr || (char *)vp->arr.arrp >= limit)
		continue ;
	    Value *newarr = (Value *)(newheap + ((char *)vp->arr.arrp - ptr)) ;
	    RemapArray (newarr, vp->arr.numelems, ptr, limit, newheap) ;
	    vp->arr.arrp = newarr ;
	} else {
	    fprintf (stderr, "?? Not location var in Lval stack\n") ;
	}
    }
}

//----------------------------------------------------------------
// RemapArray:: q[vZOgƂĊmۂĂzf[^̗vfŁA
// Âq[v̈QƂĂ̂AVq[v̈QƂ
// Ƀ|C^̒lύXB
//----------------------------------------------------------------
#define SetMark(p)\
	(((Heap::SegmentHeader *)((char *)(p) - sizeof(Heap::SegmentHeader)))->mark=1)

inline void Heap::RemapArray (Value *v, int numelems, char *ptr, char *limit, char *newheap)
{
    if (GetMark ((char *)v))
	return ;
    SetMark (v) ;
    RemapArrayElements (v, numelems, ptr, limit, newheap) ;
}

inline void Heap::RemapArrayElements (Value *v, int numelems, char *ptr, char *limit, char *newheap)
{
    int i ;

    for (i = 0 ; i < numelems ; i++) {
	if (v[i].IsString()) {
	    if (v[i].sptr < ptr || v[i].sptr >= limit)
		continue ;
	    v[i].sptr = newheap + (v[i].sptr - ptr) ;
	} else if (v[i].IsARRAY()) {
	    Value *newarr ;

	    if ((char *)v[i].arr.arrp < ptr || (char *)v[i].arr.arrp >= limit) {
		continue ;
	    }
	    newarr = (Value *)(newheap + ((char *)v[i].arr.arrp - ptr)) ;
	    RemapArray (newarr, v[i].arr.numelems, ptr, limit, newheap) ;
	    v[i].arr.arrp = newarr ;
	} else if (v[i].IsContext()) {
	    if (v[i].context.text < ptr || v[i].context.text >= limit)
		continue ;
	    v[i].context.text = newheap + (v[i].context.text - ptr) ;
	} else if (v[i].IsClass()) {
	    if ((char *)v[i].classp < ptr || (char *)v[i].classp >= limit)
		continue ;
	    Class *newclass = (Class *)((char *)v[i].classp - ptr + newheap) ;
	    RemapClass (newclass, ptr, limit, newheap) ;
	    v[i].classp = newclass ;
	} else if (v[i].IsInstance()) {
	    if ((char *)v[i].instance < ptr || (char *)v[i].classp >= limit)
		continue ;
	    Instance *newi = (Instance *)((char *)v[i].instance - ptr + newheap) ;
	    RemapInstance (newi, ptr, limit, newheap) ;
	    v[i].instance = newi ;
	}
    }
}

inline void Heap::RemapClass (Class *cl, char *ptr, char *limit, char *newheap)
{
    if (cl->superclass) {
	if (ptr <= (char *)cl->superclass && (char *)cl->superclass < limit) {
	    Class *newsc = (Class *)((char *)cl->superclass - ptr + newheap) ;
	    RemapClass (newsc, ptr, limit, newheap) ;
	    cl->superclass = newsc ;
	}
    }
    if (cl->class_template) {
	if (ptr <= (char *)cl->class_template && (char *)cl->class_template < limit)
	{
	    int i ;
	    Class::ClassMemberTemplate *newtp
		= (Class::ClassMemberTemplate *)((char *)cl->class_template - ptr + newheap) ;
#if 0
	    for (i = 0 ; i < cl->num_members ; i++) {
		Class::ClassMemberTemplate *tp = &newtp[i] ;

		if (ptr <= tp->name && tp->name < limit) {
		    tp->name = tp->name - ptr + newheap ;
		}
	    }
#endif
	    cl->class_template = newtp ;
	}
    }
    if (cl->class_vars_or_methods) {
	if (ptr <= (char *)cl->class_vars_or_methods
	    && (char *)cl->class_vars_or_methods < limit)
	{
	    Value *newcvmeth = (Value *)((char *)cl->class_vars_or_methods
					 - ptr + newheap) ;
	    RemapArray (newcvmeth, cl->num_cvs_methods,	ptr, limit, newheap) ;
	    cl->class_vars_or_methods = newcvmeth ;
	}
    }
}

inline void Heap::RemapInstance (Instance *ip, char *ptr, char *limit, char *newheap)
{
    if (GetMark((char *)ip))
	return ;

    SetMark ((char *)ip) ;
    if (ip->classp) {
	if (ptr <= (char *)ip->classp && (char *)ip->classp < limit) {
	    Class *newcp = (Class *)((char *)ip->classp - ptr + newheap) ;
	    RemapClass (newcp, ptr, limit, newheap) ;
	    ip->classp = newcp ;
	}
    }
    if (ip->instance_values) {
	if (ptr <= (char *)ip->instance_values && (char *)ip->instance_values < limit)
	{
	    Value *newiv = (Value *)((char *)ip->instance_values - ptr + newheap) ;
	    RemapArrayElements (newiv, ip->num_slots, ptr, limit, newheap) ;
	    ip->instance_values = newiv ;
	}
    }
}

//----------------------------------------------------------------
// Garbage Collection ́ASweep, Move, Remap ̂RiKōs
//	Sweep ́AA関gp̃ZOg܂Ƃ߂Ȃ
//	FreeList 𐶐B
//
//	Move ́Agp̃ZOg܂Ƃ߂āAړāA傫ȋ󂫃Z
//	Og
//
//	Remap ́A݂傫ȃq[vGA𐶐āAɂƂ
//	f[^Rs[
//----------------------------------------------------------------

//----------------------------------------------------------------
// GCSweep A関gpZOg܂Ƃ߂ăt[Xg
//
//				    freeList -----+
//						  |
//   |     |          +--------+   |     |        |   +--------+
//   +-----+          |gp  |   +-----+        +-->|gp ----+
//   |     |          +--------+   |     |            +--------+  |
//   +-----+ +-------->gp##|   +-----+ +---------->gp##|  |
//   |     | |        +--------+   |     | |          +--------+  |
//   +-----+ | +------>gp##|   +-----+ | +-------->gp##|  |
//   |     | | |      +--------+   |     | | |        +--------+  |
//   +-----+ | |      |gp  |   +-----+ | |        |gp  |<-+
//   |     |-+ |      |        |   |     |-+ |        |       ----+
//   +-----+   |      +--------+   +-----+   |        +--------+  |
//   |     |---+ +---->gp##|   |     |---+ +------>gp##|  |
//   +-----+     |    |########|   +-----+     |      |########|  |
//   |     |     |    +--------+   |     |     |      +--------+  |
//   +-----+     | +-->gp##|   +-----+     | +---->gp##|  |
//   |     |-----+ |  |########|   |     |-----+ |    |########|  |
//   +-----+       |  +--------+   +-----+       |    +--------+  |
//   |     |-------+  |gp  |   |     |-------+    |gp  |<-+
//   +-----+          +--------+   +-----+            |        |
//   |     |--+       |gp  |   |     |--+         |        |
//   +-----+  |       |        |   +-----+  |         |        |
//	      |	      +--------+            |         |        |
//            |	      |gp  |            |         |        |
//    stack   |       +--------+    stack   |         +--------+
//            +------->gp##|            +--------->gp##|
//                    |########|                      |########|
//                    |########|                      |########|
//                    |########|                      |########|
//                    +--------+                      +--------+
//                       heap                            heap
//----------------------------------------------------------------
int Heap::GCSweep (VarStack *stack)
{
    char *free_end ;
    char *areap ;
    char *heap_end = heap_area + used_size ;
    FreeSegmentHeader *hp ;
    int free_size ;

    sweep_count++ ;
    freeList = NULL ;
    numFreeList = 0 ;
    // X^bNQƂāAgp̃q[vZOg̃ZOgwb_
    // Ƀ}[NB
    ClearMark (heap_end) ;
    MarkStack (stack, 1, heap_area, heap_end) ;
#ifdef DEBUG_DUMP
    if (__DUMP) {
	printf ("---SWEEP::A---\n") ;
	Dump () ;
    }
#endif
    // ܂Aq[vGA̍ŏɂAAgpZOgXLb
    // v
    for (areap = heap_area ; areap < heap_end; areap += ((SegmentHeader *)areap)->size) {
	if (!((SegmentHeader *)areap)->mark)
	    break ;
    }
    for ( ; areap < heap_end ; ) {
	// q[v̘AgpZOg̑傫vZB
	free_size = 0 ;
	free_end = areap ;
	for (;
	     (free_end < heap_end) && !((SegmentHeader *)free_end)->mark ;
	     free_end += ((SegmentHeader *)free_end)->size)
	{
	    if (((SegmentHeader *)free_end)->type == HP_INSTANCE) {
		if (CallDisposer (free_end) < 0)
		    return -1 ;
	    }
	    ((SegmentHeader *)free_end)->type = HP_PURE ;
	    free_size += (((SegmentHeader *)free_end)->size) ;
	}
	// AgpZOgɂ܂Ƃ߂
 	numFreeList++ ;
	hp = (FreeSegmentHeader *)areap ;
	hp->mark = 0 ;
	hp->size = free_size ;
	hp->type = HP_PURE ;
	hp->next = freeList ;
	freeList = hp ;
	areap = free_end ;


	for ( ; areap < heap_end ; areap += ((SegmentHeader *)areap)->size) {
	    if (!((SegmentHeader *)areap)->mark)
		break ;
	}
    }
#ifdef DEBUG_DUMP
    if (__DUMP) {
	printf ("---SWEEP::B---\n") ;
	Dump () ;
    }
#endif
    ClearMark (heap_end) ;
#ifdef DEBUG_DUMP
    if (__DUMP) {
	printf ("---SWEEP::C---\n") ;
	Dump () ;
    }
#endif
    return 0 ;
}

//----------------------------------------------------------------
// Heap::GarbageCollect q[vGÃK[xWRNV
//
// 	Heap NXł́Aǂ̃GAɂǂ̂悤ȃf[^i[Ă邩
// 	́AǗĂȂāÁAVarStack ɂčsĂ
// 	ŁAVarStack ̒lQƂȂAK[xWRNVsB
//
//
//   |     |          +--------+     |     |            +--------+
//   +-----+          |gp  |     +-----+            |gp  |
//   |     |          +--------+     |     |            |        |
//   +-----+ +-------->gp##|     +-----+            |        |
//   |     | |        +--------+     |     |            |        |
//   +-----+ | +------>gp##|     +-----+            |        |
//   |     | | |      +--------+     |     |            |        |
//   +-----+ | |      |gp  |     +-----+            |        |
//   |     |-+ |      |        |     |     |----------+ |        |
//   +-----+   |      +--------+     +-----+          | |        |
//   |     |---+ +---->gp##|     |     |--------+ | +--------+
//   +-----+     |    |########|     +-----+        | +->gp##|
//   |     |     |    +--------+     |     |        |   +--------+
//   +-----+     | +-->gp##|     +-----+        +--->gp##|
//   |     |-----+ |  |########|     |     |------+     +--------+
//   +-----+       |  +--------+     +-----+      +----->gp##|
//   |     |-------+  |gp  |     |     |----+       |########|
//   +-----+          +--------+     +-----+    |       +--------+
//   |     |--+       |gp  |     |     |--+ +------->gp##|
//   +-----+  |       |        |     +-----+  |         |########|
//    stack   |       +--------+      stack   |         +--------+
//            +------->gp##|              +--------->gp##|
//                    |########|                        |########|
//                    |########|                        |########|
//                    |########|                        |########|
//                    +--------+                        +--------+
//                       heap                              heap
//----------------------------------------------------------------

//----------------------------------------------------------------
// GC1: v傫ȏ̋󂫃ZOgꍇɃK[xW
// RNV𒆎~ASY garbage collector
//----------------------------------------------------------------
char *Heap::GarbageCollect (VarStack *stack, int reqsize)
{
    char *move_start ;
    char *move_end ;
    char *from ;
    int  segmove ;
    char *areap ;
    char *heap_end = heap_area + used_size ;
    SegmentHeader *hp ;

    gc1_count++ ;
    reqsize += sizeof(SegmentHeader) ;

    freeList = NULL ;

#ifdef DEBUG_DUMP
    if (__DUMP) {
	stack->Dump ("STACK:: garbage collect before") ;
    }
#endif
    // X^bNQƂāAgp̃q[vZOg̃ZOgwb_
    // Ƀ}[NB
    ClearMark (heap_end) ;
    MarkStack (stack, 1, heap_area, heap_end) ;
#ifdef DEBUG_DUMP
    /**/ if (__DUMP) {
	printf ("<1> Heap Marking") ;
	Dump () ;
    }
#endif
    // ܂Aq[vGA̍ŏɂAAgpZOgXLb
    // v
    for (areap = heap_area ; areap < heap_end; areap += ((SegmentHeader *)areap)->size) {
	if (!((SegmentHeader *)areap)->mark)
	    break ;
    }
    for ( ; areap < heap_end ; ) {
	// q[v̘AgpZOg̑傫vZB
	// A̘AgpZOǧɂgpZOg̈
	// ʂƂȂB
	hp = (SegmentHeader *)areap ;
	move_start = areap ;
	segmove = 0 ;
	for (;
	     (move_start < heap_end) && !((SegmentHeader *)move_start)->mark ;
	     move_start += ((SegmentHeader *)move_start)->size)
	{
	    if (((SegmentHeader *)move_start)->type == HP_INSTANCE) {
		if (CallDisposer (move_start) < 0)
		    return NULL ;
		((SegmentHeader *)move_start)->type = HP_PURE ;
	    }
	    segmove += (((SegmentHeader *)move_start)->size) ;
	}
	if (move_start >= heap_end)
	    break ;

	// gpZOgɑAړΏۂ̎gpZOg̍Ō
	// ʒu𒲂ׂ
	for (move_end = move_start ;
	     (move_end < heap_end) && ((SegmentHeader *)move_end)->mark ;
	     move_end += ((SegmentHeader *)move_end)->size)
	    ;
	// X^bNAуq[ṽf[^ŁAړΏۂƂȂgp
	// ̈wĂ|C^Aړ̈ʒuɕtւs
	MovePointer (stack, move_start, move_end, segmove) ;
	// q[v̈撆̃f[^ړ
	for (from = move_start ; from < move_end ; from++, areap++) {
	    *areap = *from ;
	}
	// ړɂĂłAVȖgp̈̑傫ݒ
	hp = (SegmentHeader *)areap ;
	hp->mark = 0 ;
	hp->size = segmove ;
	hp->type = HP_PURE ;

	if (move_end >= heap_end) {
	    break ;
	}
	if (hp->size >= reqsize) {
#if 0
	    printf ("DEBUG:: >>> reqsize = %d(%d), segsize = %d ptr = 0x%x[ret: 0x%x]\n",
		    reqsize, reqsize - sizeof(SegmentHeader), hp->size,
		    (char *)hp, (char *)hp + sizeof(SegmentHeader)) ;
#endif
	    // ړɂĂł󂫃ZOgvꂽTCY傫
	    if (hp->size - reqsize < sizeof(SegmentHeader) + SEG_MIN_SIZE) {
		// ZOg𕪊]TȂB̂܂ܕԂ
		return (char *)hp + sizeof(SegmentHeader) ;
	    }
	    FreeSegmentHeader *hp2 ;
	    int restsize = segmove - reqsize ;
	    hp->size = reqsize ;
	    hp->type = HP_NONE ;

	    hp2 = (FreeSegmentHeader *)((char *)hp + reqsize) ;
	    hp2->size = restsize ;
	    hp2->mark = 0 ;
	    hp2->type = HP_PURE ;
	    hp2->next = 0 ;
	    numFreeList = 1 ;
	    // ZOg𕪊AgĂȂ̃ZOg
	    // freeList Ƃēo^B
	    freeList = hp2 ;
#if 0
	    printf ("DEBUG::    split(segisze=%d) retsize = %d, (restsize= %d restptr = 0x%x)\n",
		    segmove, hp->size, hp2->size, (char *)hp2) ;
	    if (__DUMP) {
		printf ("<R>") ;
		Dump () ;
		stack->Dump ("STACK:: Find Segment") ;
	    }
#endif				   
	    return (char *)hp + sizeof(SegmentHeader) ;
	}
	ClearMark (heap_end) ;
	MarkStack (stack, 1, heap_area, heap_end) ;

#ifdef DEBUG_DUMP
/**/	if (__DUMP) {
	    printf ("<2>") ;
	    Dump() ;
	    stack->Dump ("STACK:: in garbage collecting") ;
	}
#endif
    }
    // TCY̒ 
    used_size = areap - heap_area ;
    rest_size = max_size - used_size ;
#ifdef DEBUG_DUMP
/**/if (__DUMP) {
	printf ("<3>") ;
	Dump () ;
	stack->Dump ("STACK:: garbage collect after") ;
    }
#endif
    heap_end = heap_area + used_size ;
    ClearMark (heap_end) ;
#ifdef DEBUG_DUMP
/**/if (__DUMP) {
	printf ("<4>") ;
	Dump () ;
	stack->Dump ("STACK:: garbage collect clear marked") ;
    }
#endif
    return NULL ;
}

//----------------------------------------------------------------
// GC0: 󂫃ZOgCɑS܂Ƃ߂ASY garbage
// collector g Garbage collection ̂̏͏dA
// xsƁAƂ
//----------------------------------------------------------------
int Heap::GarbageCollect (VarStack *stack)
{
    char *move_start ;
    char *move_end ;
    char *from ;
    int  segmove ;
    char *areap ;
    char *heap_end = heap_area + used_size ;
    SegmentHeader *hp ;

    freeList = NULL ;
    gc_count++ ;

#ifdef DEBUG_DUMP
    if (__DUMP) {
	stack->Dump ("STACK:: garbage collect before") ;
    }
#endif
    // X^bNQƂāAq[vZOg̃ZOgwb_Ƀ}[N
    // BŃ}[NtZOgugpṽZO
    // głB
    ClearMark (heap_end) ;
    MarkStack (stack, 1, heap_area, heap_end) ;
#ifdef DEBUG_DUMP
    /**/ if (__DUMP) {
	printf ("<1> Heap Marking") ;
	Dump () ;
    }
#endif
    // ܂Aq[vGA̍ŏɂAAgpZOgXLb
    // v
    for (areap = heap_area ; areap < heap_end; areap += ((SegmentHeader *)areap)->size) {
	if (!((SegmentHeader *)areap)->mark)
	    break ;
    }
    for ( ; ; ) {
	if (areap >= heap_end)
	    break ;
	// q[v̘AgpZOg̑傫vZB
	// A̘AgpZOǧɂgpZOg̈
	// ʂƂȂB
	hp = (SegmentHeader *)areap ;
	move_start = areap ;
	segmove = 0 ;
	for ( ; move_start < heap_end && !((SegmentHeader *)move_start)->mark ;
	     move_start += ((SegmentHeader *)move_start)->size)
	{
	    if (((SegmentHeader *)move_start)->type == HP_INSTANCE) {
		if (CallDisposer (move_start) < 0)
		    return -1 ;
		((SegmentHeader *)move_start)->type = HP_PURE ;
	    }
	    segmove += ((SegmentHeader *)move_start)->size ;
	}
	if (move_start >= heap_end) {
	    break ;
	}
	// gpZOgɑAړΏۂ̎gpZOg̍Ō
	// ʒu𒲂ׂ
	for (move_end = move_start ;
	     (move_end < heap_end) && ((SegmentHeader *)move_end)->mark ;
	     move_end += ((SegmentHeader *)move_end)->size)
	    ;
	if (move_end > heap_end)
	    move_end = heap_end ;

	// X^bNAуq[ṽf[^ŁAړΏۂƂȂgp
	// ̈wĂ|C^Aړ̈ʒuɕtւs
	MovePointer (stack, move_start, move_end, segmove) ;
	// q[v̈撆̃f[^ړ
	for (from = move_start ; from < move_end ; from++, areap++) {
	    *areap = *from;
	}
	// ړɂĂłAVȖgp̈̑傫ݒ
	hp = (SegmentHeader *)areap ;
	hp->mark = 0 ;
	hp->size = segmove ;
	hp->type = HP_PURE ;

	if (move_end >= heap_end) {
	    break ;
	}
	ClearMark (heap_end) ;
	MarkStack (stack, 1, heap_area, heap_end) ;
#ifdef DEBUG_DUMP
/**/	if (__DUMP) {
	    printf ("<2>") ;
	    Dump() ;
	    stack->Dump ("STACK:: in garbage collecting") ;
	}
#endif
    }
    // TCY̒
    used_size = areap - heap_area ;
    rest_size = max_size - used_size ;
#ifdef DEBUG_DUMP
/**/if (__DUMP) {
	printf ("<3>") ;
	Dump () ;
	stack->Dump ("STACK:: garbage collect after") ;
    }
#endif
    heap_end = heap_area + used_size ;
    ClearMark (heap_end) ;
#ifdef DEBUG_DUMP
/**/if (__DUMP) {
	printf ("<4>") ;
	Dump () ;
	stack->Dump ("STACK:: garbage collect clear marked") ;
    }
#endif
    return 0 ;
}

int Heap::CallDisposer (char *entry)
{
    Instance *disposed = (Instance *)(entry + sizeof(SegmentHeader)) ;

    // disposing  ON ɂ邱ƂɂAfBX|[Up̗\̈悩
    // mۂ邱ƂɂȂB
    disposing++ ;
    if (cmmEnv->DisposeInstance (disposed) < 0)
	return -1 ;
    disposing-- ;
    // fBX|[UĂяoÎŁA\̈Ƃ̏Ԃɖ߂
    reservpt = reserved_area ;
    rest_of_reserved = reserv_size ;
    return 0 ;
}

int Heap::SegmentSize (Value *vp)
{
    char *segp ;

    if (vp->IsString ()) {
	segp = vp->sptr ;
    } else if (vp->IsARRAY()) {
	segp = (char *)vp->arr.arrp ;
    } else if (vp->IsContext()) {
	segp = vp->context.text ;
    } else {
	return -1 ;
    }
    if (segp < heap_area || segp >= heap_area + used_size) {
	return -1 ;
    }
    segp -= sizeof(SegmentHeader) ;
    return ((SegmentHeader *)segp)->size - sizeof(SegmentHeader) ;
}

//----------------------------------------------------------------
// Heap::AllocSegment CmmC^[v^̊{Iȃmۃ[`
//----------------------------------------------------------------
char *Heap::AllocSegment (int size, VarStack *stack)
{
    char *areap ;

    size = (size + 3) & ~3 ; // 4 byte alignment
    // ZOǵAŒ sizeof(FreeSegmentHeader)̑傫mۂ
    // Ȃ΂ȂȂB
    if (sizeof (SegmentHeader) + size < sizeof (FreeSegmentHeader))
	size = sizeof(FreeSegmentHeader) - sizeof (SegmentHeader) ;
    // fBX|[UśA\̈̃g
    if (disposing)
	return GetReserved (size) ;
    // ܂Aq[v̎c蕔ɋ󂫂΂Ă
    if (areap = GetArea (size))
	return areap ;
    // ꂪ߂ȂAt[XgĂ
    if (areap = GetFree (size))
	return areap ;
    // t[XgĂȂƂ́ASweep āAt[Xg
    // 
    if (GCSweep (stack) < 0)
	return NULL ;
    // łāAxt[XgĂĂ݂
    if (areap = GetFree (size))
	return areap ;
    // łʖڂȂAgpZOgul߁vċ󂫃GAm
    if (_GC1) {
	if (areap = GarbageCollect (stack, size))
	    return areap ;
    } else {
	if (GarbageCollect (stack) < 0)
	    return NULL ;
    }
    // āAq[v̍ŌォĂ
    if (areap = GetArea (size))
	return areap ;
    if (!areap && dynamic_allocate) {
	grow_count++ ;
	// q[v̈悪IɊmۂĂꍇŁAK[xWRN
	// VsĂAKvȃGAmۂłȂꍇ́A
	// ɑ傫ȃq[v̈mۂāAɃf[^꒼
	// B
	int newsize = max_size * 2 + size + RESERVED_FOR_DISPOSE ;

	Expand (newsize, stack) ;
    }
    if (areap = GetArea (size))
	return areap ;
#ifdef DEBUG_DUMP
    if (!areap) {
	fprintf (stderr, "Fatal:: Heap memory is sold out\n") ;
#ifndef MSVC_2
	exit (1) ;
#endif
    }
#endif
    return NULL ;
}

int Heap::Resize (int newsize, VarStack *stack)
{
    char *heap_end = heap_area + used_size ;

    ClearMark (heap_end) ;
    Expand (newsize, stack) ;
    heap_end = heap_area + used_size ;
    ClearMark (heap_end) ;
}

//----------------------------------------------------------------
// Heap::Expand  Vȃq[v̈mۂāAɃf[^Rs[B
//----------------------------------------------------------------
int Heap::Expand (int newsize, VarStack *stack)
{
    if (!dynamic_allocate)
	return -1 ;
    
    char *heap_end = heap_area + used_size ;
    
    if (newsize > 0 && newsize < max_size)
	return 0 ;
    if (newsize == 0)
	newsize = max_size * 2 + RESERVED_FOR_DISPOSE ;

    char *newheap = new char [newsize] ;
    int i ;

    // VɊmۂq[v̈ɁA܂ł̃q[v̈̃f[^Rs[
    // B
    for (i = 0 ; i < max_size ; i++)
    	  newheap[i] = heap_area[i] ;
#ifdef DEBUG_DUMP
    if (__DUMP) {
	printf ("## Heap Area = %x <-> %x\n",
		(int)heap_area, (int)(heap_area + max_size)) ;
	stack->Dump ("##stack :: before remap") ;
	Dump () ;
    }
#endif
    // Vȗ̈̃AhXɉāA|C^ς
    RemapPointer (heap_area, heap_area + used_size,
		  newheap, newheap + used_size, stack) ;
    delete heap_area ;
    // q[vp[^̍Đݒ
    heap_area = newheap ;
    max_size  = newsize - reserv_size ;
    rest_size = max_size - used_size ;

    reserved_area = heap_area + max_size ;
    reservpt = reserved_area ;
    rest_of_reserved = reserv_size ;

    return 0 ;
}

//----------------------------------------------------------------
// Heap::GetReserved  \̈Ƀmۂ
//----------------------------------------------------------------
char *Heap::GetReserved (int size)
{
    int area_size = sizeof (Heap::SegmentHeader) + size ;
    char *area_ptr ;

    if (rest_of_reserved < size) {
	fprintf (stderr, "Fatal: Reserved Area is exausted\n") ;
	exit (1) ;
    }
    area_ptr = reservpt + sizeof(SegmentHeader) ;
    ((SegmentHeader *)reservpt)->type = 0 ;
    ((SegmentHeader *)reservpt)->size = area_size ;
    ((SegmentHeader *)reservpt)->type = HP_NONE ;
    rest_of_reserved -= area_size ;
    reservpt += area_size ;
    return area_ptr ;
}

//----------------------------------------------------------------
// Heap::Dump  q[v_p
//----------------------------------------------------------------
void Heap::Dump()
{
    SegmentHeader *hp ;
    char *p ;
    int area_size ;
    static char *HPTYPEs[] = {"PURE", "NONE", "STRING", "ARRAY", "INSTANCE", "CLASS"} ;

//    printf ("==== Heap Dump === maxsize = %d, used = %d, rest = %d <heap_area = %lx>\n",
//	    max_size, used_size, rest_size, heap_area) ;
    Info () ;
    for (p = heap_area ; p < heap_area + used_size ; p += area_size) {
	char *segtype ;

	hp = (SegmentHeader *)p ;
	area_size = hp->size ;
	if (hp->type < HP_PURE) {
	    segtype = "!!TooLow!!" ;
	} else if (hp->type > HP_CLASS) {
	    segtype = "!!TooHigh!!" ;
	} else {
	    segtype = HPTYPEs[hp->type] ;
	}
	printf ("[%lx/%lx/%s] m(%d), s(%d/%d) n[%lx]::",
		p, p + sizeof(SegmentHeader), segtype,
		hp->mark, hp->size,
		hp->size - sizeof(SegmentHeader),
		p + hp->size) ;
	if (hp->type == HP_STRING) {
	    printf ("(%s)\n", p + sizeof(SegmentHeader)) ;
	}
#if 0
	else if (hp->type == HP_CLASS) {
	    Class *cp = (Class *)(p + sizeof(SegmentHeader)) ;

	    printf ("%s\n", cp->classnamep->name) ;
	} 
#endif
	else {
	    char *pp = p + sizeof(SegmentHeader) ;
	    int i ;
	    int dsize = hp->size - sizeof(SegmentHeader) ;
	    if (dsize > 10)
		dsize = 10 ;
	    for (i = 0 ; i < dsize ; i++, *pp++) {
		printf ("%02x ", *pp & 0xff) ;
	    }
	    printf ("\n") ;
	}
    }
    printf ("=== Free Area Info::\n") ;
    FreeSegmentHeader *fp ;
    for (fp = freeList ; fp ; fp = fp->next) {
	printf ("%08x (size = %d)\n", fp, fp->size) ;
    }
}

void Heap::BinDump ()
{
    SegmentHeader *hp ;
    char *p ;
    int col ;
    int area_size ;

    char xBuf[132] ;
    char bBuf[132] ;
    char wBuf[32] ;
    char *xp ;
    char *bp ;
    
    static char *HPTYPEs[] = {"PURE", "NONE", "STRING", "ARRAY", "INSTANCE", "CLASS"} ;

    printf ("==== Heap Dump === maxsize = %d, used = %d, rest = %d <heap_area = %lx>\n",
	    max_size, used_size, rest_size, heap_area) ;
    Info () ;
    for (col = 0, p = heap_area ; p < heap_area + max_size ; p++) {
	if (col == 0) {
	    sprintf (xBuf, "%08x: ", p) ;
	    sprintf (bBuf, "        : ") ;
	}
	sprintf (wBuf, "%02x ", (*p)&0xff) ;
	strcat (xBuf, wBuf) ;
	if (isprint(*p)) {
	    sprintf (wBuf, "%c  ", (*p)&0xff) ;
	    strcat (bBuf, wBuf) ;
	} else
	    strcat (bBuf, "-- ") ;
	col++ ;
	if (col == 16) {
	    printf ("%s\n", xBuf) ;
	    printf ("%s\n", bBuf) ;
	    col = 0 ;
	}
    }
    if (col != 0) {
	printf ("%s\n", xBuf) ;
	printf ("%s\n", bBuf) ;
    }
}

void Heap::Info ()
{
    printf ("==== Heap Info === maxsize = %d, used = %d, rest = %d <heap_area = %lx>\n",
	    max_size, used_size, rest_size, heap_area) ;
    if (freeList)
	printf ("   freeList = %x (num of free segments %d)\n", freeList, numFreeList) ;
    else
	printf ("   NO freeList\n") ;
    printf ("  Memory alloctation:  Normal %8d   FreeAlloc %8d\n",
	    alloc_count, alloc_free_count) ;
    printf ("  Garbage Collection:  Sweep  %8d   GC1  %8d   GC0       %8d\n",
	    sweep_count, gc1_count, gc_count) ;
    printf ("  Grow heap area    :         %8d\n", grow_count) ;
}

/*
 * Local variables:
 * c-indent-level: 4
 * c-continued-statement-offset: 4
 * c-brace-offset: -4
 * c-argdecl-indent: 2
 * c-label-offset:  -2
 * c-auto-newline:   nil
 * end:
 */
