/*
 * jtext.cc: Text drawing function for Ascii/SJIS mixed string.
 *						(c) Nj
 */
#include <stdio.h>
#include <X11/Xlib.h>

#include <cmm/jtext.h>

#define SHIFTJIS

#ifndef EUC
#define iskanji(c)         ((unsigned char)(c) >= 0x80 && (unsigned char)(c) <= 0x9f ||\
			    (unsigned char)(c) >= 0xe0 && (unsigned char)(c) <= 0xef)
#else
#define iskanji(c)	((unsigned char)c & 0x80)
#endif


static int ConvertSjisToJis (char *, XChar2b *) ;

void DrawMixedString (Display *disp, Drawable d,
		      GC agc, XFontStruct *afont,
		      GC jgc, XFontStruct *jfont,
		      int x, int y, char *string, int slen) 
{
    char *str = string ;
    char *ep = string + slen ;
    char *sp ;
    int len ;
    char *cp ;
    XChar2b jis_str [500] ;
    XChar2b *jp ;
    int yy ;
    
    if (afont->ascent > jfont->ascent)
	yy = y + afont->ascent ;
    else
	yy = y + jfont->ascent ;
    for ( ; str < ep ; ) {
	sp = str ;
	if (iskanji (*str)) {
	    for (str += 2 ; str < ep && iskanji(*str) ; str += 2)
		;
	    for (jp = jis_str, cp = sp ; cp < str ; cp += 2, jp++) {
#ifdef SHIFTJIS				
		(void)ConvertSjisToJis (cp, jp) ;
#endif
#ifdef EUC
		*jp = *cp & 0x7f ;
#endif				
	    }
	    len = jp - jis_str ;
	    XDrawString16 (disp, d, jgc, x, yy, jis_str, len) ;
	    x += XTextWidth16 (jfont, jis_str, len) ;
	} else {
	    for (++str ; str < ep && !iskanji(*str) ; str++)
		;
	    len = str - sp ;
	    XDrawString (disp, d, agc, x, yy, sp, len) ;
	    x += XTextWidth (afont, sp, len) ;
	}
    }
}

int MixedStringWidth (char *str, int  slen, XFontStruct *afont, XFontStruct *jfont) 
{
    int w ;
    char *sp ;
    char *ep = str + slen ;
    char *cp ;
    XChar2b jis_str[500] ;
    XChar2b *jp ;
    
    for (w = 0 ; str < ep ; ) {
	sp = str ;
	if (iskanji (*str)) {
	    for (str += 2 ; str < ep && iskanji(*str) ; str += 2)
		;
	    for (jp = jis_str, cp = sp ; cp < str ; cp += 2, jp++) {
#ifdef SHIFTJIS				
		(void)ConvertSjisToJis (cp, jp) ;
#endif
#ifdef EUC
		*jp = *cp & 0x7f ;
#endif				
	    }
	    w += XTextWidth16 (jfont, jis_str, jp - jis_str) ;
	} else {
	    for (++str ; str < ep && !iskanji(*str) ; str++)
		;
	    w += XTextWidth (afont, sp, str - sp) ;
	}
    }
    return w ;
}

#ifdef SHIFTJIS		
static int ConvertSjisToJis (char *in, XChar2b *out)
{
#define	MASK	0xff		/* 8bits mask */
#define	UNDEF	0x21
	register int	b1, b2;

	b1 = *in & MASK;	/* Shift JIS PoCg */
	b2 = *(in + 1) & MASK;	/* Shift JIS QoCg */

	if (b1 <= 0x80 || b1 >= 0xf0 ||(b1 >= 0xa0 && b1 <= 0xdf) ||
	    b2 <= 0x3f || b2 >= 0xfe || b2==0x7f) {
		out->byte1 = UNDEF;
		out->byte2 = UNDEF;
		return (1);
	}
	b1 -= (b1>=0xa0) ? 0xc1 : 0x81;
	if (b2 >= 0x9f) {
	    out->byte1 = (b1<<1) + 0x22 ;		  /* JIS PoCg */
		out->byte2 = b2 - 0x7e ;		  /* JIS QoCg */
	} else {
		out->byte1 = (b1<<1) + 0x21 ;		  /* JIS PoCg */
		out->byte2 = b2 - ((b2<=0x7e) ? 0x1f : 0x20); /* JIS QoCg */
	}
	return (0);
}
#endif

