#include <stdlib.h>
#include <cmm/cmm.h>
#include <cmm/lex.h>

struct LexBlock {
    CmmInterp::Token t ;
    CmmInterp *lexx ;
    char *buf ;
    CmmInterp::LexInfo li ;
} ;

__BeginLex (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    if (args != 1) {
	retval->Init () ;
	return 0 ;
    }
    if (!argp[0].IsString()) {
	retval->Init() ;
	return 0 ;
    }
    LexBlock *lb = new LexBlock ;
    lb->lexx = new CmmInterp (0, 0) ;
    lb->buf = new char[strlen(argp[0].sptr) + 1] ;
    strcpy (lb->buf, argp[0].sptr) ;
    
    lb->lexx->SetSource (lb->buf) ;
    *retval = (void *)lb ;
    return 0 ;
}

int __Lex (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    LexBlock *lb ;

    if (args != 1) {
	retval->Init () ;
	return 0 ;
    }
    lb = (LexBlock *)(argp[0].voidp) ;

    lb->t = lb->lexx->Lex (&lb->li) ;
    *retval = (int)(lb->t) ;
    return 0 ;
}

int __LexVal (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    LexBlock *lb ;

    if (args != 1) {
	retval->Init () ;
	return 0 ;
    }
    lb = (LexBlock *)(argp[0].voidp) ;
    if (lb->t == CmmInterp::T_Number) {
	*retval = lb->li.value ;
    } else if (lb->t == CmmInterp::T_IntNumber) {
	*retval = lb->li.ival ;
    } else if (lb->t = CmmInterp::T_String) {
	char *s = cenv->heap_area.AllocString (lb->li.sval, &cenv->varstack) ;
	*retval = s ;
    } else {
	retval->Init () ;
    }
    return 0 ;
}

int __LexPutBack (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    LexBlock *lb ;

    if (args != 1) {
	retval->Init () ;
	return 0 ;
    }
    lb = (LexBlock *)(argp[0].voidp) ;
    lb->lexx->PutBack () ;
    return 0 ;
}

int __LexError (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    LexBlock *lb ;

    Context *ctx ;
    char *mes = "== No Message ==" ;

    if (args < 1) {
	retval->Init() ;
	return 0 ;
    }
    lb = (LexBlock *)(argp[0].voidp) ;
    ctx = lb->lexx->GetContext() ;

    if (args >= 2 && argp[1].IsString()) {
	mes = argp[1].sptr ;
    }
    ctx->PrintError (mes) ;
    retval->Init() ;
    return 0 ;
}

int __LexName (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    LexBlock *lb ;

    if (args != 1) {
	retval->Init () ;
	return 0 ;
    }
    lb = (LexBlock *)(argp[0].voidp) ;
    char *bufp = cenv->heap_area.AllocString ((lb->lexx->LexName())->sptr,
					      &cenv->varstack) ;
    *retval = bufp ;
    return 0 ;
}

int __EndLex (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    LexBlock *lb ;

    if (args != 1) {
	retval->Init () ;
	return 0 ;
    }
    lb = (LexBlock *)(argp[0].voidp) ;

    delete lb->lexx ;
    delete lb->buf ;
    delete lb ;
    retval->Init () ;
    return 0 ;
}


void InstallLex (CmmInterp *cenv)
{
    cenv->AddGlobalVar ("T_Number",	(int)CmmInterp::T_Number) ;
    cenv->AddGlobalVar ("T_IntNumber",  (int)CmmInterp::T_IntNumber) ;
    cenv->AddGlobalVar ("T_Symbol",	(int)CmmInterp::T_Symbol) ;
    cenv->AddGlobalVar ("T_WorkVar",	(int)CmmInterp::T_WorkVar) ;
    cenv->AddGlobalVar ("T_Assign",	(int)CmmInterp::T_Assign) ;
    cenv->AddGlobalVar ("T_Plus",	(int)CmmInterp::T_Plus) ;
    cenv->AddGlobalVar ("T_Minus",	(int)CmmInterp::T_Minus) ;
    cenv->AddGlobalVar ("T_Mult",	(int)CmmInterp::T_Mult) ;
    cenv->AddGlobalVar ("T_Div",	(int)CmmInterp::T_Div) ;
    cenv->AddGlobalVar ("T_Mod",	(int)CmmInterp::T_Mod) ;
    cenv->AddGlobalVar ("T_Not",	(int)CmmInterp::T_Not) ;
    cenv->AddGlobalVar ("T_PP",		(int)CmmInterp::T_PP) ;
    cenv->AddGlobalVar ("T_MM",		(int)CmmInterp::T_MM) ;
    cenv->AddGlobalVar ("T_Eq",		(int)CmmInterp::T_Eq) ;
    cenv->AddGlobalVar ("T_Neq",	(int)CmmInterp::T_Neq) ;
    cenv->AddGlobalVar ("T_Lt",		(int)CmmInterp::T_Lt) ;
    cenv->AddGlobalVar ("T_Gt",		(int)CmmInterp::T_Gt) ;
    cenv->AddGlobalVar ("T_Le",		(int)CmmInterp::T_Le) ;
    cenv->AddGlobalVar ("T_Ge",		(int)CmmInterp::T_Ge) ;
    cenv->AddGlobalVar ("T_Or",		(int)CmmInterp::T_Or) ;
    cenv->AddGlobalVar ("T_And",	(int)CmmInterp::T_And) ;
    
    cenv->AddGlobalVar ("T_AddAssign",	(int)CmmInterp::T_AddAssign) ;
    cenv->AddGlobalVar ("T_SubAssign",	(int)CmmInterp::T_SubAssign) ;
    cenv->AddGlobalVar ("T_MultAssign",	(int)CmmInterp::T_MultAssign) ;
    cenv->AddGlobalVar ("T_DivAssign",	(int)CmmInterp::T_DivAssign) ;
    
    cenv->AddGlobalVar ("T_Lpar",	(int)CmmInterp::T_Lpar) ;
    cenv->AddGlobalVar ("T_Rpar",	(int)CmmInterp::T_Rpar) ;
    cenv->AddGlobalVar ("T_Lbr",	(int)CmmInterp::T_Lbr) ;
    cenv->AddGlobalVar ("T_Rbr",	(int)CmmInterp::T_Rbr) ;
    cenv->AddGlobalVar ("T_Question",	(int)CmmInterp::T_Question) ;
    cenv->AddGlobalVar ("T_Colon",	(int)CmmInterp::T_Colon) ;
    cenv->AddGlobalVar ("T_Comma",	(int)CmmInterp::T_Comma) ;

    cenv->AddGlobalVar ("T_LLBr",	(int)CmmInterp::T_LLbr) ;
    cenv->AddGlobalVar ("T_RRbr",	(int)CmmInterp::T_RRbr) ;
    
    cenv->AddGlobalVar ("T_If",		(int)CmmInterp::T_If) ;
    cenv->AddGlobalVar ("T_Else",	(int)CmmInterp::T_Else) ;
    cenv->AddGlobalVar ("T_While",	(int)CmmInterp::T_While) ;
    cenv->AddGlobalVar ("T_Func",	(int)CmmInterp::T_Func) ;
    cenv->AddGlobalVar ("T_Return",	(int)CmmInterp::T_Return) ;
    cenv->AddGlobalVar ("T_Break",	(int)CmmInterp::T_Break) ;
    cenv->AddGlobalVar ("T_Continue",	(int)CmmInterp::T_Continue) ;
    cenv->AddGlobalVar ("T_Word",	(int)CmmInterp::T_Word) ;
    
    cenv->AddGlobalVar ("T_Lbl",	(int)CmmInterp::T_Lbl) ;
    cenv->AddGlobalVar ("T_Rbl",	(int)CmmInterp::T_Rbl) ;
    cenv->AddGlobalVar ("T_For",	(int)CmmInterp::T_For) ;
    cenv->AddGlobalVar ("T_ArgCnt",	(int)CmmInterp::T_ArgCnt) ;
    cenv->AddGlobalVar ("T_ArgList",	(int)CmmInterp::T_ArgList) ;
    cenv->AddGlobalVar ("T_ArgPrefix",	(int)CmmInterp::T_ArgPrefix) ;
    
    cenv->AddGlobalVar ("T_Vardef",	(int)CmmInterp::T_Vardef) ;
    cenv->AddGlobalVar ("T_Variable",	(int)CmmInterp::T_Variable) ;
    cenv->AddGlobalVar ("T_Array",	(int)CmmInterp::T_Array) ;
    cenv->AddGlobalVar ("T_ArrElem",	(int)CmmInterp::T_ArrElem) ;
    cenv->AddGlobalVar ("T_Macro",	(int)CmmInterp::T_Macro) ;
    cenv->AddGlobalVar ("T_Load",	(int)CmmInterp::T_Load) ;
    cenv->AddGlobalVar ("T_Void",	(int)CmmInterp::T_Void) ;
    
    cenv->AddGlobalVar ("T_Eos",	(int)CmmInterp::T_Eos) ;
    cenv->AddGlobalVar ("T_Eof",	(int)CmmInterp::T_Eof) ;
    cenv->AddGlobalVar ("T_String",	(int)CmmInterp::T_String) ;
    cenv->AddGlobalVar ("T_Eval",	(int)CmmInterp::T_Eval) ;
    cenv->AddGlobalVar ("T_Err",	(int)CmmInterp::T_Err) ;

    cenv->AddGlobalVar ("T_Class",	(int)CmmInterp::T_Class) ;
    cenv->AddGlobalVar ("T_Method",	(int)CmmInterp::T_Method) ;
    cenv->AddGlobalVar ("T_Private",	(int)CmmInterp::T_Private) ;
    cenv->AddGlobalVar ("T_Public",	(int)CmmInterp::T_Public) ;
    cenv->AddGlobalVar ("T_Static",	(int)CmmInterp::T_Static) ;
    cenv->AddGlobalVar ("T_ColonColon", (int)CmmInterp::T_ColonColon) ;
    cenv->AddGlobalVar ("T_PointTo",    (int)CmmInterp::T_PointTo) ;
    cenv->AddGlobalVar ("T_New",	(int)CmmInterp::T_New) ;
    cenv->AddGlobalVar ("T_Lsh",	(int)CmmInterp::T_Lsh) ;
    cenv->AddGlobalVar ("T_Rsh",	(int)CmmInterp::T_Rsh) ;
    cenv->AddGlobalVar ("T_LshAssign",	(int)CmmInterp::T_LshAssign) ;
    cenv->AddGlobalVar ("T_RshAssign",	(int)CmmInterp::T_RshAssign) ;
    cenv->AddGlobalVar ("T_ModAssign",	(int)CmmInterp::T_ModAssign) ;

    cenv->AddBuiltin   ("BeginLex",	__BeginLex) ;
    cenv->AddBuiltin   ("Lex",		__Lex) ;
    cenv->AddBuiltin   ("LexVal",	__LexVal) ;
    cenv->AddBuiltin   ("EndLex",	__EndLex) ;
    cenv->AddBuiltin   ("LexName",	__LexName) ;
    cenv->AddBuiltin   ("LexPutBack",   __LexPutBack) ;
    cenv->AddBuiltin   ("PrintError",   __LexError) ;
}
