//----------------------------------------------------------------
// stack.cc: X^bN Cmm Interpreter
//----------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <cmm/class.h>
#include <cmm/stack.h>

#ifndef NULL
#define NULL 0
#endif

int ___DUMP = 0 ;
extern int __DUMP ;

VarStack::VarStack ()
{
    int i ;

    for (i = 0 ; i < MAXHASH ; i++)
	hashTable[i] = NULL ;

    dynamic_allocate = 0 ;
    stacksize = 0 ;
    framep = stackp = argp = 0 ;
    stack = NULL ;
}

VarStack::~VarStack()
{
    if (dynamic_allocate)
	delete stack ;
}

void VarStack::Init (Variable *stackarr, int size)
{
    stack = stackarr ;
    stacksize = size ;
    dynamic_allocate = 0 ;
}

void VarStack::Init (int size)
{
    if (dynamic_allocate && stack) {
	delete stack ;
    }
    stack = new Variable [size] ;
    lsp = stacksize = size ;
    dynamic_allocate = 1 ;
}

void VarStack::SetSymbol (int index, Symbol *symbol)
{
    Variable *pvp = &stack[index] ;
    int newhash = -1 ;
    int oldhash = -1 ;
    Variable *vlast = NULL ;
    Variable *vp ;

    if (index < stackp && pvp->sym) {
	oldhash = pvp->sym->GetHashValue () ;
    }
    if (index < stackp && symbol) {
	newhash = symbol->GetHashValue() ;
    }
    if (oldhash >= 0 && newhash >= 0 && newhash == oldhash) {
	stack[index].SetSymbol (symbol) ;
	return ;
    }
    if (oldhash >= 0) {
	/* Remove entry from old Hash link */
	for (vlast = NULL, vp = hashTable[oldhash] ;
	     vp && vp != pvp ;
	     vlast = vp, vp = vp->next) 
	    ;
	if (!vp) {
	    fprintf (stderr, "STACK Hash :: internal error\n") ;
	    return ;
	}
	if (!vlast) {
	    hashTable[oldhash] = vp->next ;
	} else {
	    vlast->next = vp->next ;
	}
    }
    if (newhash >= 0) {
	/* Insert entry to new hash link */
	if (!hashTable[newhash]) {
	    /* No Entry in new Hash Value slot */
	    hashTable[newhash] = pvp ;
	    pvp->next = NULL ;
	} else {
	    for (vlast = NULL, vp = hashTable[newhash] ;
		 vp && vp > pvp ;
		 vlast = vp, vp = vp->next) 
		;
	    if (!vlast) {
		hashTable[newhash] = pvp ;
	    } else {
		vlast->next = pvp ;
	    }
	    pvp->next = vp ;
	}
    }
    stack[index].SetSymbol(symbol) ;
}

Value *VarStack::LookupVarInFrame (Symbol *sym)
{
    int i ;

    for (i = stackp - 1 ; i >= framep ; i--) {
	if (stack[i].sym == sym)
	    return &stack[i] ;
    }
    return NULL ;
}

void VarStack::Rehash ()
{
    // nbVe[uč\B
    int i ;
    int hashv ;
    Variable *sp, *spend ;

    for (i = 0 ; i < MAXHASH ; i++)
	hashTable[i] = NULL ;
    for (sp = stack, spend = sp + stackp ; sp < spend ; sp++) {
	if (!sp->sym)
	    continue ;
	hashv = sp->sym->GetHashValue() ;
	sp->next = hashTable[hashv] ;
	hashTable[hashv] = sp ;
    }
}

//----------------------------------------------------------------
// Shrink :: stack  fromԖڂAnumshrink X^bNk߂B
//----------------------------------------------------------------
void VarStack::Shrink (int from, int numshrink)
{
    Variable *ptbase = stack + (from + numshrink) ;
    Variable *ptend = stack + stackp ;
    int hashv ;
    Variable *from_p ;
    Variable *to_p ;

    for (to_p = stack + from, from_p = ptbase ; from_p < ptend ; from_p++, to_p++) {
	if (from_p->sym) {
	    hashv = from_p->sym->GetHashValue() ;
	    if (hashTable [hashv] == from_p) {
		hashTable[hashv] = to_p ;
	    }
	    if (ptbase <= from_p->next && from_p->next < ptend) {
		from_p->next -= numshrink ;
	    }
	}
	*to_p = *from_p ;
	if (to_p->IsClass ()) {
	    //----------------------------------------------------------------
	    // class  classnamep ́Aꎩg|CgĂX
	    // ^bNGg wĂ̂ŁAk߂̈悩uv
	    // wĂ̂́A|CgȂĂ͂ȂȂ
	    //----------------------------------------------------------------
	    if (ptbase <= to_p->classp->classnamep && to_p->classp->classnamep < ptend)
	    {
		to_p->classp->classnamep -= numshrink ;
	    }
	} else if (to_p->IsLocation()) {
	    //----------------------------------------------------------------
	    // X^bŃAk߂̈悩uvwĂ|C^́A
	    // ̒lvZȂ΂ȂȂ
	    //----------------------------------------------------------------
	    if (ptbase <= (Variable *)to_p->location
		&& (Variable *)to_p->location < ptend)
	    {
		Variable *loc = (Variable *)to_p->location - numshrink ;
		to_p->location = (Value *)loc ;
	    }
	}
    }
    stackp -= numshrink ;
}

void VarStack::Fatal (char *mes)
{
    fprintf (stderr, "Fatal:: %s\n", mes) ;
    exit (30) ;
}

void VarStack::Dump (const char *s, int bottom)
{
    int i ;

    if (s)
	printf ("%s\n", s) ;
#if 0
    printf ("HashTable ====\n") ;
    for (i = 0 ; i < MAXHASH ; i++) {
	printf ("%2d: %x\n", i, hashTable[i]) ;
    }
#endif
    if (bottom < 0) {
	bottom = stackp + bottom ;
    }
    printf ("Stack ======== stackp = %d framep = %d argp = %d lsp = %d\n", stackp, framep, argp, lsp) ;
    printf ("### LVal Stack\n") ;
    for (i = stacksize - 1 ; i >= lsp ; i--) {
	printf ("[%x]%d:<%lx/%s><next=%x> ",
		stack + i, i,
		stack[i].sym,
		(stack[i].sym) ? stack[i].sym->symbol : "<TEMP>",
		stack[i].next) ;
	stack[i].Print() ;
	printf ("\n") ;
    }
    printf ("### Normal Stack\n") ;
    for (i = stackp - 1 ; i >= bottom; i--) {
	printf ("[%x]%d:<%lx/%s><next=%x> ", stack + i, i,
		stack[i].sym,
		((stack[i].sym) ? stack[i].sym->symbol : "<TEMP>"),
		stack[i].next) ;
	if (stack[i].IsArray()) {
#if 0
	    printf ("<%lx>ARRAY[%d]", stack[i].arr.arrp, stack[i].arr.numelems) ;
#endif
	    if (stack[i].arr.numelems > 10) {
		int j ;
		printf ("ARRAY<%lx> [", stack[i].arr.arrp) ;
		for (j = 0 ; j < 10 ; j++) {
		    if (j > 0)
			putchar (',') ;
		    stack[i].arr.arrp[j++].Print() ;
		}
		printf ("...]\n") ; 
	    } else {
		printf ("ARRAY<%lx> ", stack[i].arr.arrp) ;
		stack[i].Print() ;
	    }
	} else if (stack[i].IsInstance()) {
	    printf ("INSTANCE<%lx> Class<%lx> Values<%lx>",
		    stack[i].instance, stack[i].instance->classp, stack[i].instance->instance_values) ;
	    int j ;
	    int dump_save = ___DUMP ;
	    ___DUMP = 1 ;
	    printf ("[") ;
	    for (j = 0 ; j < stack[i].instance->num_slots ; j++) {
		if (j > 0)
		    putchar (',') ;
		Value *v = &(stack[i].instance->instance_values[j]) ;
		if (v->IsArray()) {
		    printf ("<array%lx>", v->arr.arrp) ;
		} else if (v->IsInstance()) {
		    printf ("<instance%lx>", v->instance) ;
		} else if (v->IsClass()) {
		    printf ("<class%lx>", v->classp) ;
		} else {
		    v->Print () ;
		}
	    }
	    printf ("]") ;
	    ___DUMP = dump_save ;
	} else {
	    stack[i].Print() ;
	}
	printf ("\n") ;
    }
}

void VarStack::Dump2 (const char *s, int lines)
{
    int i ;
    int ll ;
    char comm[30] ;

    if (s)
	printf ("%s\n", s) ;
    printf ("Stack ======== stackp = %d framep = %d argp = %d lsp = %d\n", stackp, framep, argp, lsp) ;
    printf ("### LVal Stack\n") ;
    for (i = stacksize - 1, ll = lines ; i >= lsp ; i--) {
	printf ("[%x]%d:<%lx/%s><next=%x> ", stack + i, i,
		stack[i].sym, stack[i].sym->symbol, stack[i].next) ;
	stack[i].Print() ;
	printf ("\n") ;
	if (--ll == 0) {
	    printf ("--- more(RET) or q >>") ;
	    fgets (comm, 30, stdin) ;
	    if (comm[0] == 'q')
		return ;
	    ll = lines ;
	}
    }
    printf ("### Normal Stack\n") ;
    for (ll = lines, i = stackp - 1 ; i >= 0 ; i--) {
	printf ("[%x]%d:<%lx/%s><next=%x> ", stack + i, i,
		stack[i].sym, stack[i].sym->symbol, stack[i].next) ;
	stack[i].Print() ;
	printf ("\n") ;
	if (--ll == 0) {
	    printf ("--- more(RET) or q >>") ;
	    fgets (comm, 30, stdin) ;
	    if (comm[0] == 'q')
		return ;
	    ll = lines ;
	}
    }
}

/*
 * Local variables:
 * c-indent-level: 4
 * c-continued-statement-offset: 4
 * c-brace-offset: -4
 * c-argdecl-indent: 2
 * c-label-offset:  -2
 * c-auto-newline:   nil
 * end:
 */

