//================================================================
// symbol.cc: 'Symbol Data' Manager
//						(c) Nj
//================================================================
#include <stdio.h>
#include <stdlib.h>
#include <cmm/symbol.h>

SymbolTable::SymbolTable ()
{
    int i ;
    for (i = 0 ; i < MAXHASH ; i++)
	hashTable[i] = NULL ;
}

SymbolTable::~SymbolTable () { } 

Symbol *SymbolTable::FindSymbol (int hashv, char *name)
{
    Symbol *sym ;

    sym = hashTable[hashv] ;
    while (sym) {
	if (!strcmp (sym->symbol, name))
	    return sym ;
	sym = sym->next ;
    }
    return sym ;
}

Symbol *SymbolTable::GetSymbol (char *name)
{
    int hashv = Hash(name) ;
    Symbol *sym ;

    if ((sym = FindSymbol (hashv, name))) {
	return sym ;
    }
    return CreateSymbol (hashv, name) ;
}

Symbol *SymbolTable::GetSymbol (char *name, int *hashvp)
{
    int hashv = Hash(name) ;
    Symbol *sym ;

    *hashvp = hashv ;
    if ((sym = FindSymbol (hashv, name))) {
	return sym ;
    }
    return CreateSymbol (hashv, name) ;
}

void SymbolTable::Dump ()
{
    int i ;
    Symbol *symp ;

    for (i = 0 ; i < MAXHASH ; i++) {
	symp = hashTable[i] ;
	printf ("[%3d]:", i) ;
	for (; symp ; symp = symp->next) {
	    printf (" %s", symp->symbol) ;
	}
	printf ("\n") ;
   } 
}

//================================================================
// Dynamic SymbolTable
//================================================================
DynamicSymbolTable::~DynamicSymbolTable ()
{
    SymbolPage *pagep, *delp ;

    for (pagep = poolHead ; pagep ; ) {
	delp = pagep ;
	pagep = pagep->GetNext() ;
	delete delp ;
    }
}

Symbol *DynamicSymbolTable::CreateSymbol (int hashv, char *name)
{
    Symbol *symp ;
    Symbol *slast ;
    SymbolPage *newPool ;

    if (!(symp = poolTail->AllocSymbol (name, hashv))) {
	newPool = new SymbolPage () ;
	poolTail->SetNext (newPool) ;
	poolTail = newPool ;
	symp = poolTail->AllocSymbol (name, hashv) ;
	if (!symp)
	    return NULL ;
    }
    symp->next = hashTable[hashv] ;
    hashTable[hashv] = symp ;
    return symp ;
}

int DynamicSymbolTable::PtrIsSymbol (void *ptr)
{
    SymbolPage *p ;

    for (p = poolHead ; p ; p = p->GetNext()) {
	if (p->PtrInPage ((char *)ptr))
	    return 1 ;
    }
    return 0 ;
}

SymbolPage::SymbolPage ()
{
    int i ;

    next = NULL ;
    poolSize = SYMTABLE_PAGE_SIZE ;
    poolUsed = 0 ;
    poolRest = SYMTABLE_PAGE_SIZE ;
    for (i = 0 ; i < poolSize ; i++)
	pool[i] = NULL ;
}

Symbol *SymbolPage::AllocSymbol (char *name, int hashv)
{
    int symlen  = strlen (name) ;
    int symsize = sizeof (Symbol) + symlen ;
    Symbol *symp ;

    if (symsize > poolRest)
	return NULL ;
    symp = (Symbol *)(pool + poolUsed) ;
    symp->Init (name, hashv) ;
    poolUsed += symsize ;
    poolRest -= symsize ;
    return symp ;
}

//================================================================
// Static SymbolTable
//================================================================
Symbol *StaticSymbolTable::CreateSymbol (int hashv, char *name)
{
    Symbol *symp ;
    int symsize = sizeof (Symbol) + strlen(name) ;

    if (symsize > rest) {
	fprintf (stderr, "Fatal: symboltable exausted!\n") ;
	exit (1) ;
    }
    symp = (Symbol *)(symbolPool + used) ;
    symp->Init (name, hashv) ;

    used += symsize ;
    rest -= symsize ;
    
    symp->next = hashTable[hashv] ;
    hashTable[hashv] = symp ;
    return symp ;
}

int StaticSymbolTable::PtrIsSymbol (void *ptr)
{
    return (symbolPool <= ptr &&  ptr < symbolPool + poolSize) ;
}

