//----------------------------------------------------------------
// value.cpp: Cmm C^[v^
//	Value class :: lǗ
//----------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <cmm/value.h>
#include <cmm/class.h>

#ifndef EUC
#define iskanji(c)         ((unsigned char)(c) >= 0x80 && (unsigned char)(c) <= 0x9f ||\
			    (unsigned char)(c) >= 0xe0 && (unsigned char)(c) <= 0xef)
#else
#define iskanji(c)	((unsigned char)c & 0x80)
#endif

extern int ___DUMP ;

char *Value::ForceString ()
{
    static char fstr [30] ;

    if (type == V_STRING)
	return sptr ;
    switch (type) {
      case V_REAL:
	sprintf (fstr, "%f", value) ;
	return fstr ;
      case V_INTEGER:
	sprintf (fstr, "%d", intval) ;
	return fstr ;
      case V_CLASS:	return "<Class>" ;
      case V_INSTANCE:	return "<Instance>" ;
      case V_CUSTOM:	return "<Custom>" ;
      case V_ARRAY:	return "<Array>" ;
      case V_BUILTIN:	return "<Builtin>" ;
      case V_BIMETHOD:	return "<BiMethod>" ;
      case V_VOID:	return "<void>" ;
      case V_CONTEXT:   return "<Context>" ;
      case V_SYMBOL:    return symbol->symbol ;
      default:		return "<??>" ;
    }
}

void Value::Print (char quote)
{
    Print (stdout, quote) ;
}

void Value::Write (FILE *fp = stdout)
{
    Print (fp, 1) ;
}

void Value::Print (FILE *fp, char quote)
{
    if (type == V_VOID) {
	fprintf (fp, "void") ;
    } else if (type == V_REAL) {
	fprintf (fp, "%f", value) ;
    } else if (type == V_INTEGER) {
	fprintf (fp, "%d", intval) ;
    } else if (type == V_STRING) {
	if (___DUMP) {
	    fprintf (fp, "<%lx>\"%s\"", sptr, sptr) ;
	} else {
	    if (!quote)
		fprintf (fp, "%s", sptr) ;
	    else {
		int i ;
		int splen = strlen(sptr) ;
		
		putc ('"', fp) ;
		for (i = 0 ; i < splen ; i++) {
		    char c = sptr[i] ;
		    if (iskanji (c)) {
			putc (c, fp) ;
			putc (sptr[++i], fp) ;
			continue ;
		    } 
		    switch (c) {
		      case '\n': fprintf (fp, "\\n")  ; break ;
		      case '\r': fprintf (fp, "\\r")  ; break ;
		      case '\b': fprintf (fp, "\\b")  ; break ;
		      case '\f': fprintf (fp, "\\f")  ; break ;
		      case '\t': fprintf (fp, "\\t")  ; break ;
		      case '"' : fprintf (fp, "\\\"") ; break ;
		      case '\\': fprintf (fp, "\\\\") ; break ;
		      default:
			putc (c, fp) ;
		    }
		}
		putc ('"', fp) ;
	    }
	}
    } else if (type == V_ARRAY || type == V_ACODE) {
	int i ;

	if (___DUMP) {
	    fprintf (fp, "<%lx>[", arr.arrp) ;
	} else {
	    fprintf (fp, "[") ;
	}
	for (i = 0 ; i < arr.numelems ; i++) {
	    arr.arrp[i].Print (fp, quote) ;
	    if (i < arr.numelems - 1) {
		fprintf (fp, ", ") ;
	    }
	}
	fprintf (fp, "]") ;
    } else if (type == V_SYMBOL) {
	fprintf (fp, "`%s", symbol->symbol) ;
    } else if (type == V_CONTEXT) {
	//if (__DUMP)
	//    fprintf (fp, "<%lx>", context.GetText()) ;
	fprintf (fp, "CONTEXT<%lx/%d>", context.GetText(), context.GetPC()) ;
	// fprintf (fp, "<%s>", context.GetText() + context.GetPC()) ;
    } else if (type == V_BUILTIN) {
	fprintf (fp, "<builtin func %lx>", entry) ;
    } else if (type == V_BIMETHOD) {
	fprintf (fp, "<builtin method %lx>", bimethod) ;
    } else if (type == V_CUSTOM) {
	fprintf (fp, "<custiom pointer %lx>", voidp) ;
    } else if (type == V_CLASS) {
	fprintf (fp, "<classdef %lx>", classp) ;
	if (___DUMP) {
	    classp->DumpClass () ;
	}
    } else if (type == V_INSTANCE) {
	fprintf (fp, "<instance %lx>", instance) ;
	if (___DUMP) {
	    instance->DumpInstance () ;
	}
    } else if (type == V_LOCATION) {
	fprintf (fp, "<location %lx>", location) ;
    } else {
	fprintf (fp, "??InternalBug??<this=0x%lx>", this) ;
    }
}

int CompareArray (Value& v1, Value& v2)
{
    int i ;

    if (v1.arr.numelems != v2.arr.numelems)
	return 0 ;
    for (i = 0 ; i < v1.arr.numelems ; i++) {
	if (v1.arr.arrp[i] != v2.arr.arrp[i])
	    return 0 ;
    }
    return 1 ;
}



