//
// xpack.cc : cmm X interface 
//				(c) Nj
//
#include <locale.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Xutil.h>
#include <cmm/jtext.h>

#ifdef True
#undef True 
#endif

#include <cmm/class.h>
#include <cmm/cmm.h>

#define ARGCHK(n, fname) if (args != n) { retval->Init() ;\
				      printf ("XPACK::%s: Argument Error\n", fname) ;\
					  return -1 ; }
#define ARGCHK2(n1, n2, fname) if (args != n1 && args != n2) \
				{ retval->Init() ;\
				  printf ("XPACK::%s: Argument Error\n", fname) ;\
				      return -1 ; }
#define ARGCHK3(n1, n2, n3, fname) if (args != n1 && args != n2 && args != n3) \
				{ retval->Init() ;\
				  printf ("XPACK::%s: Argument Error\n", fname) ;\
				      return -1 ; }

#define _INT(n)		(argp[n].intval)
#define _STRING(n)	(argp[n].sptr)
#define _UINT(n)	((unsigned int)argp[n].intval)
#define _ULONG(n)	((unsigned long)argp[n].intval)
#define _SHORT(n)	((short)argp[n].intval)
#define _USHORT(n)	((unsigned short)argp[n].intval)

#define _DISPLAY(n)	((Display *)argp[n].voidp)
#define _SCREEN(n)	_INT(n)
#define _WINDOW(n)	((Window   )argp[n].intval)
#define _DRAWABLE(n)	((Drawable )argp[n].intval)
#define _X(n)		_INT(n)
#define _Y(n)		_INT(n)
#define _WIDTH(n)	_UINT(n)
#define _HEIGHT(n)	_UINT(n)
#define _PIXEL(n)	_ULONG(n)
#define _PIXMAP(n)	((Pixmap)argp[n].intval)
#define _GC(n)		((GC)argp[n].intval)
#define _VISUAL_P(n)	((Visual *)argp[n].voidp)
#define _CURSOR(n)	((Cursor)argp[n].intval)

static XPoint *CreateXPoints (Value *polyp, int *npoints)
{
    int nelems = polyp->arr.numelems ;
    XPoint *pp = new XPoint [nelems] ;
    int i ;

    for (i = 0 ; i < nelems ; i++) {
	if (!polyp->arr.arrp[i].IsArray()) {
	    delete pp ;
	    return 0 ;
	}
	pp[i].x = polyp->arr.arrp[i].arr.arrp[0].intval ;
	pp[i].y = polyp->arr.arrp[i].arr.arrp[1].intval ;
    }
    *npoints = nelems ;
    return pp ;
}

//
// XOpenDisplay ()
// XOpenDisplay (display_name)
//
static int _X_OpenDisplay (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    Display *disp ;
    char *display_name = NULL ;

    setlocale (LC_ALL, "") ;
    if (args == 0) {
	display_name = NULL ;
    } else {
	if (argp[0].IsString())
	    display_name = argp[0].sval() ;
    }
    disp = XOpenDisplay (display_name) ;
    if (!disp)
	retval->Init() ;
    else
	*retval = (void *)disp ;
    return 0 ;
}

static void SetSetWindowAttribute (CmmInterp *cenv,
				   XSetWindowAttributes *atp,
				   Value *valp)
{
    atp->background_pixmap = (Pixmap)((cenv->FindMember (valp, "background_pixmap"))->intval) ;
    atp->background_pixel  = (unsigned long)((cenv->FindMember (valp, "background_pixel"))->intval) ;
    atp->border_pixmap	= (Pixmap)((cenv->FindMember(valp, "border_pixmap"))->intval) ;
    atp->border_pixel	= (unsigned long)((cenv->FindMember (valp, "border_pixel"))->intval) ;
    atp->bit_gravity	= (cenv->FindMember (valp, "bit_gravity"))->intval ;
    atp->win_gravity	= (cenv->FindMember (valp, "win_gravity"))->intval ;
    atp->backing_store	= (cenv->FindMember (valp, "backing_store"))->intval ;
    atp->backing_planes	= (unsigned long)((cenv->FindMember(valp, "backing_planes"))->intval) ;
    atp->backing_pixel	= (unsigned long)((cenv->FindMember(valp, "backing_pixel"))->intval) ;
    atp->save_under	= (Bool)((cenv->FindMember(valp, "save_under"))->intval) ;
    atp->event_mask	= (long)((cenv->FindMember(valp, "event_mask"))->intval) ;
    atp->do_not_propagate_mask	= (long)((cenv->FindMember(valp, "do_not_propagate_mask"))->intval) ;
    atp->override_redirect	= (Bool)((cenv->FindMember(valp, "override_redirect"))->intval) ;
    atp->colormap		= (Colormap)((cenv->FindMember(valp, "colormap"))->intval) ;
    atp->cursor			= (Cursor)((cenv->FindMember(valp, "cursor"))->intval) ;
}

//
// XCreateWindow (display, parent, x, y, width, height, border_width, depth,
//		  class, visual, valuemask, attributes)
//
//		 0     1       2  3  4      5       6             7
// CreateWindow (disp, parent, x, y, width, height, border_width, depth,
//		 class, visual, valuemask, attributes)
//		 8      9       10         11
static int _X_CreateWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    XSetWindowAttributes wat ;
    Window win ;
    ARGCHK(12, "XCreateWindow") ;
    Visual *visp ;

    if (argp[9].IsInteger() && argp[9].intval == CopyFromParent) {
	visp = (Visual *)CopyFromParent ;
    } else {
	visp = _VISUAL_P(9) ;
    }
   
    SetSetWindowAttribute (cenv, &wat, &argp[11]) ;
    win = XCreateWindow (_DISPLAY(0), _WINDOW(1),
			 _X(2), _Y(3), _WIDTH(4), _HEIGHT(5),
			 _UINT(6), _INT(7), _UINT(8), visp,
			 _UINT(10), &wat) ;
    *retval = (int)win ;
    return 0 ;
}

// CreateSimpleWindow (disp, parent, x, y, w, h, bdr, fg, bg) 
static int _X_CreateSimpleWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    Window win ;

    ARGCHK(9, "XCreateSimpleWindow") ;
    win = XCreateSimpleWindow (_DISPLAY(0),
			       _WINDOW(1),
			       _X(2), _Y(3), _WIDTH(4), _HEIGHT(5),
			       _UINT(6), _PIXEL(7), _PIXEL(8)) ;
    *retval = (int)win ;
    return 0 ;
}

static int _X_DefaultScreen (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "DefaultScreen") ;
    int screen = DefaultScreen (_DISPLAY(0)) ;
    *retval = screen ;
    return 0 ;
}

static int _X_RootWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "RootWindow") ;
    Window root = RootWindow (_DISPLAY(0), _SCREEN(1)) ;
    *retval = (int)root ;
    return 0 ;
}

static int _X_BlackPixel (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "BlackPixel") ;
    int pixel = BlackPixel (_DISPLAY(0), _SCREEN(1)) ;
    *retval = pixel ;
    return 0 ;
}

static int _X_WhitePixel (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "WhitePixel") ;
    int pixel = WhitePixel (_DISPLAY(0), _SCREEN(1)) ;
    *retval = pixel ;
    return 0 ;
}

static int _X_DisplayPlanes (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "DisplayPlanes") ;
    int planes = DisplayPlanes (_DISPLAY(0), _SCREEN(1)) ;
    *retval = planes ;
    return 0 ;
}

static int _X_DefaultDepth (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "DefaultDepth") ;
    int depth ;
    depth = DefaultDepth (_DISPLAY(0), _SCREEN(1)) ;
    *retval = depth ;
    return 0 ;
}

static int _X_DefaultVisual (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "DefaultVisual") ;
    Visual *visual ;
    visual = DefaultVisual (_DISPLAY(0), _SCREEN(1)) ;
    *retval = (void *)visual ;
    return 0 ;
}

static int _X_ColorValue (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    Display *display ;
    int screen ;
    char *colorname ;
    Colormap cmap ;
    XColor screen_def, exact_def ;
    
    ARGCHK2(2, 3, "XColorValue") ;

    display = _DISPLAY(0) ;
    if (args == 2) {
	screen = DefaultScreen (display) ;
	cmap = DefaultColormap (display, screen) ;
	colorname = _STRING(1) ;
    } else {
	screen = _SCREEN(1) ;
	cmap = DefaultColormap (display, screen) ;
	colorname = _STRING(2) ;
    }
    if (!XAllocNamedColor (display, cmap, colorname, &screen_def, &exact_def)) {
	*retval = (int)BlackPixel (display, screen) ;
    } else {
	*retval = (int)exact_def.pixel ;
    }
    return 0 ;
}

static int _X_PixelColor (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(4, "XPixelColor") ;

    Display *display = _DISPLAY(0) ;
    int     screen   = DefaultScreen (display) ;
    Colormap  cmap   = DefaultColormap (display, screen) ;
    XColor color ;
    
    color.red   = _INT(1) ;
    color.green = _INT(2) ;
    color.blue  = _INT(3) ;

    color.flags = DoRed|DoGreen|DoBlue ;

    if (XAllocColor (display, cmap, &color)) {
	*retval = (int)color.pixel ;
    } else {
	*retval = (int)BlackPixel (display, screen) ;
    }
    return 0 ;
}

static int _X_DrawLine (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(7, "XDrawLine") ;
    XDrawLine (_DISPLAY(0), _DRAWABLE(1), _GC(2), _X(3), _Y(4), _X(5), _Y(6)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_DrawLines (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK2(4, 5, "XDrawLines") ;
    int nelems ;
    int mode ;
    XPoint *pp ;
    if (args == 4) {
	mode = CoordModeOrigin ;
    } else {
	mode = _INT(4) ;
    }
    pp = CreateXPoints (&argp[3], &nelems) ;
    XDrawLines (_DISPLAY(0), _DRAWABLE(1), _GC(2), pp, nelems, mode) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_DrawPoint (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK (5, "XDrawPoint") ;
    XDrawPoint (_DISPLAY(0), _DRAWABLE(1), _GC(2), _X(3), _Y(4)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_DrawPoints (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK2(4, 5, "XDrawPoints") ;
   
    if (!argp[3].IsArray()) {
	*retval = -1 ;
	return 0 ;
    }
    int nelems ;
    XPoint *pp ;

    pp = CreateXPoints (&argp[3], &nelems) ;
    XDrawPoints (_DISPLAY(0), _DRAWABLE(1), _GC(2), pp, nelems,
		 ((args == 4) ? CoordModeOrigin : _INT(4))) ;
    delete pp ;
    *retval = 0 ;
    return 0 ;
}

static int _X_DrawRectangle (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(7, "XDrawRectangle") ;
    XDrawRectangle (_DISPLAY(0), _DRAWABLE(1), _GC(2),
		    _X(3), _Y(4), _WIDTH(5), _HEIGHT(6)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_FillRectangle (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(7, "XFillRectangle") ;
    XFillRectangle (_DISPLAY(0), _DRAWABLE(1), _GC(2),
		    _X(3), _Y(4), _WIDTH(5), _HEIGHT(6)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_FillPolygon (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK3(4,5,6, "XFillPolygon") ;

    if (!argp[3].IsArray()) {
	*retval = -1 ;
	return 0 ;
    }
    Value *polyp = &argp[3] ;
    int nelems ;
    XPoint *pp ;

    pp = CreateXPoints (&argp[3], &nelems) ;
    XFillPolygon (_DISPLAY(0), _DRAWABLE(1), _GC(2), pp, nelems,
		  ((args == 4) ? Nonconvex : _INT(4)),
		   (args <= 5) ? CoordModeOrigin : _INT(5)) ;
    delete pp ;
    *retval = 0 ;
    return 0 ;
}

//
//	      0	        1        2   3  4  5  6  7       8    
//  XFillArc (display, drawable, gc, x, y, w, h, angle1, angle2)
//
static int _X_FillArc (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK (9, "XFillArc") ;
    XFillArc (_DISPLAY(0),
	      _DRAWABLE(1),
	      _GC(2),
	      _X(3), _Y(4), _WIDTH(5), _HEIGHT(6),
	      _INT(7), _INT(8)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_LoadFont (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XLoadFont") ;
    Font fid = XLoadFont ((Display *)argp[0].voidp, argp[1].sptr) ;
    *retval = (int)fid ;
    return 0 ;
}

static int _X_LoadQueryFont (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XLoadQueryFont") ;
    XFontStruct *xf ;
    xf = XLoadQueryFont ((Display *)argp[0].voidp, argp[1].sptr) ;
    if (!xf) {
	retval->Init() ;
	return 0 ;
    }
    Instance *ip ;
    ip = cenv->NewInstance ("XFontStruct") ;
    if (!ip)
	return -1 ;
    cenv->SetValue (ip, "fid",       (int)xf->fid) ;
    cenv->SetValue (ip, "direction", (int)xf->direction) ;
    cenv->SetValue (ip, "ascent",    (int)xf->ascent) ;
    cenv->SetValue (ip, "descent",   (int)xf->descent) ;
    cenv->SetValue (ip, "display",   (void *)argp[0].voidp) ;
    cenv->SetValue (ip, "instance",  (void *)xf) ;

    *retval = ip ;
    return 0 ;
}

static int _X_FreeFont (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    Value *cp ;

    ARGCHK(2, "XFreeFont") ;
    if (!argp[1].IsInstance() || !cenv->FindClass (&argp[1], NULL, "XFontStruct", &cp)) {
	return cenv->ThrowSystemException (cenv->typeMismatchException) ;
    }
    XFontStruct *xf ;
    xf = (XFontStruct *)(cenv->FindMember (&argp[1], "instance"))->voidp ;
    XFreeFont ((Display *)argp[0].voidp, xf) ;
    return 0 ;
}

static int _X_TextWidth (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK2(2, 3, "XTextWidth") ;
    int len ;
    int wid ;
    XFontStruct *xf ;

    xf = (XFontStruct *)(argp[0].instance->FindMember
			 	(cenv->GetSymbol("instance")))->voidp ;
    if (args == 2)
	len = strlen (argp[1].sptr) ;
    else
	len = argp[2].intval ;
    *retval = XTextWidth (xf, argp[1].sptr, len) ;
    return 0 ;
}

static int _X_TextWidth16 (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK2(2, 3, "XTextWidth") ;
    int len ;
    int wid ;
    XFontStruct *xf ;

    xf = (XFontStruct *)(argp[0].instance->FindMember
			 	(cenv->GetSymbol("instance")))->voidp ;
    if (args == 2)
	len = strlen (argp[1].sptr) ;
    else
	len = argp[2].intval ;
    *retval = XTextWidth16 (xf, (XChar2b *)argp[1].sptr, len) ;
    return 0 ;
}

static int _X_CreateFontSet (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XCreateFontSet") ;
    char **missing_char_set ;
    int  num_missing ;
    char *defstring ;
    XFontSet fontset = XCreateFontSet ((Display *)argp[0].voidp, argp[1].sptr,
				       &missing_char_set, &num_missing, &defstring) ;
    if (!fontset) {
	int i ;
	fprintf (stderr, "Cannot open fontset\n") ;
	for (i = 0 ; i < num_missing ; i++)
	    fprintf (stderr, "	%s\n", missing_char_set[i]) ;
    }
    if (missing_char_set)
	XFreeStringList (missing_char_set) ;
    if (fontset) {
	printf ("<LOCALE>  %s (supportlocale) %s\n",
		XLocaleOfFontSet (fontset), (XSupportsLocale()) ? "YES" : "NO") ;
    }
    *retval = (void *)fontset ;
    return 0 ;
}

static int _X_FreeFontSet (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XFreeFontSet") ;
    XFreeFontSet ((Display *)argp[0].voidp, (XFontSet)argp[1].voidp) ;
    retval->Init () ;
    return 0 ;
}

static int _X_DrawMixedString (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK2(9, 10, "XDrawMixedString") ;
    XFontStruct *afont ;
    XFontStruct *jfont ;

    if (!argp[3].IsInstance())
	return 0 ;
    if (!argp[5].IsInstance())
	return 0 ;
    afont = (XFontStruct *)(argp[3].instance->FindMember
			    (cenv->GetSymbol("instance")))->voidp ;
    jfont = (XFontStruct *)(argp[5].instance->FindMember
			    (cenv->GetSymbol("instance")))->voidp ;

    DrawMixedString ((Display *)argp[0].voidp,
		     (Drawable) argp[1].intval,
		     (GC)argp[2].intval, afont,
		     (GC)argp[4].intval, jfont,
		     argp[6].intval, argp[7].intval,
		     argp[8].sptr,
		     (args == 9) ? (strlen (argp[8].sptr)): argp[9].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_MixedStringWidth (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(4, "XMixedStringWidth") ;
    int width ;
    XFontStruct *afont, *jfont ;

    afont = (XFontStruct *)(argp[2].instance->FindMember
			    (cenv->GetSymbol("instance")))->voidp ;
    jfont = (XFontStruct *)(argp[3].instance->FindMember
			    (cenv->GetSymbol("instance")))->voidp ;
    width = MixedStringWidth (argp[0].sptr, argp[1].intval, afont, jfont) ;
    *retval = width ;
    return 0 ;
}

static int _X_MbDrawString (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(8, "XmbDrawString") ;
    XmbDrawString ((Display *)argp[0].voidp,
		   (Drawable)argp[1].intval,
		   (XFontSet)argp[2].voidp,
		   (GC)argp[3].intval,
		   argp[4].intval, argp[5].intval,
		   argp[6].sptr,  argp[7].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_MbDrawImageString (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(8, "XmbDrawImageString") ;
    XmbDrawImageString ((Display *)argp[0].voidp,
			(Drawable)argp[1].intval,
			(XFontSet)argp[2].voidp,
			(GC)argp[3].intval,
			argp[4].intval, argp[5].intval,
			argp[6].sptr,  argp[7].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_DrawString (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK2(6, 7, "XDrawString") ;
    XDrawString ((Display *)argp[0].voidp,
		 (Drawable)argp[1].intval,
		 (GC)argp[2].intval,
		 argp[3].intval, argp[4].intval,
		 argp[5].sptr,
		 (args == 6) ? (strlen(argp[5].sptr)): argp[6].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_DrawString16 (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK2(6, 7, "XDrawString16") ;
    XDrawString16 ((Display *)argp[0].voidp,
		   (Drawable)argp[1].intval,
		   (GC)argp[2].intval,
		   argp[3].intval, argp[4].intval,
		   (XChar2b *)argp[5].sptr,
		   (args == 6) ? (strlen(argp[5].sptr)/2) : argp[6].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_DrawImageString (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK2(6, 7, "XDrawImageString") ;
    XDrawImageString ((Display *)argp[0].voidp,
		      (Drawable)argp[1].intval,
		      (GC)argp[2].intval,
		      argp[3].intval, argp[4].intval,
		      argp[5].sptr,
		      (args == 6) ? (strlen(argp[5].sptr)) : argp[6].intval) ;
    *retval = 0 ;
    return 0 ;
}
		 
static int _X_DrawImageString16 (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK2(6, 7, "XDrawImageString16") ;
    XDrawImageString16 ((Display *)argp[0].voidp,
			(Drawable)argp[1].intval,
			(GC)argp[2].intval,
			argp[3].intval, argp[4].intval,
			(XChar2b *)argp[5].sptr,
			(args == 6) ? (strlen(argp[5].sptr)/2) : argp[6].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_CreatePixmap (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(5, "XCreatePixmap") ;
    Pixmap pm ;
    pm = XCreatePixmap (_DISPLAY(0), _DRAWABLE(1), _WIDTH(2), _HEIGHT(3), _UINT(4)) ;
    *retval = (int)pm ;
    return 0 ;
}

static int _X_FreePixmap (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK (2, "XFreePixmap") ;
    XFreePixmap (_DISPLAY(0), _PIXMAP(1)) ;
    *retval = 0 ;
    return 0 ;
}

static char *MakeBitmapData (CmmInterp *cenv, Value *bm, int *width, int *height)
{
    int w, h ;
    Value *arrdata ;
    char *bmdata ;
    int i ; 
   
    w = (cenv->FindMember (bm, "width"))->intval ;
    h = (cenv->FindMember (bm, "height"))->intval ;
    arrdata = (cenv->FindMember (bm, "data")) ;
    bmdata = new char [arrdata->arr.numelems] ;
    for (i = 0 ; i < arrdata->arr.numelems ; i++) {
	bmdata[i] = (char)(arrdata->arr.arrp[i].intval) ;
    }
    *width = w ;
    *height = h ;
    return bmdata ;
}

static int _X_CreatePixmapFromBitmapData (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(6, "XCreatePixmapFromBitmapData") ;
    Pixmap pm ;
    int  width, height ;
    Value *pp ;
    char *data ;
    Value *cp ;
   
    retval->Init() ;
    if (!argp[2].IsInstance() || !cenv->FindClass(&argp[2], NULL, "XBitmap", &cp))
	return 0 ;
    data = MakeBitmapData (cenv, &argp[2], &width, &height) ;
    pm = XCreatePixmapFromBitmapData ((Display *)argp[0].voidp,
				      (Drawable)argp[1].intval,
				      data, width, height,
				      (unsigned long)argp[3].intval,
				      (unsigned long)argp[4].intval,
				      (unsigned int) argp[5].intval) ;
    delete data ;
    *retval = (int)pm ;
    return 0 ;
}
	 
static int _X_CreateBitmapFromData (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XCreateBitmapFromData") ;
    Pixmap pm ;
    int  width, height ;
    Value *pp ;
    char *data ;
    Value *cp ;
   
    retval->Init() ;
    if (!argp[2].IsInstance() || !cenv->FindClass(&argp[2], NULL, "XBitmap", &cp))
	return 0 ;
    data = MakeBitmapData (cenv, &argp[2], &width, &height) ;
    pm = XCreateBitmapFromData ((Display *)argp[0].voidp,
				(Drawable)argp[1].intval,
				data, width, height) ;
    delete data ;
    *retval = (int)pm ;
    return 0 ;
}

static int _X_MapWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XMapWindow") ;
    XMapWindow ((Display *)argp[0].voidp, (Window)argp[1].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_ResizeWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(4, "XResizeWindow") ;
    retval->Void() ;
    XResizeWindow ((Display *)argp[0].voidp, (Window)argp[1].intval,
		   argp[2].intval, argp[3].intval) ;
    return 0 ;
}

static int _X_MoveWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(4, "XMoveWindow") ;
    retval->Void() ;
    XMoveWindow ((Display *)argp[0].voidp, (Window)argp[1].intval,
		 argp[2].intval, argp[3].intval) ;
    return 0 ;
}

static int _X_MoveResizeWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(6, "XMoveResizeWindow") ;
    retval->Void() ;
    XMoveResizeWindow ((Display *)argp[0].voidp, (Window)argp[1].intval,
		       argp[2].intval, argp[3].intval, argp[4].intval, argp[5].intval) ;
    return 0 ;
}

static int _X_SetWindowBorderWidth (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetWindowBorderWidth") ;
    retval->Void() ;
    XSetWindowBorderWidth ((Display *)argp[0].voidp, (Window)argp[1].intval,
			  argp[2].intval) ;
    return 0 ;
}

static int _X_RaiseWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XRaiseWindow") ;
    retval->Void() ;
    XRaiseWindow (_DISPLAY(0), _WINDOW(1));
    return 0 ;
}

static int _X_LowerWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XLowerWindow") ;
    retval->Void() ;
    XRaiseWindow (_DISPLAY(0), _WINDOW(1));
    return 0 ;
}

static int _X_UnmapWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XUnmapWindow") ;
    XUnmapWindow ((Display *)argp[0].voidp, (Window)argp[1].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_ClearWindow (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XClearWindow") ;
    XClearWindow (_DISPLAY(0), _WINDOW(1)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_ClearArea (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK (7, "XClearArea") ;
    XClearArea (_DISPLAY(0), _WINDOW(1),
		_X(2), _Y(3), _WIDTH(4), _HEIGHT(5), _INT(6)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SelectInput (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSelectInput") ;
    XSelectInput ((Display *)argp[0].voidp, argp[1].intval, argp[2].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_DisplayWidth (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "DisplayWidth") ;
    *retval = DisplayWidth (_DISPLAY(0), _SCREEN(1)) ;
    return 0 ;
}

static int _X_DisplayHeight (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "DisplayHeight") ;
    *retval = DisplayHeight (_DISPLAY(0), _SCREEN(1)) ;
    return 0 ;
}

static int _X_Flush (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "XFlush") ;
    *retval = XFlush (_DISPLAY(0)) ;
    return 0 ;
}

static int _X_QLength (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "QLength") ;
    *retval = QLength (_DISPLAY(0)) ;
    return 0 ;
}

static int _X_Sync (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XSync") ;
    XSync (_DISPLAY(0), _INT(1)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_EventsQueued (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XEventsQueued") ;
    *retval= XEventsQueued (_DISPLAY(0), _INT(1)) ;
    return 0 ;
}

static int _X_Pending (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "XPending") ;
    *retval = XPending (_DISPLAY(0)) ;
    return 0 ;
}

static int _X_StoreName (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XStoreName") ;
    XStoreName (_DISPLAY(0), _WINDOW(1), _STRING(2)) ;
    *retval = 0 ;
}

//
//		 0	  1	  2	     3		 4
// XGrabPointer (display, window, eventmask, confine_to, cursor) ;
//
static int _X_GrabPointer  (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK2(5, 6, "XGrabPointer") ;
    int status ;
    if (args == 5) {
	status = XGrabPointer ((Display *)argp[0].voidp,
			       (Window)argp[1].intval,
			       False,
			       (unsigned int)argp[2].intval,
			       GrabModeAsync, GrabModeAsync,
			       (Window)argp[3].intval,
			       (Cursor)argp[4].intval,
			       CurrentTime) ;
    } else if (args == 6) {
	status = XGrabPointer ((Display *)argp[0].voidp,
			       (Window)argp[1].intval,
			       (Bool)argp[2].intval,
			       (unsigned int)argp[3].intval,
			       GrabModeAsync, GrabModeAsync,
			       (Window)argp[4].intval,
			       (Cursor)argp[5].intval,
			       CurrentTime) ;
    }
    *retval = status ;
    return 0 ;
}

static int _X_UngrabPointer (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "XUngrabPointer") ;
    XUngrabPointer ((Display *)argp[0].voidp, CurrentTime) ;
    *retval = 0 ;
    return 0 ;
}

//
//			0,	1   2     3   4   5 	6
// XCreatePixmapCursor (disp,  cur, mask, fg, bg, xhot, yhot)
//
static int _X_CreatePixmapCursor (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(7, "XCreatePixmapCursor") ;
   
    Cursor cur ;
    XColor curfg ;
    XColor curbg ;
    Display *display = _DISPLAY(0) ;
    int screen = DefaultScreen (display) ;
    Colormap cmap = DefaultColormap (display, screen) ;

    XParseColor (display, cmap, _STRING(3), &curfg) ;
    XParseColor (display, cmap, _STRING(4), &curbg) ;

    cur = XCreatePixmapCursor (display, _PIXMAP(1), _PIXMAP(2),
			       &curfg, &curbg, _X(5), _Y(6)) ;
    *retval = (int)cur ;
    return 0 ;
}

//
//  XCreateFontCursor (display, shape)
//
static int _X_CreateFontCursor (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    Cursor cursor ;
    ARGCHK(2, "XCreateFontCursor") ;
    cursor = XCreateFontCursor (_DISPLAY(0), _INT(1)) ;
    *retval = (int)cursor ;
    return 0 ;
}

//
// XFreeCursor
static int _X_FreeCursor (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    Cursor cursor ;
    ARGCHK(2, "XFreeCursor") ;
    cursor = XFreeCursor (_DISPLAY(0), _CURSOR(1)) ;
    *retval = (int)cursor ;
    return 0 ;
}

//
//	XDefineCursor (disp, win, cursor) ;
//
static int _X_DefineCursor (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XDefineCursor") ;
    *retval = 0 ;
    XDefineCursor (_DISPLAY(0), _WINDOW(1), _CURSOR(2)) ;
    return 0 ;
}

//
//	XUndefineCursor (disp, win) ;
//
static int _X_UndefineCursor (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XUndefineCursor") ;
    *retval = 0 ;
    XUndefineCursor (_DISPLAY(0), _WINDOW(1)) ;
    return 0 ;
}


static int SetGCValues (CmmInterp *cenv, XGCValues *xgcv, Value *valp)
{
    xgcv->function   = (cenv->FindMember (valp, "function"))->intval ;
    xgcv->plane_mask = (cenv->FindMember (valp, "plane_mask"))->intval ;
    xgcv->foreground = (cenv->FindMember (valp, "foreground"))->intval ;
    xgcv->background = (cenv->FindMember (valp, "background"))->intval ;
    xgcv->line_width = (cenv->FindMember (valp, "line_width"))->intval ;
    xgcv->cap_style  = (cenv->FindMember (valp, "cap_style"))->intval ;
    xgcv->join_style = (cenv->FindMember (valp, "join_style"))->intval ;
    xgcv->fill_style = (cenv->FindMember (valp, "fill_style"))->intval ;
    xgcv->fill_rule  = (cenv->FindMember (valp, "fill_rule"))->intval ;
    xgcv->arc_mode   = (cenv->FindMember (valp, "arc_mode"))->intval ;
    xgcv->tile       = (cenv->FindMember (valp, "tile"))->intval ;
    xgcv->ts_x_origin = (cenv->FindMember (valp, "ts_x_origin"))->intval ;
    xgcv->ts_y_origin = (cenv->FindMember (valp, "ts_y_origin"))->intval ;
    xgcv->font        = (cenv->FindMember (valp, "font"))->intval ;
    xgcv->subwindow_mode = (cenv->FindMember (valp, "subwindow_mode"))->intval ;
    xgcv->graphics_exposures = (cenv->FindMember (valp, "graphics_exposures"))->intval ;
    xgcv->clip_x_origin = (cenv->FindMember (valp, "clip_x_origin"))->intval ;
    xgcv->clip_y_origin = (cenv->FindMember (valp, "clip_y_origin"))->intval ;
    xgcv->clip_mask     = (cenv->FindMember (valp, "clip_mask"))->intval ;
    xgcv->dash_offset   = (cenv->FindMember (valp, "dash_offset"))->intval ;
    xgcv->dashes        = (cenv->FindMember (valp, "dashes"))->intval ;
    return 0 ;
}

static int _X_CreateGC (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(4, "XCreateGC") ;

    XGCValues xgcv ;
    GC gc ;
    if (!argp[3].IsInstance()) {
	gc = XCreateGC ((Display *)argp[0].voidp, (Window)argp[1].intval,
			(unsigned long)0, (XGCValues *)0) ;
    } else {
	SetGCValues (cenv, &xgcv, &argp[3]) ;
	gc = XCreateGC ((Display *)argp[0].voidp, (Window)argp[1].intval,
			(unsigned long)argp[2].intval, &xgcv) ;
    }
    *retval = (int)gc ;
    return 0 ;
}

static int _X_FreeGC (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK (2, "XFreeGC") ;
    XFreeGC (_DISPLAY(0), _GC(1)) ;
    *retval = 0 ;
    return 0 ;
}

#if 0
static int _X_CopyGC (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(4) ;

    XGCValues xgcv ;
    GC gc ;
    if (!argp[3].IsInstance())
	return 0 ;
    
    SetGCValues (cenv, &xgcv, &argp[3]) ;
    gc = XCopyGC ((Display *)argp[0].voidp, (Window)argp[1].intval,
		  (unsigned long)argp[2].intval, &xgcv) ;
    *retval = (int)gc ;
    return 0 ;
}
#endif

static int _X_SetWindowBackground (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetWindowBackground") ;
    XSetWindowBackground ((Display *)argp[0].voidp, (Window)argp[1].intval,
			  (unsigned long)argp[2].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SetWindowBackgroundPixmap (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetWindowBackgroundPixmap") ;
    XSetWindowBackgroundPixmap ((Display *)argp[0].voidp, (Window)argp[1].intval,
				(Pixmap)argp[2].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SetWindowBorder (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetWindowBorder") ;
    XSetWindowBorder ((Display *)argp[0].voidp, (Window)argp[1].intval,
		      (unsigned long)argp[2].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SetWindowBorderPixmap (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetWindowBorderPixmap") ;
    XSetWindowBorderPixmap ((Display *)argp[0].voidp, (Window)argp[1].intval,
			    (Pixmap)argp[2].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_ChangeGC (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(4, "XChangeGC") ;

    XGCValues xgcv ;
    if (!argp[3].IsInstance())
	return 0 ;
    
    SetGCValues (cenv, &xgcv, &argp[3]) ;
    XChangeGC ((Display *)argp[0].voidp, (GC)argp[1].intval,
	       (unsigned long)argp[2].intval, &xgcv) ;
    *retval = 0 ;
    return 0 ;
}

// XChangeWindowAttributes (display, window, valuemask, attributes)
//				0	1	2	  3
static int _X_ChangeWindowAttributes (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    XSetWindowAttributes wat ;
    Window win ;

    ARGCHK(4, "XChangeWindowAttributes") ;
    SetSetWindowAttribute (cenv ,&wat, &argp[3]) ;
    XChangeWindowAttributes (_DISPLAY(0), _WINDOW(1), _UINT(2), &wat) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SetForeground (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetForeground") ;
    XSetForeground ((Display *)argp[0].voidp, (GC)argp[1].intval,
		    (unsigned long)argp[2].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SetBackground (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetBackground") ;
    XSetBackground ((Display *)argp[0].voidp, (GC)argp[1].intval,
		    (unsigned long)argp[2].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SetFunction (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetFunction") ;
    XSetFunction ((Display *)argp[0].voidp, (GC)argp[1].intval,
		    argp[2].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SetPlaneMask (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetPlaneMask") ;
    XSetPlaneMask ((Display *)argp[0].voidp, (GC)argp[1].intval,
		   (unsigned long)argp[2].intval) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SetFont (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetFont") ;
    XSetFont (_DISPLAY(0), _GC(1), _INT(2)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SetClipOrigin (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(4, "XSetClipOrigin") ;
    XSetClipOrigin (_DISPLAY(0), _GC(1), _X(2), _Y(3)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_SetClipMask (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XSetClipMask") ;
    XSetClipMask (_DISPLAY(0), _GC(1), _PIXMAP(2)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_Bell (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(2, "XBell") ;
    XBell (_DISPLAY(0), _INT(0)) ;
    *retval = 0 ;
    return 0 ;
}

static int _X_QueryPointer (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    Window root, child ;
    int rx, ry, wx, wy ;
    unsigned int mask ;
    Instance *ip ;
    Bool status ;

    ARGCHK(2, "XQueryPointer") ;
    status = XQueryPointer (_DISPLAY(0), _WINDOW(1),
			    &root, &child, &rx, &ry, &wx, &wy, &mask) ;
    ip = cenv->NewInstance ("XQueryPointerReturn") ;
    if (!ip) {
	retval->Init() ;
	return -1 ;
    }
    cenv->SetValue (ip, "in_window", (int)status) ;
    cenv->SetValue (ip, "root", (int)root) ;
    cenv->SetValue (ip, "child", (int)child) ;
    cenv->SetValue (ip, "root_x", rx) ;
    cenv->SetValue (ip, "root_y", ry) ;
    cenv->SetValue (ip, "win_x",  wx) ;
    cenv->SetValue (ip, "win_y",  wy) ;
    cenv->SetValue (ip, "mask", (int)mask) ;
    *retval = ip ;
    return 0 ;
}

// XTranslateCoordinates (display, src_w, dest_w, src_x, src_y)
//			  0	       1       2      3      4
//	returns [x, y, window] ;
//
static int _X_TranslateCoordinates (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    int dstx, dsty ;
    Window child ;
    Bool   ret ; 
    ARGCHK(5, "XTranslateCoordinates") ;
    Value arr ;

    if (!XTranslateCoordinates (_DISPLAY(0), _WINDOW(1), _WINDOW(2),
				_INT(3), _INT(4), &dstx, &dsty, &child))
    {
	retval->Init() ;
    } else {
	arr = cenv->AllocArray(3) ;
	arr[0] = dstx ;
	arr[1] = dsty ;
	arr[2] = (int)child ;
	*retval = arr ;
    }
    return 0 ;
}

static int _X_SetXAnyEvent (CmmInterp *cenv, Instance *ip, XEvent *xev)
{
    cenv->SetValue (ip, "type", (int)xev->xany.type) ;
    cenv->SetValue (ip, "serial", (int)xev->xany.serial) ;
    cenv->SetValue (ip, "display", (void *)xev->xany.display) ;
    cenv->SetValue (ip, "window", (int)xev->xany.window) ;
    return 0 ;
}

static int _X_NextEvent (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "XNextEvent") ;
    
    XEvent event ;
    Value x ;
    XNextEvent ((Display *)argp[0].voidp, &event) ;
    Instance *ip ;

    switch (event.type) {
      case KeyPress:
      case KeyRelease:
	ip = cenv->NewInstance ("XKeyEvent") ;
	if (!ip) {
	    x.Init () ;
	} else {
	    _X_SetXAnyEvent (cenv, ip, &event) ;
	    cenv->SetValue (ip, "root", 	(int)event.xkey.window) ;
	    cenv->SetValue (ip, "subwindow",	(int)event.xkey.subwindow) ;
	    cenv->SetValue (ip, "time",		(int)event.xkey.time) ;
	    cenv->SetValue (ip, "x",		(int)event.xkey.x) ;
	    cenv->SetValue (ip, "y",		(int)event.xkey.y) ;
	    cenv->SetValue (ip, "x_root",	(int)event.xkey.x_root) ;
	    cenv->SetValue (ip, "y_root",	(int)event.xkey.y_root) ;
	    cenv->SetValue (ip, "state",	(int)event.xkey.state) ;
	    cenv->SetValue (ip, "keycode",	(int)event.xkey.keycode) ;
	    cenv->SetValue (ip, "same_screen",	(int)event.xkey.same_screen) ;
	    x = ip ;
	}
	break;
      case ButtonPress:
      case ButtonRelease:
	ip = cenv->NewInstance ("XButtonEvent") ;
	if (!ip) {
	    x.Init () ;
	} else {
	    _X_SetXAnyEvent (cenv, ip, &event) ;
	    cenv->SetValue (ip, "root",		(int)event.xbutton.window) ;
	    cenv->SetValue (ip, "subwindow",	(int)event.xbutton.subwindow) ;
	    cenv->SetValue (ip, "time",		(int)event.xbutton.time) ;
	    cenv->SetValue (ip, "x",		(int)event.xbutton.x) ;
	    cenv->SetValue (ip, "y",		(int)event.xbutton.y) ;
	    cenv->SetValue (ip, "x_root",	(int)event.xbutton.x_root) ;
	    cenv->SetValue (ip, "y_root",	(int)event.xbutton.y_root) ;
	    cenv->SetValue (ip, "state",	(int)event.xbutton.state) ;
	    cenv->SetValue (ip, "button",	(int)event.xbutton.button) ;
	    cenv->SetValue (ip, "same_screen",	(int)event.xbutton.same_screen) ;
	    x = ip ;
	}
	break;
      case MotionNotify:
	ip = cenv->NewInstance ("XMotionEvent") ;
	if (!ip) {
	    x.Init () ;
	} else {
	    _X_SetXAnyEvent (cenv, ip, &event) ;
	    cenv->SetValue (ip, "root",	(int)event.xmotion.window) ;
	    cenv->SetValue (ip, "subwindow",	(int)event.xmotion.subwindow) ;
	    cenv->SetValue (ip, "time",	(int)event.xmotion.time) ;
	    cenv->SetValue (ip, "x",		(int)event.xmotion.x) ;
	    cenv->SetValue (ip, "y",		(int)event.xmotion.y) ;
	    cenv->SetValue (ip, "x_root",	(int)event.xmotion.x_root) ;
	    cenv->SetValue (ip, "y_root",	(int)event.xmotion.y_root) ;
	    cenv->SetValue (ip, "state",	(int)event.xmotion.state) ;
	    cenv->SetValue (ip, "is_hint",	(int)event.xmotion.is_hint) ;
	    cenv->SetValue (ip, "same_screen",(int)event.xmotion.same_screen) ;
	    x = ip ;
	}
	break;
      case Expose:
	ip = cenv->NewInstance ("XExposeEvent") ;
	if (!ip) {
	    x.Init () ;
	} else {
	    _X_SetXAnyEvent (cenv, ip, &event) ;
	    cenv->SetValue (ip, "x",		(int)event.xexpose.x) ;
	    cenv->SetValue (ip, "y",		(int)event.xexpose.y) ;
	    cenv->SetValue (ip, "width",	(int)event.xexpose.width) ;
	    cenv->SetValue (ip, "height",	(int)event.xexpose.height) ;
	    cenv->SetValue (ip, "count",	(int)event.xexpose.count) ;
	    x = ip ;
	}
	break ;
      case EnterNotify:
      case LeaveNotify:
	ip = cenv->NewInstance ("XCrossingEvent") ;
	_X_SetXAnyEvent (cenv, ip, &event) ;
	cenv->SetValue (ip, "root",	   (int)event.xcrossing.root) ;
	cenv->SetValue (ip, "subwindow", (int)event.xcrossing.subwindow) ;
	cenv->SetValue (ip, "time",	   (int)event.xcrossing.time) ;
	cenv->SetValue (ip, "x",	   (int)event.xcrossing.x) ;
	cenv->SetValue (ip, "y",	   (int)event.xcrossing.y) ;
	cenv->SetValue (ip, "x_root",	   (int)event.xcrossing.x_root) ;
	cenv->SetValue (ip, "y_root",	   (int)event.xcrossing.y_root) ;
	cenv->SetValue (ip, "mode",	   (int)event.xcrossing.mode) ;
	cenv->SetValue (ip, "detail",    (int)event.xcrossing.detail) ;
	cenv->SetValue (ip, "same_screen", (int)event.xcrossing.same_screen) ;
	cenv->SetValue (ip, "focus",	(int)event.xcrossing.focus) ;
	cenv->SetValue (ip, "state",  (int)event.xcrossing.state) ;
	x = ip ;
	break ;
      case FocusIn:
      case FocusOut:
	ip = cenv->NewInstance ("XFocusChangeEvent") ;
	_X_SetXAnyEvent (cenv, ip, &event) ;
	cenv->SetValue (ip, "mode", (int)event.xfocus.mode) ;
	cenv->SetValue (ip, "detail", (int)event.xfocus.detail) ;
	break ;
      case PropertyNotify:
	ip = cenv->NewInstance ("XPropertyEvent") ;
	_X_SetXAnyEvent (cenv, ip, &event) ;
	cenv->SetValue (ip, "atom",  (int)event.xproperty.atom) ;
	cenv->SetValue (ip, "time",  (int)event.xproperty.time) ;
	cenv->SetValue (ip, "state", (int)event.xproperty.state) ;
	x = ip ;
	break ;
      case ConfigureNotify:
	ip = cenv->NewInstance ("XConfigureEvent") ;
	_X_SetXAnyEvent (cenv, ip, &event) ;
	cenv->SetValue (ip, "window_configured", (int)event.xconfigure.window) ;
	cenv->SetValue (ip, "display", (int)event.xconfigure.display) ;
	cenv->SetValue (ip, "x", (int)event.xconfigure.x) ;
	cenv->SetValue (ip, "y", (int)event.xconfigure.y) ;
	cenv->SetValue (ip, "width", (int)event.xconfigure.width) ;
	cenv->SetValue (ip, "height", (int)event.xconfigure.height) ;
	cenv->SetValue (ip, "border_width", (int)event.xconfigure.border_width) ;
	cenv->SetValue (ip, "above", (int)event.xconfigure.above) ;
	cenv->SetValue (ip, "override_redirect", (int)event.xconfigure.override_redirect) ;
	x = ip ;
	break ;
      default:
	ip = cenv->NewInstance ("XAnyEvent") ;
	if (!ip) {
	    x.Init () ;
	} else {
	    _X_SetXAnyEvent (cenv, ip, &event) ;
	    x = ip ;
	}
	break ;
    }
    *retval = x ;
    return 0 ;
}

// XLookupString (event)
static int _X_LookupString (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK (1, "XLookupString") ;
    XKeyEvent event ;
    
    Instance *kev = argp[0].instance ;
    event.type   = kev->FindMember (cenv->GetSymbol("type"))->intval ;
    event.serial = kev->FindMember (cenv->GetSymbol("serial"))->intval ;
    event.display = (Display *)kev->FindMember (cenv->GetSymbol("display"))->pointer ;
    event.window = (Window)((kev->FindMember(cenv->GetSymbol("window")))->intval) ;
    event.root   = (Window)((kev->FindMember(cenv->GetSymbol("root")))->intval) ;
    event.subwindow = (Window)((kev->FindMember(cenv->GetSymbol("subwindow")))->intval) ;
    event.time   = ((kev->FindMember(cenv->GetSymbol("time")))->intval) ;
    event.x      = ((kev->FindMember(cenv->GetSymbol("x")))->intval) ;
    event.y      = ((kev->FindMember(cenv->GetSymbol("y")))->intval) ;
    event.x_root = ((kev->FindMember(cenv->GetSymbol("x_root")))->intval) ;
    event.y_root = ((kev->FindMember(cenv->GetSymbol("y_root")))->intval) ;
    event.state  = ((kev->FindMember(cenv->GetSymbol("state")))->intval) ;
    event.keycode = ((kev->FindMember(cenv->GetSymbol("keycode")))->intval) ;
    event.same_screen = ((kev->FindMember(cenv->GetSymbol("same_screen")))->intval) ;

    char buffer[128] ;
    KeySym keysym ;
    XComposeStatus status ;
    int len ;
    
    len = XLookupString (&event, buffer, 128, &keysym, &status) ;

    Value arr0, *arr0p ;

    arr0 = cenv->AllocArray (2) ;
    arr0p = cenv->varstack.TmPush (arr0) ;

    if (len == 0)
	*retval = "" ;
    char *sp ;
    buffer[len] = '\0' ;
    (*arr0p)[0] = cenv->heap_area.AllocString (buffer, &cenv->varstack) ;
    (*arr0p)[1] = (int)keysym ;
    *retval = cenv->varstack.VPop () ;
    return 0 ;
}

static int _X_IsKeypadKey (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "IsKeypadKey") ;
    if (IsKeypadKey (argp[0].intval))
	*retval = 1 ;
    else
	retval->Init () ;
    return 0 ;
}

static int _X_IsPrivateKeypadKey (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "IsPrivateKeypadKey") ;
    if (IsPrivateKeypadKey (argp[0].intval))
	*retval = 1 ;
    else
	retval->Init () ;
    return 0 ;
}

static int _X_IsCursorKey (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "IsCursorKey") ;
    if (IsCursorKey (argp[0].intval))
	*retval = 1 ;
    else
	retval->Init () ;
    return 0 ;
}

static int _X_IsPFKey (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "IsPFKey") ;
    if (IsPFKey (argp[0].intval))
	*retval = 1 ;
    else
	retval->Init() ;
    return 0 ;
}

static int _X_IsFunctionKey (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "IsFunctionKey") ;
    if (IsFunctionKey (argp[0].intval))
	*retval = 1 ;
    else
	retval->Init() ;
    return 0 ;
}

static int _X_IsMiscFunctionKey (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "IsMiscFunctionKey") ;
    if (IsMiscFunctionKey (argp[0].intval))
	*retval = 1 ;
    else
	retval->Init() ;
    return 0 ;
}

static int _X_IsModifierKey (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(1, "IsModifierKey") ;
    if (IsModifierKey (argp[0].intval))
	*retval = 1 ;
    else
	retval->Init() ;
    return 0 ;
}

//		0      1      2    3   4     5    6 7    8     9
// XCopyArea (display, src, dest, gc, srcx, srcy, w, h, destx, desty) 
static int _X_CopyArea (CmmInterp *cenv, int args, Variable *argp, Value *retval)
{
    ARGCHK(10, "XCopyArea") ;
    XCopyArea (_DISPLAY(0), _DRAWABLE(1), _DRAWABLE(2), _GC(3),
	       _X(4), _Y(5), _WIDTH(6), _HEIGHT(7), _X(8), _Y(9)) ;
    *retval = 0 ;
    return 0 ;
}

//			 1,	  2,	    3,		4	    5		
// bitmap->CreatePixmap (display, drawable, foreground, background, depth)
//
static int _XBitmap_CreatePixmap
	(CmmInterp *cenv, Value *thisp, int args, Variable *argp, Value *retval)
{
    ARGCHK(6, "XBitmap::CreatePixmap") ;
    char *data ;
    int  width, height ;
    Pixmap pm ;

    data = MakeBitmapData (cenv, thisp, &width, &height) ;
    if (!data) {
	printf ("XBitmap::CreatePixmap cannot create data");
	return 0 ;
    }
    pm = XCreatePixmapFromBitmapData ((Display *)argp[1].voidp,
				      (Drawable)argp[2].intval,
				      data, width, height,
				      (unsigned long)argp[3].intval,
				      (unsigned long)argp[4].intval,
				      (unsigned int) argp[5].intval) ;
    delete data ;
    *retval = (int)pm ;
    return 0 ;
}

//			   1,	    2,
//   bitmap->CreateBitmap (display, drawable)
//
static int _XBitmap_CreateBitmap
	(CmmInterp *cenv, Value *thisp, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XBitmap::CreateBitmap") ;
    char *data ;
    int  width, height ;
    Pixmap pm ;

    data = MakeBitmapData (cenv, thisp, &width, &height) ;
    if (!data) {
	printf ("XBitmap::CreatePixmap cannot create data");
	return 0 ;
    }
    pm = XCreateBitmapFromData ((Display *)argp[1].voidp,
				(Drawable)argp[2].intval,
				data, width, height) ;
    delete data ;
    *retval = (int)pm ;
    return 0 ;
}

//
// fs->Load (disp, name)
//
static int _XFontStruct_Load
	(CmmInterp *cenv, Value *thisp, int args, Variable *argp, Value *retval)
{
    ARGCHK(3, "XFontStruct::Load") ;
    XFontStruct *xf ;
    xf = XLoadQueryFont ((Display *)argp[1].voidp, argp[2].sptr) ;
    if (!xf) {
	retval->Init() ;
	return 0 ;
    }
    cenv->SetValue (thisp, "fid",       (int)xf->fid) ;
    cenv->SetValue (thisp, "direction", (int)xf->direction) ;
    cenv->SetValue (thisp, "ascent",    (int)xf->ascent) ;
    cenv->SetValue (thisp, "descent",   (int)xf->descent) ;
    cenv->SetValue (thisp, "instance",  (void *)xf) ;
    cenv->SetValue (thisp, "display",   (void *)argp[1].voidp) ;
    *retval = 1 ;
    return 0 ;
}

//
// fs->Free ()
//
static int _XFontStruct_Free
	(CmmInterp *cenv, Value *thisp, int args, Variable *argp, Value *retval)
{
    Display *display ;
    XFontStruct *xf ;
    display = (Display *)(cenv->FindMember (thisp, "display"))->voidp ;
    xf = (XFontStruct *)(cenv->FindMember (thisp, "instance"))->voidp ;
    XFreeFont (display, xf) ;
    return 0 ;
}

char xClassDef [] = 
"class XGCValues {
	public  function, plane_mask, foreground, background, \n\
		line_width, line_style,	cap_style, join_style, fill_style,\n\
		fill_rule, arc_mode,	tile, stipple,\n\
		ts_x_origin, ts_y_origin, font,\n\
		subwindow_mode,	graphics_exposures,\n\
		clip_x_origin, clip_y_origin, clip_mask, dash_offset, dashes;\n\
} ;\n\
class XEvent {	public type ;} ;\n\
class XAnyEvent : XEvent { public serial, send_event, display, window ;}\n\
class XKeyEvent : XAnyEvent {\n\
    public root, subwindow, time, x, y, x_root, y_root, state, keycode, same_screen ;\n\
} ;\n\
class XButtonEvent : XAnyEvent {\n\
	public root, subwindow, time, x, y, x_root, y_root, state, button, same_sreen ;\n\
} ;\n\
class XMotionEvent: XAnyEvent {\n\
	public root, subwindow, time, x, y, x_root, y_root, state, is_hint, same_sreen ;\n\
} ;\n\
class XClossingEvent : XAnyEvent {\n\
	public root, subwindow, time, x, y, x_root, y_root, mode, detail,\n\
		same_sreen, focus, state ;\n\
}\n\
class XFocusChangeEvent : XAnyEvent {\n\
	public mode, detial ;\n\
};\n\
class XExposeEvent : XAnyEvent {\n\
	public x, y, width, height, count ;\n\
} ;\n\
class XGraphicsExposeEvent : XAnyEvent {\n\
	public x, y, width, height, count, major_code, minor_code ;\n\
};\n\
class XPropertyEvent : XAnyEvent {\n\
	public atom, time, state ;\n\
};\n\
class XCrossingEvent : XAnyEvent {\n\
	public root, subwindow, time, x, y, x_root, y_root, \n\
		mode, detail, focus, state ;\n\
};\n\
class XFocusChangeEvent : XAnyEvent {\n\
	public mode, detail ;
};\n\
class XConfigureEvent : XAnyEvent {\n\
	public window_configured ;
	public x, y, width, height, border_width ;
	public above, override_redirect ;
};\n\
class XBitmap {\n\
    public width, height ;\n\
    public xhot,  yhot ;\n\
    public data ;\n\
    method Set (w, h, data) {\n\
	var w_in_byte ;\n\
	var size ;\n\
	width = w ;\n\
	height = h;\n\
	w_in_byte = (w + 7)/8 ;\n\
	xhot = yhot = 0 ;\n\
	size = w_in_byte * h ;\n\
	if (!is_array (data))\n\
	    return ;\n\
	if (length(data) != size)\n\
	    return ;\n\
	->data = data ;\n\
    } ;\n\
    method SetHotSpot (x, y) {\n\
	xhot = x ;\n\
	yhot = y ;\n\
    }\n\
    method CreatePixmap ;
    method CreateBitmap ;
} ;\n\
class XFontStruct {\n\
    public fid;\n\
    public direction;\n\
    public ascent;\n\
    public descent;\n\
    private display ;\n\
    private instance ;\n\
    method Load ;
    method Free ;
} ;\n\
class XSetWindowAttributes {\n\
    public background_pixmap;\n\
    public background_pixel;\n\
    public border_pixmap;\n\
    public border_pixel;\n\
    public bit_gravity;\n\
    public win_gravity;\n\
    public backing_store;\n\
    public backing_planes;\n\
    public backing_pixel;\n\
    public save_under;\n\
    public event_mask;\n\
    public do_not_propagate_mask;\n\
    public override_redirect;\n\
    public colormap;\n\
    public cursor;\n\
} ;\n\
class XQueryPointerReturn {\n\
    public in_window ;\n\
    public root ;\n\
    public child ;\n\
    public root_x, root_y ;\n\
    public win_x, win_y ;\n\
    public mask ;\n\
}\n\
" ;

int InstallX (CmmInterp *calc)
{
    // EventMasks
    calc->AddGlobalVar ("NoEventMask",	 (int)NoEventMask) ;
    calc->AddGlobalVar ("KeyPressMask",	 (int)KeyPressMask) ;
    calc->AddGlobalVar ("KeyReleaseMask",	(int)KeyReleaseMask) ;
    calc->AddGlobalVar ("ButtonPressMask",	(int)ButtonPressMask) ;
    calc->AddGlobalVar ("ButtonReleaseMask",	(int)ButtonReleaseMask) ;
    calc->AddGlobalVar ("EnterWindowMask",	 (int)EnterWindowMask) ;
    calc->AddGlobalVar ("LeaveWindowMask",	 (int)LeaveWindowMask) ;
    calc->AddGlobalVar ("PointerMotionMask",	 (int)PointerMotionMask) ;
    calc->AddGlobalVar ("PointerMotionHintMask",	(int) PointerMotionHintMask) ;
    calc->AddGlobalVar ("Button1MotionMask",	 (int)Button1MotionMask) ;
    calc->AddGlobalVar ("Button2MotionMask",	 (int)Button2MotionMask) ;
    calc->AddGlobalVar ("Button3MotionMask",	 (int)Button3MotionMask) ;
    calc->AddGlobalVar ("Button4MotionMask",	 (int)Button4MotionMask) ;
    calc->AddGlobalVar ("Button5MotionMask",	 (int)Button5MotionMask) ;
    calc->AddGlobalVar ("ButtonMotionMask",	 (int)ButtonMotionMask) ;
    calc->AddGlobalVar ("KeymapStateMask",	 (int)KeymapStateMask) ;
    calc->AddGlobalVar ("ExposureMask",	 (int)ExposureMask) ;
    calc->AddGlobalVar ("VisibilityChangeMask",	 (int)VisibilityChangeMask) ;
    calc->AddGlobalVar ("StructureNotifyMask",	 (int)StructureNotifyMask) ;
    calc->AddGlobalVar ("ResizeRedirectMask",	 (int)ResizeRedirectMask) ;
    calc->AddGlobalVar ("SubstructureNotifyMask",	(int)SubstructureNotifyMask) ;
    calc->AddGlobalVar ("SubstructureRedirectMask",	 (int)SubstructureRedirectMask) ;
    calc->AddGlobalVar ("FocusChangeMask",	 (int)FocusChangeMask) ;
    calc->AddGlobalVar ("PropertyChangeMask",	 (int)PropertyChangeMask) ;
    calc->AddGlobalVar ("ColormapChangeMask",	 (int)ColormapChangeMask) ;
    calc->AddGlobalVar ("OwnerGrabButtonMask",	 (int)OwnerGrabButtonMask) ;
    // Events
    calc->AddGlobalVar ("KeyPress",      KeyPress) ;
    calc->AddGlobalVar ("KeyRelease",    KeyRelease) ;
    calc->AddGlobalVar ("ButtonPress",   ButtonPress) ;
    calc->AddGlobalVar ("ButtonRelease", ButtonRelease) ;
    calc->AddGlobalVar ("MotionNotify",  MotionNotify) ;
    calc->AddGlobalVar ("EnterNotify",   EnterNotify) ;
    calc->AddGlobalVar ("LeaveNotify",	 LeaveNotify) ;
    calc->AddGlobalVar ("FocusIn",	 FocusIn) ;
    calc->AddGlobalVar ("FocusIn",	 FocusIn) ;
    calc->AddGlobalVar ("FocusOut",	 FocusOut) ;
    calc->AddGlobalVar ("KeymapNotify",	 KeymapNotify) ;
    calc->AddGlobalVar ("Expose",	 Expose) ;
    calc->AddGlobalVar ("GraphicsExpose",	 GraphicsExpose) ;
    calc->AddGlobalVar ("NoExpose",	 NoExpose) ;
    calc->AddGlobalVar ("VisibilityNotify",	 VisibilityNotify) ;
    calc->AddGlobalVar ("CreateNotify",	 CreateNotify) ;
    calc->AddGlobalVar ("DestroyNotify",	 DestroyNotify) ;
    calc->AddGlobalVar ("UnmapNotify",	 UnmapNotify) ;
    calc->AddGlobalVar ("MapNotify",	 MapNotify) ;
    calc->AddGlobalVar ("MapRequest",	 MapRequest) ;
    calc->AddGlobalVar ("ReparentNotify",	 ReparentNotify) ;
    calc->AddGlobalVar ("ConfigureNotify",	 ConfigureNotify) ;
    calc->AddGlobalVar ("ConfigureRequest",	 ConfigureRequest) ;
    calc->AddGlobalVar ("GravityNotify",	 GravityNotify) ;
    calc->AddGlobalVar ("ResizeRequest",	 ResizeRequest) ;
    calc->AddGlobalVar ("CirculateNotify",	 CirculateNotify) ;
    calc->AddGlobalVar ("CirculateRequest",	 CirculateRequest) ;
    calc->AddGlobalVar ("PropertyNotify",	 PropertyNotify) ;
    calc->AddGlobalVar ("SelectionClear",	 SelectionClear) ;
    calc->AddGlobalVar ("SelectionRequest",	 SelectionRequest) ;
    calc->AddGlobalVar ("SelectionNotify",	 SelectionNotify) ;
    calc->AddGlobalVar ("ColormapNotify",	 ColormapNotify) ;
    calc->AddGlobalVar ("ClientMessage",	 ClientMessage) ;
    calc->AddGlobalVar ("MappingNotify",	 MappingNotify) ;
    //
    calc->AddGlobalVar ("Button1",	 Button1) ;
    calc->AddGlobalVar ("Button2",	 Button2) ;
    calc->AddGlobalVar ("Button3",	 Button3) ;
    calc->AddGlobalVar ("Button4",	 Button4) ;
    calc->AddGlobalVar ("Button5",	 Button5) ;
    
    //
    calc->AddGlobalVar ("NotifyNormal",	 NotifyNormal) ;
    calc->AddGlobalVar ("NotifyGrab",	 NotifyGrab) ;
    calc->AddGlobalVar ("NotifyUngrab",	 NotifyUngrab) ;
    calc->AddGlobalVar ("NotifyWhileGrabbed",	 NotifyWhileGrabbed) ;
    calc->AddGlobalVar ("NotifyHint",	NotifyHint) ;
    //
    calc->AddGlobalVar ("ShiftMask",	ShiftMask) ;
    calc->AddGlobalVar ("LockMask",	LockMask) ;
    calc->AddGlobalVar ("ControlMask",	ControlMask) ;
    calc->AddGlobalVar ("Mod1Mask",	Mod1Mask) ;
    calc->AddGlobalVar ("Mod2Mask",	Mod2Mask) ;
    calc->AddGlobalVar ("Mod3Mask",	Mod3Mask) ;
    calc->AddGlobalVar ("Mod4Mask",	Mod4Mask) ;
    calc->AddGlobalVar ("Mod5Mask",	Mod5Mask) ;
    //
    calc->AddGlobalVar ("Button1Mask",  Button1Mask) ;
    calc->AddGlobalVar ("Button2Mask",  Button2Mask) ;
    calc->AddGlobalVar ("Button3Mask",	Button3Mask) ;
    calc->AddGlobalVar ("Button4Mask",  Button4Mask) ;
    calc->AddGlobalVar ("Button5Mask",	Button5Mask) ;
    calc->AddGlobalVar ("AnyModifier",	AnyModifier) ;

    //
    calc->AddGlobalVar ("NotifyHint",	 NotifyHint) ;
		       
    // Notify detail
    calc->AddGlobalVar ("NotifyAncestor",	 NotifyAncestor) ;
    calc->AddGlobalVar ("NotifyVirtual",	 NotifyVirtual) ;
    calc->AddGlobalVar ("NotifyInferior",	 NotifyInferior) ;
    calc->AddGlobalVar ("NotifyNonlinear",	 NotifyNonlinear) ;
    calc->AddGlobalVar ("NotifyNonlinearVirtual",	 NotifyNonlinearVirtual) ;
    calc->AddGlobalVar ("NotifyPointer",	 NotifyPointer) ;
    calc->AddGlobalVar ("NotifyPointerRoot",	 NotifyPointerRoot) ;
    calc->AddGlobalVar ("NotifyDetailNone",	 NotifyDetailNone) ;

    calc->AddGlobalVar ("VisibilityUnobscured",	 VisibilityUnobscured) ;
    calc->AddGlobalVar ("VisibilityPartiallyObscured",	 VisibilityPartiallyObscured) ;
    calc->AddGlobalVar ("VisibilityFullyObscured",	 VisibilityFullyObscured) ;

    // Circulation request
    calc->AddGlobalVar ("PlaceOnTop",	 PlaceOnTop) ;
    calc->AddGlobalVar ("PlaceOnBottom",	 PlaceOnBottom) ;

    // protocol families
    calc->AddGlobalVar ("FamilyInternet",	 FamilyInternet) ;
    calc->AddGlobalVar ("FamilyDECnet",	 FamilyDECnet) ;
    calc->AddGlobalVar ("FamilyChaos",	 FamilyChaos) ;

    // Property notification
    calc->AddGlobalVar ("PropertyNewValue",	 PropertyNewValue) ;
    calc->AddGlobalVar ("PropertyDelete",	 PropertyDelete) ;

    // Color Map notification 
    calc->AddGlobalVar ("ColormapUninstalled",	 ColormapUninstalled) ;
    calc->AddGlobalVar ("ColormapInstalled",	 ColormapInstalled) ;

    // GrabPointer, GrabButton, GrabKeyboard, GrabKey Modes
    calc->AddGlobalVar ("GrabModeSync",	 GrabModeSync) ;
    calc->AddGlobalVar ("GrabModeAsync",	 GrabModeAsync) ;

    // GrabPointer, GrabKeyboard reply status
    calc->AddGlobalVar ("GrabSuccess",	 GrabSuccess) ;
    calc->AddGlobalVar ("AlreadyGrabbed",	 AlreadyGrabbed) ;
    calc->AddGlobalVar ("GrabInvalidTime",	 GrabInvalidTime) ;
    calc->AddGlobalVar ("GrabNotViewable",	 GrabNotViewable) ;
    calc->AddGlobalVar ("GrabFrozen",	 GrabFrozen) ;

    // AllowEvents modes
    calc->AddGlobalVar ("AsyncPointer",	 AsyncPointer) ;
    calc->AddGlobalVar ("SyncPointer",	 SyncPointer) ;
    calc->AddGlobalVar ("ReplayPointer",	 ReplayPointer) ;
    calc->AddGlobalVar ("AsyncKeyboard",	 AsyncKeyboard) ;
    calc->AddGlobalVar ("SyncKeyboard",	 SyncKeyboard) ;
    calc->AddGlobalVar ("ReplayKeyboard",	 ReplayKeyboard) ;
    calc->AddGlobalVar ("AsyncBoth",	 AsyncBoth) ;
    calc->AddGlobalVar ("SyncBoth",	 SyncBoth) ;

    // Used in SetInputFocus, GetInputFocus 
    calc->AddGlobalVar ("RevertToNone",	 RevertToNone) ;
    calc->AddGlobalVar ("RevertToPointerRoot",	 RevertToPointerRoot) ;
    calc->AddGlobalVar ("RevertToParent",	 RevertToParent) ;

    calc->AddGlobalVar ("InputOutput",	 InputOutput) ;
    calc->AddGlobalVar ("InputOnly",	 InputOnly) ;

    calc->AddGlobalVar ("CWBackPixmap",	 (int)CWBackPixmap) ;
    calc->AddGlobalVar ("CWBackPixel",	 (int)CWBackPixel) ;
    calc->AddGlobalVar ("CWBorderPixmap",	 (int)CWBorderPixmap) ;
    calc->AddGlobalVar ("CWBorderPixel",	 (int)CWBorderPixel) ;
    calc->AddGlobalVar ("CWBitGravity",	 (int)CWBitGravity) ;
    calc->AddGlobalVar ("CWWinGravity",	 (int)CWWinGravity) ;
    calc->AddGlobalVar ("CWBackingStore",	 (int)CWBackingStore) ;
    calc->AddGlobalVar ("CWBackingPlanes",	 (int)CWBackingPlanes) ;
    calc->AddGlobalVar ("CWBackingPixel",	 (int)CWBackingPixel) ;
    calc->AddGlobalVar ("CWOverrideRedirect",	 (int)CWOverrideRedirect) ;
    calc->AddGlobalVar ("CWSaveUnder",	 (int)CWSaveUnder) ;
    calc->AddGlobalVar ("CWEventMask",	 (int)CWEventMask) ;
    calc->AddGlobalVar ("CWDontPropagate",	 (int)CWDontPropagate) ;
    calc->AddGlobalVar ("CWColormap",	 (int)CWColormap) ;
    calc->AddGlobalVar ("CWCursor",	 (int)CWCursor) ;

    // ConfigureWindow structure
    calc->AddGlobalVar ("CWX",	 (int)CWX) ;
    calc->AddGlobalVar ("CWY",	 (int)CWY) ;
    calc->AddGlobalVar ("CWWidth",	 (int)CWWidth) ;
    calc->AddGlobalVar ("CWHeight",	 (int)CWHeight) ;
    calc->AddGlobalVar ("CWBorderWidth",	(int)CWBorderWidth) ;
    calc->AddGlobalVar ("CWSibling",	 (int)CWSibling) ;
    calc->AddGlobalVar ("CWStackMode",	 (int)CWStackMode) ;

    // Bit Gravity
    calc->AddGlobalVar ("ForgetGravity",	 ForgetGravity) ;
    calc->AddGlobalVar ("NorthWestGravity",	 NorthWestGravity) ;
    calc->AddGlobalVar ("NorthGravity",	 NorthGravity) ;
    calc->AddGlobalVar ("NorthEastGravity",	 NorthEastGravity) ;
    calc->AddGlobalVar ("WestGravity",	 WestGravity) ;
    calc->AddGlobalVar ("CenterGravity",	 CenterGravity) ;
    calc->AddGlobalVar ("EastGravity",	 EastGravity) ;
    calc->AddGlobalVar ("SouthWestGravity",	 SouthWestGravity) ;
    calc->AddGlobalVar ("SouthGravity",	 SouthGravity) ;
    calc->AddGlobalVar ("SouthEastGravity",	 SouthEastGravity) ;
    calc->AddGlobalVar ("StaticGravity",	 StaticGravity) ;

    // Window gravity + bit gravity above
    calc->AddGlobalVar ("UnmapGravity",	 UnmapGravity) ;

    // Used in CreateWindow for backing-store hint
    calc->AddGlobalVar ("NotUseful",	 NotUseful) ;
    calc->AddGlobalVar ("WhenMapped",	 WhenMapped) ;
    calc->AddGlobalVar ("Always",	 Always) ;

    // Used in GetWindowAttributes reply
    calc->AddGlobalVar ("IsUnmapped",	 IsUnmapped) ;
    calc->AddGlobalVar ("IsUnviewable",	 IsUnviewable) ;
    calc->AddGlobalVar ("IsViewable",	 IsViewable) ;

    // Used in ChangeSaveSet
    calc->AddGlobalVar ("SetModeInsert",	 SetModeInsert) ;
    calc->AddGlobalVar ("SetModeDelete",	 SetModeDelete) ;

    // Used in ChangeCloseDownMode
    calc->AddGlobalVar ("DestroyAll",	 DestroyAll) ;
    calc->AddGlobalVar ("RetainPermanent",	 RetainPermanent) ;
    calc->AddGlobalVar ("RetainTemporary",	 RetainTemporary) ;

    // Window stacking method (in configureWindow)
    calc->AddGlobalVar ("Above",	 Above) ;
    calc->AddGlobalVar ("Below",	 Below) ;
    calc->AddGlobalVar ("TopIf",	 TopIf) ;
    calc->AddGlobalVar ("BottomIf",	 BottomIf) ;
    calc->AddGlobalVar ("Opposite",	 Opposite) ;

    // Circulation direction
    calc->AddGlobalVar ("RaiseLowest",	 RaiseLowest) ;
    calc->AddGlobalVar ("LowerHighest",	 LowerHighest) ;

    // Property modes
    calc->AddGlobalVar ("PropModeReplace",	 PropModeReplace) ;
    calc->AddGlobalVar ("PropModePrepend",	 PropModePrepend) ;
    calc->AddGlobalVar ("PropModeAppend",	 PropModeAppend) ;

    //
    // GRAPHICS DEFINITIONS
    //

    /* graphics functions, as in GC.alu */
    calc->AddGlobalVar ("GXclear",	 GXclear) ;
    calc->AddGlobalVar ("GXand",	 GXand) ;
    calc->AddGlobalVar ("GXandReverse",	 GXandReverse) ;
    calc->AddGlobalVar ("GXcopy",	 GXcopy) ;
    calc->AddGlobalVar ("GXandInverted",	 GXandInverted) ;
    calc->AddGlobalVar ("GXnoop",	 GXnoop) ;
    calc->AddGlobalVar ("GXxor",	 GXxor) ;
    calc->AddGlobalVar ("GXor",	 GXor) ;
    calc->AddGlobalVar ("GXnor",	 GXnor) ;
    calc->AddGlobalVar ("GXequiv",	 GXequiv) ;
    calc->AddGlobalVar ("GXinvert",	 GXinvert) ;
    calc->AddGlobalVar ("GXorReverse",	 GXorReverse) ;
    calc->AddGlobalVar ("GXcopyInverted",	 GXcopyInverted) ;
    calc->AddGlobalVar ("GXorInverted",	 GXorInverted) ;
    calc->AddGlobalVar ("GXnand",	 GXnand) ;
    calc->AddGlobalVar ("GXset",	 GXset) ;

    /* LineStyle */
    calc->AddGlobalVar ("LineSolid",	 LineSolid) ;
    calc->AddGlobalVar ("LineOnOffDash",	 LineOnOffDash) ;
    calc->AddGlobalVar ("LineDoubleDash",	 LineDoubleDash) ;

    /* capStyle */
    calc->AddGlobalVar ("CapNotLast",	 CapNotLast) ;
    calc->AddGlobalVar ("CapButt",	 CapButt) ;
    calc->AddGlobalVar ("CapRound",	 CapRound) ;
    calc->AddGlobalVar ("CapProjecting",	 CapProjecting) ;

    /* joinStyle */
    calc->AddGlobalVar ("JoinMiter",	 JoinMiter) ;
    calc->AddGlobalVar ("JoinRound",	 JoinRound) ;
    calc->AddGlobalVar ("JoinBevel",	 JoinBevel) ;

    /* fillStyle */
    calc->AddGlobalVar ("FillSolid",	 FillSolid) ;
    calc->AddGlobalVar ("FillTiled",	 FillTiled) ;
    calc->AddGlobalVar ("FillStippled",	 FillStippled) ;
    calc->AddGlobalVar ("FillOpaqueStippled",	 FillOpaqueStippled) ;

    /* fillRule */
    calc->AddGlobalVar ("EvenOddRule",	 EvenOddRule) ;
    calc->AddGlobalVar ("WindingRule",	 WindingRule) ;

    /* subwindow mode */
    calc->AddGlobalVar ("ClipByChildren",	 ClipByChildren) ;
    calc->AddGlobalVar ("IncludeInferiors",	 IncludeInferiors) ;

    /* SetClipRectangles ordering */
    calc->AddGlobalVar ("Unsorted",	 Unsorted) ;
    calc->AddGlobalVar ("YSorted",	 YSorted) ;
    calc->AddGlobalVar ("YXSorted",	 YXSorted) ;
    calc->AddGlobalVar ("YXBanded",	 YXBanded) ;

    /* CoordinateMode for drawing routines */
    calc->AddGlobalVar ("CoordModeOrigin",	 CoordModeOrigin) ;
    calc->AddGlobalVar ("CoordModePrevious",	 CoordModePrevious) ;

    /* Polygon shapes */
    calc->AddGlobalVar ("Complex",	 Complex) ;
    calc->AddGlobalVar ("Nonconvex",	 Nonconvex) ;
    calc->AddGlobalVar ("Convex",	 Convex) ;

    /* Arc modes for PolyFillArc */
    calc->AddGlobalVar ("ArcChord",	 ArcChord) ;
    calc->AddGlobalVar ("ArcPieSlice",	 ArcPieSlice) ;

    /* GC components: masks used in CreateGC, CopyGC, ChangeGC, OR'ed into
       GC.stateChanges */
    calc->AddGlobalVar ("GCFunction",	 (int)GCFunction) ;
    calc->AddGlobalVar ("GCPlaneMask",	 (int)GCPlaneMask) ;
    calc->AddGlobalVar ("GCForeground",	 (int)GCForeground) ;
    calc->AddGlobalVar ("GCBackground",	 (int)GCBackground) ;
    calc->AddGlobalVar ("GCLineWidth",	 (int)GCLineWidth) ;
    calc->AddGlobalVar ("GCLineStyle",	 (int)GCLineStyle) ;
    calc->AddGlobalVar ("GCCapStyle",	 (int)GCCapStyle) ;
    calc->AddGlobalVar ("GCJoinStyle",	 (int)GCJoinStyle) ;
    calc->AddGlobalVar ("GCFillStyle",	 (int)GCFillStyle) ;
    calc->AddGlobalVar ("GCFillRule",	 (int)GCFillRule) ;
    calc->AddGlobalVar ("GCTile",	 (int)GCTile) ;
    calc->AddGlobalVar ("GCStipple",	 (int)GCStipple) ;
    calc->AddGlobalVar ("GCTileStipXOrigin",	 (int)GCTileStipXOrigin) ;
    calc->AddGlobalVar ("GCTileStipYOrigin",	 (int)GCTileStipYOrigin) ;
    calc->AddGlobalVar ("GCFont",	 (int)GCFont) ;
    calc->AddGlobalVar ("GCSubwindowMode",	 (int)GCSubwindowMode) ;
    calc->AddGlobalVar ("GCGraphicsExposures",	 (int)GCGraphicsExposures) ;
    calc->AddGlobalVar ("GCClipXOrigin",	 (int)GCClipXOrigin) ;
    calc->AddGlobalVar ("GCClipYOrigin",	 (int)GCClipYOrigin) ;
    calc->AddGlobalVar ("GCClipMask",	 (int)GCClipMask) ;
    calc->AddGlobalVar ("GCDashOffset",	 (int)GCDashOffset) ;
    calc->AddGlobalVar ("GCDashList",	 (int)GCDashList) ;
    calc->AddGlobalVar ("GCArcMode",	 (int)GCArcMode) ;

    /*****************************************************************
     * FONTS 
     *****************************************************************/
    /* used in QueryFont -- draw direction */
    calc->AddGlobalVar ("FontLeftToRight",	 FontLeftToRight) ;
    calc->AddGlobalVar ("FontRightToLeft",	 FontRightToLeft) ;
    calc->AddGlobalVar ("FontChange",	 FontChange) ;

    /*****************************************************************
     *  IMAGING 
     *****************************************************************/
    /* ImageFormat -- PutImage, GetImage */
    calc->AddGlobalVar ("XYBitmap",	 XYBitmap) ;
    calc->AddGlobalVar ("XYPixmap",	 XYPixmap) ;
    calc->AddGlobalVar ("ZPixmap",	 ZPixmap) ;

    /*****************************************************************
     *  COLOR MAP STUFF 
     *****************************************************************/
    /* For CreateColormap */
    calc->AddGlobalVar ("AllocNone",	 AllocNone) ;
    calc->AddGlobalVar ("AllocAll",	 AllocAll) ;
    /* Flags used in StoreNamedColor, StoreColors */
    calc->AddGlobalVar ("DoRed",	 DoRed) ;
    calc->AddGlobalVar ("DoGreen",	 DoGreen) ;
    calc->AddGlobalVar ("DoBlue",	 DoBlue) ;

    /*****************************************************************
     * CURSOR STUFF
     *****************************************************************/
    /* QueryBestSize Class */
    calc->AddGlobalVar ("CursorShape",	 CursorShape) ;
    calc->AddGlobalVar ("TileShape",	 TileShape) ;
    calc->AddGlobalVar ("StippleShape",	 StippleShape) ;

    /***************************************************************** 
     * KEYBOARD/POINTER STUFF
     *****************************************************************/
    calc->AddGlobalVar ("AutoRepeatModeOff",	 AutoRepeatModeOff) ;
    calc->AddGlobalVar ("AutoRepeatModeOn",	 AutoRepeatModeOn) ;
    calc->AddGlobalVar ("AutoRepeatModeDefault",	 AutoRepeatModeDefault) ;
    calc->AddGlobalVar ("LedModeOff",	 LedModeOff) ;
    calc->AddGlobalVar ("LedModeOn",	 LedModeOn) ;

    /* masks for ChangeKeyboardControl */
    calc->AddGlobalVar ("KBKeyClickPercent",	 (int)KBKeyClickPercent) ;
    calc->AddGlobalVar ("KBBellPercent",	 (int)KBBellPercent) ;
    calc->AddGlobalVar ("KBBellPitch",	 (int)KBBellPitch) ;
    calc->AddGlobalVar ("KBBellDuration",	 (int)KBBellDuration) ;
    calc->AddGlobalVar ("KBLed",	 (int)KBLed) ;
    calc->AddGlobalVar ("KBLedMode",	 (int)KBLedMode) ;
    calc->AddGlobalVar ("KBKey",	 (int)KBKey) ;
    calc->AddGlobalVar ("KBAutoRepeatMode",	 (int)KBAutoRepeatMode) ;

    calc->AddGlobalVar ("MappingSuccess",	 MappingSuccess) ;
    calc->AddGlobalVar ("MappingBusy",	 MappingBusy) ;
    calc->AddGlobalVar ("MappingFailed",	 MappingFailed) ;
    calc->AddGlobalVar ("MappingModifier",	 MappingModifier) ;
    calc->AddGlobalVar ("MappingKeyboard",	 MappingKeyboard) ;
    calc->AddGlobalVar ("MappingPointer",	 MappingPointer) ;

    /*****************************************************************
     * SCREEN SAVER STUFF 
     *****************************************************************/
    calc->AddGlobalVar ("DontPreferBlanking",	 DontPreferBlanking) ;
    calc->AddGlobalVar ("PreferBlanking",	 PreferBlanking) ;
    calc->AddGlobalVar ("DefaultBlanking",	 DefaultBlanking) ;
    calc->AddGlobalVar ("DisableScreenSaver",	 DisableScreenSaver) ;
    calc->AddGlobalVar ("DisableScreenInterval",	 DisableScreenInterval) ;

    calc->AddGlobalVar ("DontAllowExposures",	 DontAllowExposures) ;
    calc->AddGlobalVar ("AllowExposures",	 AllowExposures) ;
    calc->AddGlobalVar ("DefaultExposures",	 DefaultExposures) ;

    calc->AddGlobalVar ("ScreenSaverReset",	 ScreenSaverReset) ;
    calc->AddGlobalVar ("ScreenSaverActive",	 ScreenSaverActive) ;

    /*****************************************************************
     * HOSTS AND CONNECTIONS
     *****************************************************************/
    /* for ChangeHosts */
    calc->AddGlobalVar ("HostInsert",	 HostInsert) ;
    calc->AddGlobalVar ("HostDelete",	 HostDelete) ;

    calc->AddGlobalVar ("EnableAccess",	 EnableAccess) ;
    calc->AddGlobalVar ("DisableAccess",	 DisableAccess) ;

    /* Display classes  used in opening the connection 
     * Note that the statically allocated ones are even numbered and the
     * dynamically changeable ones are odd numbered */
    calc->AddGlobalVar ("StaticGray",	 StaticGray) ;
    calc->AddGlobalVar ("GrayScale",	 GrayScale) ;
    calc->AddGlobalVar ("StaticColor",	 StaticColor) ;
    calc->AddGlobalVar ("PseudoColor",	 PseudoColor) ;
    calc->AddGlobalVar ("TrueColor",	 TrueColor) ;
    calc->AddGlobalVar ("DirectColor",	 DirectColor) ;

    /* Byte order  used in imageByteOrder and bitmapBitOrder */
    calc->AddGlobalVar ("LSBFirst",	 LSBFirst) ;
    calc->AddGlobalVar ("MSBFirst",	 MSBFirst) ;
    /*****************************************************************
     * RESERVED RESOURCE AND CONSTANT DEFINITIONS
     *****************************************************************/
    calc->AddGlobalVar ("None",			(int)None) ;
    calc->AddGlobalVar ("ParentRelative",	(int)ParentRelative) ;
    calc->AddGlobalVar ("CopyFromParent",	(int)CopyFromParent) ;
    calc->AddGlobalVar ("PointerWindow",	(int)PointerWindow) ;
    calc->AddGlobalVar ("InputFocus",		(int)InputFocus) ;
    calc->AddGlobalVar ("PointerRoot",		(int)PointerRoot) ;
    calc->AddGlobalVar ("AnyPropertyType",	(int)AnyPropertyType) ;
    calc->AddGlobalVar ("AnyKey",		(int)AnyKey) ;
    calc->AddGlobalVar ("AnyButton",		(int)AnyButton) ;
    calc->AddGlobalVar ("AllTemporary",		(int)AllTemporary) ;
    calc->AddGlobalVar ("CurrentTime",		(int)CurrentTime) ;
    calc->AddGlobalVar ("NoSymbol",		(int)NoSymbol) ;

    calc->AddGlobalVar ("AllPlanes",		(int)AllPlanes) ;
    //
    // XLib interface
    //
    calc->AddBuiltin ("XOpenDisplay", _X_OpenDisplay) ;
    calc->AddBuiltin ("XCreateWindow", _X_CreateWindow) ;
    calc->AddBuiltin ("XCreateSimpleWindow", _X_CreateSimpleWindow) ;
    calc->AddBuiltin ("XSelectInput",	_X_SelectInput) ;
    calc->AddBuiltin ("XNextEvent",	_X_NextEvent) ;
    calc->AddBuiltin ("DefaultVisual",  _X_DefaultVisual) ;
    calc->AddBuiltin ("DefaultScreen",	_X_DefaultScreen) ;
    calc->AddBuiltin ("BlackPixel",	_X_BlackPixel) ;
    calc->AddBuiltin ("WhitePixel",	_X_WhitePixel) ;
    calc->AddBuiltin ("XMapWindow",	_X_MapWindow) ;
    calc->AddBuiltin ("XUnmapWindow",	_X_UnmapWindow) ;
    calc->AddBuiltin ("RootWindow",	_X_RootWindow) ;
    calc->AddBuiltin ("XCreateGC",	_X_CreateGC) ;
    calc->AddBuiltin ("XFreeGC",	_X_FreeGC) ;
#if 0
    calc->AddBuiltin ("XCopyGC",	_X_CopyGC) ;
#endif
    calc->AddBuiltin ("XChangeGC",      _X_ChangeGC) ;
    calc->AddBuiltin ("XLoadFont",      _X_LoadFont) ;
    calc->AddBuiltin ("XDrawLine",      _X_DrawLine) ;
    calc->AddBuiltin ("XDrawLines",     _X_DrawLines) ;
    calc->AddBuiltin ("XDrawRectangle", _X_DrawRectangle) ;
    calc->AddBuiltin ("XDrawString",    _X_DrawString) ;
    calc->AddBuiltin ("XDrawString16",  _X_DrawString16) ;
    calc->AddBuiltin ("XDrawImageString", _X_DrawImageString) ;
    calc->AddBuiltin ("XDrawImageString16", _X_DrawImageString16) ;
    calc->AddBuiltin ("DisplayWidth",  _X_DisplayWidth) ;
    calc->AddBuiltin ("DisplayHeight", _X_DisplayHeight) ;
    calc->AddBuiltin ("XResizeWindow", _X_ResizeWindow) ;
    calc->AddBuiltin ("XMoveWindow",   _X_MoveWindow) ;
    calc->AddBuiltin ("XMoveResizeWindow",     _X_MoveResizeWindow) ;
    calc->AddBuiltin ("XSetWindowBorderWidth", _X_SetWindowBorderWidth) ;
    calc->AddBuiltin ("XColorValue",    _X_ColorValue) ;
    calc->AddBuiltin ("XPixelColor",	_X_PixelColor) ;
    calc->AddBuiltin ("DisplayPlanes",  _X_DisplayPlanes) ;
    calc->AddBuiltin ("XFillRectangle", _X_FillRectangle) ;
    calc->AddBuiltin ("XFillPolygon",   _X_FillPolygon) ;
    calc->AddBuiltin ("XSetForeground", _X_SetForeground) ;
    calc->AddBuiltin ("XSetBackground", _X_SetBackground) ;
    calc->AddBuiltin ("XSetFunction",   _X_SetFunction) ;
    calc->AddBuiltin ("XSetPlaneMask",  _X_SetPlaneMask) ;
    calc->AddBuiltin ("XSetFont",	_X_SetFont) ;
    calc->AddBuiltin ("XSetClipOrigin", _X_SetClipOrigin) ;
    calc->AddBuiltin ("XSetClipMask",	_X_SetClipMask) ;
    calc->AddBuiltin ("DefaultDepth",   _X_DefaultDepth) ;
    calc->AddBuiltin ("XmbDrawString",  _X_MbDrawString) ;
    calc->AddBuiltin ("XmbDrawImageString", _X_MbDrawImageString) ;
    calc->AddBuiltin ("XCreateFontSet", _X_CreateFontSet) ;
    calc->AddBuiltin ("XFreeFontSet",   _X_FreeFontSet) ;
    calc->AddBuiltin ("XLoadQueryFont", _X_LoadQueryFont) ;
    calc->AddBuiltin ("XFreeFont",	_X_FreeFont) ;
    calc->AddBuiltin ("XTextWidth",	_X_TextWidth) ;
    calc->AddBuiltin ("XTextWidth16",	_X_TextWidth16) ;
    calc->AddBuiltin ("XDrawMixedString", _X_DrawMixedString) ;
    calc->AddBuiltin ("XMixedStringWidth", _X_MixedStringWidth) ;
    calc->AddBuiltin ("XChangeWindowAttributes", _X_ChangeWindowAttributes) ;
    calc->AddBuiltin ("XSetWindowBorder", _X_SetWindowBorder) ;
    calc->AddBuiltin ("XSetWindowBorderPixmap", _X_SetWindowBorderPixmap) ;
    calc->AddBuiltin ("XSetWindowBackground", _X_SetWindowBackground) ;
    calc->AddBuiltin ("XSetWindowBackgroundPixmap", _X_SetWindowBackgroundPixmap) ;
    calc->AddBuiltin ("XGrabPointer",	_X_GrabPointer) ;
    calc->AddBuiltin ("XUngrabPointer", _X_UngrabPointer) ;
    calc->AddBuiltin ("XCopyArea",      _X_CopyArea) ;
    calc->AddBuiltin ("QLength",	_X_QLength) ;
    calc->AddBuiltin ("XSync",		_X_Sync) ;
    calc->AddBuiltin ("XEventsQueued",  _X_EventsQueued) ;
    calc->AddBuiltin ("XPending",	_X_Pending) ;
    calc->AddBuiltin ("XFlush",		_X_Flush) ;
    calc->AddBuiltin ("XDrawPoint",	_X_DrawPoint) ;
    calc->AddBuiltin ("XDrawPoints",    _X_DrawPoints) ;
    calc->AddBuiltin ("XCreatePixmap",	_X_CreatePixmap) ;
    calc->AddBuiltin ("XFreePixmap",	_X_FreePixmap) ;
    calc->AddBuiltin ("XCreatePixmapCursor", _X_CreatePixmapCursor) ;
    calc->AddBuiltin ("XDefineCursor",	    _X_DefineCursor) ;
    calc->AddBuiltin ("XUndefineCursor",    _X_UndefineCursor) ;
    calc->AddBuiltin ("XFillArc",	    _X_FillArc) ;
    calc->AddBuiltin ("XBell",		_X_Bell) ;
    calc->AddBuiltin ("XClearWindow",   _X_ClearWindow) ;
    calc->AddBuiltin ("XClearArea",	_X_ClearArea) ;
    calc->AddBuiltin ("XQueryPointer",  _X_QueryPointer) ;
    calc->AddBuiltin ("XLookupString",  _X_LookupString) ;
    calc->AddBuiltin ("IsKeypadKey",    _X_IsKeypadKey) ;
    calc->AddBuiltin ("IsPrivateKeypadKey", _X_IsPrivateKeypadKey) ;
    calc->AddBuiltin ("IsPFKey",	_X_IsPFKey) ;
    calc->AddBuiltin ("IsFunctionKey",	_X_IsFunctionKey) ;
    calc->AddBuiltin ("IsMiscFunctionKey", _X_IsMiscFunctionKey) ;
    calc->AddBuiltin ("IsModifierKey", _X_IsModifierKey) ;
    calc->AddBuiltin ("XRaiseWindow", _X_RaiseWindow) ;
    calc->AddBuiltin ("XLowerWindow", _X_LowerWindow) ;
    calc->AddBuiltin ("XTranslateCoordinates", _X_TranslateCoordinates) ;
    calc->AddBuiltin ("XCreateFontCursor", _X_CreateFontCursor) ;
    calc->AddBuiltin ("XFreeFont", 	   _X_FreeFont) ;
    calc->AddBuiltin ("XStoreName",	   _X_StoreName) ;

    calc->Execute (xClassDef) ;
    calc->AddBiMethod ("XBitmap", "CreatePixmap", _XBitmap_CreatePixmap) ;
    calc->AddBiMethod ("XBitmap", "CreateBitmap", _XBitmap_CreateBitmap) ;
    calc->AddBiMethod ("XFontStruct", "Load",	  _XFontStruct_Load) ;
    calc->AddBiMethod ("XFontStruct", "Free",	  _XFontStruct_Free) ;
    return 0 ;
} ;
